import json

from django import forms
from django.contrib import admin, messages
from django.contrib.auth.admin import UserAdmin as DjangoUserAdmin
from django.contrib.auth.forms import UserChangeForm as DjangoUserChangeForm
from django.forms import JSONField
from django.forms.fields import InvalidJSONInput
from django.utils.translation import gettext_lazy as _
from rest_framework.authtoken.admin import TokenAdmin

from pount.apps.api.models import Item, Project, Referential, ReferentialData, Set, Template, User


@admin.action(description=_("Toggle project creation privilege"))
def toggle_project_creator(modeladmin, request, queryset):
    promoted_users = []
    for user in queryset:
        user.settings["project_creator"] = not user.is_project_creator
        user.save()
        promoted_users.append(str(user))
    messages.add_message(
        request,
        messages.SUCCESS,
        f"Project creator status changed for {', '.join(promoted_users)}",
    )


class UserChangeForm(DjangoUserChangeForm):
    project_creator = forms.BooleanField(
        required=False,
        label=_("Project creator"),
        help_text=_("Designates whether the user can create projects."),
    )

    def save(self, commit=True):
        project_creator = self.cleaned_data.get("project_creator", False)
        user = super().save(commit=False)
        user.settings["project_creator"] = project_creator
        user.save()
        return user


@admin.register(User)
class UserAdmin(DjangoUserAdmin):
    list_display = [
        "username",
        "name",
        "email",
        "is_staff",
        "project_creator",
    ]
    fieldsets = (
        (None, {"fields": ("username", "password")}),
        (_("Personal info"), {"fields": ("first_name", "last_name", "email")}),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "project_creator",
                    "is_staff",
                    "is_superuser",
                )
            },
        ),
        (_("Important dates"), {"fields": ("last_login", "date_joined")}),
    )
    readonly_fields = DjangoUserAdmin.readonly_fields + (
        "username",
        "email",
        "last_login",
        "date_joined",
    )
    actions = [toggle_project_creator]
    form = UserChangeForm

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        try:
            form.base_fields["project_creator"].initial = obj.is_project_creator
        except KeyError:
            pass
        return form

    @admin.display(
        description=_("Project creator"),
        boolean=True,
    )
    def project_creator(self, instance):
        return instance.is_project_creator

    def has_add_permission(self, request):
        return False


class PrettyJSONField(JSONField):
    widget_settings = forms.Textarea(attrs={"rows": 20, "cols": 80})

    def prepare_value(self, value):
        if isinstance(value, InvalidJSONInput):
            return value
        return json.dumps(value, indent=4)


class ProjectForm(forms.ModelForm):
    class Meta:
        field_classes = {
            "info": PrettyJSONField,
        }
        widgets = {"info": PrettyJSONField.widget_settings}


@admin.register(Project)
class ProjectAdmin(admin.ModelAdmin):
    form = ProjectForm

    def delete_model(self, request, obj):
        super().delete_model(request, obj)


class TemplateForm(forms.ModelForm):
    class Meta:
        field_classes = {
            "fields": PrettyJSONField,
            "messages": PrettyJSONField,
        }
        widgets = {
            "fields": PrettyJSONField.widget_settings,
            "messages": PrettyJSONField.widget_settings,
            "private_fields": PrettyJSONField.widget_settings,
        }


@admin.register(Template)
class TemplateAdmin(admin.ModelAdmin):
    form = TemplateForm


class SetForm(forms.ModelForm):
    class Meta:
        field_classes = {
            "settings": PrettyJSONField,
        }
        widgets = {
            "settings": PrettyJSONField.widget_settings,
        }


@admin.register(Set)
class SetAdmin(admin.ModelAdmin):
    form = SetForm


class ItemForm(forms.ModelForm):
    class Meta:
        field_classes = {
            "metadata": PrettyJSONField,
            "viewer": PrettyJSONField,
            "settings": PrettyJSONField,
        }
        widgets = {
            "metadata": PrettyJSONField.widget_settings,
            "viewer": PrettyJSONField.widget_settings,
            "settings": PrettyJSONField.widget_settings,
        }


@admin.register(Item)
class ItemAdmin(admin.ModelAdmin):
    form = ItemForm


class ReferentialDataForm(forms.ModelForm):
    class Meta:
        field_classes = {
            "definition": PrettyJSONField,
            "languages": PrettyJSONField,
            "lists": PrettyJSONField,
        }
        widgets = {
            "definition": PrettyJSONField.widget_settings,
            "languages": PrettyJSONField.widget_settings,
            "lists": PrettyJSONField.widget_settings,
        }


class ReferentialDataInline(admin.StackedInline):
    model = ReferentialData
    form = ReferentialDataForm


@admin.register(Referential)
class ReferentialAdmin(admin.ModelAdmin):
    list_display = ["name", "scope", "project"]
    list_filter = ["scope", "creator"]
    search_fields = ["name", "description", "project__name"]

    fields = ["name", "description", "project", "scope", "source", "creator", "created_at", "updated_at"]
    readonly_fields = ["created_at", "updated_at"]
    inlines = [ReferentialDataInline]


TokenAdmin.raw_id_fields = ["user"]
