from __future__ import annotations

import logging
from typing import Any

from django.conf import settings

from pount.apps.api.libs.structure import structure_reference
from pount.apps.api.models import User

logger = logging.getLogger(__file__)


def get_affiliation_info(user_attributes: dict[str, Any]) -> tuple[str, dict[str, Any]]:
    affiliations: list[str] = user_attributes.pop("affiliations", [])  # only multivalued attribute

    # If attributes are lists, convert them to strings
    cleaned_attributes: dict[str, Any] = {
        key: user_attributes[key][0]
        for key in user_attributes
        if isinstance(user_attributes[key], list) or user_attributes[key]
    }

    try:  # assure primary is the first affiliation
        primary_index = affiliations.index(cleaned_attributes.pop("primary_affiliation", ""))
        if primary_index:
            affiliations.insert(0, affiliations.pop(primary_index))
    except ValueError:
        pass  # index error: empty affiliations or unknown primary

    email: str = cleaned_attributes.pop("email", "")

    affiliation_info: dict = {
        "institution": "Université de Strasbourg",
        "research_labs": [],
        "affiliations": [],
    }

    # For each affiliation code, get the structure info
    for structure_code in affiliations:
        try:
            current_structure = structure_reference.get(structure_code)
        except ValueError:
            logger.exception(f"Structure code {structure_code} not found")
            continue

        if current_structure.is_research_unit:
            labs = affiliation_info["research_labs"]
            labs.append(current_structure.to_dict())
        else:
            affiliations = affiliation_info["affiliations"]
            affiliations.append(current_structure.long_name)

    cleaned_attributes["info"] = affiliation_info

    return email, cleaned_attributes


def update_or_create_user(user_attributes: dict[str, Any]) -> None:
    """
    user_attributes is a dictionary of the form:
    {
        "username": "username",
        "first_name": ["first_name"],
        "last_name": ["last_name"],
        "email": ["email"],
        "affiliations": ["udsTopUnitCode"],
        "primary_affiliation": ["udsPrimaryTopUnitCode"],
    }
    """
    # Callback on CAS login (settings.CAS_RESPONSE_CALLBACK)
    username: str = settings.CAS_USERNAME_FORMAT(user_attributes.pop("username"))

    try:
        user = User.objects.get(username=username)

        # Get fresh user info if corresponding flag is set in `settings` field
        if user.settings.get("refresh_info", None):
            email, cleaned_attributes = get_affiliation_info(user_attributes)
            user.email = email
            user.first_name = cleaned_attributes["first_name"]
            user.last_name = cleaned_attributes["last_name"]
            cleaned_attributes["info"]["bio"] = user.info.get("bio", "")
            cleaned_attributes["info"]["other_affiliations"] = user.info.get("other_affiliations", [])
            user.info = cleaned_attributes["info"]
            user.settings["refresh_info"] = False
            user.save()

    except User.DoesNotExist:
        email, cleaned_attributes = get_affiliation_info(user_attributes)
        User.objects.create_user(username, email, **cleaned_attributes)
