from __future__ import annotations

import os
from io import BytesIO
from zipfile import ZipFile

from PIL.Image import Image

from pount.apps.api.storage_backends import ItemMediaFileStorage, ThumbnailFileStorage


def item_mediafile_fullname(name: str, item, ignore_location=False) -> str:
    """The filename is prefixed with the item_id"""
    if item.pk is None:
        raise ValueError("Item is not yet saved")
    if ignore_location:
        return f"{str(item.pk)}/{name}"
    return f"{ItemMediaFileStorage.location}/{str(item.pk)}/{name}"


def item_mediafile_filename(name: str) -> str:
    """In serializer, we strip the item_id and keep only the filename"""
    return name.split("/")[-1]


def sizeof_format(num, suffix="B"):
    if num is None:
        return ""
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return "{:.1f} {}{}".format(num, "Yi", suffix)


def make_thumbnail(image: Image, size) -> Image:
    thumbnail = image.copy()
    thumbnail.thumbnail(size)

    return thumbnail


def resize_min_height(image: Image, size) -> Image | None:
    """
    resize image keeping width/height ratio to minimal height

    @param image: Image to resize
    @param size: tuple(None, min_height)
    @return: resized image
    """
    width, height = image.size
    min_width, min_height = size
    if height < min_height:
        # image height is less than min_height, we don't keep it
        return None
    new_width = int((width * min_height) / height)
    resized_image = image.resize((new_width, min_height))
    return resized_image


def upload_thumbnail(file: Image, filename, storage: ThumbnailFileStorage):
    with BytesIO() as output:
        file.save(output, format="PNG")
        filename = storage.save(filename, output)

    return filename


def delete_rti_files(item) -> None:
    # Delete all RTI MediaFiles related to an Item
    from pount.apps.api.models.item import MediaFile

    files_to_replace = MediaFile.objects.filter(item=item, info__is_rti=True)
    # Delete each MediaFile as we want to delete S3 file too
    for _file in files_to_replace:
        _file.delete()


def upload_rti_zipfile(zfile: ZipFile, item, component: str | None):
    from pount.apps.api.models.item import MediaFile

    storage = ItemMediaFileStorage()

    # We first delete all MediaFiles related to item if they exist before
    # uploading new ones
    delete_rti_files(item)

    for filename in zfile.namelist():
        extracted_file = zfile.open(filename)
        key = os.path.join(f"{item.pk}/rti/", filename)
        storage.save(key, extracted_file)

        # We create a MediaFile to keep track of the file
        mediafile = MediaFile.objects.create(
            component=component,
            filename=f"rti/{filename}",
            item=item,
            info={
                "is_rti": True,
                "content_type": "application/json" if filename == "info.json" else "unknown",
            },
            is_hidden=False if filename == "info.json" else True,
        )
        if filename == "info.json":
            mediafile_to_return = mediafile

    return mediafile_to_return
