from typing import Union
from uuid import UUID

from django.contrib.auth import get_user_model
from django.core import signing
from django.core.signing import BadSignature, loads
from django.utils.translation import gettext_lazy as _

from pount.apps.api.libs.email import send_email
from pount.apps.api.models import Invitation

User = get_user_model()


def make_invitation_token(email: str, project_id: Union[UUID, str], user: User = None) -> str:
    data = {
        "email": email,
        "project_id": str(project_id),
        "user_id": str(user.id) if user else None,
    }
    return signing.dumps(data)


def validate_invitation(invitation: Invitation) -> tuple[dict, bool]:
    if invitation.is_expired:
        invitation.delete()
        return {"error_text": "invitation_expired"}, True
    try:
        data = verify_invitation(invitation.token, invitation.id)
    except BadSignature:
        return {"error_text": "invitation_not_valid"}, True
    return data, False


def verify_invitation(token: str, invitation_id: UUID) -> dict:
    data: dict = loads(token)
    data.update({"invitation_id": str(invitation_id)})
    return data


def send_invitation_email(email, project, invitation_url):
    subject = _("Invitation to join the project %(project_name)s") % {"project_name": project.name}
    context = {
        "project_name": project.name,
        "url": invitation_url,
        "manager": "Project manager",
    }
    template = "email/user/invite_user"
    send_email(subject, template, context, email)


def send_internal_invitation_email(email, project):
    subject = _("Invitation to join the project %(project_name)s") % {"project_name": project.name}
    context = {
        "project_name": project.name,
        "manager": "Project manager",
    }
    template = "email/user/invite_internal_user"
    send_email(subject, template, context, email)
