import csv
import io

from django.contrib.auth import get_user_model
from django.core.files import File

from pount.apps.api.libs.metadata import Metadata, MetadataType
from pount.apps.api.libs.metadata.processors import (
    MetadataProcessor,
    ReferentialValuesProcessor,
    SelectValuesProcessor,
)
from pount.apps.api.models import Item, Set, Template
from pount.exceptions import CSVImportError, CSVMetadataError, MetadataOptionValueNotFound

User = get_user_model()


def import_items_from_csv(csv_file: File, into_set: Set, creator: User) -> int:
    created_items = 0
    row_number = 0
    try:
        metadata_parser = Metadata([], template_fields=into_set.template.fields)
        reader = csv.DictReader(io.StringIO(csv_file.read().decode("utf-8")))

        # The header must contains the ids of all the template required fields
        missing_fields = {_.id for _ in metadata_parser.template_fields.required_fields} - set(reader.fieldnames)
        if missing_fields:
            from django.utils.translation import gettext_lazy as _

            raise CSVImportError(
                error=_(
                    "Required field(s) missing in CSV file: %(fields)s"
                    % {
                        "fields": ", ".join(
                            metadata_parser.template_fields.get_field_by_id(_id).name for _id in missing_fields
                        )
                    }
                ),
                line=0,
            )

        # Fist line (after headers) contains field label, skip it
        next(reader)
        for row in reader:
            row_number += 1
            item_metadata = metadata_parser.from_csv_row(row, row_number=row_number)
            item_metadata = process_metadata(item_metadata, into_set.template)
            Item.objects.create(
                description="",
                metadata=item_metadata,
                set=into_set,
                creator=creator,
            )
            created_items += 1
    except (csv.Error, UnicodeDecodeError, CSVMetadataError, MetadataOptionValueNotFound) as e:
        from django.utils.translation import gettext_lazy as _

        if isinstance(e, CSVImportError):
            error = _("CSV import error at line %(line)d: %(err)s") % {"line": row_number, "err": e.error}
        else:
            error = _("CSV import error at line %(line)d: %(err)s") % {"line": row_number, "err": str(e)}
        raise CSVImportError(
            error=error,
            line=row_number,
        )
    return created_items


def process_metadata(metadata: MetadataType, template: Template) -> MetadataType:
    processors: list[MetadataProcessor] = [
        ReferentialValuesProcessor(),
        SelectValuesProcessor(template.messages),
    ]
    for processor in processors:
        metadata = processor.process(metadata, template.fields)
    return metadata
