from urllib.parse import parse_qsl, unquote, urlencode, urlparse, urlunsplit

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from pount.apps.api.libs.email import send_email
from pount.apps.api.models.login import Ticket
from pount.apps.api.models.user import AuthenticationProvider
from pount.apps.api.serializers.jwt import TokenObtainSerializer

User = get_user_model()


def create_ticket(user: User, provider: AuthenticationProvider) -> Ticket:
    ticket = Ticket.objects.create(
        user=user,
        provider=provider,
    )
    return ticket


def add_ticket_to_url(url, ticket_id):
    """
    Add ticket id to redirection URL, merge it into possibly existing
    querystring params
    """
    parsed_url = urlparse(unquote(url))
    querystrings = dict(parse_qsl(parsed_url.query))
    querystrings.update({"ticket": ticket_id})
    return urlunsplit(
        (
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            urlencode(querystrings, doseq=True),
            "",
        )
    )


def make_jwt_token_pair(user: User, request):
    """
    Generate a JWT refresh and access token for a user
    """
    serializer = TokenObtainSerializer(data={}, context={"user": user, "request": request})
    if serializer.is_valid(raise_exception=True):
        return serializer.validated_data


def send_reset_password_email(user: User, request):
    subject = _("Password reset request")
    template = "email/login/password_reset"
    token = user.settings["password_reset"]["token"]
    rid = user.settings["password_reset"]["reset_id"]
    if request.data.get("url") is None:
        raise ValueError("Missing URL in request")
    context = {"reset_url": f"{request.data.get('url')}?rid={rid}&token={token}"}
    send_email(subject, template, context, user.email, request=request)
