from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar
from uuid import UUID

from django.utils.translation import gettext_lazy as _

from pount.exceptions import CSVMetadataError, MetadataOptionValueNotFound, MetadataValidationError

if TYPE_CHECKING:
    from pount.apps.api.models import Template


class TemplateEntry:
    def __init__(self, entry: dict):
        self._data: dict[str, Any] = {}
        self.parse_entry(entry)

    def parse_entry(self, entry: dict) -> None:
        for k, v in entry.items():
            self._data[k] = v

    @property
    def repeatable(self) -> bool:
        return self._data.get("repeatable", False)

    @property
    def multiple(self) -> bool:
        return self._data.get("multiple", False)

    @property
    def required(self) -> bool:
        return self._data.get("required", False)

    @property
    def component(self) -> str:
        return self._data.get("component", None)

    def has_option(self, name) -> bool:
        """Check if `name` is one of the entries in self._data["options"]"""
        if not name and not self.required:
            # An emtpy option is a valid option if it is not required
            return True
        try:
            for option in self.options:
                if option.get("value", "") == name:
                    return True
            return False
        except AttributeError:
            return False

    def get_option_label(self, option_id) -> str:
        try:
            return next(
                option["label"] for option in self.options if option["id"] == option_id or option["value"] == option_id
            )
        except StopIteration:
            return ""

    def get_option_id(self, label: str) -> str | None:
        # Returns the id of the option with the given label
        try:
            return next(option["id"] for option in self.options if option["label"] == label)
        except StopIteration:
            return ""

    def get_options(self):
        return [(option["id"], option.get("label", "")) for option in self.options]

    def __getattr__(self, item) -> Any:
        try:
            return self._data[item]
        except KeyError:
            if item == "name":
                # If name is not set, use the id as name
                return self._data["id"]
            if item == "label":
                # If label is not set, use the name as label
                return self.name
            raise AttributeError

    def __str__(self) -> str:
        return self._data.get("name", "Unknown name")


class MetadataEntry:
    REQUIRED_FIELDS = ["id", "name", "value"]

    def __init__(self, entry: dict[str, Any]) -> None:
        self._data: dict[str, Any] = {}
        self.parse_entry(entry)

    def __str__(self):
        return f"{self.name}: {str(self.value)}"

    def parse_entry(self, entry: dict[str, Any]):
        for k, v in entry.items():
            self._data[k] = v
        for k in MetadataEntry.REQUIRED_FIELDS:
            # Validate required keys
            if k not in self._data.keys():
                raise ValueError(_('Missing key "%(key)s" for metadata entry') % {"key": k})

    @property
    def id(self) -> str:
        return self._data["id"]

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def required(self) -> bool:
        return self._data.get("required", False)

    @property
    def multiple(self) -> bool:
        return self._data.get("multiple", False)

    @property
    def repeatable(self) -> bool:
        return self._data.get("repeatable", False)

    @property
    def is_option(self) -> bool:
        if self.template and hasattr(self.template, "options"):
            return bool(self.template.options)
        return self._data.get("is_option", False)

    @property
    def is_referential_value(self) -> bool:
        if self.template and hasattr(self.template, "component") and self.template.component == "ReferentialSelect":
            return True
        return False

    @property
    def template(self) -> TemplateEntry:
        return self._data.get("template", None)

    @property
    def label(self) -> dict[str, dict[str, str]] | None:
        return self._data.get("label", None)

    @property
    def value(self) -> Any:
        return self._data.get("value", None)

    @property
    def value_to_index(self) -> Any:
        # Options and referentials store the ID in the value, we need to search the real value in the labels
        # If translated, we may have a value for each language, for the moment we return the first one
        if label := self.label:
            values_to_index = []
            value = self.value if isinstance(self.value, list) else list(self.value)
            # For each value
            for val in value:
                labels = {f"opt_{lang}": label[val][lang] for lang in label[val].keys()}
                labels["opt_id"] = val
                values_to_index.append(labels)
            return values_to_index
        return self.value

    def as_dict(self) -> dict[str, Any]:
        """
        Return a dict with the required fields and those that are not falsy
        """
        fields = [
            "id",
            "name",
            "is_option",
            "repeatable",
            "multiple",
            "required",
            "value",
            "label",
            "level_id",
            "referential_id",
        ]
        return {
            field_name: getattr(self, field_name)
            for field_name in fields
            if field_name in MetadataEntry.REQUIRED_FIELDS or getattr(self, field_name)
        }

    def set_value(self, new_value: Any) -> None:
        value = self._validate(new_value)
        self._data["value"] = value

    def set_label(self, value: Any) -> None:
        self._data["label"] = value

    def add_template_entry(self, template_entry: TemplateEntry):
        self._data["template"] = template_entry

    def __getattr__(self, item: str) -> Any:
        try:
            return self._data[item]
        except KeyError:
            # When importing from CSV, these keys are not set, they will be set from the template when serializing
            if item in ["referential_id", "level_id"] and self.template:
                return getattr(self.template, item, None)
            return None

    T = TypeVar("T")

    def _validate(self, value: T) -> T:
        if self.template:
            if self.required:
                value = self._validate_required(value)
            if self.repeatable:
                value = self._validate_repeatable(value)
            if hasattr(self.template, "options"):
                value = self._validate_options(value)
        return value

    def _validate_repeatable(self, value: T) -> T:
        """
        Repeated fields can not have more than options["max_repeat"] values
        """
        if not isinstance(value, list):
            raise MetadataValidationError(
                error=_("Repeatable field %(field_name)s value should be a list" % {"field_name": self.name})
            )
        if len(value) > (int(self.template.max_repeat) + 1):
            raise MetadataValidationError(
                error=_(
                    "Too many values for %(field_name)s: max_repeat is %(max_repeat)d"  # noqa: E501
                    % {
                        "field_name": self.name,
                        "max_repeat": self.template.max_repeat,
                    }
                )
            )
        return value

    def _validate_options(self, value: T) -> T:
        if self.template.options:
            if isinstance(value, str):
                if not self.template.has_option(value):
                    raise MetadataValidationError(
                        error=_(
                            "Invalid option %(value)s for %(field_name)s" % {"value": value, "field_name": self.name}
                        )
                    )
            if isinstance(value, list):
                for _item in value:
                    if isinstance(_item, str):
                        if not self.template.has_option(_item):
                            raise MetadataValidationError(
                                error=_(
                                    "Invalid option %(value)s for %(field_name)s"  # noqa: E501
                                    % {"value": _item, "field_name": self.name}
                                )
                            )
                    elif isinstance(_item, list):
                        if not all(self.template.has_option(_opt.strip()) for _opt in _item):
                            raise MetadataValidationError(
                                error=_("Invalid option(s) for %(field_name)s" % {"field_name": self.name})
                            )
        return value

    def _validate_required(self, value: T) -> T:
        if not self.required:
            return value
        missing_error = _("Missing required value for %(field_name)s" % {"field_name": self.name})
        if not value:
            raise MetadataValidationError(error=missing_error)
        if isinstance(value, str):
            if not (value.strip()):
                raise MetadataValidationError(error=missing_error)
        if isinstance(value, list):
            if len(value) == 0:
                raise MetadataValidationError(error=missing_error)
        return value


class MetadataIterator:
    def __init__(self, entries):
        self._entries = entries
        self._index = 0

    def __next__(self) -> MetadataEntry:
        try:
            entry = self._entries[self._index]
            self._index += 1
            return entry
        except IndexError:
            raise StopIteration


class TemplateFields:
    LAYOUT_FIELDS = [
        "UsfAccordion",
        "UsfCol",
        "UsfContent",
        "UsfPanel",
        "UsfRow",
        "ReferentialComponent",
    ]

    def __init__(self, template_fields: list):
        self._entries: dict[str, TemplateEntry] = {}
        self.parse_fields(template_fields)

    def __iter__(self) -> MetadataIterator:
        return MetadataIterator(list(self._entries.values()))

    def __len__(self) -> int:
        return len(self._entries)

    @classmethod
    def from_template(cls, template: Template) -> TemplateFields:
        return cls(template.fields)

    def parse_fields(self, template_fields: list):
        def get_field(field):
            try:
                name = field.get("name", None) or field["id"]
                if field.get("component", "") not in TemplateFields.LAYOUT_FIELDS:
                    # Ignore layout fields
                    self._entries[name] = TemplateEntry(field)
            except KeyError:
                pass
            if field.get("fields", None):
                for subfield in field["fields"]:
                    get_field(subfield)

        for field in template_fields:
            get_field(field)

    def get_field(self, name) -> TemplateEntry | None:
        try:
            return self._entries[name]
        except KeyError:
            return None

    def get_field_by_id(self, field_id) -> TemplateEntry | None:
        try:
            fields_with_id = [_ for _ in self if _.id == field_id]
            field = fields_with_id[0]
        except IndexError:
            field = None
        return field

    def __getattr__(self, item: str) -> TemplateEntry:
        try:
            return self._entries[item]
        except KeyError:
            raise AttributeError

    @property
    def fieldnames(self) -> list[str]:
        return [_ for _ in self._entries]

    @property
    def required_fields(self) -> list[TemplateEntry]:
        return [_ for _ in self._entries.values() if _.required]


class Metadata:
    def __init__(self, metadata: list, template_fields: list = None) -> None:
        self._entries = {}
        self.parse_metadata(metadata)
        self.template_fields = None
        if template_fields:
            self.parse_template_fields(template_fields)

    def __iter__(self) -> MetadataIterator:
        return MetadataIterator(list(self._entries.values()))

    def __len__(self) -> int:
        return len(self._entries)

    def __getattr__(self, item) -> MetadataEntry:
        try:
            return self._entries[item]
        except KeyError:
            raise AttributeError

    def to_representation(self) -> list[dict[str, Any]]:
        if self.template_fields:
            return [self._entries[_field_name].as_dict() for _field_name in self.template_fields.fieldnames]
        return []

    def parse_metadata(self, metadata: list) -> None:
        for m in metadata:
            entry = MetadataEntry(m)
            self._entries[entry.name] = entry

    def parse_template_fields(self, template_fields: list) -> None:
        self.template_fields = TemplateFields(template_fields)
        self.update_or_create_entries()

    def update_or_create_entries(self) -> None:
        for field in self.template_fields:
            if self._entries.get(field.name):
                self.update_entry(field)
            else:
                self.create_entry(field)

    def update_entry(self, field: TemplateEntry) -> None:
        self._entries[field.name].add_template_entry(field)

    def create_entry(self, field: TemplateEntry) -> None:
        self._entries[field.name] = MetadataEntry(
            {
                "id": field.id,
                "name": field.name,
                "value": None,
                "is_option": False,
                "multiple": field.multiple,
                "required": field.required,
                "repeatable": field.repeatable,
                "template": field,
            }
        )

    def get_entry_by_id(self, id: str) -> MetadataEntry | None:
        return next((_e for _e in self._entries.values() if _e.id == id), None)

    def get_entry_by_name(self, name: str) -> MetadataEntry | None:
        return next((_e for _e in self._entries.values() if _e.name == name), None)

    def from_csv_row(self, row: dict, row_number: int):
        """
        @param row: A dictionary representing a row in CSV file
        @param row_number: An integer representing the row number in the CSV file
        @return: The representation of the processed row as a dictionary
        """
        for field_id, field_value in row.items():
            self.process_csv_field(field_id, field_value.strip(), row_number)

        return self.to_representation()

    def process_csv_field(self, field_id: str, field_value: str, row_number: int) -> None:
        """
        @param field_id: The ID of the field being processed.
        @param field_value: The value of the field being processed.
        @param row_number: The row number of the CSV file being processed.
        @return: None.
        """
        entry: MetadataEntry = self.get_entry_by_id(field_id)
        if entry is None:
            return

        try:
            new_value = self.get_entry_value_from_csv(entry, field_value)
            entry.set_value(new_value)
            self._entries[entry.name] = entry
        except (MetadataValidationError, MetadataOptionValueNotFound) as error:
            raise CSVMetadataError(
                error=error.error,
                line=row_number,
            )

    def get_entry_value_from_csv(self, entry: MetadataEntry, field_value: str) -> str | list[str] | list[list[str]]:
        """
        @param entry: The MetadataEntry object containing information about the field in the CSV file.
        @param field_value: The value of the field from the CSV file.
        @return: The processed value of the field based on the MetadataEntry.
        """

        if entry.repeatable and not entry.multiple:
            field_value = [_val.strip() for _val in field_value.split(";")]
        if entry.multiple and not entry.repeatable:
            field_value = [_val.strip() for _val in field_value.split("|")]
        if entry.multiple and entry.repeatable:
            field_value = [
                [_multiple.strip() for _multiple in _repeated.split(";")] for _repeated in field_value.split("|")
            ]

        # When importing CSV, selects and referential entries are values that must be changed to their ids
        if entry.template.component in [
            "UsfSelect",
            "UsfRadioGroup",
        ]:
            # Look up the id of the selected value in the entry.template options
            # value can be a list or a string
            if isinstance(field_value, str):
                field_value = [field_value]
            field_value = [self.lookup_option_id_from_value(entry, _val) for _val in field_value]

        if entry.template.component in ["ReferentialSelect"]:
            # Look up the id of the selected value in the referential's node list
            if isinstance(field_value, str):
                field_value = [field_value]
            field_value = [self.lookup_referential_node_id_from_value(entry, _val) for _val in field_value]

        # We keep only values that are not "" or None
        if isinstance(field_value, list):
            field_value = [val for val in field_value if val]
            # If empty list, return None, unless the field is repeatable
            if not field_value and not entry.repeatable:
                field_value = None

        return field_value

    def lookup_option_id_from_value(self, entry: MetadataEntry, value: str | list[str]) -> str | list[str] | None:
        if isinstance(value, str):
            option_id: str | None = entry.template.get_option_id(value)
        else:
            option_id: list[str] = [self.lookup_option_id_from_value(entry, _val) for _val in value]
        if not option_id and entry.template.required:
            raise MetadataOptionValueNotFound(
                error=_("Value <%(value)s> not found in option list for <%(entry_name)s>")
                % {"value": value, "entry_name": entry.name},
            )
        return option_id

    @staticmethod
    def lookup_referential_node_id_from_value(entry: MetadataEntry, value: str) -> str | None:
        from pount.apps.api.libs.referential import get_referential_node_id_for_node_value

        if not value:
            return None
        # Look up the id of the selected value in the referential values
        rid = entry.template.referential_id
        lid = entry.template.level_id
        node_id = get_referential_node_id_for_node_value(UUID(rid), UUID(lid), value)
        return str(node_id)


MetadataType = list[dict[str, Any]]
