from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any

from pount.apps.api.libs.metadata import Metadata, MetadataEntry, MetadataType
from pount.apps.api.libs.referential import get_node_translations


class MetadataProcessor(ABC):
    @abstractmethod
    def process(self, metadata: MetadataType, template_fields: list) -> MetadataType:
        pass


class ReferentialValuesProcessor(MetadataProcessor):
    """
    The `ReferentialValuesProcessor` class is responsible for processing metadata and setting translated labels for
    entries associated with a referential.

    Example Usage:
    -------------
    metadata = [...]  # list of metadata fields
    template_fields = [...]  # list of fields of the template

    processor = ReferentialValuesProcessor()
    processed_metadata = processor.process(metadata, template_fields)

    """

    @staticmethod
    def translate_node(node_id: str, referential_id: str, level_id: str) -> dict[str, str]:
        """
        @param node_id: The ID of the node to be translated.
        @param referential_id: The ID of the referential from which to fetch translations.
        @param level_id: The ID of the level in the referential to which the node belongs.
        @return: A dictionary containing translations for the specified node. The keys are language codes and the
                values are the translated strings.
        """
        return get_node_translations(referential_id=referential_id, node_id=node_id)

    def get_entry_labels(self, entry: MetadataEntry) -> dict[str, dict[str, str]]:
        """
        @param entry: A MetadataEntry object containing information about a specific entry in the metadata.
        @return: A dictionary where the keys are node_ids and the values are dictionaries containing label information.
        """
        label = {}
        for node_id in entry.value:
            # value is a list of ids
            label.update(
                {
                    node_id: self.translate_node(
                        node_id=node_id,
                        referential_id=entry.referential_id,
                        level_id=entry.level_id,
                    )
                }
            )
        return label

    def process(self, metadata: MetadataType, template_fields: list) -> MetadataType:
        """
        Process the given metadata to set labels for entries with specific template fields.

        @param metadata: The metadata to process.
        @param template_fields: The list of template fields to consider.
        @return: The processed metadata.

        @rtype: MetadataType
        """
        metadata = Metadata(metadata=metadata, template_fields=template_fields)
        for entry in metadata:
            if entry.value and entry.template is not None and entry.template.component == "ReferentialSelect":
                if isinstance(entry.value, str):
                    entry.set_value(list(entry.value))
                entry.set_label(self.get_entry_labels(entry))
        return metadata.to_representation()


class SelectValuesProcessor(MetadataProcessor):
    """
    Add the translated values from template.messages or the default template option label for each option and each
    available language and store them in the metadata.label entry
    """

    def __init__(self, messages: dict[str, Any]):
        self.messages = messages

    @cached_property
    def available_languages(self) -> list[str]:
        try:
            return list(self.messages["locales"].keys())
        except KeyError:
            return []

    def translate_option(self, option_id: str, language: str, value: str) -> str:
        """
        Translates the option label for a given option ID, language, and default value.

        @param option_id: The ID of the option to translate.
        @param language: The language code for the translation.
        @param value: The default value if no translation is found.
        @return: The translated option label or the default value.
        """
        option = self.messages.get("locales", {}).get(language, {}).get(option_id, {}).get("label", None) or value
        return option

    def process(self, metadata: MetadataType, template_fields: list) -> MetadataType:
        """
        Process metadata by setting translated labels for select values.

        :param metadata: The metadata to process.
        :type metadata: MetadataType
        :param template_fields: The list of template fields.
        :type template_fields: list
        :return: The processed metadata.
        :rtype: MetadataType
        """
        metadata = Metadata(metadata=metadata, template_fields=template_fields)
        available_languages = self.available_languages
        for entry in metadata:
            if entry.template.component in ["UsfRadioGroup", "UsfSelect"]:
                label = {}
                values = entry.value
                if not values:
                    continue
                if isinstance(values, str):
                    values = [values]
                    entry.set_value(list(values))
                for value in values:
                    label[value] = {}
                    default_label = entry.template.get_option_label(value)
                    for language_code in available_languages:
                        label[value][language_code] = self.translate_option(value, language_code, default_label)
                entry.set_label(label)

        return metadata.to_representation()
