from uuid import UUID

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from notifications import settings as notifications_settings
from swapper import load_model

from pount.apps.api.models.group import Group
from pount.apps.api.models.item import Item
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set

Notification = load_model("notifications", "Notification")
User = get_user_model()

VERB_STRINGS = {
    "project_join": _("has joined project"),
    "project_create": _("has created project"),
    "project_update": _("has updated project"),
    "group_update": _("has updated group"),
    "group_delete": _("has deleted group"),
    "group_create": _("has created group"),
    "group_user_add": _("has added user to group"),
    "group_user_remove": _("has removed user from group"),
    "set_create": _("has created set"),
    "set_update": _("has updated set"),
    "item_create": _("has created item"),
    "item_update": _("has updated item"),
    "group_contribute_to_set": _("added group to contributors of set"),
    "group_uncontribute_to_set": _("removed group from contributors of set"),
}


def notify_join_project(actor: User, project: Project):
    return notify(
        sender=actor,
        recipient=get_project_managers(project),
        verb="project_join",
        action_object=actor,
        target=project,
        level=Notification.LEVELS.success,
        project=project,
    )


def notify_project_create(actor: User, project: Project):
    return notify(
        sender=actor,
        recipient=User.objects.admins(),
        verb="project_create",
        target=project,
        level=Notification.LEVELS.success,
        project=project,
    )


def notify_project_update(actor: User, project: Project):
    return notify(
        sender=actor,
        recipient=get_project_managers(project),
        verb="project_update",
        target=project,
        level=Notification.LEVELS.info,
        project=project,
    )


def notify_group(actor: User, group: Group, verb):
    """
    Notify group managers
    """
    return notify(
        sender=actor,
        recipient=get_project_managers(group.project),
        verb=verb,
        target=group,
        level=Notification.LEVELS.info,
        project=group.project,
    )


def notify_group_membership(actor: User, user: User, group: Group, verb):
    """
    Notify group managers and users
    """
    return notify(
        sender=actor,
        recipient=(
            get_project_managers(group.project) | group.users.all() | User.objects.filter(pk=user.pk)
        ).distinct(),
        verb=verb,
        action_object=user,
        target=group,
        level=Notification.LEVELS.info,
        project=group.project,
    )


def notify_set(actor: User, set: Set, verb):
    return notify(
        sender=actor,
        recipient=(get_project_members(set.project)),
        verb=verb,
        action_object=set,
        level=Notification.LEVELS.info,
        project=set.project,
        set=set,
    )


def notify_item(actor: User, item: Item, verb):
    set: Set = item.set
    return notify_set(actor, set, verb)


def notify_group_contribution_to_set(actor: User, group: Group, set: Set, verb: str):
    return notify(
        sender=actor,
        recipient=(get_project_managers(group.project) | group.users.all()).distinct(),
        verb=verb,
        action_object=group,
        target=set,
        level=Notification.LEVELS.success,
        project=group.project,
        set=set,
    )


def get_project_managers(project: Project):
    return project.managers


def get_project_members(project: Project):
    return project.members


def notify(**kwargs):
    """
    Customized version of notifications.base.models.notify_handler()
    taking advantage of Project and Set foreign keys for easy filtering
    """

    def recipient_is_actor(recipient, actor):
        if isinstance(recipient, User) and recipient == actor:
            return True
        if isinstance(recipient, (UUID, str)) and (str(recipient) == str(actor.id)):
            return True
        return False

    EXTRA_DATA = notifications_settings.get_config()["USE_JSONFIELD"]

    recipient = kwargs.pop("recipient")
    actor = kwargs.pop("sender")
    optional_objs = [(kwargs.pop(opt, None), opt) for opt in ("target", "action_object")]
    public = bool(kwargs.pop("public", True))
    verb = kwargs.pop("verb", "")
    description = kwargs.pop("description", None)
    timestamp = kwargs.pop("timestamp", timezone.now())
    level = kwargs.pop("level", Notification.LEVELS.info)
    project = kwargs.pop("project", None)
    set = kwargs.pop("set", None)

    # Check if User or Group
    if isinstance(recipient, Group):
        recipients = recipient.user_set.all()
    elif isinstance(recipient, (QuerySet, list)):
        recipients = recipient
    else:
        recipients = [recipient]

    new_notifications = []

    for recipient in recipients:
        if recipient_is_actor(recipient, actor):
            # don't send notification to actor
            continue
        newnotify = Notification(
            recipient=recipient,
            actor_content_type=ContentType.objects.get_for_model(actor),
            actor_object_id=actor.pk,
            verb=str(verb),
            public=public,
            description=description,
            timestamp=timestamp,
            level=level,
            project=project,
            set=set,
        )

        for obj, opt in optional_objs:
            if obj is not None:
                setattr(newnotify, "%s_object_id" % opt, obj.pk)
                setattr(
                    newnotify,
                    "%s_content_type" % opt,
                    ContentType.objects.get_for_model(obj),
                )

        if kwargs and EXTRA_DATA:
            newnotify.data = kwargs

        newnotify.save()
        new_notifications.append(newnotify)

    return new_notifications


def translated_verb(verb):
    return _(VERB_STRINGS[verb])
