from __future__ import annotations

import itertools
import uuid
from typing import Any, Optional
from uuid import UUID

from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from iso_language_codes import language_autonym
from pymongo.collection import Collection
from pymongo.database import Database

from pount.apps.api.libs.template import update_items_using_template
from pount.apps.api.models import Referential as ReferentialModel
from pount.apps.api.models import ReferentialData, Template
from pount.apps.api.models.referential import DEFAULT_LANGUAGE
from pount.decorators import cached
from pount.exceptions import LevelNotFoundError, MetadataOptionValueNotFound
from pount.mongo import mongo_db

CACHE_TTL = None  # Infinite timeout: valid until cache is invalidated

NodeListType = list[dict[str, str]]
ParamType = dict[str, Optional[str]]


def share_referential(referential: ReferentialModel, new_scope: int):
    scope: int = referential.scope

    if scope == ReferentialModel.Scope.SCOPE_BASE:
        raise ValueError(_("Can not change a base referential's scope"))

    if new_scope < scope:
        # Make public or make global
        if new_scope == ReferentialModel.Scope.SCOPE_BASE:
            referential.project = None
        referential.scope = new_scope
    else:
        # Make private to project, only if not used in other project
        if referential.projects.exclude(id=referential.id).exists():
            raise ValueError(_("Referential is used in other project"))
        referential.scope = new_scope

    referential.save(update_fields=["scope", "project", "updated_at"])
    return referential


def cache_key_fn(referential: ReferentialModel):
    """Callback function passed to decorator to make a cache key prefix"""
    return f"{referential._meta.model_name}:{referential.pk}"


@cached(timeout=CACHE_TTL, key_prefix=cache_key_fn, key_object="referential")
def get_default_language(referential: ReferentialModel) -> str:
    try:
        default_language: str = ReferentialData.objects.filter(referential=referential).values_list(
            "languages__default", flat=True
        )[0]
        default_language = default_language or DEFAULT_LANGUAGE
    except IndexError:
        default_language: str = DEFAULT_LANGUAGE
    return default_language


def build_node_list(referential: Referential, params: ParamType) -> NodeListType:
    """Extract list of nodes for a level, optionally filtered by parent, for storing in Referential.lists"""
    nodes = referential.nodes_for_level(**params)
    node_list = [{"id": _.id, "name": _.name, "parents": _.parents, "ancestors": _.ancestors} for _ in nodes]
    return node_list


def get_node_list(referential: ReferentialModel, params: ParamType) -> NodeListType:
    default_language = get_default_language(referential=referential)
    language: str = params.get("language", default_language)
    if language not in get_languages_list(referential=referential, code_only=True):
        language = default_language

    match = {
        "$match": {
            "referential_id": referential.id,
            "level_id": uuid.UUID(params["level_id"]),
        },
    }

    for key in ("parent", "ancestor"):
        if value := params.get(key):
            match["$match"][key + "s"] = uuid.UUID(value)

    query = [
        match,
        {"$sort": {"order": 1}},
        {
            "$project": {
                "_id": 0,
                "id": 1,
                "label": {"$ifNull": [f"$label.{language}", "$name"]},
                "parents": {"$ifNull": ["$parents", []]},
            }
        },
    ]

    nodes = mongo_db().referential_nodes.aggregate(query)
    return nodes


def get_referential_node_id_for_node_value(referential_id: UUID | str, level_id: UUID | str, value: str) -> UUID:
    # Given a value, referential_id and level_id, return the node's ID
    if not isinstance(referential_id, UUID):
        referential_id = UUID(referential_id)
    if not isinstance(level_id, UUID):
        level_id = UUID(level_id)
    node = mongo_db().referential_nodes.find_one(
        {"referential_id": referential_id, "level_id": level_id, "name": value}
    )
    if node is None:
        raise MetadataOptionValueNotFound(
            _("Value <%(value)s> not found in referential") % {"value": value},
        )
    return node["id"]


def build_level_list(referential: Referential) -> NodeListType:
    """Extract list of levels for storing in Referential.lists"""
    levels = referential.levels
    level_list = [{"id": _.id, "name": _.name} for _ in levels]

    return level_list


def get_level_list(referential: ReferentialModel, params: ParamType) -> list[dict[str, Any]]:
    language: str = params.get("language", get_default_language(referential=referential))

    query = [
        {"$match": {"referential_id": referential.id}},
        {"$sort": {"order": 1}},
        {
            "$project": {
                "_id": 0,
                "id": 1,
                "label": {"$ifNull": [f"$label.{language}", "$name"]},
            }
        },
    ]
    levels = mongo_db().referential_levels.aggregate(query)
    return levels


@cached(timeout=CACHE_TTL, key_prefix=cache_key_fn, key_object="referential")
def get_languages_list(*, referential: ReferentialModel, code_only: bool = False) -> list[dict[str, str] | str]:
    languages: list[str] = ReferentialData.objects.values_list("languages__available", flat=True).filter(
        referential=referential
    )[0]

    if code_only:
        languages_list = [_l for _l in languages]
    else:
        languages_list: list[{str, str}] = [{"code": _l, "name": language_autonym(_l)} for _l in languages]

    return languages_list


def get_node_translations(referential_id: str, node_id: str) -> dict[str, str]:
    translations_field = f"languages__translations__{node_id}"
    name_field = f"definition__nodes__{node_id}__name"

    referential_translations = ReferentialData.objects.values(translations_field, name_field).get(
        referential_id=referential_id
    )

    languages = ReferentialData.objects.values("languages__available").get(referential_id=referential_id)

    translations = {}
    for lang in [_lang for _lang in languages.get("languages__available", [])]:
        try:
            translation = referential_translations[translations_field][lang]
        except TypeError:
            translation = ""
        translations[lang] = translation or referential_translations[name_field]
    return translations


def generate_node_lists(referential: ReferentialModel) -> None:
    mongo: Database = mongo_db()
    levels_collection: Collection = mongo.referential_levels
    nodes_collection: Collection = mongo.referential_nodes
    translations = Translations(DEFAULT_LANGUAGE, referential.languages.get("translations", {}))
    parsed_referential: Referential = Referential.from_definition(
        referential.definition,
        language=DEFAULT_LANGUAGE,
        translations=translations,
    )
    languages = get_languages_list(referential=referential, code_only=True)
    levels = build_level_list(parsed_referential)

    # First delete all levels and nodes before inserting
    delete_referential_in_mongodb(referential)

    for level_order, level in enumerate(levels):
        data = {
            "id": uuid.UUID(level["id"]),
            "name": level["name"],
            "order": level_order,
            "referential_id": referential.id,
        }
        _add_label(data, languages, level, translations)
        levels_collection.replace_one({"id": uuid.UUID(level["id"])}, data, upsert=True)

        nodes = build_node_list(parsed_referential, {"level_id": level["id"], "parent": None})
        for node_order, node in enumerate(nodes):
            data = {
                "id": uuid.UUID(node["id"]),
                "name": node["name"],
                "level_id": uuid.UUID(level["id"]),
                "referential_id": referential.id,
                "order": node_order,
            }

            if node["parents"]:
                data["parents"] = [uuid.UUID(_node) for _node in node["parents"]]
            if node["ancestors"]:
                data["ancestors"] = [uuid.UUID(_node) for _node in node["ancestors"]]

            _add_label(data, languages, node, translations)

            nodes_collection.replace_one({"id": uuid.UUID(node["id"])}, data, upsert=True)


def _add_label(data: dict[str, Any], languages: list[str], level: list[dict[str, str]], translations: Translations):
    label = {}
    for lang in languages:
        label[lang] = translations.t(level["id"], lang) or level["name"]
    if label:
        data["label"] = label


def delete_referential_in_mongodb(referential: ReferentialModel) -> None:
    mongo: Database = mongo_db()
    mongo.referential_levels.delete_many({"referential_id": referential.id})
    mongo.referential_nodes.delete_many({"referential_id": referential.id})


def clear_cache_for_referential(referential: ReferentialModel) -> None:
    if hasattr(cache, "delete_pattern"):
        cache.delete_pattern(f"referential:{referential.id}:*")


def update_items_using_referential(referential_id: UUID) -> None:
    """Update metadata of all items using a referential"""

    templates = Template.objects.filter(used_referentials=referential_id)
    for template in templates:
        update_items_using_template(template)


class Translations:
    def __init__(self, language: str, translations: dict[str, dict[str, str]]):
        self.language = language
        self.translations = translations

    def t(self, node_id: str, language: str = None):
        language = language or self.language
        try:
            return self.translations.get(node_id, None).get(language, None)
        except AttributeError:
            return None


class Node:
    id: str
    name: str
    order: int
    parents: list[str]
    children: list[str]
    ancestors: list[str]

    def __init__(
        self,
        id: str,
        name: str,
        order: int,
        children: list[str] = None,
        parents: list[str] = None,
        ancestors: list[str] = None,
    ):
        self.id = id
        self.name = name
        self.order = order
        self.children = children
        self.parents = parents
        self.ancestors = ancestors or []

    @property
    def serialized(self) -> dict[str, Any]:
        serialized_node = {
            "id": self.id,
            "name": self.name,
            "level": self.order,
        }
        if self.children:
            serialized_node["children"] = self.children
        if self.parents:
            serialized_node["parents"] = self.parents
        return serialized_node

    def build(self, referential: Referential, nodes: dict[str, dict]) -> None:
        if self.children:
            for child in self.children:
                _node_definition = nodes[child]
                node = Node(
                    id=_node_definition["id"],
                    name=referential.trans(_node_definition["id"], _node_definition["name"]),
                    order=_node_definition["level"],
                    children=_node_definition.get("children", []),
                    parents=_node_definition.get("parents", []),
                    ancestors=self.ancestors.copy() + [self.id],
                )
                node.build(referential=referential, nodes=nodes)
                referential.levels[node.order].nodes.append(node)


class Level:
    id: str
    name: str
    order: int
    nodes: list[Node]

    def __init__(self, id: str, name: str, order: int = 0):
        self.name = name
        self.id = id
        self.order = order
        self.nodes = []

    @property
    def serialized(self) -> dict[str, str]:
        return {
            "id": self.id,
            "name": self.name,
        }

    def build(self, roots: list, nodes: dict[str, dict], referential: Referential, order: int = 0) -> None:
        for root in roots:
            try:
                _node_definition = nodes[root]
            except KeyError:
                continue
            node = Node(
                id=_node_definition["id"],
                name=referential.trans(_node_definition["id"], _node_definition["name"]),
                order=_node_definition["level"],
                children=_node_definition.get("children", []),
                parents=_node_definition.get("parents", []),
            )
            node.build(referential=referential, nodes=nodes)
            self.nodes.append(node)

    def get_nodes(self, parent: str = None) -> list[Node]:
        if parent:
            return [node for node in self.nodes if parent in node.parents]
        else:
            return self.nodes


class Referential:
    def __init__(self):
        self.levels: list[Level] = []
        self.roots: list[str] = []
        self.translations: Translations | None = None

    @classmethod
    def from_definition(
        cls, definition: dict, language: str = DEFAULT_LANGUAGE, translations: Translations = None
    ) -> Referential:
        ref = cls()
        ref.translations = translations or Translations(language, {})
        ref.parse(definition)
        return ref

    @property
    def serialized(self) -> dict:
        nodes = itertools.chain.from_iterable([_level.nodes for _level in self.levels])
        return {
            "nodes": {_node.id: _node.serialized for _node in nodes},
            "roots": [_root for _root in self.roots],
            "levels": [_level.serialized for _level in self.levels],
        }

    def parse(self, definition: dict) -> None:
        self.roots = definition["roots"]
        for level_number, level_definition in enumerate(definition["levels"]):
            self.levels.append(
                Level(
                    id=level_definition["id"],
                    name=self.trans(level_definition["id"], level_definition["name"]),
                    order=level_number,
                )
            )
        try:
            self.levels[0].build(self.roots, definition["nodes"], referential=self)
        except IndexError:
            pass

    def nodes_for_level(self, level_id: str, language: str = None, parent: str = None) -> list[Node]:
        level = self.get_level(level_id)
        return level.get_nodes(parent=parent)

    def get_level(self, level_id):
        try:
            level = [_l for _l in self.levels if _l.id == level_id][0]
        except IndexError:
            raise LevelNotFoundError(f"Level '{level_id}' not found.")
        return level

    def trans(self, node_id: str, default: str) -> str:
        if not self.translations:
            return default
        try:
            return self.translations.t(node_id) or default
        except KeyError:
            return default
