from pount.apps.api.libs.metadata import TemplateFields
from pount.apps.api.models import Set, Template

DEFAULT_SEARCH_CONFIG = {
    "mapping_type": None,
    "use_in_search": False,
    "use_in_aggregation": False,
}

DATACITE_SEARCH_CONFIG = {
    "creator": {
        "mapping_type": "strings",
        "use_in_search": True,
        "use_in_aggregation": False,
        "locked": True,
    },
    "title": {
        "mapping_type": "strings",
        "use_in_search": True,
        "use_in_aggregation": False,
        "locked": True,
    },
    "publisher": {
        "mapping_type": "strings",
        "use_in_search": True,
        "use_in_aggregation": False,
        "locked": True,
    },
    "publicationYear": {
        "mapping_type": "longs",
        "use_in_search": False,
        "use_in_aggregation": True,
        "locked": True,
    },
    "resourceType": {
        "mapping_type": "strings",
        "use_in_search": True,
        "use_in_aggregation": False,
        "locked": True,
    },
    "resourceTypeGeneral": {
        "mapping_type": "strings",
        "use_in_search": False,
        "use_in_aggregation": True,
        "locked": True,
    },
}


def get_mapped_field_names(search_config) -> list[str]:
    try:
        names = [field["name"] for field in search_config if field.get("name")]
    except TypeError:
        names = []
    return names


def get_template_field_names(template_fields: TemplateFields) -> list[str]:
    return [str(field) for field in template_fields]


def update_set_search_config_on_template_save(template: Template):
    """
    When we save a template, fields may have been added or removed, we may need to
    add or remove them to the fields available in the search config
    """
    set_model: Set = template.set
    template_fields = TemplateFields.from_template(template)
    template_field_names = get_template_field_names(template_fields)
    configured_field_names = get_mapped_field_names(set_model.search_config)

    added_fields: list[str] = list(set(template_field_names) - set(configured_field_names))
    removed_fields: list[str] = list(set(configured_field_names) - set(template_field_names))

    # Delete removed fields
    if len(removed_fields):
        set_model.search_config = [
            config for config in set_model.search_config if config["name"] not in removed_fields
        ]

    # Add fields added to the template with a default config
    for added_field in added_fields:
        template_field = template_fields.get_field(added_field)
        try:
            label = template_field.label
        except (AttributeError, KeyError):
            # If a field doesn't have a name, we won't index it
            continue
        set_model.search_config.append(
            {
                "name": added_field,
                "label": {"en": label, "fr": label},
                "config": DEFAULT_SEARCH_CONFIG,
            }
        )

    set_model.search_config = add_or_replace_datacite_fields(set_model.search_config)
    set_model.search_config = order_fields_in_template_order(set_model.search_config, template_fields)

    set_model.save()

    # If fields were added or removed, reindex the items in the set
    if len(added_fields) or len(removed_fields):
        #  @todo reindex set items ? only if removed fields ?
        # If added we probably don't want to reindex, default config has no
        # mapping type
        pass


def add_or_replace_datacite_fields(search_config: list[dict]) -> list[dict]:
    # datacite fields are replaced by default config or added if not already present
    datacite_fields = list(DATACITE_SEARCH_CONFIG.keys())

    for _field in search_config:
        if _field["name"] in DATACITE_SEARCH_CONFIG:
            _field["config"] = DATACITE_SEARCH_CONFIG[_field["name"]]
            datacite_fields.remove(_field["name"])

    if datacite_fields:
        # those are not present in the search config
        fields_to_add = [
            {"name": field, "config": DATACITE_SEARCH_CONFIG[field], "label": {"en": field, "fr": field}}
            for field in datacite_fields
        ]
        search_config.extend(fields_to_add)

    return search_config


def order_fields_in_template_order(search_config: list[dict], template_fields: TemplateFields) -> list[dict]:
    try:
        template_field_order = {field: idx for idx, field in enumerate(template_fields.fieldnames)}
        search_config = sorted(search_config, key=lambda field: template_field_order[field["name"]])
    except KeyError:
        pass
    return search_config
