from __future__ import annotations

import csv
import uuid
from uuid import UUID

from django.contrib.auth import get_user_model
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _

from pount.apps.api.libs.metadata import TemplateFields
from pount.apps.api.models import Item
from pount.apps.api.models.template import Template
from pount.exceptions import TemplateFieldError

User = get_user_model()


def clone_template_for_set(from_template_id: uuid, set_title: str, project_id: uuid, creator: User) -> Template:
    from_template: Template = Template.objects.get(pk=from_template_id)
    if from_template.is_local and not from_template.project_id == project_id:
        raise ValueError(_("Invalid template"))

    cloned_template_id = uuid.uuid4()
    cloned_template_name = f"Template {cloned_template_id} of {set_title}"
    return Template.objects.create(
        id=cloned_template_id,
        name=cloned_template_name,
        project_id=project_id,
        scope=Template.SCOPE_SET,
        creator=creator,
        fields=from_template.fields,
        messages=from_template.messages,
    )


def get_template_fieldnames(template: Template) -> dict:
    """
    Return dict of template fields
    @param template: Template
    @return: dict of fields with id as key and name as value
    """
    fieldnames = {}
    template_fields = TemplateFields.from_template(template)
    for f in template_fields.fieldnames:
        opts = []
        field = getattr(template_fields, f)
        if getattr(field, "repeatable", False):
            opts.append("*")
        if getattr(field, "multiple", False):
            opts.append("x")
        if getattr(field, "required", False):
            opts.append("o")
        fieldnames[field.id] = f"{field.name} {', '.join(opts)}".strip()

    return fieldnames


def export_template_as_csv(output: HttpResponse, template: Template):
    """
    Get CSV of template fields
    @param output: Response object
    @param template: the template object
    @return: CSV, first row with ids, second row with names
    """
    fields = get_template_fieldnames(template)
    field_ids = list(fields.keys())
    writer = csv.DictWriter(output, dialect="excel", fieldnames=field_ids)
    writer.writeheader()
    writer.writerow(fields)
    return output


def get_referentials_used_in_template(fields: list[dict]) -> list[uuid.UUID]:
    try:
        return [uuid.UUID(_f["referential_id"]) for _f in fields if _f.get("component") == "ReferentialComponent"]
    except (TypeError, ValueError):
        # Could not convert to UUID
        raise TemplateFieldError(_("Referential ID could not be converted to UUID"))


def update_items_using_template(template: UUID | Template) -> None:
    from pount.apps.api.libs.item import process_metadata

    if isinstance(template, (UUID, str)):
        try:
            template = Template.objects.get(pk=template)
        except Template.DoesNotExist:
            return
    for item in Item.objects.filter(set__template=template):
        updated_metadata = process_metadata(item.metadata, template)
        item.metadata = updated_metadata
        item.save(update_fields=["metadata"])
