import datetime as dt
import logging

from django.conf import settings
from django.core.management import BaseCommand
from django.utils import timezone

from pount.apps.api.models.item import Item
from pount.apps.api.models.notification import Notification
from pount.apps.api.models.set import Set

logger = logging.getLogger("cron")


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--days",
            type=int,
            help="Delete objects soft deleted more than <days> ago.",
        )

    def handle(self, *args, **options):
        """
        Delete Item and Set objects that have been soft-deleted more than
        x days ago. Delete also also associated MediaFile objects as well as
        thumbnail files

        Delay can be set in settings file or passed as argument
        If no delay is provided, 30 days is used by default
        """

        delay = options.get("days", None)
        if delay is None:
            delay = getattr(settings, "HARD_DELETE_AFTER_DAYS", 30)

        delay_ago = timezone.now() - dt.timedelta(days=delay)

        items = Item.deleted_objects.filter(deleted_at__lt=delay_ago)
        for item in items:
            for file in item.media_files.all():
                logger.info(f"MediaFile <{file.filename}> deleted")
                file.delete()
            item.delete_thumbnails()
            logger.info(f"Item <{item.id} - {item.title}> deleted")
            item.delete(hard_delete=True)

        sets = Set.deleted_objects.filter(deleted_at__lt=delay_ago)
        for set in sets:
            template = set.template
            set.delete_thumbnails()
            logger.info(f"Set <{set.id} - {set.title}> deleted")
            set.delete(hard_delete=True)
            logger.info(f"Template <{template.id} associated with set <{set.id}> deleted")  # noqa: E501

        # Delete read notifications generated before delay_ago
        num_deleted, objects_deleted = Notification.objects.read().filter(timestamp__lt=delay_ago).delete()
        logger.info(f"{num_deleted} read notifications deleted")
