import boto3
from botocore.exceptions import ClientError
from django.conf import settings
from django.core.management import BaseCommand, CommandError

from pount.apps.api.libs.file import sizeof_format
from pount.apps.api.storage_backends import CORS_CONFIG


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = boto3.client(
            "s3",
            aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
            endpoint_url=settings.AWS_S3_ENDPOINT_URL,
        )

    def add_arguments(self, parser):
        parser.add_argument("action", help="Action to perform: list, create, delete")
        parser.add_argument("-b", "--bucket", type=str, help="Name of the bucket")
        parser.add_argument("-f", "--file", type=str, help="Name of file")

    def handle(self, *args, **options):
        if options["action"] == "list":
            bucket_name = options.get("bucket", None)
            if bucket_name is not None:
                """List objects in bucket"""
                try:
                    self.list_bucket_objects(bucket_name)
                except KeyError:
                    self.stdout.write(self.style.WARNING(f"Bucket {bucket_name} is empty"))
            else:
                """List the buckets"""
                self.list_buckets()
        elif options["action"] == "create":
            """Create a new bucket"""
            if options["bucket"] is None:
                raise CommandError("Bucket name not given")
            bucket_name = options["bucket"]
            self._create_bucket(bucket_name)
            self.stdout.write(f"Bucket {bucket_name} created")
            self.stdout.write("Bucket CORS : " + str(self._get_bucket_cors_config(bucket_name)))
        elif options["action"] == "delete":
            bucket_name = options.get("bucket", None)
            file_name = options.get("file", None)
            if bucket_name is None:
                raise CommandError("Bucket name not given")
            if file_name is None:
                # Delete bucket after user confirmation
                if input(f"Are you sure you want to delete bucket {bucket_name}?\n").lower() in ["y", "yes"]:
                    try:
                        self.client.delete_bucket(Bucket=bucket_name)
                        self.stdout.write(self.style.WARNING(f"Bucket {bucket_name} deleted"))
                    except ClientError as e:
                        if e.response["Error"]["Code"] == "BucketNotEmpty":
                            self.delete_files_and_bucket(bucket_name)
                        else:
                            raise CommandError(e)
                else:
                    # Operation cancelled
                    self.stdout.write(self.style.WARNING("Operation cancelled"))
            else:
                # Delete file afer user confirmation
                self._delete_object(bucket_name, file_name, confirm=True)
        elif options["action"] == "set_cors":
            bucket_name = options.get("bucket", None)
            if bucket_name is None:
                raise CommandError("Bucket name not given")
            self._set_bucket_cors_config(bucket_name)
            self.stdout.write(str(self._get_bucket_cors_config(bucket_name)))
        elif options["action"] == "get_cors":
            bucket_name = options.get("bucket", None)
            if bucket_name is None:
                raise CommandError("Bucket name not given")
            self.stdout.write(str(self._get_bucket_cors_config(bucket_name)))
        else:
            self.stderr.write(f"Unknown action {options['action']}")

    def _create_bucket(self, bucket_name: str):
        response = self.client.create_bucket(Bucket=bucket_name)
        self._set_bucket_cors_config(bucket_name)
        return response

    def _set_bucket_cors_config(self, bucket_name: str):
        self.client.put_bucket_cors(
            Bucket=bucket_name,
            CORSConfiguration=CORS_CONFIG,
        )

    def _get_bucket_cors_config(self, bucket_name: str):
        response = self.client.get_bucket_cors(Bucket=bucket_name)
        return response["CORSRules"]

    def list_buckets(self):
        response = self.client.list_buckets()
        for bucket in response["Buckets"]:
            self.stdout.write(f" * {bucket['Name']} (Created {bucket['CreationDate']})")

    def list_bucket_objects(self, bucket_name):
        response = self.client.list_objects(Bucket=bucket_name)
        for obj in response["Contents"]:
            self.stdout.write(self.style.SUCCESS(f"* {obj['Key']}"))
            self.stdout.write(f"   Mod: {obj['LastModified']} - Size: {sizeof_format(int(obj['Size']))}")  # noqa: E501

    def delete_files_and_bucket(self, bucket_name):
        if input("Bucket is not empty, delete all files and bucket?\n").lower() in ["y", "yes"]:
            response = self.client.list_objects(Bucket=bucket_name)
            for obj in response["Contents"]:
                self._delete_object(
                    bucket_name=bucket_name,
                    file_name=obj["Key"],
                    confirm=False,
                )
            self.client.delete_bucket(Bucket=bucket_name)
            self.stdout.write(self.style.WARNING(f"Bucket {bucket_name} deleted"))
        else:
            raise CommandError("Operation cancelled")

    def _delete_object(self, bucket_name, file_name, confirm=True):
        if confirm and input(f"Are you sure you want to delete file {file_name}?\n").lower() not in ["y", "yes"]:
            raise CommandError("Operation cancelled")
        self.client.delete_object(Bucket=bucket_name, Key=file_name)
        self.stdout.write(self.style.WARNING(f"{file_name} deleted"))
