import uuid
from functools import partial

from django.db import models


class NullableEmailField(models.EmailField):
    """
    Subclass of the EmailField that allows empty strings to be stored as NULL

    https://stackoverflow.com/a/1934764
    """

    description = "EmailField that stores NULL but returns ''"

    def from_db_value(self, value, expression, connection):
        """
        Gets value right out of the db and changes it if it's ``None``
        """
        if value is None:
            return ""
        else:
            return value

    def to_python(self, value):
        """
        Gets value right out of the db or an instance,
        and changes it if its ``None``.
        """
        if isinstance(value, models.EmailField):
            return value
        if value is None:
            return ""
        return ""

    def get_prep_value(self, value):
        """
        Catches value right before sending it to db
        """
        if value == "":
            return None
        else:
            return value


UUIDAutoField = partial(models.UUIDField, primary_key=True, default=uuid.uuid4, editable=False)
CreatedAtField = partial(models.DateTimeField, auto_now_add=True, editable=False)
UpdatedAtField = partial(models.DateTimeField, auto_now=True)
