import datetime as dt
from typing import Optional

from django.conf import settings as project_settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.db.models import JSONField, UniqueConstraint

from pount.apps.api.libs.citation import make_item_citation
from pount.apps.api.libs.file import item_mediafile_fullname
from pount.apps.api.models import Group, mixins
from pount.apps.api.models.fields import CreatedAtField, UpdatedAtField, UUIDAutoField
from pount.apps.api.models.set import Set
from pount.apps.api.storage_backends import ItemMediaFileStorage

User = get_user_model()


class ItemQuerySet(mixins.SoftDeleteQuerySet):
    def for_set(self, set: Set):
        return self.filter(set=set)

    def public(self):
        return self.filter(is_public=True, set__is_public=True, set__project__is_public=True)

    def is_public_or_user_is_project_member(self, user: User):
        qs = self.filter(models.Q(is_public=True) | models.Q(set__project__groups__in=user.project_groups.all()))
        return qs.distinct()

    def is_user_creator_or_project_manager(self, user: User):
        qs = self.filter(
            models.Q(creator=user) | models.Q(set__project__groups__in=user.project_groups.filter(name=Group.MANAGERS))
        )
        return qs.distinct()


class Item(mixins.ThumbnailUrlMixin, mixins.SoftDeletableModel):
    id = UUIDAutoField()
    title = models.CharField(max_length=255, null=False)
    description = models.TextField(null=True, blank=True)
    is_public = models.BooleanField(default=False)
    publication_date = models.DateTimeField(null=True, blank=True)
    online_at = models.DateTimeField(null=True, blank=True)
    created_at = CreatedAtField()
    updated_at = UpdatedAtField()
    metadata = JSONField()
    set = models.ForeignKey(Set, related_name="items", on_delete=models.CASCADE)
    creator = models.ForeignKey(
        project_settings.AUTH_USER_MODEL,
        related_name="items",
        null=False,
        on_delete=models.CASCADE,
    )
    settings = JSONField(null=True)
    perms = GenericRelation("Permission", related_query_name="item")

    objects = mixins.SoftDeleteManager.from_queryset(ItemQuerySet)()
    deleted_objects = mixins.DeletedObjectsManager.from_queryset(ItemQuerySet)()
    all_objects = mixins.AllObjectsManager.from_queryset(ItemQuerySet)()

    class Meta:
        ordering = ["title"]

    def __str__(self):
        return self.title

    def is_user_editor(self, user: User):
        return bool(self.creator == user or self.set.project.is_user_manager(user))

    def save(self, *args, **kwargs):
        if self.settings is None:
            self.settings = {}
        self.settings["citation"] = make_item_citation(self.metadata)
        # Sync item.title with title in metadata
        self.title = next((_m for _m in self.metadata if _m["name"] == "title"), {}).get("value", "")
        return super().save(*args, **kwargs)

    @property
    def citation(self) -> str:
        try:
            return self.settings["citation"]
        except (TypeError, KeyError):
            # save() computes and stores the citation in self.settings
            self.save()
            return self.settings["citation"]

    @property
    def can_be_shared(self) -> bool:
        return self.is_public and self.set.is_public and self.set.project.is_public


class MediaFile(models.Model):
    class Component(models.TextChoices):
        AUDIO = "Audio"
        DOWNLOAD = "Download"
        IMAGE = "Image"
        MODEL = "Model"
        PDF = "Pdf"
        RTI = "Rti"
        VIDEO = "Video"

    id = UUIDAutoField()
    filename = models.CharField(max_length=255, null=False)
    item = models.ForeignKey(Item, on_delete=models.CASCADE, related_name="media_files")
    info = JSONField(null=True)
    is_hidden = models.BooleanField(default=False)
    is_visualized = models.BooleanField(default=False)
    is_downloadable = models.BooleanField(default=False)
    component = models.CharField(max_length=25, choices=Component.choices, null=True, blank=True, default=None)
    created_at = CreatedAtField()
    updated_at = UpdatedAtField()

    class Meta:
        constraints = [UniqueConstraint(fields=["filename", "item"], name="unique_filename_per_item")]

    @property
    def content_type(self) -> Optional[str]:
        try:
            return self.info["content_type"]
        except (TypeError, KeyError):
            return None

    @property
    def size(self):
        try:
            return self.info["size"]
        except (TypeError, KeyError):
            return None

    def delete(self, *args, **kwargs):
        self.delete_file()
        super().delete(*args, **kwargs)

    def delete_file(self):
        ItemMediaFileStorage().delete(item_mediafile_fullname(self.filename, self.item, ignore_location=True))

    @property
    def url(self):
        return self._get_url()

    @property
    def public_url(self):
        url = self.info.get("public_url", None)
        now = dt.datetime.now()
        if isinstance(url, dict) and url.get("expires", 0) > now.timestamp():
            # stored public is not expired, use it
            return url["url"]
        # generate presigned url valid for 1 year and store it
        presigned_url = self._get_url(expire=ItemMediaFileStorage.PUBLIC_EXPIRATION_DELAY)
        self.info["public_url"] = {
            "url": presigned_url,
            "expires": now.timestamp() + ItemMediaFileStorage.PUBLIC_EXPIRATION_DELAY,
        }
        self.save(update_fields=["info"])
        return presigned_url

    def _get_url(self, expire=None):
        return ItemMediaFileStorage().url(
            item_mediafile_fullname(
                self.filename,
                self.item,
                ignore_location=True,
            ),
            expire=expire,
        )
