import datetime as dt

from django.db import models
from django.utils import timezone

from pount.apps.api.storage_backends import ThumbnailFileStorage


class ThumbnailUrlMixin:
    SIZES = ["s", "m", "l", "embed"]

    def thumbnail_url(self, size, expiration=None):
        if size not in self.SIZES:
            raise ValueError("Not a valid size")
        thumbnail_field = self._get_thumbnail_field()
        if thumbnail_field is None:
            return None
        thumbnail = thumbnail_field.get(size, None)

        if expiration is not None:
            # return URL with defined expiration
            try:
                return ThumbnailFileStorage().url(thumbnail["filename"], expire=expiration)
            except TypeError:
                return None

        now = dt.datetime.now()
        if isinstance(thumbnail, dict) and thumbnail.get("expires", 0) > now.timestamp():
            return thumbnail.get("url", None)
        try:
            presigned_url = ThumbnailFileStorage().url(
                thumbnail["filename"],
                expire=ThumbnailFileStorage.EXPIRATION_DELAY,
            )
        except TypeError:
            return None
        thumbnail["url"] = presigned_url
        thumbnail["expires"] = now.timestamp() + ThumbnailFileStorage.EXPIRATION_DELAY

        if hasattr(self, "settings"):
            self.settings["thumbnails"][size] = thumbnail
        else:
            self.info["thumbnails"][size] = thumbnail
        self.save()
        return presigned_url

    def delete_thumbnails(self):
        thumbnail_field = self._get_thumbnail_field()
        if thumbnail_field is None:
            return
        for size in self.SIZES:
            try:
                filename = thumbnail_field[size]["filename"]
                ThumbnailFileStorage().delete(filename)
            except (KeyError, TypeError):
                pass

    def _get_thumbnail_field(self):
        try:
            if hasattr(self, "settings"):
                thumbnail = self.settings["thumbnails"]
            else:
                thumbnail = self.info["thumbnails"]
            return thumbnail
        except (KeyError, TypeError):
            return None


class SoftDeletableModel(models.Model):
    is_deleted = models.BooleanField(default=False, db_index=True, editable=False)
    deleted_at = models.DateTimeField(default=None, blank=True, null=True, editable=False)

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False, hard_delete=False):
        if hard_delete:
            return super().delete(using=using, keep_parents=keep_parents)

        from pount.apps.api.signals import pre_soft_delete

        pre_soft_delete.send(sender=self.__class__, instance=self)

        self.is_deleted = True
        self.deleted_at = timezone.now()
        self.save(update_fields=["is_deleted", "deleted_at"])
        return 1, {self._meta.label: 1}


class SoftDeleteManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(is_deleted=False)


class DeletedObjectsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(is_deleted=True)


class AllObjectsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset()


class SoftDeleteQuerySet(models.QuerySet):
    def delete(self):
        """
        Soft-delete objects in QuerySet
        We update every object in the QuerySet in turn because we want
        to emit update signals used to update ElasticSearch indexes
        """
        for obj in self:
            obj.delete()
        return self.count(), {self.model._meta.label: self.count()}
