import typing

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.db.models import JSONField
from django.utils.translation import gettext_lazy as _

from pount.apps.api.models.fields import CreatedAtField, UpdatedAtField, UUIDAutoField
from pount.apps.api.models.mixins import ThumbnailUrlMixin

if typing.TYPE_CHECKING:
    from pount.apps.api.models.referential import Referential

User = get_user_model()


class ProjectQuerySet(models.QuerySet):
    def public(self):
        return self.filter(is_public=True)

    def is_public_or_user_is_project_member(self, user: User):
        return self.filter(models.Q(groups__in=user.project_groups.all()) | models.Q(is_public=True)).distinct()

    def for_member(self, user: User):
        return self.filter(models.Q(groups__in=user.project_groups.all()))

    def for_manager(self, user: User):
        from pount.apps.api.models.group import Group

        return self.filter(groups__in=user.project_groups.filter(name=Group.MANAGERS))

    def annotate_with_is_manager(self, user: User):
        from pount.apps.api.models.group import Group

        manager_subquery = Group.objects.filter(name=Group.MANAGERS, users=user, project=models.OuterRef("pk"))
        return self.annotate(is_manager=models.Exists(manager_subquery))

    def annotate_with_set_count(self):
        return self.annotate(num_sets=models.Count("sets", filter=models.Q(sets__is_deleted=False)))


class Project(ThumbnailUrlMixin, models.Model):
    id = UUIDAutoField()
    name = models.CharField(max_length=255)
    description = models.TextField()
    is_public = models.BooleanField(default=False)
    created_at = CreatedAtField()
    updated_at = UpdatedAtField()
    info = JSONField(default=dict)
    perms = GenericRelation("Permission", related_query_name="project")
    external_referentials = models.ManyToManyField("Referential", related_name="projects")

    objects = ProjectQuerySet.as_manager()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ["name"]

    def is_user_manager(self, user: User) -> bool:
        from pount.apps.api.models.group import Group

        return self.groups.filter(
            name=Group.MANAGERS,
            users=user,
        ).exists()

    @property
    def managers(self):
        from pount.apps.api.models.group import Group

        return self.groups.get(name=Group.MANAGERS).users.all()

    @property
    def members(self):
        from pount.apps.api.models.group import Group

        return User.objects.filter(project_groups__in=self.groups.filter(name=Group.MEMBERS))

    def add_member(self, user: User):
        from pount.apps.api.models.group import Group

        return self.groups.get(name=Group.MEMBERS).users.add(user)

    def add_manager(self, user: User):
        from pount.apps.api.models.group import Group

        if not self.groups.filter(name=Group.MEMBERS, users=user).exists():
            raise ValueError(_("User is not a project member."))
        return self.groups.get(name=Group.MANAGERS).users.add(user)

    def public_sets(self):
        if self.is_public is False:
            return None
        return self.sets.filter(is_public=True)

    def add_external_referential(self, referential: "Referential"):
        from pount.apps.api.models.referential import Referential

        if referential.scope == Referential.Scope.SCOPE_PROJECT:
            raise ValueError(_("Can not use a private referential"))
        if referential.scope == Referential.Scope.SCOPE_PUBLIC and (referential.project_id == self.id):
            # It's not an external referential!
            pass
        self.external_referentials.add(referential)
