from typing import Union
from uuid import UUID

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import JSONField, Model, Q, UniqueConstraint
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from rules import is_authenticated
from rules.contrib.models import RulesModel

from pount.apps.api.models import Project
from pount.apps.api.models.fields import CreatedAtField, UpdatedAtField, UUIDAutoField
from pount.apps.api.rules import predicates
from pount.apps.api.tasks import (
    handle_build_node_lists_on_referential_save,
    handle_update_item_metadata_on_referential_save,
)

User = get_user_model()

DEFAULT_LANGUAGE = "fr"


class ReferentialQuerySet(models.QuerySet):
    def create_with_node_list(self, **kwargs):
        """For tests, node lists are built synchronously"""

        languages = kwargs.pop("languages", {})
        definition = kwargs.pop("definition", {})
        lists = kwargs.pop("lists", {})
        referential = Referential(**kwargs)
        referential.languages = languages
        referential.definition = definition
        referential.lists = lists
        referential.save(force_insert=True, using=self.db, build_node_lists=False)
        return referential

    def public(self):
        return self.filter(
            current_version=True,
            scope__in=[
                Referential.Scope.SCOPE_BASE,
                Referential.Scope.SCOPE_PUBLIC,
            ],
        )

    def for_project(self, project: Project, all_versions: bool = False):
        qs = self.filter(
            scope=Referential.Scope.SCOPE_PROJECT,
            project=project,
        )
        if all_versions is False:
            qs = self.current_version()
        return qs

    def current_version(self):
        return self.filter(current_version=True)

    def public_or_for_project(self, project: Union[Project, UUID, str], user: User):
        # return base refs, public refs and private refs from project (if user is a member of)
        return self.filter(
            Q(scope__in=[Referential.Scope.SCOPE_BASE, Referential.Scope.SCOPE_PUBLIC])
            | Q(
                scope=Referential.Scope.SCOPE_PROJECT,
                project=project,
                project__in=Project.objects.for_member(user),
            )
        )


class Referential(RulesModel):
    class Scope(models.IntegerChoices):
        SCOPE_BASE = 1, _("Base scope")
        SCOPE_PUBLIC = 2, _("Public scope")
        SCOPE_PROJECT = 4, _("Project scope")

    id = UUIDAutoField()
    name = models.CharField(_("Name"), max_length=255, db_index=True)
    description = models.TextField(_("Description"), null=True, blank=True)
    scope = models.IntegerField(
        choices=Scope.choices,
        null=False,
        blank=False,
        db_index=True,
        help_text="1: Global, 2: Public, 4: Project",
    )

    version = models.CharField(_("Version"), max_length=20, db_index=True, null=True, blank=True)
    version_label = models.CharField(_("Version label"), max_length=100, blank=True, null=True)
    root_version = models.ForeignKey("self", on_delete=models.CASCADE, null=True, blank=True)
    current_version = models.BooleanField(
        default=False,
    )
    source = models.CharField(_("Source"), max_length=255)

    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="referentials",
        null=True,
        on_delete=models.SET_NULL,
    )
    project = models.ForeignKey(
        Project,
        related_name="referentials",
        null=True,  # can be null if it is a base referential scope = Scope.SCOPE_BASE
        blank=True,
        on_delete=models.RESTRICT,
    )
    created_at = CreatedAtField()
    updated_at = UpdatedAtField()

    objects = ReferentialQuerySet.as_manager()

    class Meta:
        verbose_name = _("Referential")
        verbose_name_plural = _("Referentials")
        ordering = ["name"]
        constraints = [
            # A name must be unique but can have several versions (in which case names are the same)
            UniqueConstraint(
                fields=[
                    "name",
                ],
                name="%(app_label)s_%(class)s_name_unique",
                condition=models.Q(root_version__isnull=True),
            ),
            UniqueConstraint(
                fields=[
                    "name",
                    "version",
                ],
                name="%(app_label)s_%(class)s_name_version_unique",
            ),
        ]
        rules_permissions = {
            "add": is_authenticated & predicates.is_manager,
            "view": is_authenticated & predicates.can_view_referential,
            "change": is_authenticated & predicates.can_change_referential,
            "delete": is_authenticated & predicates.can_delete_referential,
            "use": is_authenticated & predicates.is_manager,
            "view_nodes": predicates.can_view_referential_nodes,
        }

    def __str__(self) -> str:
        return self.display_name

    def set_current_version(self):
        other_versions = self.objects.filter(
            ~Q(id=self.id),
            Q(id=self.root_version) | Q(root_version=self.root_version),
        )
        other_versions.update(current_version=False, updated_at=now())
        self.current_version = True
        self.save(update_fields=["current_version", "updated_at"])

    @property
    def display_name(self) -> str:
        if self.version:
            return f"{self.name}.{self.version}"
        else:
            return self.name

    @property
    def is_public(self) -> bool:
        return self.scope in [self.Scope.SCOPE_BASE, self.Scope.SCOPE_PUBLIC]

    @property
    def definition(self) -> dict:
        return self.data.definition

    @definition.setter
    def definition(self, definition: dict) -> None:
        try:
            self.data.definition = definition
        except ReferentialData.DoesNotExist:
            self.data = ReferentialData(referential=self, definition=definition)

    @property
    def languages(self) -> dict:
        return self.data.languages

    @languages.setter
    def languages(self, languages: dict) -> None:
        try:
            self.data.languages = languages
        except ReferentialData.DoesNotExist:
            self.data = ReferentialData(referential=self, languages=languages)

    @property
    def lists(self):
        return self.data.lists

    @lists.setter
    def lists(self, lists: dict) -> None:
        try:
            self.data.lists = lists
        except ReferentialData.DoesNotExist:
            self.data = ReferentialData(referential=self, lists=lists)

    def save(self, *args, **kwargs):
        build_node_lists = kwargs.pop("build_node_lists", True)
        super().save(*args, **kwargs)
        try:
            self.data.save()
        except ReferentialData.DoesNotExist:
            self.data = ReferentialData.objects.create(referential=self)
        if build_node_lists is True:
            handle_build_node_lists_on_referential_save.delay(self.id)
            handle_update_item_metadata_on_referential_save.delay(self.id)


class ReferentialData(Model):
    id = UUIDAutoField()
    referential = models.OneToOneField(Referential, on_delete=models.CASCADE, related_name="data")
    languages = JSONField(_("Languages"), blank=True, default=dict)
    definition = JSONField(_("Definition of Referential"), blank=True, default=dict)
    lists = JSONField(_("Available lists"), blank=True, default=dict)
