from django.conf import settings as project_settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.db.models import JSONField

from pount.apps.api.models import mixins
from pount.apps.api.models.fields import CreatedAtField, UpdatedAtField, UUIDAutoField
from pount.apps.api.models.permission import Role
from pount.apps.api.models.project import Project
from pount.apps.api.models.template import Template

User = get_user_model()


class SetQuerySet(mixins.SoftDeleteQuerySet):
    def public(self):
        return self.filter(is_public=True, project__is_public=True)

    def is_public_or_user_is_project_member(self, user: User):
        qs = self.filter(models.Q(project__groups__in=user.project_groups.all()) | models.Q(is_public=True))
        return qs.distinct()

    def annotate_with_item_count(self):
        return self.annotate(
            num_items=models.Count("items", filter=models.Q(items__is_deleted=False)),
            num_public_items=models.Count(
                "items",
                filter=models.Q(items__is_deleted=False, items__is_public=True),
            ),
        )


class Set(mixins.ThumbnailUrlMixin, mixins.SoftDeletableModel):
    id = UUIDAutoField()
    title = models.CharField(max_length=255)
    description = models.TextField()
    is_public = models.BooleanField(default=False)
    publication_date = models.DateTimeField(null=True)
    online_at = models.DateTimeField(null=True)
    created_at = CreatedAtField()
    updated_at = UpdatedAtField()
    creator = models.ForeignKey(
        project_settings.AUTH_USER_MODEL,
        related_name="sets",
        null=False,
        on_delete=models.PROTECT,
    )
    publisher = models.CharField(max_length=255)
    template = models.OneToOneField(Template, related_name="set", on_delete=models.CASCADE)
    project = models.ForeignKey(
        Project,
        related_name="sets",
        null=False,
        on_delete=models.CASCADE,
    )
    settings = JSONField(null=True)
    search_config = JSONField(null=False, default=list)
    perms = GenericRelation("Permission", related_query_name="set")

    objects = mixins.SoftDeleteManager.from_queryset(SetQuerySet)()
    deleted_objects = mixins.DeletedObjectsManager.from_queryset(SetQuerySet)()
    all_objects = mixins.AllObjectsManager.from_queryset(SetQuerySet)()

    class Meta:
        ordering = ["title"]

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs) -> None:
        if not self.search_config:
            from pount.apps.api.libs.set import add_or_replace_datacite_fields

            self.search_config = add_or_replace_datacite_fields([])
        super().save(*args, **kwargs)

    def delete(self, using=None, keep_parents=False, hard_delete=False):
        if hard_delete:
            self.template.delete()
        super().delete(using=using, keep_parents=keep_parents, hard_delete=hard_delete)

    def is_user_contributor(self, user: User):
        return self.perms.filter(role=Role.objects.get_contributor(), group__users=user).exists()

    def public_items(self):
        if self.is_public is False or self.project.is_public is False:
            return None
        return self.items.filter(is_public=True)

    @property
    def items_index_name(self) -> str:
        return f"items-{self.id.hex}"
