from django.conf import settings
from django.db import models
from django.db.models import JSONField
from django.utils.translation import gettext_lazy as _

from pount.apps.api.models.fields import CreatedAtField, UpdatedAtField, UUIDAutoField
from pount.apps.api.models.project import Project
from pount.apps.api.tasks import handle_update_item_metadata_on_template_save

DELETED_SUFFIX = " [DELETED]"


class TemplateQuerySet(models.QuerySet):
    def globals(self):
        return self.filter(scope__in=(Template.SCOPE_BASE, Template.SCOPE_GLOBAL))

    def locals(self):
        return self.filter(scope__in=(Template.SCOPE_PROJECT, Template.SCOPE_SET))


class Template(models.Model):
    SCOPE_BASE = 1
    SCOPE_GLOBAL = 2
    SCOPE_PROJECT = 3
    SCOPE_SET = 4
    SCOPES = (
        (SCOPE_BASE, "base"),
        (SCOPE_GLOBAL, "global"),
        (SCOPE_PROJECT, "project"),
        (SCOPE_SET, "set"),
    )
    id = UUIDAutoField()
    name = models.CharField(max_length=255, unique=True)
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="templates",
        null=True,
        on_delete=models.SET_NULL,
    )
    fields = JSONField(_("Formbuilder data"), default=list)
    messages = JSONField(_("Translation strings"), null=True, default=dict)
    private_fields = JSONField(null=True, default=list)
    project = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name="templates",
        null=True,
        blank=True,
    )
    scope = models.SmallIntegerField(
        choices=SCOPES,
        default=SCOPE_PROJECT,
        db_index=True,
        help_text="1: Base,  2: Global, 3: Project, 4: Set",
    )
    used_referentials = models.ManyToManyField("Referential", related_name="templates")
    created_at = CreatedAtField()
    updated_at = UpdatedAtField()

    class Meta:
        ordering = ["name"]

    objects = TemplateQuerySet.as_manager()

    def __str__(self):
        return self.name

    @property
    def is_global(self):
        return self.scope in [self.SCOPE_BASE, self.SCOPE_GLOBAL]

    @property
    def is_local(self):
        return self.scope in [self.SCOPE_PROJECT, self.SCOPE_SET]

    def save(self, *args, **kwargs):
        from pount.apps.api.libs.set import update_set_search_config_on_template_save
        from pount.apps.api.libs.template import get_referentials_used_in_template

        referentials = get_referentials_used_in_template(self.fields)
        super().save(*args, **kwargs)
        # Update list of referentials used by template
        self.used_referentials.set(referentials)

        # Update set.search_config
        if self.scope == self.SCOPE_SET and hasattr(self, "set") and self.set is not None:
            update_set_search_config_on_template_save(self)

        # Update metadata for items using this template
        handle_update_item_metadata_on_template_save.delay(template_id=self.id)
