import typing
from datetime import timedelta
from enum import Enum
from typing import Optional
from uuid import UUID

from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.db.models import JSONField
from django.utils import timezone

from pount.apps.api.models.fields import NullableEmailField, UUIDAutoField
from pount.apps.api.models.mixins import ThumbnailUrlMixin
from pount.exceptions import PountError

if typing.TYPE_CHECKING:
    from pount.apps.api.models.project import Project


class AuthenticationProvider(Enum):
    CAS = "CAS"
    LOCAL = "LOCAL"


class UserManager(BaseUserManager):
    def _create_user(self, username, email, password, **extra_fields):
        if email is not None:
            email = email.strip().lower()
        username = User.normalize_username(username)
        user = User(username=username, email=email, **extra_fields)
        user.set_password(password)
        user.save()
        return user

    def create_user(self, username, email=None, password=None, **extra_fields):
        authentication_provider = extra_fields.pop("authentication_provider", AuthenticationProvider.LOCAL.value)
        if not username:
            raise ValueError("The given username must be set")
        if username and username.split("/")[0] not in [_.value for _ in AuthenticationProvider]:
            username = f"{authentication_provider}/{username}"
        if email == "":
            email = None
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(username, email, password, **extra_fields)

    def create_superuser(self, username, email=None, password=None, **extra_fields):
        authentication_provider = extra_fields.pop("authentication_provider", AuthenticationProvider.CAS.value)
        if not username:
            raise ValueError("The given username must be set")
        username = f"{authentication_provider}/{username}"
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)
        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")
        return self._create_user(username, email, password, **extra_fields)

    def get_by_natural_key(self, username, authentication_provider=None):
        if authentication_provider is None:
            authentication_provider = AuthenticationProvider.LOCAL.value
        username = f"{authentication_provider}/{username}"
        return super().get_by_natural_key(username)

    def admins(self):
        return self.filter(is_active=True, is_superuser=True)


class User(ThumbnailUrlMixin, AbstractUser):
    """
    Yeah, there's a built-in User model for authentication in Django.
    However it is "highly recommended" to use a custom user model, because
    updating this custom model is quite easy compared to updating the default
    User model which is quite a PITA.
    @see https://docs.djangoproject.com/en/3.0/topics/auth/customizing/
         #using-a-custom-user-model-when-starting-a-project
    """

    id = UUIDAutoField()
    email = NullableEmailField(blank=True, unique=True, null=True)
    info = JSONField(default=dict)
    settings = JSONField(default=dict)
    # +permissions, see .permission.Permission

    objects = UserManager()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # storing the name on init allows us to determine later if it's been changed
        self.__original_name = self.name

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.__original_name = self.name

    @property
    def name(self) -> str:
        if self.last_name or self.first_name:
            return f"{self.first_name or ''} {self.last_name or ''}".strip()
        return self.base_username

    @property
    def name_has_changed(self) -> bool:
        return self.__original_name != self.name

    @property
    def authentication_provider(self) -> Optional[str]:
        if self.username is None:
            return None
        return self.username.split("/")[0]

    @property
    def base_username(self) -> Optional[str]:
        """Get the username without the authentication provider"""
        try:
            return self.username.split("/")[1]
        except (IndexError, AttributeError):
            return self.username

    @property
    def is_project_creator(self) -> bool:
        return self.settings.get("project_creator", False)

    @property
    def is_external(self) -> bool:
        return self.authentication_provider != AuthenticationProvider.CAS.value

    def update_on_next_connection(self):
        if self.is_external:
            raise PountError("Available for internal users only")
        self.settings["refresh_info"] = True
        self.save(update_fields=["settings"])

    def is_project_manager(self, project: "Project") -> bool:
        from pount.apps.api.models import Group

        return self.project_groups.filter(name=Group.MANAGERS, project=project).exists()

    def is_project_member(self, project: Optional["Project"] = None) -> bool:
        """User is a member of a project (or of any project if project arg is None)"""
        from pount.apps.api.models import Group, Project

        if project and isinstance(project, (str, UUID, Project)):
            return self.project_groups.filter(name=Group.MEMBERS, project=project).exists()
        return self.project_groups.filter(name=Group.MEMBERS).exists()


class Invitation(models.Model):
    """
    External users must be invited to create an account.
    The email and project_id are stored in the token.
    """

    id = UUIDAutoField()
    token = models.TextField(null=False, blank=False)
    expires_at = models.DateTimeField()

    def save(self, *args, **kwargs):
        if self.expires_at is None:
            self.expires_at = timezone.now() + timedelta(days=30)
        return super().save(*args, **kwargs)

    @property
    def is_expired(self) -> bool:
        return self.expires_at < timezone.now()
