from rest_framework import permissions

from pount.apps.api import models, rules


class ObjectPermission(permissions.BasePermission):
    def _make_perm_name(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            action = "view"
        if request.method == "POST":
            action = "add"
        if request.method in ["PUT", "PATCH"]:
            action = "change"
        if request.method == "DELETE":
            action = "delete"
        app_label = view.queryset.model._meta.app_label
        if not hasattr(self, "model_name"):
            raise AssertionError(f"{self.__class__.__name__} has no model_name attribute")
        return f"{app_label}.{action}_{self.model_name}"

    def has_permission(self, request, view):
        if not bool(request.user and request.user.is_authenticated):
            return False
        if request.method == "POST":
            perm = self._make_perm_name(request, view)
            return request.user.has_perm(perm)
        return True

    def has_object_permission(self, request, view, obj):
        perm = self._make_perm_name(request, view)
        return request.user.has_perm(perm, obj)


class ProjectPermission(ObjectPermission):
    model_name = "project"

    def has_object_permission(self, request, view, obj):
        if request.method == "DELETE" and view.action == "request_deletion":
            return request.user.has_perm(rules.PROJECT_REQUEST_DELETE, obj)
        return super().has_object_permission(request, view, obj)


class SetPermission(ObjectPermission):
    model_name = "set"

    def has_permission(self, request, view):
        if request.method == "POST":
            # Special case: is checked in view
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if view.action == "thumbnail":
            return request.user.has_perm(rules.SET_EDIT, obj)
        if view.action == "import_csv":
            return request.user.has_perm(rules.SET_EDIT, obj)
        return super().has_object_permission(request, view, obj)


class ItemPermission(ObjectPermission):
    model_name = "item"

    def has_permission(self, request, view):
        if request.method == "POST":
            # Special case: is checked in view
            return True
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        if view.action == "partial_update" and request.method == "PATCH":
            # To move an Item into another Set, the user must be manager
            # of the destination Set
            destination_set_id = request.data.get("set_id", "")
            if destination_set_id and (destination_set_id != str(obj.id)):
                destination_set = models.Set.objects.get(pk=destination_set_id)
                return request.user.has_perm(rules.SET_EDIT, destination_set)
        if view.action == "thumbnail":
            return request.user.has_perm(rules.ITEM_EDIT, obj)
        return super().has_object_permission(request, view, obj)


class GroupPermissions(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        return request.user.has_perm(rules.GROUP_EDIT, obj)


class TemplatePermission(ObjectPermission):
    model_name = "template"

    def has_permission(self, request, view):
        if request.method == "POST":
            # is checked in view
            return True
        if view.action == "list":
            return request.user.has_perm(rules.TEMPLATE_LIST_GLOBALS)
        return super().has_permission(request, view)

    def has_object_permission(self, request, view, obj):
        return super().has_object_permission(request, view, obj)
