from __future__ import annotations

import typing

import rules

from pount.apps.api.models.group import Group
from pount.apps.api.models.project import Project

if typing.TYPE_CHECKING:
    from pount.apps.api.models import Item, Referential, Set, Template, User


@rules.predicate
def is_creator(user: User, obj) -> bool:
    if not hasattr(obj, "creator"):
        return False
    return obj.creator == user


@rules.predicate
def is_owner(user: User, obj) -> bool:
    if not hasattr(obj, "owner"):
        return False
    return obj.owner == user


@rules.predicate
def is_public(user: User, obj) -> bool:
    if not hasattr(obj, "is_public"):
        return False
    return obj.is_public


@rules.predicate
def is_project_creator(user: User) -> bool:
    return user.is_project_creator


@rules.predicate
def is_manager(user: User) -> bool:
    """User is a manager of any project"""
    return user.project_groups.filter(
        name=Group.MANAGERS,
    ).exists()


is_authenticated = rules.is_authenticated
is_superadmin = rules.is_superuser

is_project_public = is_public


@rules.predicate
def is_project_manager(user: User, project: Project) -> bool:
    return user.is_project_manager(project)


@rules.predicate
def is_project_member(user: User, project: Project) -> bool:
    return user.project_groups.filter(
        project=project,
    ).exists()


@rules.predicate
def is_group_manager(user: User, group: Group) -> bool:
    return Group.objects.filter(
        project=group.project,
        name=Group.MANAGERS,
        users=user,
    ).exists()


@rules.predicate
def is_set_project_manager(user: User, set: Set) -> bool:
    """Is a manager of the set's project"""
    return is_project_manager(user, set.project)


is_set_creator = is_creator


@rules.predicate
def is_set_project_member(user: User, set: Set) -> bool:
    """Is a member of the set's project"""
    return is_project_member(user, set.project)


@rules.predicate
def is_set_contributor(user: User, set: Set) -> bool:
    """Is a member of a Group associated to the set"""
    return set.is_user_contributor(user)


@rules.predicate
def is_item_public(user: User, item: Item) -> bool:
    set_public = is_public(user, item.set)
    return set_public and is_public(user, item)


@rules.predicate
def is_item_project_member(user: User, item: Item) -> bool:
    return is_set_project_member(user, item.set)


@rules.predicate
def is_item_project_manager(user: User, item: Item) -> bool:
    return is_project_manager(user, item.set.project)


is_item_creator = is_creator

is_set_public = is_public


@rules.predicate
def is_template_project_member(user: User, template: Template) -> bool:
    result = template.is_global or is_project_member(user, template.project)
    return result


@rules.predicate
def is_template_project_manager(user: User, template: Template) -> bool:
    return is_project_manager(user, template.project)


@rules.predicate
def is_template_global(user: User, template: Template) -> bool:
    return template.is_global


@rules.predicate
def is_template_local(user: User, template: Template) -> bool:
    return template.is_local


@rules.predicate
def is_template_unused(user: User, template: Template) -> bool:
    try:
        template.set
    except Exception:
        return True
    return False


@rules.predicate
def is_referential_project_member(user: User, referential: Referential) -> bool:
    if project := referential.project:
        return is_project_member(user, project)
    else:
        return is_superadmin(user)


@rules.predicate
def is_referential_manager(user: User, referential: Referential) -> bool:
    if project := referential.project:
        return is_project_manager(user, project)
    else:
        return is_superadmin(user)


can_add_referential = is_authenticated | is_manager
can_view_referential = is_public | is_referential_project_member
can_change_referential = is_referential_manager
can_delete_referential = is_referential_manager
can_view_referential_nodes = is_public | (is_authenticated & is_referential_project_member)
