from django.db import IntegrityError
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from pount.apps.api.libs.group import add_set_to_group
from pount.apps.api.libs.html import sanitize
from pount.apps.api.libs.notifications import notify_group_contribution_to_set
from pount.apps.api.models.group import Group
from pount.apps.api.models.permission import Role
from pount.apps.api.models.set import Set
from pount.apps.api.models.user import User
from pount.apps.api.serializers.set import SetTitleSerializer


class GroupMemberSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["id", "name", "username", "email"]


class GroupSerializer(serializers.ModelSerializer):
    users = GroupMemberSerializer(many=True, read_only=True)
    sets = serializers.SerializerMethodField(read_only=True)

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        super().__init__(*args, **kwargs)

    class Meta:
        model = Group
        fields = [
            "id",
            "name",
            "description",
            "is_default_group",
            "users",
            "sets",
        ]

    @extend_schema_field(SetTitleSerializer(many=True))
    def get_sets(self, instance: Group):
        sets = Set.objects.filter(perms__group=instance, perms__role=Role.objects.get_contributor())
        serializer = SetTitleSerializer(sets, many=True)
        return serializer.data

    def validate_description(self, value: str) -> str:
        cleaned_description = sanitize(value)
        return cleaned_description


class CreateGroupSerializer(GroupSerializer):
    users = GroupMemberSerializer(read_only=True, many=True)
    user_ids = serializers.ListSerializer(child=serializers.UUIDField(), allow_empty=True, write_only=True)
    sets = serializers.SerializerMethodField(read_only=True)
    set_ids = serializers.ListSerializer(child=serializers.UUIDField(), allow_empty=True, write_only=True)
    project_id = serializers.UUIDField(write_only=True)

    class Meta:
        model = Group
        fields = [
            "id",
            "name",
            "description",
            "is_default_group",
            "users",
            "user_ids",
            "sets",
            "set_ids",
            "project_id",
        ]

    def create(self, validated_data):
        try:
            group = Group.objects.create_for_project(
                name=validated_data.get("name"),
                description=validated_data.get("description"),
                project=self.project,
            )
            if validated_data.get("user_ids"):
                group.users.add(*validated_data.get("user_ids"))
            if validated_data.get("set_ids"):
                user = self.context["request"].user
                for id in validated_data.get("set_ids"):
                    set = Set.objects.get(pk=id)
                    add_set_to_group(group, set)
                    notify_group_contribution_to_set(
                        actor=user,
                        group=group,
                        set=set,
                        verb="group_contribute_to_set",
                    )
        except IntegrityError:
            raise ValidationError(_("Group name already in use for this project."))
        return group
