from typing import Optional

from botocore.exceptions import ClientError
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.reverse import reverse

from pount.apps.api import rules
from pount.apps.api.libs.file import item_mediafile_filename, item_mediafile_fullname, sizeof_format
from pount.apps.api.libs.html import sanitize
from pount.apps.api.libs.item import process_metadata
from pount.apps.api.libs.notifications import notify_item
from pount.apps.api.models.item import Item, MediaFile
from pount.apps.api.models.set import Set
from pount.apps.api.serializers.mixins import ThumbnailUrlMixin
from pount.apps.api.serializers.user import CreatorSerializer
from pount.apps.api.storage_backends import ItemMediaFileStorage
from pount.apps.api.validators import JSONSchemaValidator

User = get_user_model()


class NestedMediaFileSerializer(serializers.ModelSerializer):
    updated_at = serializers.DateTimeField(read_only=True)
    name = serializers.SerializerMethodField(read_only=True)
    url = serializers.SerializerMethodField(read_only=True)
    size = serializers.SerializerMethodField(read_only=True)
    info = serializers.JSONField(validators=[JSONSchemaValidator(schema="mediafile-info.schema.json")], required=False)

    class Meta:
        model = MediaFile
        fields = [
            "id",
            "url",
            "name",
            "info",
            "is_visualized",
            "is_downloadable",
            "component",
            "updated_at",
            "content_type",
            "size",
        ]

    def get_url(self, instance: MediaFile) -> Optional[str]:
        if instance.info.get("is_rti"):
            url = reverse(
                "item-get-rti",
                kwargs={
                    "pk": str(instance.item.id),
                    "filename": item_mediafile_filename(instance.filename),
                },
            )
            return self.context["request"].build_absolute_uri(url)
        try:
            return instance.url
        except ClientError:
            return None

    def get_name(self, instance: MediaFile) -> str:
        return item_mediafile_filename(instance.filename)

    def get_size(self, instance: MediaFile) -> str:
        return sizeof_format(instance.size)


class MediaFileSerializer(NestedMediaFileSerializer):
    filename = serializers.CharField()
    name = serializers.SerializerMethodField(read_only=True)
    info = serializers.JSONField(read_only=True)
    component = serializers.CharField(required=False, allow_null=True)

    class Meta(NestedMediaFileSerializer.Meta):
        model = MediaFile
        fields = [
            "id",
            "filename",
            "name",
            "info",
            "is_visualized",
            "is_downloadable",
            "component",
            "url",
            "content_type",
            "size",
        ]

    def get_name(self, instance: MediaFile) -> str:
        return item_mediafile_filename(instance.filename)

    def create(self, validated_data):
        try:
            item = Item.objects.get(pk=self.context["item_id"])
            file_object = ItemMediaFileStorage().get_file_object(
                item_mediafile_fullname(validated_data.get("filename"), item)
            )
            mediafile = MediaFile.objects.create(
                item_id=self.context["item_id"],
                info={
                    "content_type": file_object.content_type,
                    "size": file_object.content_length,
                },
                **validated_data,
            )
            return mediafile
        except ClientError:
            raise serializers.ValidationError("S3 ClientError")

    def validate(self, attrs):
        if self.instance:
            is_visualized = attrs.get("is_visualized") or self.instance.is_visualized
            component = attrs.get("component") or self.instance.component
        else:
            is_visualized = attrs.get("is_visualized")
            component = attrs.get("component")
        if is_visualized and not component:
            raise serializers.ValidationError({"component": _("Component must be set if is_visualized is true")})
        return attrs


class UpdateMediaFileSerializer(MediaFileSerializer):
    filename = serializers.CharField(read_only=True)
    name = serializers.SerializerMethodField(read_only=True)
    info = serializers.JSONField(read_only=True)

    class Meta(MediaFileSerializer.Meta):
        fields = [
            "id",
            "filename",
            "name",
            "info",
            "is_visualized",
            "is_downloadable",
            "component",
            "url",
            "content_type",
            "size",
        ]


class ItemSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = Set
        fields = ["id", "title", "template_id", "project_id"]


class ItemSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    creator = CreatorSerializer(read_only=True)
    set = ItemSetSerializer(read_only=True)
    set_id = serializers.UUIDField(write_only=True)
    can_edit = serializers.SerializerMethodField()
    media_files = serializers.SerializerMethodField(read_only=True)
    thumbnails = serializers.SerializerMethodField(read_only=True)
    metadata = serializers.JSONField(
        validators=[JSONSchemaValidator(schema="item-metadata.schema.json")],
        help_text=_("See `item-metadata.schema.json`"),
    )
    updated_at = serializers.DateTimeField(read_only=True)

    class Meta:
        model = Item
        fields = [
            "id",
            "title",
            "description",
            "metadata",
            "is_public",
            "can_be_shared",
            "set",
            "set_id",
            "creator",
            "media_files",
            "can_edit",
            "thumbnails",
            "citation",
            "updated_at",
        ]

    def create(self, validated_data):
        user: User = self.context["request"].user
        instance = super().create(
            {
                **validated_data,
                "creator": user,
            }
        )
        notify_item(user, instance, "item_create")
        return instance

    def update(self, instance, validated_data):
        instance = super().update(instance, validated_data)
        notify_item(self.context["request"].user, instance, "item_update")
        return instance

    def get_can_edit(self, instance: Item) -> bool:
        user = self.context["request"].user
        return user.has_perm(rules.ITEM_EDIT, instance)

    @extend_schema_field(NestedMediaFileSerializer(many=True))
    def get_media_files(self, instance: Item):
        mediafiles = MediaFile.objects.filter(item=instance, is_hidden=False)
        serializer = NestedMediaFileSerializer(instance=mediafiles, many=True, context=self.context)
        return serializer.data

    def validate_description(self, value: str) -> str:
        cleaned_description = sanitize(value)
        return cleaned_description

    def validate_metadata(self, value):
        try:
            template = self.instance.set.template
        except AttributeError:
            set_id = self.initial_data["set_id"]
            set = Set.objects.get(pk=set_id)
            template = set.template
        metadata = process_metadata(value, template)
        return metadata


class ItemListSerializer(ItemSerializer):
    class Meta(ItemSerializer.Meta):
        fields = [
            "id",
            "title",
            "description",
            "is_public",
            "updated_at",
            "can_edit",
            "thumbnails",
        ]
