from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from pount.apps.api.libs.html import sanitize
from pount.apps.api.libs.notifications import notify_project_create, notify_project_update
from pount.apps.api.models.group import Group
from pount.apps.api.models.permission import Role
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.serializers.mixins import ThumbnailUrlMixin

User = get_user_model()


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["id", "name", "email"]


class CompactSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = Set
        fields = ["id", "title"]


class ProjectSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    managers = serializers.SerializerMethodField()
    is_manager = serializers.BooleanField(read_only=True)
    thumbnails = serializers.SerializerMethodField()
    num_sets = serializers.BooleanField(read_only=True)

    def create(self, validated_data):
        instance = super().create(validated_data)
        self.add_creator_to_default_groups(instance)
        notify_project_create(self.context["request"].user, instance)
        return instance

    def update(self, instance, validated_data):
        instance = super().update(instance, validated_data)
        notify_project_update(self.context["request"].user, instance)
        return instance

    @extend_schema_field(UserSerializer(many=True))
    def get_managers(self, instance: Project):
        managers = instance.managers
        serializer = UserSerializer(managers, many=True)
        return serializer.data

    class Meta:
        model = Project
        fields = [
            "id",
            "managers",
            "is_manager",
            "thumbnails",
            "name",
            "description",
            "is_public",
            "updated_at",
            "num_sets",
        ]

    def add_creator_to_default_groups(self, instance: Project):
        user = self.context["request"].user
        instance.add_member(user)
        instance.add_manager(user)

    def validate_description(self, value: str) -> str:
        cleaned_description = sanitize(value)
        return cleaned_description


class ProjectListSerializer(ProjectSerializer):
    class Meta(ProjectSerializer.Meta):
        fields = [
            "id",
            "name",
            "description",
            "is_public",
            "updated_at",
            "is_manager",
            "thumbnails",
            "num_sets",
        ]


class ProjectNameSerializer(ProjectSerializer):
    class Meta(ProjectSerializer.Meta):
        fields = [
            "id",
            "name",
        ]


class ProjectGroupSerializer(serializers.ModelSerializer):
    users = UserSerializer(read_only=True, many=True)
    sets = serializers.SerializerMethodField()

    @extend_schema_field(CompactSetSerializer(many=True))
    def get_sets(self, instance: Group):
        permissions = instance.perms.filter(
            role__name=Role.CONTRIBUTOR,
            content_type=ContentType.objects.get_for_model(Set),
        )
        sets = [permission.content_object for permission in permissions if permission.content_object]
        serializer = CompactSetSerializer(many=True, instance=sets)
        return serializer.data

    class Meta:
        model = Group
        fields = [
            "id",
            "name",
            "description",
            "is_default_group",
            "users",
            "sets",
        ]


class ProjectGroupsSerializer(serializers.ModelSerializer):
    groups = ProjectGroupSerializer(read_only=True, many=True)

    class Meta:
        model = Project
        fields = ["id", "name", "groups"]
