from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from pount.apps.api.libs.html import sanitize
from pount.apps.api.models import Project, Referential, Template
from pount.apps.api.serializers.project import ProjectNameSerializer
from pount.apps.api.validators import JSONSchemaValidator

User = get_user_model()


class ReferentialCreatorSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["id", "name"]


class ReferentialProjectSerializer(serializers.ModelSerializer):
    class Meta:
        model = Project
        fields = ["id", "name"]


class ReferentialSerializer(serializers.ModelSerializer):
    project = ReferentialProjectSerializer()

    class Meta:
        model = Referential
        fields = [
            "id",
            "display_name",
            "description",
            "languages",
            "project",
            "updated_at",
        ]

    def validate_description(self, value: str) -> str:
        cleaned_description = sanitize(value)
        return cleaned_description


class ReferentialListSerializer(ReferentialSerializer):
    project = ReferentialProjectSerializer()

    is_manager = serializers.SerializerMethodField(
        help_text=_("The current user is a manager of the referential's project")
    )
    is_external = serializers.SerializerMethodField(
        help_text=_("The referential is defined in another project or is a base referential")
    )

    class Meta(ReferentialSerializer.Meta):
        fields = [
            "id",
            "name",
            "description",
            "scope",
            "project",
            "is_manager",
            "is_external",
            "created_at",
            "updated_at",
        ]

    def get_is_manager(self, obj: Referential) -> bool:
        user: User = self.context["request"].user
        if obj.project is None:
            return user.is_superuser or False
        return user.is_project_manager(project=obj.project)

    def get_is_external(self, obj: Referential) -> bool:
        if not obj.project:
            return True

        return bool(str(obj.project_id) != str(self.context.get("project_id", "")))


class ReferentialDetailSerializer(ReferentialSerializer):
    creator = ReferentialCreatorSerializer(read_only=True)

    class Meta(ReferentialSerializer.Meta):
        fields = [
            "id",
            "project",
            "creator",
            "name",
            "description",
            "scope",
            "source",
            "languages",
            "definition",
            "created_at",
            "updated_at",
        ]


class ReferentialEditSerializer(ReferentialSerializer):
    definition = serializers.JSONField(validators=[JSONSchemaValidator(schema="referential-definition.schema.json")])
    project_id = serializers.UUIDField(write_only=True, required=False, allow_null=True)
    project = ReferentialProjectSerializer(read_only=True)
    source = serializers.CharField(read_only=True)
    languages = serializers.JSONField(validators=[JSONSchemaValidator(schema="referential-languages.schema.json")])

    class Meta(ReferentialSerializer.Meta):
        fields = [
            "id",
            "name",
            "description",
            "definition",
            "project_id",
            "project",
            "scope",
            "source",
            "languages",
        ]

    def create(self, validated_data) -> Referential:
        user: User = self.context.get("request").user
        validated_data["creator"] = user
        validated_data["current_version"] = True
        validated_data["root_version"] = None
        if validated_data["scope"] != Referential.Scope.SCOPE_BASE:
            validated_data["source"] = user.name
        try:
            instance = super().create(validated_data)
        except IntegrityError as e:
            if "api_referential_name_unique" in str(e):
                raise serializers.ValidationError({"name": _("A referential with this name already exists")})
        return instance

    def validate(self, attrs):
        if attrs.get("scope") == Referential.Scope.SCOPE_BASE:
            if attrs.get("project_id"):
                raise serializers.ValidationError({"project_id": _("project_id is not allowed for base referential")})
        if attrs.get("scope") in [Referential.Scope.SCOPE_PROJECT, Referential.Scope.SCOPE_PUBLIC]:
            if not attrs.get("project_id"):
                raise serializers.ValidationError({"project_id": _("project_id is required")})
        return attrs


class ReferentialLevelsSerializer(serializers.Serializer):
    id = serializers.UUIDField()
    label = serializers.CharField()

    class Meta:
        fields = [
            "id",
            "label",
        ]


class ReferentialNodesSerializer(serializers.Serializer):
    id = serializers.UUIDField()
    label = serializers.CharField()
    parents = serializers.ListSerializer(child=serializers.UUIDField())

    class Meta:
        fields = [
            "id",
            "label",
            "parents",
        ]


class ReferentialLanguagesSerializer(serializers.Serializer):
    code = serializers.CharField(read_only=True, min_length=2, max_length=2, help_text="2 letter ISO code")
    name = serializers.CharField(read_only=True)

    class Meta:
        fields = [
            "code",
            "name",
        ]


class TemplateListSerializer(serializers.ModelSerializer):
    project = ProjectNameSerializer(read_only=True)

    class Meta:
        model = Template
        fields = [
            "id",
            "name",
            "scope",
            "project",
        ]
