from datetime import datetime as dt

from django.utils.translation import gettext_lazy as _
from elasticsearch_dsl.response import Hit, Response
from rest_framework import serializers


class DetailedSavedSearchSerializer(serializers.Serializer):
    id = serializers.UUIDField(read_only=True)
    name = serializers.CharField()
    saved_at = serializers.SerializerMethodField(read_only=True)
    document = serializers.CharField(read_only=True)
    query = serializers.JSONField()

    class Meta:
        fields = ["id", "name", "saved_at", "document", "query"]

    def get_saved_at(self, obj) -> str:
        ts = obj["ts"]
        savedAt = dt.fromtimestamp(ts)
        return savedAt.isoformat()


class CompactSavedSearchSerializer(serializers.Serializer):
    id = serializers.UUIDField(read_only=True)
    name = serializers.CharField()
    saved_at = serializers.SerializerMethodField(read_only=True)

    class Meta:
        fields = ["id", "name", "saved_at"]

    def get_saved_at(self, obj) -> str:
        ts = obj["ts"]
        savedAt = dt.fromtimestamp(ts)
        return savedAt.isoformat()


class HitsSerializer(serializers.Serializer):
    value = serializers.IntegerField(read_only=True, help_text=_("Number of results"))
    relation = serializers.ChoiceField(
        read_only=True,
        choices=[("eq", "eq"), ("gte", "gte")],
        help_text=_("`eq` if the number of results is accurate, `gte` if it is a lower bound"),
    )


class SearchHitListSerializer(serializers.ListSerializer):
    def to_representation(self, instance) -> list["SearchHitSerializer"]:
        """
        We return a list of hits using a serializer that is dependent on the type of document
        """
        hits = []
        for hit in instance:
            if hit.meta["index"].startswith("items-"):
                hits.append(ItemHitSerializer(hit).data)
            elif hit.meta["index"].startswith("project"):
                hits.append(ProjectHitSerializer(hit).data)
            elif hit.meta["index"].startswith("set"):
                hits.append(SetHitSerializer(hit).data)
        return hits


class SearchHitSerializer(serializers.Serializer):
    doc_type = serializers.CharField(read_only=True)
    id = serializers.SerializerMethodField(read_only=True)
    score = serializers.SerializerMethodField(read_only=True)
    title = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    is_public = serializers.BooleanField(read_only=True)
    thumbnail = serializers.CharField(read_only=True)

    class Meta:
        list_serializer_class = SearchHitListSerializer

    def get_id(self, hit: Hit) -> str:
        return hit.meta["id"]

    def get_score(self, hit: Hit) -> float:
        return hit.meta["score"]

    def get_creator(self, hit: Hit) -> dict[str, str]:
        return {
            "id": hit.creator["id"],
            "name": hit.creator["name"],
        }


class ProjectHitSerializer(SearchHitSerializer):
    title = serializers.CharField(read_only=True, source="name")
    managers = serializers.ListSerializer(read_only=True, child=serializers.CharField())
    sets_count = serializers.IntegerField(read_only=True)
    public_sets_count = serializers.IntegerField(read_only=True)
    items_count = serializers.IntegerField(read_only=True)
    public_items_count = serializers.IntegerField(read_only=True)
    thumbnail = serializers.CharField(read_only=True)


class SetHitSerializer(SearchHitSerializer):
    creator = serializers.SerializerMethodField(read_only=True)
    items_count = serializers.IntegerField(read_only=True)
    public_items_count = serializers.IntegerField(read_only=True)
    thumbnail = serializers.CharField(read_only=True)


class ItemHitSerializer(SearchHitSerializer):
    creator = serializers.SerializerMethodField(read_only=True)
    thumbnail = serializers.CharField(read_only=True)


class AggretationSerializer(serializers.Serializer):
    def to_representation(self, instance):
        return instance.to_dict()


class SearchResultSerializer(serializers.Serializer):
    query = serializers.SerializerMethodField(read_only=True, help_text=_("The search query"))
    hits = serializers.SerializerMethodField(read_only=True, help_text=_("The number of results"))
    results = serializers.SerializerMethodField(
        read_only=True, help_text=_("The results (serializer used depends on the type of the doc that is serialized)")
    )
    aggregations = serializers.SerializerMethodField(read_only=True, help_text=_("The aggregations"))

    def get_hits(self, response: Response) -> HitsSerializer:
        return HitsSerializer(instance=response.hits.total).data

    def get_query(self, response: Response) -> str:
        return self.context.get("query", "")

    def get_results(self, response: Response) -> SearchHitListSerializer(child=SearchHitSerializer()):
        return SearchHitSerializer(instance=response.hits, many=True).data

    def get_aggregations(self, response: Response):
        agg_keys = response.aggregations.to_dict().keys()
        filternames = [
            {bucket: filter.to_dict().get("terms").get("field")}
            for bucket, filter in self.context["aggregation_fields"].items()
        ]
        aggregations = {
            agg_name: AggretationSerializer(instance=response.aggregations[agg_name], many=True).data
            for agg_name in agg_keys
        }
        return {
            "filternames": filternames,
            "aggregations": aggregations,
        }
