from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from pount.apps.api.libs.html import sanitize
from pount.apps.api.libs.notifications import notify_set
from pount.apps.api.libs.template import clone_template_for_set
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.serializers.mixins import ThumbnailUrlMixin
from pount.apps.api.serializers.template import TemplateNameSerializer
from pount.apps.api.serializers.user import CreatorSerializer
from pount.apps.search.elasticsearch.index import MappingTypes

User = get_user_model()


class SetNestedProjectSerializer(serializers.ModelSerializer):
    is_manager = serializers.SerializerMethodField()

    class Meta:
        model = Project
        fields = ["id", "name", "is_public", "is_manager"]

    def get_is_manager(self, instance: Project) -> bool:
        return instance.is_user_manager(self.context["request"].user)


class SetSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    creator = CreatorSerializer(many=False, read_only=True)
    project = SetNestedProjectSerializer(many=False, read_only=True)
    project_id = serializers.UUIDField(write_only=True)
    template = TemplateNameSerializer(many=False, read_only=True)
    template_id = serializers.UUIDField(write_only=True, required=False)
    from_template_id = serializers.UUIDField(write_only=True)
    can_contribute = serializers.SerializerMethodField(read_only=True)
    thumbnails = serializers.SerializerMethodField(read_only=True)
    num_items = serializers.BooleanField(read_only=True)
    num_public_items = serializers.BooleanField(read_only=True)

    class Meta:
        model = Set
        fields = [
            "id",
            "title",
            "description",
            "is_public",
            "publication_date",
            "online_at",
            "template",
            "template_id",
            "from_template_id",
            "publisher",
            "creator",
            "can_contribute",
            "project",
            "project_id",
            "created_at",
            "updated_at",
            "thumbnails",
            "num_items",
            "num_public_items",
        ]

    def create(self, validated_data):
        template_to_clone_id = validated_data.pop("from_template_id")
        template = None
        if template_to_clone_id:
            try:
                template = clone_template_for_set(
                    template_to_clone_id,
                    validated_data.get("title"),
                    validated_data.get("project_id"),
                    self.context["request"].user,
                )
            except (Template.DoesNotExist, ValueError):
                raise serializers.ValidationError({"from_template_id": _("Invalid template.")})
        instance: Set = super().create(
            {
                **validated_data,
                "template_id": template.id,
                "creator": self.context["request"].user,
            }
        )
        notify_set(self.context["request"].user, instance, "set_create")
        return instance

    def update(self, instance: Set, validated_data):
        if "project_id" in validated_data.keys():
            # A Set's project_id can not be changed
            del validated_data["project_id"]
        instance = super().update(instance, validated_data)
        notify_set(self.context["request"].user, instance, "set_update")
        return instance

    def get_can_contribute(self, instance: Set) -> bool:
        user = self.context["request"].user
        return bool(instance.project.is_user_manager(user) or instance.is_user_contributor(user))

    def validate_description(self, value: str) -> str:
        cleaned_description = sanitize(value)
        return cleaned_description


class SetListSerializer(SetSerializer):
    template_id = serializers.UUIDField()

    class Meta(SetSerializer.Meta):
        fields = [
            "id",
            "title",
            "description",
            "is_public",
            "can_contribute",
            "updated_at",
            "thumbnails",
            "template_id",
            "num_items",
        ]


class SetTitleSerializer(SetSerializer):
    class Meta(SetSerializer.Meta):
        fields = [
            "id",
            "title",
        ]


class IndexSettingsSerializer(serializers.Serializer):
    mapping_type = serializers.ChoiceField(
        choices=MappingTypes.choices,
        default=None,
        allow_null=True,
        help_text=_("Type of ElasticSearch mapping"),
    )
    use_in_search = serializers.BooleanField(default=False)
    use_in_aggregation = serializers.BooleanField(default=False)
    locked = serializers.BooleanField(default=False, read_only=True)


class SearchConfigSerializer(serializers.Serializer):
    name = serializers.CharField(help_text=_("Name of the metadata field"))
    label = serializers.DictField(child=serializers.CharField(help_text=_("Label for the given language")))
    config = IndexSettingsSerializer(help_text=_("Configuration of the search settings"))
