from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from pount.apps.api.models import AuthenticationProvider, Project, User
from pount.apps.api.serializers.mixins import ThumbnailUrlMixin
from pount.apps.api.validators import PasswordValidator


class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = User
        fields = [
            "url",
            "id",
            "username",
            "first_name",
            "last_name",
            "email",
            "info",
        ]


class CreatorSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["id", "name", "username", "email"]


class ProfileSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    thumbnails = serializers.SerializerMethodField(read_only=True)
    is_admin = serializers.BooleanField(read_only=True, source="is_superuser")
    is_contributor = serializers.BooleanField(read_only=True, source="is_project_member")

    class Meta:
        model = User
        read_only_fields = ["username", "email"]
        fields = [
            "id",
            "username",
            "first_name",
            "last_name",
            "name",
            "email",
            "info",
            "thumbnails",
            "authentication_provider",
            "base_username",
            "is_project_creator",
            "is_contributor",
            "is_external",
            "is_admin",
        ]


class ChangeEmailSerializer(serializers.ModelSerializer):
    def save(self, **kwargs):
        user: User = self.instance
        authentication_provider = user.authentication_provider
        if authentication_provider != AuthenticationProvider.CAS.value:
            username = f"{authentication_provider}/{self.validated_data.get('email')}"
            self.instance.username = username
        self.instance.email = self.validated_data.get("email")
        self.instance.save()
        return self.instance

    class Meta:
        model = User
        read_only_fields = ["id", "username", "base_username"]
        fields = ["id", "username", "email", "base_username"]


class ChangePasswordSerializer(serializers.ModelSerializer):
    old_password = serializers.CharField(write_only=True, label=_("Old password"))
    new_password = serializers.CharField(
        write_only=True,
        label=_("New password"),
        min_length=8,
        validators=[PasswordValidator()],
    )
    new_password_2 = serializers.CharField(write_only=True, label=_("Repeat new password"), min_length=8)

    def update(self, instance, validated_data):
        if not instance.authentication_provider == AuthenticationProvider.LOCAL.value:
            msg = _("You can only change the password of a local account")
            raise serializers.ValidationError({"old_password": msg})
        instance.set_password(validated_data["new_password"])
        instance.save()
        return instance

    def validate(self, attrs):
        old_password = attrs.get("old_password", None)
        if old_password and not self.context["request"].user.check_password(old_password):
            msg = _("Invalid password")
            raise serializers.ValidationError({"old_password": msg})
        if attrs.get("new_password") != attrs.get("new_password_2"):
            msg = _("Your new passwords do not match")
            raise serializers.ValidationError({"new_password_2": msg})
        if attrs.get("new_password") == attrs.get("old_password"):
            msg = _("Your new password should be different from the old password")
            raise serializers.ValidationError({"new_password": msg})
        return attrs

    class Meta:
        model = User
        fields = ["old_password", "new_password", "new_password_2"]


class ResetPasswordSerializer(ChangePasswordSerializer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fields.pop("old_password")


class UserCreationSerializer(UserSerializer):
    username = serializers.CharField(required=False)

    def __init__(self, *args, **kwargs):
        self.invitation = kwargs.pop("invitation", None)
        super().__init__(*args, **kwargs)
        self.fields["new_password"] = serializers.CharField(
            label=_("Password"),
            write_only=True,
            min_length=8,
            validators=[PasswordValidator()],
        )
        self.fields["new_password2"] = serializers.CharField(
            label=_("Password (repeat)"),
            write_only=True,
        )

    def create(self, validated_data):
        validated_data.pop("new_password2")
        password = validated_data.pop("new_password")
        user: User = User.objects.create_user(validated_data["email"], password=password, **validated_data)
        project: Project = self.context.get("project")
        project.add_member(user)
        return user

    def validate(self, attrs):
        """Check if provided data is the same as in the invitation"""
        if attrs.get("email") != self.invitation["email"]:
            raise serializers.ValidationError("Email is not the same as in the invitation")
        if attrs.get("new_password") != attrs.get("new_password2"):
            msg = _("Your passwords do not match")
            raise serializers.ValidationError({"new_password": msg})
        return attrs
