import django.dispatch
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from pount.apps.api import models
from pount.apps.api.libs.referential import clear_cache_for_referential, delete_referential_in_mongodb
from pount.apps.api.models.template import DELETED_SUFFIX

pre_soft_delete = django.dispatch.Signal()


@receiver(post_save, sender=models.project.Project)
def create_default_group_for_project(sender, instance: models.project.Project, created, **kwargs):
    """
    Create a default Members and Managers group when a new project is created
    """
    if created:
        # Create a default Members group and assign the Member role to it
        _create_project_group_and_permission(instance, models.group.Group.MEMBERS, models.permission.Role.MEMBER)
        # Create a default "Managers" group and assign the Manager role to it
        _create_project_group_and_permission(
            instance,
            models.group.Group.MANAGERS,
            models.permission.Role.MANAGER,
        )


@receiver(pre_soft_delete, sender=models.set.Set)
def change_template_name_when_set_is_soft_deleted(sender, instance: models.set.Set, **kwargs):
    """
    When a Set is SOFT deleted change the name of the related Template to reflect that change
    """
    template = instance.template
    template.name += DELETED_SUFFIX
    template.save(update_fields=["name"])


@receiver(post_delete, sender=models.Referential)
def delete_mongo_and_cache_objects_on_referential_delete(sender, instance: models.Referential, **kwargs):
    """
    Delete documents in MongoDB referencing the referential
    """
    delete_referential_in_mongodb(referential=instance)
    clear_cache_for_referential(referential=instance)


def _create_project_group_and_permission(project, group_name, role_name):
    group = models.group.Group.objects.create_for_project(name=group_name, description=group_name, project=project)
    models.permission.Permission.objects.create(
        group=group,
        content_object=project,
        role=models.permission.Role.objects.get(name=role_name),
    )
