from storages.backends.s3 import S3Storage

from pount.exceptions import FileNameExistsError

CORS_CONFIG = {
    "CORSRules": [
        {
            "AllowedMethods": [
                "GET",
                "POST",
                "HEAD",
            ],
            "AllowedOrigins": [
                "*",
            ],
            "ExposeHeaders": ["*"],
            "AllowedHeaders": [
                "Content-Type",
                "Authorization",
                "Range",
            ],
        }
    ]
}


class MediaStorage(S3Storage):
    location = "media"
    file_overwrite = True

    def get_file_object(self, key):
        file_object = self.bucket.Object(key)
        return file_object


class ItemMediaFileStorage(MediaStorage):
    POST_EXPIRATION_DELAY = 60 * 60  # One hour
    EXPIRATION_DELAY = 60 * 60  # One hour
    PUBLIC_EXPIRATION_DELAY = 60 * 60 * 24 * 365  # One year

    location = "items"
    file_overwrite = False

    def url(self, name, **kwargs):
        expire = kwargs.pop("expire", self.EXPIRATION_DELAY)
        return super().url(name, expire=expire, parameters=kwargs)

    def content_type(self, name):
        return self.bucket.Object(name).content_type

    def post_url(self, file_name, file_type):
        if self.file_overwrite is False and self.exists(file_name):
            raise FileNameExistsError()

        # https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-HTTPPOSTConstructPolicy.html
        url = self.bucket.meta.client.generate_presigned_post(
            Bucket=self.bucket.name,
            Key=file_name,
            Fields={
                "acl": "public-read",
                "Content-Type": file_type,
            },
            Conditions=[
                {"acl": "public-read"},
                {"Content-Type": file_type},
            ],
            ExpiresIn=self.POST_EXPIRATION_DELAY,
        )
        return url

    def exists(self, name: str) -> bool:
        # we remove location: super().exists() adds it again
        if name.startswith(self.location):
            name = name.lstrip(self.location + "/")
        return super().exists(name)


class ThumbnailFileStorage(MediaStorage):
    EXPIRATION_DELAY = 60 * 60 * 24 * 30  # One month

    location = "thumbnails"

    def get_object_parameters(self, name):
        params = super().get_object_parameters(name)
        params.update({"CacheControl": "public, max-age=86400"})
        return params
