from pathlib import Path
from unittest.mock import patch

from pount.apps.api.libs.item import import_items_from_csv
from pount.apps.api.models import Item, Project, Set, User
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.tests import TestCase


class SetupTest(TestCase):
    @classmethod
    def setUpClass(cls):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_roles()
            initialize_templates()
            super().setUpClass()


class OptionalOptionTest(SetupTest):
    # A list of options may not be required

    fixtures = [
        Path(__file__).resolve(strict=True).parent / "fixtures/creator.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/project.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/template.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/set.json",
    ]

    def test_csv_import_with_no_value_for_optional_options_field(self):
        manager = User.objects.get(pk="abbe5407-91a4-4e99-aede-cdd461bae581")
        project = Project.objects.get(pk="742dd373-2c61-4809-bd44-b65e71b15956")
        project.add_member(manager)
        project.add_manager(manager)
        set_model = Set.objects.get(pk="0e02edfe-bbf8-4644-a2dd-5d59f715d065")
        with open(Path(__file__).resolve(strict=True).parent / "fixtures/import.csv", "rb") as csv_file:
            number_of_items = import_items_from_csv(csv_file, set_model, manager)

        self.assertEqual(number_of_items, 1)
        self.assertEqual(Item.objects.count(), 1)
        created_item = Item.objects.first()
        optional_field_metadata = next((_ for _ in created_item.metadata if _.get("name") == "optionsfacult"), {})
        self.assertEqual(
            optional_field_metadata["value"],
            None,
        )


class OptionalReferentialOptionTest(SetupTest):
    # A referential option is optional it may not be in the import data

    fixtures = [
        Path(__file__).resolve(strict=True).parent / "fixtures/creator.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/project.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/referential.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/template-ref.json",
        Path(__file__).resolve(strict=True).parent / "fixtures/set-ref.json",
    ]

    def setUp(self):
        super().setUp()
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()

    def tearDown(self):
        super().tearDown()
        self.patcher1.stop()

    def test_csv_import_with_no_value_for_referential_options_field(self):
        manager = User.objects.get(pk="abbe5407-91a4-4e99-aede-cdd461bae581")
        project = Project.objects.get(pk="742dd373-2c61-4809-bd44-b65e71b15956")
        project.add_member(manager)
        project.add_manager(manager)
        set_model = Set.objects.get(pk="4b1dd820-8593-43e6-a327-983ec7453a0f")
        with open(Path(__file__).resolve(strict=True).parent / "fixtures/import-ref.csv", "rb") as csv_file:
            number_of_items = import_items_from_csv(csv_file, set_model, manager)
        self.assertEqual(number_of_items, 1)
        self.assertEqual(Item.objects.count(), 1)
        created_item = Item.objects.first()
        optional_field_metadata = next((_ for _ in created_item.metadata if _.get("name") == "Level"), {})
        self.assertDictEqual(
            optional_field_metadata,
            {
                "id": "9dnalg8vabb",
                "name": "Level",
                "value": None,
                "level_id": "614f17d5-d346-4b09-ae06-9f4b71c6584f",
                "referential_id": "96c41d6e-4772-42f1-935e-527bdc954072",
            },
        )
