from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models import AuthenticationProvider

User = get_user_model()


class ChangePasswordTest(APITestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            "user@pount.org",
            email="user@pount.org",
            password="SecretPassword",
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )
        self.client.force_login(self.user)

    def make_request(self, data):
        response = self.client.patch(
            reverse("user-profile-change-password"),
            data,
            format="json",
        )
        return response

    def test_old_password_must_match(self):
        response = self.make_request(
            {
                "old_password": "WrongSecretPassword",
                "new_password": "NewSecretPassword",
                "new_password_2": "NewSecretPassword",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["old_password"][0], "Invalid password")

    def test_two_passwords_must_match(self):
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "NewSecretPassword",
                "new_password_2": "DifferentNewSecret",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.data["new_password_2"][0],
            "Your new passwords do not match",
        )

    def test_password_must_not_be_too_short(self):
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "ab",
                "new_password_2": "ab",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.data["new_password"][0],
            "Ensure this field has at least 8 characters.",
        )

    def test_new_password_must_be_different_from_old_password(self):
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "SecretPassword",
                "new_password_2": "SecretPassword",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.data["new_password"][0],
            "Your new password should be different from the old password",
        )

    def test_password_change_needs_authentication(self):
        self.client.logout()
        response = self.make_request({})
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_change_password(self):
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "NewSecretPassword",
                "new_password_2": "NewSecretPassword",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_new_password_can_not_be_entirely_numeric(self):
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "0123456789",
                "new_password_2": "0123456789",
            }
        )
        self.assertEqual(
            response.data["new_password"][0],
            "This password is entirely numeric.",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_new_password_can_not_be_too_common(self):
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "qwerty123",
                "new_password_2": "qwerty123",
            }
        )
        self.assertEqual(response.data["new_password"][0], "This password is too common.")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_only_local_users_can_change_password(self):
        self.user = User.objects.create_user(
            "cas@pount.org",
            email="cas@pount.org",
            password="SecretPasssword",
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.client.force_login(self.user)
        response = self.make_request(
            {
                "old_password": "SecretPassword",
                "new_password": "NewSecretPassword",
                "new_password_2": "NewSecretPassword",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
