import smtplib
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.core import mail
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.libs.group import add_set_to_group
from pount.apps.api.models import Notification
from pount.apps.api.models.group import Group
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.models.permission import Permission, Role
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template

User = get_user_model()


class TestSetup:
    def setUp(self) -> None:
        initialize_roles()
        self.user = User.objects.create_user("username", email="user@pount.org")
        self.manager = User.objects.create_user("manager")
        self.project = Project.objects.create(
            name="Project name",
            description="Project description",
        )
        self.project.add_member(self.user)
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.group = Group.objects.create_for_project(
            name="Other group",
            description="Group description",
            project=self.project,
        )


class CreateGroupTest(TestSetup, APITestCase):
    def test_group_name_is_required(self):
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse(
                "group-create-for-project",
            ),
            {"project_id": str(self.project.id)},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["name"][0], "This field is required.")

    def make_create_request(self, group_name, user, members=[], set_ids=[]):
        self.client.force_login(user)
        data = {
            "name": group_name,
            "description": "Group description",
            "project_id": self.project.id,
            "user_ids": members,
            "set_ids": set_ids,
        }
        response = self.client.post(
            reverse("group-create-for-project"),
            data,
            format="json",
        )
        return response

    def test_create_group_for_project_requires_manager_privileges(self):
        response = self.make_create_request("Group name", self.user)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self.make_create_request("Group name", self.manager)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["name"], "Group name")

    def test_group_name_is_unique_for_project(self):
        response = self.make_create_request("Other group", self.manager)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data[0], "Group name already in use for this project.")

    def test_cannot_use_members_as_group_name(self):
        response = self.make_create_request(Group.MEMBERS, self.manager)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data[0], "Group name already in use for this project.")

    def test_cannot_user_managers_as_group_name(self):
        response = self.make_create_request(Group.MANAGERS, self.manager)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data[0], "Group name already in use for this project.")

    def test_create_group_with_users(self):
        user_a = User.objects.create_user("user_a")
        user_b = User.objects.create_user("user_b")
        self.project.add_member(user_a)
        self.project.add_member(user_b)
        members = [str(user_a.id), str(user_b.id)]
        response = self.make_create_request("New group", self.manager, members=members)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_group = Group.objects.get(pk=response.data["id"])
        self.assertIn(user_a, created_group.users.all())
        self.assertIn(user_b, created_group.users.all())
        self.assertEqual(len(user_a.notifications.unread()), 2)
        self.assertEqual(len(user_b.notifications.unread()), 2)
        # no notification is sent to actor (manager)
        self.assertEqual(len(self.manager.notifications.unread()), 0)

    def test_create_group_with_sets(self):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_templates()
        template = Template.objects.first()
        set = Set.objects.create(
            title="Set title",
            project=self.project,
            creator=self.manager,
            template=template,
        )
        response = self.make_create_request("New group", self.manager, set_ids=[str(set.id)])
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertListEqual([_["id"] for _ in response.data["sets"]], [str(set.id)])


class GroupDetailTest(TestSetup, APITestCase):
    def test_view_group_detail_requires_manager_privileges(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.client.force_login(self.manager)
        response = self.client.get(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_detail_view_keys(self):
        self.client.force_login(self.manager)

        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_templates()
        set = Set.objects.create(
            title="Set title",
            description="",
            project=self.project,
            creator=self.manager,
            template=Template.objects.first(),
        )
        add_set_to_group(self.group, set)
        response = self.client.get(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            list(response.data.keys()),
            ["id", "name", "description", "is_default_group", "users", "sets"],
        )

    def test_detail_view_lists_sets(self):
        self.client.force_login(self.manager)
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_templates()
        set = Set.objects.create(
            title="Set title",
            description="",
            project=self.project,
            creator=self.manager,
            template=Template.objects.first(),
        )
        add_set_to_group(self.group, set)
        response = self.client.get(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(len(response.data["sets"]), 1)
        self.assertEqual(response.data["sets"][0]["id"], str(set.id))


class DeleteGroupTest(TestSetup, APITestCase):
    def test_delete_group_requires_manager_privilege(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_manager_can_delete_group(self):
        self.client.force_login(self.manager)
        othermanager = User.objects.create_user("othermanager")
        self.project.add_member(othermanager)
        self.project.add_manager(othermanager)
        response = self.client.delete(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Group.DoesNotExist):
            self.group.refresh_from_db()
        self.assertEqual(len(othermanager.notifications.unread()), 1)

    def test_group_with_members_can_be_deleted(self):
        self.group.users.add(self.user)
        self.client.force_login(self.manager)
        response = self.client.delete(reverse("group-detail", kwargs={"pk": self.group.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)


class AddUserToGroupTest(TestSetup, APITestCase):
    def test_user_to_add_must_be_active(self):
        user = User.objects.create_user("user_to_add", is_active=False)
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse("group-add-user", kwargs={"pk": self.group.id}),
            {
                "user_id": user.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data[0], "User is not active.")

    def test_user_to_add_must_be_a_project_member(self):
        user = User.objects.create_user("user_to_add")
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse("group-add-user", kwargs={"pk": self.group.id}),
            {
                "user_id": user.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data[0], "User is not a project member.")

    def test_add_user(self):
        user = User.objects.create_user("user_to_add")
        self.project.add_member(user)
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse("group-add-user", kwargs={"pk": self.group.id}),
            {
                "user_id": user.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["users"][0]["id"], str(user.id))
        self.assertEqual(len(user.notifications.unread()), 1)


class RemoveUserFromGroupTest(TestSetup, APITestCase):
    def make_delete_response(self, user, group):
        self.client.force_login(user)
        response = self.client.delete(
            reverse("group-remove-user", kwargs={"pk": group.id}),
            {
                "user_id": self.user.id,
            },
        )
        return response

    def test_remove_user_from_group_requires_manager_privileges(self):
        self.group.users.add(self.user)
        user = User.objects.create_user("otheruser")

        response = self.make_delete_response(user, self.group)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            str(response.data["detail"]),
            "You do not have permission to perform this action.",
        )

        response = self.make_delete_response(self.manager, self.group)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_remove_user_returns_group(self):
        members = self.project.groups.get(name=Group.MEMBERS)
        response = self.make_delete_response(self.manager, members)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["group"]["id"], members.id)
        self.assertListEqual(
            list(response.data["group"].keys()),
            ["id", "name", "description", "is_default_group", "users", "sets"],
        )

    def test_cannot_remove_last_manager(self):
        managers_group = self.project.groups.get(name=Group.MANAGERS)
        self.client.force_login(self.manager)
        response = self.client.delete(
            reverse("group-remove-user", kwargs={"pk": managers_group.id}),
            {
                "user_id": self.manager.id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data[0]), "Group must have at least one manager.")

    def test_remove_user_from_members_group_removes_him_from_whole_project(
        self,
    ):
        members_group = self.project.groups.get(name=Group.MEMBERS)
        self.client.force_login(self.manager)
        response = self.make_delete_response(self.manager, members_group)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            User.objects.filter(project_groups__project=self.project, id=self.user.id).count(),
            0,
        )
        # Actor (manager) is not notified
        self.assertEqual(len(self.manager.notifications.unread()), 0)
        self.assertEqual(len(self.user.notifications.unread()), 1)

    def test_user_is_deactivated_when_removed_from_last_group(self):
        group = self.project.groups.get(name=Group.MEMBERS)
        response = self.make_delete_response(self.manager, group)
        self.user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertFalse(self.user.is_active)

    def test_user_is_not_deactivated_when_member_of_another_project(self):
        other_project: Project = Project.objects.create(
            name="Other project",
            description="Project description",
        )
        other_project.add_member(self.user)
        other_project.add_member(self.manager)
        other_project.add_manager(self.manager)
        group = self.project.groups.get(name=Group.MEMBERS)
        response = self.make_delete_response(self.manager, group)
        self.user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(self.user.is_active)

    def test_remove_user_from_project_sends_email(self):
        members_group = self.project.groups.get(name=Group.MEMBERS)
        self.client.force_login(self.manager)
        response = self.make_delete_response(self.manager, members_group)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].subject,
            f"[pount] You have left project {self.project.name}",
        )

    def test_remove_user_from_project_tolerance_to_email_error(self):
        members_group = self.project.groups.get(name=Group.MEMBERS)
        self.client.force_login(self.manager)
        with patch(
            "pount.apps.api.libs.group.send_email",
            side_effect=smtplib.SMTPException,
        ):
            response = self.make_delete_response(self.manager, members_group)
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        with patch(
            "pount.apps.api.libs.group.send_email",
            side_effect=ConnectionRefusedError,
        ):
            response = self.make_delete_response(self.manager, members_group)
            self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_remove_user_from_project_when_user_has_no_email_address(self):
        self.user.email = None
        self.user.save()
        members_group = self.project.groups.get(name=Group.MEMBERS)
        self.client.force_login(self.manager)
        response = self.make_delete_response(self.manager, members_group)
        self.assertEqual(response.status_code, status.HTTP_200_OK)


class ManageContributionRoleToGroupTest(TestSetup, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        super().setUp()
        initialize_templates()
        template = Template.objects.first()
        self.set = Set.objects.create(
            title="Set title",
            description="description",
            project=self.project,
            creator=self.user,
            publisher="Publisher",
            template=template,
        )

    def test_add_contributor_role_to_group(self):
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse("group-add-set", kwargs={"pk": self.group.id}),
            {
                "set_id": self.set.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(
            Permission.objects.filter(
                role=Role.objects.get_contributor(),
                group=self.group,
                set__id=self.set.id,
            ).count(),
            1,
        )

    def test_add_contributor_role_to_group_sends_notification_to_group_members_and_managers(  # noqa: E501
        self,
    ):
        self.group.users.add(self.user)
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse("group-add-set", kwargs={"pk": self.group.id}),
            {
                "set_id": self.set.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(
            Notification.objects.count(),
            (self.project.managers | self.group.users.all())
            .exclude(id=self.manager.id)  # actor (manager) is not notified
            .distinct()
            .count(),
        )

    def test_must_be_manager_to_add_contributor_role_to_group(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("group-add-set", kwargs={"pk": self.group.id}),
            {
                "set_id": self.set.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_superadmin_can_add_contributor_role_to_group(self):
        admin = User.objects.create_superuser("admin")
        self.client.force_login(admin)
        response = self.client.post(
            reverse("group-add-set", kwargs={"pk": self.group.id}),
            {
                "set_id": self.set.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(
            Permission.objects.filter(
                role=Role.objects.get_contributor(),
                group=self.group,
                set__id=self.set.id,
            ).count(),
            1,
        )

    def test_remove_contributor_role_from_group(self):
        self.client.force_login(self.manager)
        self._create_permission()
        response = self._make_delete_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Permission.DoesNotExist):
            self._get_permission()

    def test_remove_contributor_role_from_group_sends_notification_to_managers_and_group_members(  # noqa: E501
        self,
    ):
        self.client.force_login(self.manager)
        self._create_permission()
        response = self._make_delete_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(
            Notification.objects.count(),
            (self.project.managers | self.group.users.all())
            .exclude(id=self.manager.id)  # actor (manager) is not notified
            .distinct()
            .count(),
        )

    def test_member_can_not_remove_contributor_role_to_group(self):
        self.client.force_login(self.user)
        self._create_permission()
        response = self._make_delete_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertIsInstance(self._get_permission(), Permission)

    def test_superadmin_can_remove_contributor_role_to_group(self):
        admin = User.objects.create_superuser("admin")
        self.client.force_login(admin)
        response = self._make_delete_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Permission.DoesNotExist):
            self._get_permission()

    def _make_delete_response(self):
        response = self.client.delete(
            reverse("group-remove-set", kwargs={"pk": self.group.id}),
            {
                "set_id": self.set.id,
            },
            format="json",
        )
        return response

    def _create_permission(self):
        Permission.objects.create(
            group=self.group,
            content_object=self.set,
            role=Role.objects.get_contributor(),
        )

    def _get_permission(self):
        return Permission.objects.get(
            group=self.group,
            role=Role.objects.get_contributor(),
            set__id=self.set.id,
        )
