import smtplib
import uuid
from datetime import timedelta
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.core import mail
from django.utils import timezone
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.libs.invite import make_invitation_token
from pount.apps.api.models.group import Group
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.models.project import Project
from pount.apps.api.models.user import AuthenticationProvider, Invitation

User = get_user_model()


class CreateInvitationTest(APITestCase):
    def setUp(self) -> None:
        initialize_roles()
        self.project = Project.objects.create(name="Pount Project")
        self.manager = User.objects.create_user("manager")
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.member = User.objects.create_user("member")
        self.project.add_member(self.member)
        self.project_id = uuid.uuid4()
        self.client.force_login(self.manager)

    def test_create_invitation_requires_valid_email(self):
        response = self.client.post(
            reverse("user-invite"),
            {
                "project_id": self.project.id,
                "email": "this_is_not_an_email",
                "url": "https://pount.org/invitation",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "invalid_email")

    def test_create_invitation_requires_existing_project_id(self):
        fake_id = uuid.uuid4()
        response = self.client.post(
            reverse("user-invite"),
            {
                "project_id": fake_id,
                "email": "user@punt.org",
                "url": "https://pount.org/invitation",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_invitation_is_created_for_user_and_project_id(self):
        response = self.client.post(
            reverse("user-invite"),
            {
                "project_id": self.project.id,
                "email": "user@pount.org",
                "url": "https://pount.org/invitation",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(len(Invitation.objects.all()), 1)

    def test_invitation_requires_manager_privilege(self):
        self.client.force_login(self.member)
        response = self.client.post(
            reverse("user-invite"),
            {
                "project_id": self.project.id,
                "email": "user@pount.org",
                "url": "https://pount.org/invitation",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(len(Invitation.objects.all()), 0)

    def test_email_is_sent_with_id_and_token_in_link(self):
        response = self.client.post(
            reverse("user-invite"),
            {
                "project_id": self.project.id,
                "email": "user@pount.org",
                "url": "https://pount.org/invitation",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(len(mail.outbox), 1)
        sent_email = mail.outbox[0]
        invitation = Invitation.objects.first()
        oid = urlsafe_base64_encode(force_bytes(invitation.id))
        self.assertIn(
            f"https://pount.org/invitation?oid={oid}&amp;token={invitation.token}",  # noqa:E501
            sent_email.body,
        )
        self.assertIn("user@pount.org", sent_email.recipients())

    @patch("pount.apps.api.views.user.logger")
    @patch(
        "pount.apps.api.libs.email.send_mail",
        side_effect=smtplib.SMTPException,
    )
    def test_error_is_logged_if_send_email_fails(self, send_email_mock, logger_mock):
        response = self.client.post(
            reverse("user-invite"),
            {
                "project_id": self.project.id,
                "email": "user@pount.org",
                "url": "https://pount.org/invitation",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)
        send_email_mock.assert_called_once()
        logger_mock.exception.assert_called_once()


class ValidateInvitationTest(APITestCase):
    def setUp(self) -> None:
        self.email = "user@pount.org"
        initialize_roles()
        self.project = Project.objects.create(name="Pount project")
        self.token = make_invitation_token(self.email, self.project.id)
        self.invitation = Invitation.objects.create(
            token=self.token,
        )
        self.oid = urlsafe_base64_encode(force_bytes(self.invitation.id))
        self.manager = User.objects.create_user("manager")
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)

    def test_validate_invitation_requires_valid_oid(self):
        oid = urlsafe_base64_encode(force_bytes(uuid.uuid4()))
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": oid, "token": self.token},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.data["error"], "invitation_not_found")

    def test_validate_invitation_requires_non_empty_oid(self):
        oid = urlsafe_base64_encode(force_bytes(""))
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": oid, "token": self.token},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.data["error"], "invitation_not_found")

    def test_validate_invitation_requires_valid_token(self):
        token = make_invitation_token("bad@pount.org", self.project.id)
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": self.oid, "token": token},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.data["error"], "invitation_not_found")

    def test_validate_invitation_requires_non_expired_token(self):
        self.invitation.expires_at = timezone.now() - timedelta(days=1)
        self.invitation.save()
        self.invitation.refresh_from_db()
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": self.oid, "token": self.token},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "invitation_expired")

    def test_validate_invitation_token_signature_must_be_valid(self):
        token = self.token + "changed"
        self.invitation.token = token
        self.invitation.save()
        self.invitation.refresh_from_db()
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": self.oid, "token": token},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "invitation_not_valid")

    def test_validate_invitation_returns_email_project_id_and_invitation_id(
        self,
    ):
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": self.oid, "token": self.token},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["result"], "valid")
        self.assertEqual(response.data["data"]["email"], self.email)
        self.assertEqual(response.data["data"]["invitation_id"], str(self.invitation.id))
        self.assertEqual(response.data["data"]["project_id"], str(self.project.id))
        self.assertListEqual(list(response.data.keys()), ["result", "data"])
        self.assertListEqual(
            list(response.data["data"].keys()),
            ["email", "project_id", "user_id", "invitation_id"],
        )

    def test_validate_invitation_for_existing_user_adds_user_to_project_default_group(  # noqa:E501
        self,
    ):
        user = User.objects.create_user(self.email, self.email)
        token = make_invitation_token(self.email, self.project.id, user)
        invitation = Invitation.objects.create(
            token=token,
        )
        oid = urlsafe_base64_encode(force_bytes(invitation.id))
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": oid, "token": token},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual("added_to_project", response.data["result"])
        self.assertEqual(len(user.project_groups.filter(project=self.project)), 1)
        self.assertEqual(len(self.manager.notifications.unread()), 1)
        self.assertEqual(self.manager.notifications.unread()[0].project.id, self.project.id)

    def test_validate_invitation_for_existing_inactive_user_activates_user(
        self,
    ):  # noqa:E501
        user = User.objects.create_user(self.email, self.email, is_active=False)
        token = make_invitation_token(self.email, self.project.id, user)
        invitation = Invitation.objects.create(token=token)
        oid = urlsafe_base64_encode(force_bytes(invitation.id))
        response = self.client.post(
            reverse("user-validate-invitation"),
            {"oid": oid, "token": token},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual("added_to_project", response.data["result"])
        user.refresh_from_db()
        self.assertTrue(user.is_active)


class UserSignupTest(APITestCase):
    def setUp(self) -> None:
        initialize_roles()
        self.project = Project.objects.create(name="Project name")
        self.manager = User.objects.create_user("manager")
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.invitation = Invitation.objects.create(
            token=make_invitation_token(
                "user@pount.org",
                self.project.id,
            )
        )
        self.signup_data = {
            "invitation_id": str(self.invitation.id),
            "first_name": "First name",
            "last_name": "Last name",
            "email": "user@pount.org",
            "new_password": "SecretPassword",
            "new_password2": "SecretPassword",
        }

    def test_signup_invited_user(self):
        response = self.client.post(
            reverse("user-signup-from-invitation"),
            self.signup_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        user = get_user_model().objects.get_by_natural_key("user@pount.org")
        self.assertEqual(response.data["id"], str(user.id))
        self.assertTrue(user.check_password("SecretPassword"))
        self.assertEqual(len(self.manager.notifications.unread()), 1)

    def test_signup_invited_user_requires_repeated_password(self):
        signup_data = self.signup_data
        signup_data["new_password2"] = "NotTheSamePassword"
        response = self.client.post(
            reverse("user-signup-from-invitation"),
            signup_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["new_password"][0], "Your passwords do not match")

    def test_signup_must_be_made_with_same_email_as_in_invitation(self):
        self.signup_data["email"] = "other@pount.org"
        response = self.client.post(
            reverse("user-signup-from-invitation"),
            self.signup_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.data["non_field_errors"][0],
            "Email is not the same as in the invitation",
        )

    def test_signup_uses_the_email_as_username(self):
        created_user = self._signup_user()
        self.assertEqual(created_user.base_username, self.signup_data["email"])

    def test_created_user_is_added_to_project_default_group(self):
        created_user = self._signup_user()
        groups = created_user.project_groups.filter(project=self.project)
        self.assertEqual(len(groups), 1)
        self.assertEqual(groups[0].name, Group.MEMBERS)
        self.assertEqual(str(groups[0]), Group.MEMBERS)

    def _signup_user(self) -> User:
        response = self.client.post(
            reverse("user-signup-from-invitation"),
            self.signup_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_user = get_user_model().objects.get(pk=response.data["id"])
        return created_user

    @patch("pount.apps.api.views.user.validate_invitation")
    def test_signup_fails_if_invitation_is_not_valid(self, validate_mock):
        validate_mock.return_value = {"error_text": "validation_error_text"}, True
        response = self.client.post(
            reverse("user-signup-from-invitation"),
            self.signup_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "validation_error_text")


class InternalUserInvitation(APITestCase):
    def setUp(self) -> None:
        initialize_roles()
        self.user = User.objects.create_user(
            "username",
            email="user@pount.org",
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.project = Project.objects.create(name="Pount project")
        self.manager = User.objects.create_user("manager")
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.member = User.objects.create_user("member")
        self.project.add_member(self.member)
        self.client.force_login(self.manager)

    def test_invite_user_requires_existing_project(self):
        project_id = str(uuid.uuid4())
        response = self.client.post(
            reverse("user-invite-internal-user"),
            {
                "username": self.user.username,
                "project_id": project_id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_invite_user_requires_existing_username(self):
        response = self.client.post(
            reverse("user-invite-internal-user"),
            {
                "username": "non_existent",
                "project_id": self.project.id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_invite_internal_user_adds_user_to_default_project_group(self):
        response = self.client.post(
            reverse("user-invite-internal-user"),
            {
                "username": self.user.base_username,
                "project_id": self.project.id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            len(
                self.user.project_groups.filter(
                    project=self.project,
                    name=Group.MEMBERS,
                )
            ),
            1,
        )
        self.assertEqual(response.data["user"]["username"], self.user.username)
        self.assertEqual(response.data["email"], "sent")
        self.assertEqual(len(self.manager.notifications.unread()), 1)

    def test_invite_internal_user_reports_email_not_sent_error(self):
        with patch(
            "pount.apps.api.views.user.send_internal_invitation_email",
            side_effect=smtplib.SMTPException,
        ):
            response = self.client.post(
                reverse("user-invite-internal-user"),
                {
                    "username": self.user.base_username,
                    "project_id": self.project.id,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data["email"], "error")
        with patch(
            "pount.apps.api.views.user.send_internal_invitation_email",
            side_effect=ConnectionRefusedError,
        ):
            response = self.client.post(
                reverse("user-invite-internal-user"),
                {
                    "username": self.user.base_username,
                    "project_id": self.project.id,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data["email"], "error")

    def test_invite_internal_user_with_no_email_reports_that_no_email_is_sent(
        self,
    ):
        self.user.email = None
        self.user.save()
        response = self.client.post(
            reverse("user-invite-internal-user"),
            {
                "username": self.user.base_username,
                "project_id": self.project.id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["email"], "no")

    def test_invite_internal_user_requires_manager_privileges(self):
        self.client.force_login(self.member)
        response = self.client.post(
            reverse("user-invite-internal-user"),
            {
                "username": self.user.base_username,
                "project_id": self.project.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_invite_internal_user_sends_email_if_user_has_address(self):
        response = self.client.post(
            reverse("user-invite-internal-user"),
            {
                "username": self.user.base_username,
                "project_id": self.project.id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(mail.outbox), 1)
        self.assertIn(self.user.email, mail.outbox[0].recipients())
        self.assertEqual(
            mail.outbox[0].subject,
            f"[pount] Invitation to join the project {self.project.name}",
        )
