import random
import string
from unittest.mock import patch

from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models.group import Group
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.models.item import Item, MediaFile
from pount.apps.api.models.permission import Permission, Role
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.tests.factories.template import TemplateFactory

User = get_user_model()


class SetUpMixin:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_templates()
        initialize_roles()
        self.creator = User.objects.create_user("creator")
        self.member = User.objects.create_user("member")
        self.manager = User.objects.create_user("manager")
        self.other_creator = User.objects.create_user("other_creator")
        self.template = Template.objects.get(name="datacite")
        self.project = Project.objects.create(
            name="Project name",
            description="Project description",
        )
        self.project.add_member(self.creator)
        self.project.add_member(self.member)
        self.project.add_member(self.other_creator)
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.set = Set.objects.create(
            title="Set title",
            description="Set description",
            creator=self.creator,
            publisher="Unistra",
            project=self.project,
            template=TemplateFactory(),
        )
        self.public_set = Set.objects.create(
            title="Public set",
            description="Set description",
            is_public=True,
            creator=self.creator,
            publisher="Unistra",
            project=self.project,
            template=TemplateFactory(),
        )
        group = Group.objects.create(
            name="Set contributors",
            project=self.project,
            settings=[],
        )
        group.users.add(self.creator, self.other_creator)
        Permission.objects.create(
            group=group,
            role=Role.objects.get_contributor(),
            content_object=self.set,
        )
        for i in range(5):
            Item.objects.create(
                title=f"Item {i}",
                description=f"Item {i} description",
                set=self.set,
                creator=self.creator,
                metadata=[
                    {
                        "id": "".join(random.choice(string.ascii_lowercase) for _ in range(9)),
                        "name": "title",
                        "value": f"Item {i}",
                        "multiple": False,
                        "repeatable": False,
                    }
                ],
            )


@patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
class CanEditFlagTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        items = self.set.items.all()
        self.creator_item = items[0]
        self.other_item = items[1]
        self.other_item.creator = self.member
        self.other_item.save()

    def test_item_creator_can_edit(self, update_mock):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.creator_item.pk}))
        self.assertTrue(response.data["can_edit"])

    def test_non_creator_can_not_edit(self, update_mock):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.other_item.pk}))
        self.assertFalse(response.data["can_edit"])

    def test_project_manager_can_edit(self, update_mock):
        self.client.force_login(self.manager)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.creator_item.pk}))
        self.assertTrue(response.data["can_edit"])


class UpdateItemTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_update_item_needs_authentication(self):
        response = self.client.patch(
            reverse("item-detail", kwargs={"pk": self.set.items.first().pk}),
            {},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def _make_patch_request(self, item=None):
        if item is None:
            item = self.set.items.first()
        response = self.client.patch(
            reverse("item-detail", kwargs={"pk": item.pk}),
            {
                "description": "Changed description",
            },
            format="json",
        )
        return response

    def test_update_item_by_project_manager(self):
        self.client.force_login(self.manager)
        response = self._make_patch_request()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["description"], "Changed description")

    def test_creator_can_patch_item(self):
        self.client.force_login(self.creator)
        response = self._make_patch_request()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["description"], "Changed description")

    def test_item_update_calls_metadata_processors(self):
        self.client.force_login(self.creator)
        item = self.set.items.first()
        with patch("pount.apps.api.serializers.item.process_metadata") as process_metadata:
            process_metadata.return_value = []
            self.client.patch(
                reverse("item-detail", kwargs={"pk": item.pk}),
                {
                    "metadata": [],
                },
                format="json",
            )
        process_metadata.assert_called_with([], self.set.template)

    def test_project_member_can_not_patch_item(self):
        self.client.force_login(self.member)
        response = self._make_patch_request()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_item_adds_notification(self):
        self.client.force_login(self.creator)
        response = self._make_patch_request()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.manager.notifications.unread().count(), 1)
        # Actor (creator) is not notified
        self.assertEqual(self.creator.notifications.unread().count(), 0)
        self.assertEqual(self.member.notifications.unread().count(), 1)

    def test_item_title_is_set_from_metadata(self):
        item = self.set.items.first()
        item.metadata = [
            {
                "id": "".join(random.choice(string.ascii_lowercase) for _ in range(9)),
                "name": "title",
                "value": "The real title",
                "multiple": False,
                "repeatable": False,
            }
        ]
        item.save()
        self.client.force_login(self.creator)
        response = self._make_patch_request(item)
        item.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(item.title, "The real title")

    def test_patch_item_s_set_id_checks_user_s_manager_status(self):
        """
        When moving Item to another Set, check if user is manager of
        destination Set's project
        """
        item = self.set.items.first()
        other_project = Project.objects.create(
            name="Other project name",
            description="Project User is only a member of",
        )
        other_project.add_member(self.manager)
        set = Set.objects.create(
            title="Set title",
            description="Set description",
            creator=self.creator,
            publisher="Unistra",
            project=other_project,
            template=self.template,
        )
        self.client.force_login(self.manager)
        response = self.client.patch(
            reverse("item-detail", kwargs={"pk": item.pk}),
            {
                "set_id": str(set.id),
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        other_project.add_manager(self.manager)
        response = self.client.patch(
            reverse("item-detail", kwargs={"pk": item.pk}),
            {
                "set_id": str(set.id),
            },
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_validate_metadata_json_schema(self):
        item = self.set.items.first()
        self.client.force_login(self.manager)
        response = self.client.patch(
            reverse("item-detail", kwargs={"pk": item.pk}),
            {
                "metadata": [
                    {
                        "unknown_key": "some value",
                    }
                ]
            },
            format="json",
        )
        self.assertEqual(
            response.status_code,
            status.HTTP_400_BAD_REQUEST,
        )
        self.assertIn(
            "JSON Schema validation error",
            str(response.data["metadata"]),
        )


class ViewItemTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.public_item = Item.objects.create(
            title="Public item",
            set=self.public_set,
            creator=self.creator,
            is_public=True,
            metadata=[],
        )
        self.private_item = self.set.items.first()

    def test_view_item_needs_authentication(self):
        self.client.logout()
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_project_member_can_view_public_item(self):
        self.client.force_login(self.member)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_project_member_can_view_private_item(self):
        self.client.force_login(self.member)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.private_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_authenticated_user_can_view_public_item(self):
        user = User.objects.create_user("user")
        self.client.force_login(user)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_authenticated_user_can_not_view_private_item(self):
        user = User.objects.create_user("user")
        self.client.force_login(user)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.private_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_set_must_be_public_to_view_public_item(self):
        self.public_item.set = self.set
        self.public_item.save()
        user = User.objects.create_user("user")
        self.client.force_login(user)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_set_contributor_can_view_public_item(self):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_set_contributor_can_view_private_item(self):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.private_item.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_item_has_can_be_shared_flag(self):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.private_item.pk}))
        self.assertFalse(response.data["can_be_shared"])

    def test_can_be_shared_flag_for_public_item_public_set_public_project(
        self,
    ):
        self.client.force_login(self.creator)
        project = self.public_item.set.project
        project.is_public = True
        project.save(update_fields=["is_public"])
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertTrue(response.data["can_be_shared"])

    def test_can_be_shared_flag_for_public_item(self):
        self.client.force_login(self.creator)
        # Public item, public set, public project
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertFalse(response.data["can_be_shared"])

    def test_can_be_shared_flag_for_public_item_private_set(self):
        self.client.force_login(self.creator)
        # Public item, private set
        self.public_set.is_public = False
        self.public_set.save(update_fields=["is_public"])
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertFalse(response.data["can_be_shared"])

    def test_can_be_shared_flag_for_public_item_public_set_private_project(
        self,
    ):
        self.client.force_login(self.creator)
        # Public item, public set, private project
        project = self.public_item.set.project
        # project.is_public = True
        # project.save(update_fields=["is_public"])
        self.assertFalse(project.is_public)
        response = self.client.get(reverse("item-detail", kwargs={"pk": self.public_item.pk}))
        self.assertFalse(response.data["can_be_shared"])


class ListItemsTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_list_items_needs_authentication(self):
        response = self.client.get(reverse("item-list"), format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_items(self):
        self.client.force_login(self.member)
        response = self.client.get(reverse("item-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_items_is_paginated(self):
        self.client.force_login(self.member)
        response = self.client.get(reverse("item-list"), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 5)


class CreateItemTest(SetUpMixin, APITestCase):
    def test_create_item_needs_authentication(self):
        self.client.logout()
        response = self.client.post(reverse("item-list"), {}, format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_create_item_if_only_project_member(self):
        response = self._make_create_request(self.member)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_item_with_contributor_privilege(self):
        response = self._make_create_request(self.creator)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["set"]["id"], str(self.set.id))

    def test_create_item_with_manager_privilege(self):
        response = self._make_create_request(self.manager)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["set"]["id"], str(self.set.id))

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def _make_create_request(self, user, update_mock):
        self.client.force_login(user)
        response = self.client.post(
            reverse("item-list"),
            {
                "title": "Item title",
                "metadata": [],
                "set_id": self.set.id,
            },
            format="json",
        )
        return response

    def test_create_item_adds_notification(self):
        response = self._make_create_request(self.creator)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.manager.notifications.unread().count(), 1)
        # Actor (creator) is not notified
        self.assertEqual(self.creator.notifications.unread().count(), 0)
        self.assertEqual(self.member.notifications.unread().count(), 1)

    def test_create_item_calls_metadata_processors(self):
        with patch("pount.apps.api.serializers.item.process_metadata") as process_metadata:
            process_metadata.return_value = []
            self._make_create_request(self.manager)
        process_metadata.assert_called_with([], self.set.template)


class DeleteItemTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.item = Item.objects.first()

    def _get_response(self):
        response = self.client.delete(reverse("item-detail", kwargs={"pk": self.item.id}))
        return response

    def test_delete_item_needs_authentication(self):
        self.client.logout()
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_delete_item_not_allowed_for_project_member(self):
        self.client.force_login(self.member)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_item_allowed_for_creator(self):
        self.client.force_login(self.creator)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_item_allowed_for_project_manager(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_deleted_item_not_returned_by_default_manager(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Item.DoesNotExist):
            Item.objects.get(pk=self.item.id)

    def test_deleted_item_is_soft_deleted(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        # Item is still present in the database, marked as deleted
        self.item.refresh_from_db()
        self.assertTrue(self.item.is_deleted)
        self.assertIsNotNone(self.item.deleted_at)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_delete_item_does_not_delete_mediafiles(self, bucket_mock):
        bucket_mock.Object.return_value.delete.return_value = None
        mf1 = MediaFile.objects.create(
            item=self.item,
            filename="pount1.png",
            info={"content_type": "image/png", "size": 1024},
        )
        mf2 = MediaFile.objects.create(
            item=self.item,
            filename="pount2.png",
            info={"content_type": "image/png", "size": 1024},
        )
        self.item.media_files.add(mf1, mf2)
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(MediaFile.objects.filter(item=self.item).count(), 2)
        self.assertEqual(bucket_mock.Object.return_value.delete.call_count, 0)

    def test_bulk_delete_items_for_manager(self):
        items = self.set.items.all()
        self.client.force_login(self.manager)
        response = self.client.delete(
            reverse("item-list"),
            {"item_ids": [_.pk for _ in items]},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(Item.objects.count(), 0)
        self.assertEqual(Item.deleted_objects.count(), 5)

    def test_bulk_delete_items_for_creator_only_deletes_own_items(self):
        item = self.set.items.first()
        other_item = Item.objects.create(
            title="Other creator item",
            description="",
            set=self.set,
            creator=self.other_creator,
            metadata=[],
        )
        self.client.force_login(self.creator)
        response = self.client.delete(
            reverse("item-list"),
            {"item_ids": [item.pk, other_item.pk]},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        item.refresh_from_db()
        other_item.refresh_from_db()
        self.assertTrue(item.is_deleted)
        self.assertFalse(other_item.is_deleted)

    def test_bulk_delete_items_not_allowed_for_project_member(self):
        self.client.force_login(self.member)
        response = self.client.delete(
            reverse("item-list"),
            {"item_ids": [_.id for _ in self.set.items.all()]},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(Item.objects.count(), 5)

    def test_bulk_delete_items_needs_authentication(self):
        self.client.logout()
        response = self.client.delete(
            reverse("item-list"),
            {"item_ids": [_.id for _ in self.set.items.all()]},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
