from pathlib import Path

from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models.item import Item
from pount.apps.api.models.project import Project
from pount.apps.api.tests.api.test_item import SetUpMixin

User = get_user_model()


class XmlExportTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        project2 = Project.objects.create(
            name="Other Project",
            description="",
        )
        self.project2_member = User.objects.create_user("member2")
        metadata = [
            {
                "id": "6qgq1netsxh",
                "name": "creator",
                "value": ["Creator Name"],
                "multiple": False,
                "is_option": False,
                "repeatable": True,
            },
            {
                "id": "5vxode5mjsh",
                "name": "title",
                "value": "The item's title",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
            {
                "id": "unym14mu68",
                "name": "publisher",
                "value": "Publisher",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
            {
                "id": "m1w8jje0wsh",
                "name": "publicationYear",
                "value": "2021",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
            {
                "id": "2qwhbanmbsc",
                "name": "resourceTypeGeneral",
                "value": "Model",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
            {
                "id": "yxuullxsicj",
                "name": "resourceType",
                "value": "3D Model",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
        ]
        self.item = Item.objects.first()
        self.item.metadata = metadata
        self.item.save()
        project2.add_member(self.project2_member)
        project2.add_manager(self.project2_member)
        file = open(
            Path(__file__).resolve(strict=True).parent.parent / "fixtures/datacite_metadata.xml",
        )
        self.xml = file.read()
        file.close()

    def _get_response(self):
        response = self.client.get(reverse("item-export-datacite", kwargs={"pk": self.item.id}))
        return response

    def test_view_unauthorized_for_anonymous_user(self):
        self.client.logout()
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_item_creator_can_export_xml(self):
        self.client.force_login(self.creator)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_project_manager_can_export_xml(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_project_member_can_export_xml(self):
        self.client.force_login(self.member)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_other_project_member_can_not_export_xml(self):
        self.client.force_login(self.project2_member)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_view_returns_exported_datacite_xml(self):
        self.client.force_login(self.creator)
        response = self._get_response()
        self.assertEqual(response.get("content-type"), "application/xml")
        self.assertXMLEqual(response.content.decode(), self.xml)
