import logging
from pathlib import Path
from unittest.mock import patch

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.libs.metadata import Metadata
from pount.apps.api.models import Item, Template
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.tests.factories.set import SetForManagerFactory
from pount.apps.api.tests.factories.template import TemplateForSetFactory
from pount.apps.api.tests.factories.user import UserFactory


class ImportItemFromCSVTest(APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_roles()
        initialize_templates()
        datacite = Template.objects.get(name="datacite")
        self.member = UserFactory(username="member")
        self.manager = UserFactory(username="manager")
        self.set = SetForManagerFactory(members=[self.member, self.manager], managers=self.manager)
        self.template = TemplateForSetFactory(fields=datacite.fields, creator=self.manager, set=self.set)
        self.set.template = self.template
        self.set.save()

    def _get_response(self, set=None, filename=None):
        set = set or self.set
        filename = filename or "import-item.csv"
        with open(
            Path(__file__).resolve(strict=True).parent.parent / f"fixtures/csv/{filename}",
        ) as file:
            response = self.client.post(
                reverse("item-import-csv"),
                {
                    "set_id": set.id,
                    "csv": file,
                },
            )
        return response

    def test_view_is_authenticated(self):
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_import_is_forbidden_for_members(self):
        self.client.force_login(self.member)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_import_is_allowed_for_managers(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.set.items.count(), 1)
        self.assertEqual(response.data["created_items"], 1)

    def test_a_file_must_be_posted(self):
        self.client.force_login(self.manager)
        response = self.client.post(
            reverse("item-import-csv"),
            {
                "set_id": self.set.id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            "Missing file",
            response.data["error"],
        )

    def test_file_must_be_csv(self):
        self.client.force_login(self.manager)
        # A non CSV file should fail
        with open(
            Path(__file__).resolve(strict=True).parent.parent / "fixtures/pount.jpg",
            "rb",
        ) as file:
            with self.assertLogs(level=logging.ERROR) as ctx:
                response = self.client.post(
                    reverse("item-import-csv"),
                    {
                        "set_id": self.set.id,
                        "csv": file,
                    },
                )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertIn("Error processing CSV import", str(ctx.output))

    def test_item_is_created_with_repeatable_field(self):
        self.client.force_login(self.manager)
        response = self._get_response(filename="import-repeatable.csv")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        item = Item.objects.get(set=self.set)
        self.assertEqual(item.is_public, False)
        self.assertEqual(item.title, "The item's title")
        md = Metadata(item.metadata)
        self.assertEqual(
            "Lucy Carter, Robin Hall",
            ", ".join(md.creator.value),
        )

    def test_import_fails_if_required_field_is_missing(self):
        self.client.force_login(self.manager)
        with patch("pount.apps.api.views.item.logger"):
            # patch logger to avoid getting the output
            response = self._get_response(filename="import-missing-required.csv")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertEqual(
                "CSV import error at line 2: Missing required value for title",
                response.data["error"],
            )

    def test_import_fails_with_invalid_options(self):
        self.client.force_login(self.manager)
        with patch("pount.apps.api.views.item.logger"):
            response = self._get_response(filename="import-options.csv")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertEqual(
                "CSV import error at line 1: Value <Invalid> not found in option list for <resourceTypeGeneral>",  # noqa: E501
                response.data["error"],
            )

    def test_import_fails_if_required_field_is_missing_in_header(self):
        """
        First line of CSV file must contain the ids of
        all the template's required fields
        """
        self.client.force_login(self.manager)
        with patch("pount.apps.api.views.item.logger"):
            response = self._get_response(filename="import-missing-fields-ids.csv")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.set.items.count(), 0)
        self.assertEqual(
            "Required field(s) missing in CSV file: publisher",
            response.data["error"],
        )
