from pathlib import Path
from unittest.mock import patch

from django.contrib.auth import get_user_model
from rest_framework.reverse import reverse

from pount.apps.api.libs.file import item_mediafile_filename, sizeof_format
from pount.apps.api.libs.group import add_set_to_group
from pount.apps.api.models import init
from pount.apps.api.models.group import Group
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.models.item import Item, MediaFile
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.tests.factories.item import ItemForMemberFactory
from pount.apps.api.tests.factories.user import UserFactory
from pount.tests import APITestCase

User = get_user_model()


class SetUpMixin:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        init.initialize_templates()
        init.initialize_roles()
        self.member = User.objects.create_user("member")
        self.creator = User.objects.create_user("creator")
        self.manager = User.objects.create_user("user")
        self.template = Template.objects.get(name="datacite")
        self.project = Project.objects.create(
            name="Project name",
            description="Project description",
        )
        group = Group.objects.create_for_project(name="Group", description="", project=self.project)
        self.project.add_member(self.member)
        self.project.add_member(self.creator)
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.set = Set.objects.create(
            title="Set title",
            description="Set description",
            creator=self.creator,
            publisher="Unistra",
            project=self.project,
            template=self.template,
        )
        add_set_to_group(group, self.set)
        self.item = Item.objects.create(
            title="Item title",
            description="Item description",
            set=self.set,
            creator=self.creator,
            metadata=[],
        )

    def _uploaded_file(self):
        uploaded_file = open(
            Path(__file__).resolve(strict=True).parent.parent / "fixtures/pount.jpg",
            "rb",
        )
        return uploaded_file


class CreateMediaFileTest(SetUpMixin, APITestCase):
    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_create_mediafile_requires_authentication(self, storage_mock):
        response = self._make_post_request(storage_mock)
        self.response_unauthorized(response)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_project_member_cannot_create_mediafile(self, storage_mock):
        response = self._make_post_request(storage_mock, user=self.member)
        self.response_forbidden(response)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_item_creator_can_create_mediafile(self, storage_mock):
        storage_mock.name = "pount_test"
        response = self._make_post_request(storage_mock, user=self.creator)
        self.response_created(response)
        self.assertListEqual(
            list(response.data.keys()),
            [
                "id",
                "filename",
                "name",
                "info",
                "is_visualized",
                "is_downloadable",
                "component",
                "url",
                "content_type",
                "size",
            ],
        )

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_manager_can_create_mediafile(self, storage_mock):
        storage_mock.name = "pount_test"
        response = self._make_post_request(storage_mock, user=self.manager)
        self.response_created(response)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_content_type_and_size_are_stored_on_mediafile(self, storage_mock):
        storage_mock.name = "pount_test"
        response = self._make_post_request(storage_mock, user=self.creator)
        mediafile = MediaFile.objects.get(pk=response.data["id"])
        self.assertEqual(mediafile.info["content_type"], "image/png")
        self.assertEqual(mediafile.info["size"], 1024)
        self.assertEqual(response.data["content_type"], "image/png")
        self.assertEqual(response.data["size"], sizeof_format(1024))
        self.assertEqual(response.data["name"], item_mediafile_filename(mediafile.filename))

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_component_is_required_if_is_visualized_is_true(self, storage_mock):
        response = self._make_post_request(
            storage_mock,
            data={
                "filename": "pount.png",
                "is_visualized": True,
            },
            user=self.manager,
        )
        self.response_bad_request(response)
        self.assertIn("Component must be set", str(response.data["component"]))

    def _make_post_request(self, storage_mock, data: dict = None, user: User = None):
        data = data or {
            "filename": "pount.png",
            "is_visualized": True,
            "component": MediaFile.Component.IMAGE,
        }
        storage_mock.Object.return_value.content_type = "image/png"
        storage_mock.Object.return_value.content_length = 1024
        response = self.post(
            reverse("mediafile-list", kwargs={"items_pk": self.item.pk}),
            data=data,
            user=user,
        )
        return response


class UpdateMediaFileTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.mediafile = MediaFile.objects.create(
            item=self.item,
            filename="pount.png",
            info={"content_type": "image/png", "size": 1024},
        )

    def test_update_mediafile_needs_authentication(self):
        response = self._make_response(user=None)
        self.response_unauthorized(response)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_project_member_cannot_update_mediafile(self, storage_mock):
        response = self._make_response(user=self.member)
        self.response_forbidden(response)

    def _make_response(self, data: dict = None, user: User = None):
        data = data or {}
        response = self.patch(
            reverse(
                "mediafile-detail",
                kwargs={"items_pk": self.item.pk, "pk": self.mediafile.pk},
            ),
            data=data,
            user=user,
        )
        return response

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_file_can_be_visualized_if_component_is_set(self, storage):
        storage.name = "pount_test"
        self.mediafile.component = MediaFile.Component.IMAGE
        self.mediafile.save()
        response = self._make_response(data={"is_visualized": True}, user=self.creator)
        self.response_ok(response)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_file_can_be_visualized_if_component_is_provided(self, storage):
        storage.name = "pount_test"
        response = self._make_response(
            data={"is_visualized": True, "component": MediaFile.Component.IMAGE}, user=self.creator
        )
        self.response_ok(response)

    def test_file_can_not_be_visualized_if_component_is_not_set(self):
        self.mediafile.component = None
        self.mediafile.save()
        response = self._make_response(data={"is_visualized": True}, user=self.creator)
        self.response_bad_request(response)
        self.assertIn("Component must be set", str(response.data["component"]))


class DeleteMediaFileTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.mediafile = MediaFile.objects.create(
            item=self.item,
            filename="pount.png",
            info={"content_type": "image/png", "size": 1024},
        )

    def test_delete_mediafile_needs_authentication(self):
        response = self._make_response()
        self.response_unauthorized(response)

    def test_project_member_cannot_delete_mediafile(self):
        response = self._make_response(user=self.member)
        self.response_forbidden(response)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.bucket")
    def test_delete_mediafile(self, bucket_mock):
        bucket_mock.Object.return_value.delete.return_value = None
        response = self._make_response(user=self.creator)
        self.response_no_content(response)
        bucket_mock.Object.return_value.delete.assert_called_once()

    def _make_response(self, user: User = None):
        response = self.delete(
            reverse(
                "mediafile-detail",
                kwargs={"items_pk": self.item.pk, "pk": self.mediafile.pk},
            ),
            user=user,
        )
        return response


class GetPostUrlTest(APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock):
        initialize_roles()
        self.member = UserFactory(username="member")
        self.item = ItemForMemberFactory(members=self.member, creator=self.member)

    def test_get_post_url_view_is_authenticated(self):
        response = self.post(reverse("item-post-url", kwargs={"pk": str(self.item.id)}))
        self.response_forbidden(response)

    def test_filename_parameter_is_required(self):
        response = self.post(
            reverse("item-post-url", kwargs={"pk": str(self.item.id)}),
            {
                "content_type": "image/png",
            },
            user=self.member,
        )
        self.response_bad_request(response)

    def test_content_type_parameter_is_required(self):
        response = self.post(
            reverse("item-post-url", kwargs={"pk": str(self.item.id)}),
            {
                "filename": "name.ext",
            },
            user=self.member,
        )
        self.response_bad_request(response)

    def test_storage_post_url_is_returned(self):
        with patch("pount.apps.api.views.item.ItemMediaFileStorage") as storage:
            storage.return_value.post_url.return_value = "https://pount.unistra.fr/post_url"
            response = self.post(
                reverse("item-post-url", kwargs={"pk": str(self.item.id)}),
                {
                    "filename": "name.ext",
                    "content_type": "image/png",
                },
                user=self.member,
            )
            self.response_ok(response)
            self.assertEqual(response.data, "https://pount.unistra.fr/post_url")
            self.assertEqual(storage.return_value.post_url.call_count, 1)
