import uuid
from datetime import datetime, timedelta

import jwt
from django.contrib.auth import REDIRECT_FIELD_NAME, get_user_model
from django.contrib.sessions.models import Session
from django.core import mail, signing
from django.urls import reverse
from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase
from rest_framework_simplejwt.settings import api_settings

from pount.apps.api.models import AuthenticationProvider, Ticket
from pount.apps.api.tests.util import create_user

User = get_user_model()


class LocalLoginTest(APITestCase):
    def setUp(self) -> None:
        self.user = create_user("user", False, password="password")

    def test_local_login_route(self):
        response = self.client.post(
            "/api/login/local/",
            {
                "username": "user",
                "password": "password",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(list(response.data.keys()), ["refresh", "access"])

    def test_user_must_be_active(self):
        """
        Is tested in djangorestframework_simplejwt, test here in case the
        library's behaviour changes
        """
        self.user.is_active = False
        self.user.save()
        self.user.refresh_from_db()
        response = self.client.post(
            "/api/login/local/",
            {
                "username": "user",
                "password": "password",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)


class CASLoginTest(APITestCase):
    def test_login_view_redirects_to_cas(self):
        response = self.client.get("/api/login/cas/", follow=False)
        redirect_to = f"{reverse('django_cas:login')}?{REDIRECT_FIELD_NAME}=/api/login/cas/"  # noqa:501
        self.assertRedirects(response, redirect_to, fetch_redirect_response=False)

    def test_redirect_url_must_be_passed_as_parameter(self):
        user = create_user("user", False)
        self.client.force_login(user)
        response = self.client.get(
            "/api/login/cas/",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_creates_authentication_ticket_for_authenticated_user(self):
        user = create_user("user", False)
        self.client.force_login(user)
        response = self.client.get("/api/login/cas/?redirect=http://redirect.url/")
        ticket = Ticket.objects.get(user=user)
        self.assertRedirects(
            response,
            f"http://redirect.url/?ticket={ticket.id}",
            fetch_redirect_response=False,
        )


class LogoutViewTest(APITestCase):
    def test_redirects_to_provided_url(self):
        redirect_to = "https://url_to_foll.ow"
        response = self.client.get(f"/api/logout/?redirect={redirect_to}", follow=False)
        self.assertRedirects(response, redirect_to, fetch_redirect_response=False)

    def test_redirect_url_is_mandatory(self):
        response = self.client.get(
            "/api/logout/",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_user_session_is_flushed(self):
        user = create_user("user", False)
        self.client.force_login(user)
        self.assertEqual(Session.objects.count(), 1)
        self.client.get("/api/logout/?redirect=somewhere", follow=False)
        self.assertEqual(Session.objects.count(), 0)


class JWTTokenTest(APITestCase):
    def setUp(self) -> None:
        self.user = create_user("user", False)

    def _create_ticket(self, expiration: datetime = None):
        self.ticket = Ticket.objects.create(
            user=self.user,
            provider=AuthenticationProvider.CAS.value,
            expires_at=expiration,
        )

    def test_ticket_id_is_required_in_request(self):
        response = self.client.post(reverse("login-token"), {}, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_ticket_must_not_be_expired(self):
        self._create_ticket(expiration=(timezone.now() - timedelta(seconds=120)))
        response = self.client.post(reverse("login-token"), {"ticket": self.ticket.id}, format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        with self.assertRaises(Ticket.DoesNotExist):
            # Expired ticket is deleted
            self.ticket.refresh_from_db()

    def test_ticket_must_exist(self):
        ticket_id = uuid.uuid4()
        response = self.client.post(reverse("login-token"), {"ticket": ticket_id}, format="json")
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_generate_jwt_token_pair_for_valid_ticket(self):
        self._create_ticket()
        response = self.client.post(reverse("login-token"), {"ticket": self.ticket.id}, format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(list(response.data.keys()), ["refresh", "access"])

    def test_jwt_token_contains_user_id_and_login_provider(self):
        self._create_ticket()
        response = self.client.post(reverse("login-token"), {"ticket": self.ticket.id}, format="json")
        access_token = response.data["access"]
        token_data = jwt.decode(
            str(access_token),
            key=api_settings.VERIFYING_KEY,
            algorithms=["RS256"],
        )
        self.assertEqual(token_data["user_id"], str(self.user.id))

    def test_user_must_be_active(self):
        self.user.is_active = False
        self.user.save()
        self.user.refresh_from_db()
        self._create_ticket()
        response = self.client.post(reverse("login-token"), {"ticket": self.ticket.id}, format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_ticket_is_deleted_after_token_is_generated(self):
        self._create_ticket()
        response = self.client.post(reverse("login-token"), {"ticket": self.ticket.id}, format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        with self.assertRaises(Ticket.DoesNotExist):
            self.ticket.refresh_from_db()


class PasswordResetTest(APITestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            "user@pount.org",
            password="secret",
            email="user@pount.org",
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )

    def make_request(self, data={}):
        response = self.client.post(
            reverse("login-password-reset-initialize"),
            data,
            format="json",
        )
        return response

    def test_email_must_exist(self):
        response = self.make_request(
            {
                "email": "another.email@pount.org",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_password_reset_request_must_contain_url(self):
        response = self.make_request(
            {
                "email": self.user.email,
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "missing_url")

    def test_reset_password_request_sends_email_with_token(self):
        response = self.make_request(
            {
                "email": self.user.email,
                "url": "/reset-password",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.user.refresh_from_db()
        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertEqual(email.recipients()[0], self.user.email)
        rid = self.user.settings["password_reset"]["reset_id"]
        token = self.user.settings["password_reset"]["token"]
        url = f"/reset-password?rid={rid}&token={token}"
        self.assertIn(url, email.body)


class PasswordResetSigninTest(APITestCase):
    def setUp(self) -> None:
        self.token = signing.dumps({"email": "user@pount.org"}, salt="password-reset")
        self.reset_id = str(uuid.uuid4())
        self.user = User.objects.create_user(
            "user@pount.org",
            password="secret",
            email="user@pount.org",
            authentication_provider=AuthenticationProvider.LOCAL.value,
            settings={
                "password_reset": {
                    "reset_id": self.reset_id,
                    "token": self.token,
                }
            },
        )

    def make_request(self, data={}):
        response = self.client.post(
            reverse("login-password-reset-signin"),
            data,
            format="json",
        )
        return response

    def test_invalid_email_returns_forbidden(self):
        response = self.make_request({"token": signing.dumps({"email": "other@pount.org"}, salt="password-reset")})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_request_must_contain_token(self):
        response = self.make_request(
            {
                "rid": self.reset_id,
            }
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_request_must_contain_reset_id(self):
        response = self.make_request(
            {
                "token": self.token,
            }
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reset_returns_pair_of_jwt_tokens(self):
        response = self.make_request(
            {
                "token": self.token,
                "rid": self.reset_id,
                "new_password": "SecretPassword",
                "new_password2": "SecretPassword",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("refresh", response.data.keys())
        self.assertIn("access", response.data.keys())
