from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.libs import notifications
from pount.apps.api.models import Group, Project
from pount.apps.api.models.init import initialize_roles


class NotificationsTest(APITestCase):
    def setUp(self) -> None:
        initialize_roles()
        self.manager = get_user_model().objects.create_user("manager")
        self.user = get_user_model().objects.create_user("username")
        self.project = Project.objects.create(
            name="Project name",
            description="Project description",
            is_public=False,
        )
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        custom_group = Group.objects.create(
            name="Custom group",
            description="Custom group description",
            project=self.project,
            settings={},
        )
        custom_group.users.set([self.user, self.manager])
        self.notifications = [
            notifications.notify_group(
                self.manager,
                custom_group,
                "group_create",
            ),
            notifications.notify_group_membership(
                self.manager,
                self.user,
                custom_group,
                "group_user_add",
            ),
        ]

    def test_list_notifications(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("notification-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 1)

    def test_mark_notification_read(self):
        self.client.force_login(self.user)
        list_response = self.client.get(reverse("notification-list"))
        self.assertEqual(list_response.status_code, status.HTTP_200_OK)
        mark_read_response = self.client.patch(
            reverse(
                "notification-mark-read",
                kwargs={"pk": list_response.data["results"][0]["id"]},
            )
        )
        self.assertEqual(mark_read_response.status_code, status.HTTP_200_OK)
        self.assertEqual(mark_read_response.data["count"], 0)

    def test_mark_all_notifications_read_requires_authentication(self):
        response = self.client.patch(reverse("notification-mark-all-read"))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_mark_all_notifications_read_for_user(self):
        self.client.force_login(self.user)
        self.assertEqual(self.user.notifications.unread().count(), 1)
        response = self.client.patch(reverse("notification-mark-all-read"))
        self.assertEqual(self.user.notifications.unread().count(), 0)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
