from pathlib import Path
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.core import mail
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models import Permission, Role, Set, Template
from pount.apps.api.models.group import Group
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.models.project import Project
from pount.apps.api.views.mixins import ThumbnailMixin

User = get_user_model()


class SetUpMixin:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_roles()
        self.user = User.objects.create_user("username")
        self.creator = User.objects.create_user("creator", settings={"project_creator": True})
        self.manager = User.objects.create_user("manager")
        self.admin = User.objects.create_superuser("admin", email="admin@pount.org")
        self.project = Project.objects.create(
            name="Project name",
            description="Project description",
            is_public=False,
        )
        self.project.add_member(self.user)
        self.project.add_member(self.manager)
        self.project.add_member(self.creator)
        self.project.add_manager(self.manager)
        self.project.add_manager(self.creator)

        other_user = User.objects.create_user("other_user", settings={"project_creator": True})
        self.other_project = Project.objects.create(
            name="Other project name",
            description="Other project description",
            is_public=False,
        )
        self.other_project.add_member(other_user)
        self.other_project.add_manager(other_user)


class DeleteProjectTest(SetUpMixin, APITestCase):
    def _get_response(self, add_project_ids=[]):
        project_ids = [self.project.id]
        if add_project_ids:
            project_ids += add_project_ids
        return self.client.delete(
            reverse("project-request-deletion"),
            {"project_ids": project_ids},
            format="json",
        )

    def test_delete_request_needs_authentication(self):
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_delete_request_allowed_for_project_manager(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_request_sets_project_as_private(self):
        self.client.force_login(self.manager)
        self.project.is_public = True
        self.project.save()
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.project.refresh_from_db()
        self.assertFalse(self.project.is_public)

    def test_delete_request_sends_email_to_admin(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        outbox = mail.outbox
        self.assertEqual(len(outbox), 1)
        email = outbox[0]
        self.assertTemplateUsed(response, "email/project/request_deletion.txt")
        self.assertEqual(email.to, [self.admin.email])

    def test_delete_request_not_allowed_if_project_has_non_deleted_sets(self):
        self._add_set(False)
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "project_not_empty")

    def test_delete_request_allowed_if_project_has_deleted_set(self):
        self._add_set(True)
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_other_project_is_not_deleted(self):
        self.other_project.is_public = True
        self.other_project.save()
        self.client.force_login(self.manager)
        response = self._get_response([self.other_project.id])
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(len(mail.outbox), 1)

    def _add_set(self, deleted=False, project=None):
        project = project or self.project
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_templates()
            Set.objects.create(
                title="Set title",
                description="",
                creator=self.manager,
                is_deleted=deleted,
                project=project,
                template=Template.objects.first(),
            )


class CreateProjectTest(SetUpMixin, APITestCase):
    def create_project(self, user):
        self.client.force_login(user)
        response = self.client.post(
            reverse("project-list"),
            {
                "name": "Created project name",
                "description": "Project description",
                "is_public": False,
            },
            format="json",
        )
        return response

    def test_project_creator_can_create_a_project(self):
        response = self.create_project(self.creator)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        project = Project.objects.get(name="Created project name")
        self.assertEqual(response.data["id"], str(project.id))
        self.assertEqual(project.is_public, False)

    def test_creator_is_added_to_managers_group(self):
        self.create_project(self.creator)
        project = Project.objects.get(name="Created project name")
        self.assertIn(self.creator, project.managers)

    def test_creator_is_added_to_members_group(self):
        self.create_project(self.creator)
        project = Project.objects.get(name="Created project name")
        self.assertIn(self.creator, project.members)

    def test_admin_gets_notification(self):
        self.create_project(self.creator)
        project = Project.objects.get(name="Created project name")
        notifications = self.admin.notifications.unread()
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].project, project)
        self.assertEqual(notifications[0].actor, self.creator)

    def test_user_without_project_creator_privileges_cannot_create_project(
        self,
    ):
        self.creator.settings["project_creator"] = False
        self.creator.save()
        self.creator.refresh_from_db()
        response = self.create_project(self.creator)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_view_list_needs_authenticated_user(self):
        response = self.client.get(
            reverse("project-list"),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_of_projects_can_be_viewed_by_authenticated_user(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("project-list"),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def change_project(self, user):
        self.client.force_login(user)
        response = self.client.patch(
            reverse("project-detail", kwargs={"pk": self.project.id}),
            {
                "name": "New project name",
            },
            format="json",
        )
        return response

    def test_manager_can_edit_project(self):
        response = self.change_project(self.manager)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.project.refresh_from_db()
        self.assertEqual(self.project.name, "New project name")

    def test_managers_get_notification_on_project_edit(self):
        co_manager = User.objects.create_user("comanager")
        self.project.add_member(co_manager)
        self.project.add_manager(co_manager)
        self.change_project(self.manager)
        # Actor (manager) is not notified
        self.assertEqual(len(self.manager.notifications.unread()), 0)
        self.assertEqual(len(co_manager.notifications.unread()), 1)

    def test_member_cannot_edit_project(self):
        response = self.change_project(self.user)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def delete_project(self, user):
        self.client.force_login(user)
        response = self.client.delete(
            reverse("project-detail", kwargs={"pk": self.project.id}),
            format="json",
        )
        return response

    def test_creator_cannot_delete_project(self):
        response = self.delete_project(self.creator)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_manager_cannot_delete_project(self):
        response = self.delete_project(self.manager)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_member_cannot_delete_project(self):
        response = self.delete_project(self.user)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class ListProjectsForMembers(SetUpMixin, APITestCase):
    def test_user_can_list_projects_he_is_a_member_of(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("project-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["id"], str(self.project.id))

    def test_list_is_paginated(self):
        self.client.force_login(self.manager)
        response = self.client.get(reverse("project-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            ["count", "next_page", "previous_page", "results"],
            list(response.data.keys()),
        )

    def test_list_contains_only_projects_the_user_is_a_member_of(self):
        self.client.force_login(self.manager)
        project = Project.objects.create(
            name="Other project name",
            description="Project description",
            is_public=False,
        )
        response = self.client.get(reverse("project-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertNotIn(str(project.id), [_["id"] for _ in response.data["results"]])

    def test_list_does_not_contain_public_projects_the_user_is_not_a_member_of(
        self,
    ):
        self.client.force_login(self.manager)
        project = Project.objects.create(
            name="Other project name",
            description="Project description",
            is_public=True,
        )
        response = self.client.get(reverse("project-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertNotIn(str(project.id), [_["id"] for _ in response.data["results"]])

    def test_filter_project_on_is_manager(self):
        self.client.force_login(self.manager)
        response = self.client.get(f"{reverse('project-list')}?is_manager=true")
        self.assertEqual(len(response.data["results"]), 1)
        user = User.objects.create_user("not_manager")
        self.project.add_member(user)
        self.client.force_login(user)
        response = self.client.get(f"{reverse('project-list')}?is_manager=true")
        self.assertEqual(len(response.data["results"]), 0)
        response = self.client.get(f"{reverse('project-list')}?is_manager=false")
        self.assertEqual(len(response.data["results"]), 1)


class ProjectManagersTest(SetUpMixin, APITestCase):
    def test_project_retrieve_contains_list_of_project_managers(self):
        """Project details contains the list of the managers"""
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("project-detail", kwargs={"pk": self.project.pk}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["managers"]), 2)
        self.assertEqual(
            sorted(_["id"] for _ in response.data["managers"]),
            sorted([str(self.creator.id), str(self.manager.id)]),
        )


class ProjectGroupsTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.user1 = User.objects.create_user("user_one")
        self.user2 = User.objects.create_user("user_two")
        self.group_a = Group.objects.create_for_project(name="Group A", description="", project=self.project)
        self.group_a.users.add(self.user1, self.user2)
        self.group_b = Group.objects.create_for_project(name="Group B", description="", project=self.project)
        self.group_b.users.add(self.user1, self.user)
        self.project.add_member(self.user1)
        self.project.add_member(self.user2)

    def test_get_groups_requires_member_privilege(self):
        response = self.client.get(reverse("project-groups", kwargs={"pk": self.project.id}))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_groups_for_project_returns_groups_and_users(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("project-groups", kwargs={"pk": self.project.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], str(self.project.id))
        self.assertEqual(response.data["name"], self.project.name)
        self.assertEqual(len(response.data["groups"]), 4)
        self.assertListEqual(
            sorted(_["name"] for _ in response.data["groups"]),
            sorted(["Group A", "Group B", "Members", "Managers"]),
        )
        for group in response.data["groups"]:
            if group["name"] == "Members":
                self.assertListEqual(
                    sorted(_["name"] for _ in group["users"]),
                    sorted(
                        [
                            self.user1.name,
                            self.user2.name,
                            self.user.name,
                            self.manager.name,
                            self.creator.name,
                        ]
                    ),
                )
            if group["name"] == "Managers":
                self.assertListEqual(
                    sorted(_["name"] for _ in group["users"]),
                    sorted(
                        [
                            self.manager.name,
                            self.creator.name,
                        ]
                    ),
                )
            if group["name"] == "Group A":
                self.assertListEqual(
                    sorted(_["name"] for _ in group["users"]),
                    sorted([self.user1.name, self.user2.name]),
                )
            if group["name"] == "Group B":
                self.assertListEqual(
                    sorted(_["name"] for _ in group["users"]),
                    sorted([self.user1.name, self.user.name]),
                )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_groups_for_project_returns_sets_with_contribution_access(self, update_mock):
        self.client.force_login(self.user)
        initialize_templates()
        template = Template.objects.first()
        set = Set.objects.create(
            title="Set title",
            description="description",
            project=self.project,
            creator=self.user,
            publisher="Publisher",
            template=template,
        )
        Permission.objects.create(
            group=self.group_a,
            content_object=set,
            role=Role.objects.get_contributor(),
        )
        response = self.client.get(reverse("project-groups", kwargs={"pk": self.project.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], str(self.project.id))
        self.assertEqual(response.data["name"], self.project.name)
        self.assertEqual(len(response.data["groups"]), 4)
        group = [g for g in response.data["groups"] if g["name"] == self.group_a.name][0]
        self.assertEqual(len(group["sets"]), 1)
        self.assertEqual(group["sets"][0]["title"], set.title)
        self.assertListEqual(list(group["sets"][0].keys()), ["id", "title"])


class ListProjectTemplatesTest(SetUpMixin, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        super().setUp()
        initialize_templates()
        Template.objects.create(
            name="Local template",
            scope=Template.SCOPE_PROJECT,
            project=self.project,
            creator=self.manager,
        )
        other_project = Project.objects.create(name="Other project", description="Description", is_public=True)
        Template.objects.create(
            name="Other project template",
            scope=Template.SCOPE_PROJECT,
            project=other_project,
            creator=self.creator,
        )

    def test_compact_list_global_and_local_templates(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("project-list-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 2)
        self.assertListEqual(list(response.data[0].keys()), ["id", "name", "scope"])

    def test_detailed_list_of_local_templates(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("project-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)

    def test_must_be_project_member_to_list_templates(self):
        # unauthenticated user has no access
        response = self.client.get(
            reverse("project-list-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        # member ot another project has no access
        otheruser = User.objects.create_user("otheruser")
        self.other_project.add_member(otheruser)
        self.client.force_login(otheruser)
        response = self.client.get(
            reverse("project-list-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self.client.get(
            reverse("project-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class TestProjectThumbnail(SetUpMixin, APITestCase):
    @patch("pount.apps.api.models.mixins.ThumbnailUrlMixin.thumbnail_url")
    @patch("pount.apps.api.storage_backends.ThumbnailFileStorage.save")
    def test_upload(self, storage_mock, url_mock):
        storage_mock.return_value = "thumbnail_filename.png"
        url_mock.return_value = "http://s3.pount.org/generated_thumbnail_filename.png"
        self.client.force_login(self.creator)
        f = open(
            Path(__file__).resolve(strict=True).parent.parent / "fixtures/thumbnail.png",
            "rb",
        )
        response = self.client.post(
            reverse("project-thumbnail", kwargs={"pk": self.project.id}),
            {
                "file": f,
            },
        )
        f.close()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(storage_mock.call_count, len(ThumbnailMixin.SIZES))
        self.assertEqual(
            response.data["thumbnails"]["s"],
            "http://s3.pount.org/generated_thumbnail_filename.png",
        )
