from unittest.mock import patch

from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models import Item, Project, Set, Template
from pount.apps.api.models.init import initialize_roles, initialize_templates

User = get_user_model()


class TestSetup:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_templates()
        initialize_roles()
        self.user = User.objects.create_user("user")
        self.project = Project.objects.create(
            name="Project name",
            description="Project description",
            is_public=False,
        )
        self.set = Set.objects.create(
            title="Set title",
            description="",
            is_public=False,
            template=Template.objects.first(),
            creator=self.user,
            project=self.project,
        )
        self.item = Item.objects.create(
            title="Item title",
            description="",
            is_public=False,
            creator=self.user,
            metadata=[],
            set=self.set,
        )
        self.project.add_member(self.user)
        self.project.add_manager(self.user)


class SearchExportTest(TestSetup, APITestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_export_simple_search_view_is_authenticated(self):
        response = self.client.post(reverse("search-export"), format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_export_simple_search_view_returns_csv(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("search-export"),
            {"type": "project", "ids": str(self.project.id)},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response["content-type"], "text/csv")

    def test_export_simple_search_view_requires_ids_parameter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("search-export"),
            {"type": "project"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.content, b"IDs not given")

    def test_export_simple_search_view_requires_type_parameter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("search-export"),
            {"ids": str(self.project.id)},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.content, b"Not a valid type")

    def test_export_simple_search_view_for_set(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("search-export"),
            {"type": "set", "ids": str(self.set.id)},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_export_simple_search_view_for_item(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("search-export"),
            {"type": "item", "ids": str(self.item.id)},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_ids_must_be_uuids(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("search-export"),
            {"type": "project", "ids": "ab,cd,ef,3"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.content, b"Not valid UUIDs")
