import uuid
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.utils import timezone
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.libs.group import add_set_to_group
from pount.apps.api.models import Item, init
from pount.apps.api.models.permission import Group
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.tests.factories.template import TemplateFactory

User = get_user_model()


class SetupSetsMixin:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        init.initialize_templates()
        init.initialize_roles()
        self.template = Template.objects.get(name="datacite", scope=Template.SCOPE_BASE)
        self.project = Project.objects.create(
            name="Project",
            description="Description",
        )
        self.creator = User.objects.create_user("creator")
        self.project.add_member(self.creator)
        self.manager = User.objects.create_user("manager")
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)
        self.set = Set.objects.create(
            title="Set title",
            description="Set description",
            publisher="Unistra",
            template=TemplateFactory(),
            project=self.project,
            creator=self.creator,
        )


class ListSetTest(SetupSetsMixin, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        super().setUp()
        self.other_user = User.objects.create_user("other_user")
        self.public_set = Set.objects.create(
            title="Public set title",
            description="Public set description",
            publisher="Unistra",
            is_public=True,
            template=TemplateFactory(),
            project=self.project,
            creator=self.creator,
        )
        self.private_set = Set.objects.create(
            title="Private set title",
            description="Private set description",
            publisher="Unistra",
            is_public=False,
            template=TemplateFactory(),
            project=self.project,
            creator=self.creator,
        )

    def test_list_sets_needs_authentication(self):
        self.client.logout()
        response = self.client.get(reverse("set-list"), format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_items(self):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-list"), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_tiems_is_paginated(self):
        self.client.force_login(self.creator)
        response = self.client.get(f"{reverse('set-list')}?page_size=2", format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 3)
        self.assertEqual(response.data["next_page"], 2)
        self.assertEqual(len(response.data["results"]), 2)

    def test_non_member_user_can_list_public_sets(self):
        self.client.force_login(self.other_user)
        response = self.client.get(reverse("set-list"), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn(
            str(self.public_set.id),
            [_["id"] for _ in response.data["results"]],
        )

    def test_non_member_user_can_not_list_private_sets(self):
        self.client.force_login(self.other_user)
        response = self.client.get(reverse("set-list"), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertNotIn(
            str(self.private_set.id),
            [_["id"] for _ in response.data["results"]],
        )


class CreateSetTest(SetupSetsMixin, APITestCase):
    def _get_create_response(self, **kwargs):
        self.project.add_member(self.creator)
        self.project.add_manager(self.creator)
        self.client.force_login(self.creator)
        data = {
            "title": "Set title",
            "description": "Set description",
            "publisher": "Set publisher",
            "from_template_id": self.template.id,
            "project_id": self.project.id,
        }
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            response = self.client.post(
                reverse("set-list"),
                data={**data, **kwargs},
                format="json",
            )
        return response

    def test_manager_can_create_set(self):
        response = self._get_create_response()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["creator"]["id"], str(self.creator.id))

    def test_create_set_with_copy_of_global_template(self):
        response = self._get_create_response()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            response.data["template"]["name"],
            f"Template {response.data['template']['id']} of {response.data['title']}",  # noqa: E501
        )
        created_template = Template.objects.get(pk=response.data["template"]["id"])
        self.assertEqual(
            created_template.project_id,
            self.project.id,
        )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_create_set_with_copy_of_project_template(self, update_mock):
        template = Template.objects.create(
            name="Project template",
            scope=Template.SCOPE_PROJECT,
            project=self.project,
            fields=[{"test": "test"}],
            messages={"test": "test"},
        )
        response = self._get_create_response(from_template_id=template.id)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            response.data["template"]["name"],
            f"Template {response.data['template']['id']} of {response.data['title']}",  # noqa: E501
        )
        cloned_template = Template.objects.get(pk=response.data["template"]["id"])
        self.assertEqual(template.fields, cloned_template.fields)
        self.assertEqual(template.messages, cloned_template.messages)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_cannot_create_set_with_clone_of_template_of_another_project(self, update_mock):
        project = Project.objects.create(name="Other project", description="Description")
        template = Template.objects.create(
            name="Project template",
            scope=Template.SCOPE_PROJECT,
            project=project,
        )
        response = self._get_create_response(from_template_id=template.id)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data["from_template_id"]), "Invalid template.")

    def test_cannot_clone_non_existent_template(self):
        Project.objects.create(name="Other project", description="Description")
        response = self._get_create_response(from_template_id=str(uuid.uuid4()))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data["from_template_id"]), "Invalid template.")

    def test_project_member_cannot_create_set(self):
        self.client.force_login(self.creator)
        response = self.client.post(
            reverse("set-list"),
            {
                "title": "Set title",
                "description": "Set description",
                "publisher": "Set publisher",
                "template_id": self.template.id,
                "project_id": self.project.id,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_set_requires_valid_project(self):
        self.project.add_member(self.creator)
        self.project.add_manager(self.creator)
        self.client.force_login(self.creator)
        response = self.client.post(
            reverse("set-list"),
            {
                "title": "Set title",
                "description": "Set description",
                "publisher": "Set publisher",
                "template_id": self.template.id,
                "project_id": str(uuid.uuid4()),
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class UpdateSetTest(SetupSetsMixin, APITestCase):
    def test_project_member_cannot_update_set(self):
        self.client.force_login(self.creator)
        response = self.client.patch(
            reverse("set-detail", kwargs={"pk": self.set.id}),
            {
                "title": "New set title",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_manager_can_update_set(self):
        self.project.add_member(self.creator)
        self.project.add_manager(self.creator)
        self.client.force_login(self.creator)
        response = self.client.patch(
            reverse("set-detail", kwargs={"pk": self.set.id}),
            {"title": "New set title"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.set.refresh_from_db()
        self.assertEqual(self.set.title, "New set title")

    def test_project_id_can_not_be_changed(self):
        """
        A Set can not be moved from one Project into another Project
        """
        other_project = Project.objects.create(
            name="Other project title",
            description="Other project description",
        )
        other_project.add_member(self.creator)
        other_project.add_manager(self.creator)
        self.project.add_member(self.creator)
        self.project.add_manager(self.creator)
        self.client.force_login(self.creator)
        response = self.client.patch(
            reverse("set-detail", kwargs={"pk": self.set.id}),
            {"title": "New set title", "project_id": other_project.id},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.set.refresh_from_db()
        self.assertEqual("New set title", self.set.title)
        self.assertNotEqual(self.set.project_id, other_project.id)
        self.assertEqual(self.set.project_id, self.project.id)


class ViewSetTest(SetupSetsMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.viewer = User.objects.create_user("viewer")

    def test_anonymous_user_can_not_view_published_set(self):
        response = self.client.get(
            reverse("set-detail", kwargs={"pk": self.set.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_authenticated_user_can_view_published_set(self):
        self.set.is_public = True
        self.set.save()
        self.client.force_login(self.viewer)
        response = self.client.get(
            reverse("set-detail", kwargs={"pk": self.set.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_authenticated_user_can_not_view_unpublished_set(self):
        self.set.is_public = False
        self.set.save()
        self.client.force_login(self.viewer)
        response = self.client.get(
            reverse("set-detail", kwargs={"pk": self.set.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_project_member_can_view_published_set(self):
        self.project.add_member(self.viewer)
        self.client.force_login(self.viewer)
        self.set.is_public = True
        self.set.save()
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            sorted(list(response.data.keys())),
            sorted(
                [
                    "id",
                    "title",
                    "description",
                    "is_public",
                    "publication_date",
                    "online_at",
                    "template",
                    "publisher",
                    "creator",
                    "can_contribute",
                    "project",
                    "created_at",
                    "updated_at",
                    "thumbnails",
                    "num_items",
                    "num_public_items",
                ]
            ),
        )

    def test_project_member_can_view_unpublished_set(self):
        self.project.add_member(self.viewer)
        self.client.force_login(self.viewer)
        self.set.is_public = False
        self.set.save()
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_project_info_indicates_if_user_is_manager(self):
        self.project.add_member(self.creator)
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertFalse(response.data["project"]["is_manager"])
        self.project.add_manager(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertTrue(response.data["project"]["is_manager"])

    def test_manager_has_can_contributor_flag_set(self):
        self.project.add_manager(self.creator)
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertTrue(response.data["can_contribute"])

    def test_member_has_can_contributor_flag_not_set(self):
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertFalse(response.data["can_contribute"])

    def test_contributor_has_can_contribute_flag_set(self):
        group = Group.objects.create(
            name="Contributors",
            project=self.project,
            settings=[],
        )
        group.users.add(self.creator)
        add_set_to_group(group, self.set)
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertTrue(response.data["can_contribute"])

    def test_list_serializer_tells_if_set_has_items(self):
        Item.objects.create(
            title="Item title",
            description="",
            set=self.set,
            creator=self.creator,
            metadata=[],
        )
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertEqual(response.data["num_items"], True)

    def test_list_serializer_tells_if_set_has_public_items(self):
        item = Item.objects.create(
            title="Item title",
            description="",
            set=self.set,
            creator=self.creator,
            is_public=True,
            metadata=[],
        )
        self.client.force_login(self.creator)
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertEqual(response.data["num_public_items"], True)

        item.is_public = False
        item.save(update_fields=["is_public"])
        item.refresh_from_db()
        response = self.client.get(reverse("set-detail", kwargs={"pk": self.set.id}), format="json")
        self.assertEqual(response.data["num_public_items"], False)


class DeleteSetTest(SetupSetsMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def get_response(self, update_mock):
        return self.client.delete(reverse("set-detail", kwargs={"pk": self.set.id}))

    def test_delete_set_needs_authentication(self):
        self.client.logout()
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_delete_set_not_allowed_for_project_member(self):
        member = User.objects.create_user("member")
        self.project.add_member(member)
        self.client.force_login(member)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_set_not_allowed_for_creator(self):
        self.client.force_login(self.creator)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_set_allowed_for_manager(self):
        self.client.force_login(self.manager)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_deleted_item_not_returned_by_default_manager(self):
        self.client.force_login(self.manager)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Set.DoesNotExist):
            Set.objects.get(pk=self.set.id)

    def test_deleted_set_is_soft_deleted(self):
        self.client.force_login(self.manager)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.set.refresh_from_db()
        self.assertTrue(self.set.is_deleted)
        self.assertIsNotNone(self.set.deleted_at)

    def test_non_empty_set_cannot_be_deleted(self):
        Item.objects.create(
            title="Item title",
            description="",
            set=self.set,
            creator=self.creator,
            metadata=[],
        )
        self.client.force_login(self.manager)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "set_not_empty")

    def test_set_with_deleted_item_can_be_deleted(self):
        Item.objects.create(
            title="Deleted item",
            description="",
            set=self.set,
            creator=self.creator,
            metadata=[],
            is_deleted=True,
            deleted_at=timezone.now(),
        )
        self.client.force_login(self.manager)
        response = self.get_response()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_bulk_delete_set_is_soft_deleted(self, update_mock):
        self.client.force_login(self.manager)

        response = self.client.delete(reverse("set-list"), {"set_ids": [self.set.id]}, format="json")
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        deleted_set = Set.deleted_objects.get(pk=self.set.pk)
        self.assertEqual(self.set.id, deleted_set.id)

    def test_bulk_delete_set_fails_if_set_contains_item(self):
        self.client.force_login(self.manager)
        Item.objects.create(
            title="Item title",
            description="",
            metadata=[],
            creator=self.creator,
            set=self.set,
        )
        response = self.client.delete(
            reverse("set-list"),
            {
                "set_ids": [self.set.id],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["error"], "set_not_empty")
