import io
from unittest.mock import patch

from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.tests.factories.set import SetForManagerFactory
from pount.apps.api.tests.factories.template import TemplateForSetFactory
from pount.apps.api.tests.factories.user import UserFactory

User = get_user_model()


class SetUpMixin:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_roles()
        self.user = User.objects.create_user("user")
        self.member = User.objects.create_user("member")
        self.manager = User.objects.create_user("manager")
        self.project = Project.objects.create(name="Project name", description="Description")
        self.project.add_member(self.member)
        self.project.add_member(self.manager)
        self.project.add_manager(self.manager)


class CreateTemplateTest(SetUpMixin, APITestCase):
    def _make_create_response(self, **kwargs):
        self.client.force_login(self.manager)
        data = {
            "name": "Template name",
            "fields": [],
            "messages": {},
            "private_fields": ["private_field"],
            "scope": None,
            "project_id": self.project.id,
        }

        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            response = self.client.post(
                reverse("template-list"),
                {**data, **kwargs},
                format="json",
            )
        return response

    def test_manager_can_create_template(self):
        response = self._make_create_response(scope=Template.SCOPE_PROJECT)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_project_template_requires_template(self):
        response = self._make_create_response(
            project_id=None,
            scope=Template.SCOPE_PROJECT,
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_non_global_template_must_be_linked_to_project(self):
        response = self._make_create_response(project_id=None, scope=Template.SCOPE_PROJECT)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        response = self._make_create_response(project_id=None, scope=Template.SCOPE_SET)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_template_names_must_be_unique(self, update_mock):
        Template.objects.create(
            name="Template name",
            scope=Template.SCOPE_GLOBAL,
        )
        response = self._make_create_response(project_id=self.project.id, scope=Template.SCOPE_PROJECT)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_manager_cannot_update_global_template(self, update_mock):
        self.client.force_login(self.manager)
        template = Template.objects.create(
            name="Template",
            scope=Template.SCOPE_GLOBAL,
        )
        response = self.client.patch(
            reverse("template-detail", kwargs={"pk": template.id}),
            {"name": "New name"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_serialized_fields_in_response(self):
        self.client.force_login(self.manager)
        response = self._make_create_response(scope=Template.SCOPE_PROJECT)
        self.assertListEqual(
            sorted(list(response.data.keys())),
            sorted(
                [
                    "id",
                    "name",
                    "fields",
                    "messages",
                    "private_fields",
                    "created_at",
                    "updated_at",
                    "scope",
                    "project",
                    "creator",
                ]
            ),
        )


class ListGlobalTemplatesTest(SetUpMixin, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        super().setUp()
        initialize_templates()
        self.project_template = Template.objects.create(
            name="Project template",
            project=self.project,
            scope=Template.SCOPE_PROJECT,
            creator=self.user,
        )
        self.set_template = Template.objects.create(
            name="Set template",
            project=self.project,
            scope=Template.SCOPE_SET,
            creator=self.user,
        )

    def test_list_global_templates(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("template-list"),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)
        self.assertListEqual(sorted(_["name"] for _ in response.data), ["datacite"])

    def test_list_global_template_needs_authentication(self):
        self.client.logout()
        response = self.client.get(
            reverse("template-list"),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_project_templates(self):
        self.client.force_login(self.manager)
        response = self.client.get(
            reverse("project-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 2)
        self.assertListEqual(
            sorted(_["name"] for _ in response.data),
            ["Project template", "Set template"],
        )

    def test_list_project_templates_needs_project_membership(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("project-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.project.add_member(self.user)
        response = self.client.get(
            reverse("project-templates", kwargs={"pk": self.project.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_retrieve_global_template(self):
        self.client.force_login(self.user)
        datacite = Template.objects.get(name="datacite")
        response = self.client.get(
            reverse("template-detail", kwargs={"pk": datacite.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["name"], "datacite")

    def test_retrieve_project_template(self):
        self.client.force_login(self.member)
        response = self.client.get(
            reverse("template-detail", kwargs={"pk": self.set_template.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["name"], self.set_template.name)

    def test_retrieve_project_template_needs_project_membership(self):
        other_user = User.objects.create_user("other_user")
        other_project: Project = Project.objects.create(name="Other project", description="Description")
        other_project.add_member(other_user)
        self.client.force_login(other_user)
        response = self.client.get(
            reverse("template-detail", kwargs={"pk": self.set_template.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class UpdateTemplateTest(SetUpMixin, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        super().setUp()
        self.superadmin = User.objects.create_superuser("admin")
        self.template = Template.objects.create(
            name="template",
            scope=Template.SCOPE_PROJECT,
            project=self.project,
        )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_global_templates_can_not_be_modified(self, update_mock):
        initialize_templates()
        self.client.force_login(self.manager)
        datacite = Template.objects.get(name="datacite")
        response = self.client.patch(
            reverse("template-detail", kwargs={"pk": datacite.id}),
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_only_manager_can_update_local_template(self, update_mock):
        self.client.force_login(self.member)
        response = self.client.patch(
            reverse("template-detail", kwargs={"pk": self.template.id}),
            {"name": "New name"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.client.force_login(self.manager)
        response = self.client.patch(
            reverse("template-detail", kwargs={"pk": self.template.id}),
            {"name": "New name"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["name"], "New name")

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_scope_is_not_updated(self, update_mock):
        self.client.force_login(self.manager)
        response = self.client.patch(
            reverse("template-detail", kwargs={"pk": self.template.id}),
            {"scope": Template.SCOPE_SET},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["scope"], Template.SCOPE_PROJECT)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_datacite_template_can_not_be_updated(self, update_mock):
        initialize_templates()
        self.client.force_login(self.manager)
        datacite = Template.objects.get(name="datacite")
        response = self.client.patch(
            reverse("template-detail", kwargs={"pk": datacite.id}),
            {"name": "Datacite with a new name?"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class DeleteTemplateTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.superadmin = User.objects.create_superuser("admin")

        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            self.template = Template.objects.create(
                name="template",
                scope=Template.SCOPE_PROJECT,
                project=self.project,
            )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_datacite_template_can_not_be_deleted(self, update_mock):
        initialize_templates()
        self.client.force_login(self.superadmin)
        datacite = Template.objects.get(name="datacite")
        response = self.client.delete(reverse("template-detail", kwargs={"pk": datacite.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_superadmin_can_delete_global_template(self):
        self.template.scope = Template.SCOPE_GLOBAL
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            self.template.save()
        self.client.force_login(self.superadmin)
        response = self.client.delete(reverse("template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Template.DoesNotExist):
            self.template.refresh_from_db()

    def test_superadmin_can_delete_base_template(self):
        self.template.scope = Template.SCOPE_BASE
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            self.template.save()
        self.client.force_login(self.superadmin)
        response = self.client.delete(reverse("template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Template.DoesNotExist):
            self.template.refresh_from_db()

    def test_member_can_not_delete_project_template(self):
        self.client.force_login(self.member)
        response = self.client.delete(reverse("template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_manager_can_not_delete_template_if_used_in_set(self):
        Set.objects.create(
            title="Set title",
            description="description",
            project=self.project,
            template=self.template,
            creator=self.manager,
        )
        self.client.force_login(self.manager)
        response = self.client.delete(reverse("template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_manager_can_delete_template_if_not_used_in_set(self):
        self.client.force_login(self.manager)
        response = self.client.delete(reverse("template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)


class ExportTemplateAsCSVTest(APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_roles()
        initialize_templates()
        datacite = Template.objects.get(name="datacite")
        self.member = UserFactory(username="member")
        self.manager = UserFactory(username="manager")
        set = SetForManagerFactory(members=[self.member, self.manager], managers=self.manager)
        self.template = TemplateForSetFactory(fields=datacite.fields, creator=self.manager, set=set)

    def _get_response(self):
        response = self.client.get(reverse("template-export-as-csv", kwargs={"pk": str(self.template.id)}))
        return response

    def test_view_is_authenticated(self):
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_view_is_forbidden_for_members(self):
        self.client.force_login(self.member)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_view_is_ok_for_manager(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_view_returns_csv_file(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(response["Content-Type"], "text/csv")

    def test_downloadable_csv_filename_format(self):
        self.client.force_login(self.manager)
        response = self._get_response()
        self.assertEqual(
            response["Content-Disposition"],
            f'attachment; filename="{self.template.name}.csv"',
        )

    def test_fieldnames_in_csv(self):
        self.client.force_login(self.manager)
        with patch("pount.apps.api.libs.template.get_template_fieldnames") as get_fieldnames:
            get_fieldnames.return_value = {"id1": "Field A", "id2": "Field B"}
            response = self._get_response()
            body = io.BytesIO(response.content)
            data1 = body.readline().rstrip().decode()
            data2 = body.readline().rstrip().decode()
            self.assertEqual(get_fieldnames.call_count, 1)
            # First line contains the fields' ids
            self.assertEqual("id1,id2", data1)
            # Second line contains the fields' names
            self.assertEqual("Field A,Field B", data2)
