import uuid
from io import BytesIO
from pathlib import Path
from unittest.mock import patch

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models.init import initialize_roles
from pount.apps.api.models.item import MediaFile
from pount.apps.api.tests.factories.item import ItemFactory, ItemForMemberFactory
from pount.apps.api.tests.factories.user import UserFactory


class RITFileUploadTest(APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_roles()
        self.member = UserFactory(username="member")
        self.item = ItemForMemberFactory(members=self.member, creator=self.member)
        self.client.force_login(self.member)

    def test_upload_view_is_authenticated(self):
        self.client.logout()
        file = BytesIO()
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(self.item.pk)}),
            {
                "file": file,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_user_must_be_project_member(self):
        file = BytesIO()
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            item = ItemFactory()
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(item.pk)}),
            {
                "file": file,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def _get_file(self, filename):
        file = open(
            Path(__file__).resolve(strict=True).parent.parent / f"fixtures/rti/{filename}.zip",
            "rb",
        )
        return file

    def test_uploaded_file_must_contain_info_json(self):
        file = self._get_file("file-without-infojson")
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(self.item.pk)}),
            {
                "file": file,
            },
        )
        file.close()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data, "Missing info.json")

    def test_upload_request_must_contain_file(self):
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(self.item.pk)}),
            {},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data, "Missing file")

    def test_item_id_must_be_valid(self):
        file = self._get_file("file")
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(uuid.uuid4())}),
            {"file": file},
        )
        file.close()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_uploaded_file_must_be_a_zipfile(self):
        file = open(
            Path(__file__).resolve(strict=True).parent.parent / "fixtures/pount.jpg",
            "rb",
        )
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(self.item.pk)}),
            {
                "file": file,
            },
        )
        file.close()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data, "File is not a zip file")

    @patch("pount.apps.api.models.item.MediaFile.delete_file")
    @patch("pount.apps.api.libs.file.ItemMediaFileStorage")
    def test_previous_files_are_deleted(self, storage, delete_file):
        file_obj = MediaFile.objects.create(filename="rti/filename.ext", item=self.item, info={"is_rti": True})

        file = self._get_file("file")
        response = self.client.post(
            reverse("mediafile-upload-rti", kwargs={"items_pk": str(self.item.pk)}),
            {
                "file": file,
            },
        )
        file.close()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(storage.return_value.save.call_count, 3)
        with self.assertRaises(MediaFile.DoesNotExist):
            file_obj.refresh_from_db()
        self.assertEqual(delete_file.call_count, 1)

    def test_destroy_view_is_authenticated(self):
        self.client.logout()
        mediafile = MediaFile.objects.create(filename="rti/filename.ext", item=self.item, info={"is_rti": True})
        response = self.client.delete(
            reverse(
                "mediafile-detail",
                kwargs={
                    "items_pk": str(self.item.id),
                    "pk": str(mediafile.pk),
                },
            )
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_destroy_view_deletes_files(self):
        infojson = MediaFile.objects.create(
            filename="rti/info.json",
            item=self.item,
            info={"is_rti": True},
            is_hidden=False,
        )
        mediafile = MediaFile.objects.create(
            filename="rti/file.ext",
            item=self.item,
            info={"is_rti": True},
            is_hidden=True,
        )
        with patch("pount.apps.api.models.item.MediaFile.delete_file") as delete_file:
            response = self.client.delete(
                reverse(
                    "mediafile-detail",
                    kwargs={
                        "items_pk": str(self.item.id),
                        "pk": str(infojson.pk),
                    },
                )
            )
            self.assertEqual(delete_file.call_count, 2)
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(MediaFile.DoesNotExist):
            infojson.refresh_from_db()
        with self.assertRaises(MediaFile.DoesNotExist):
            mediafile.refresh_from_db()

    def test_download_uploaded_file(self):
        mediafile = MediaFile.objects.create(
            filename="rti/info.json",
            item=self.item,
            info={"is_rti": True},
            is_hidden=False,
        )
        with patch("pount.apps.api.views.item.ItemMediaFileStorage") as storage:
            storage.return_value.open.return_value = BytesIO()
            response = self.client.get(
                reverse(
                    "item-get-rti",
                    kwargs={
                        "pk": str(self.item.pk),
                        "filename": mediafile.filename.split("/")[1],
                    },
                )
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(storage.return_value.open.call_count, 1)
