from unittest.mock import patch

from django.test import TestCase

from pount.apps.api.libs.metadata import Metadata, MetadataEntry, TemplateEntry
from pount.apps.api.models import Template
from pount.apps.api.models.init import initialize_templates
from pount.exceptions import CSVMetadataError


class MetadataTest(TestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_templates()
        self.datacite = Template.objects.get(name="datacite")
        self.referential_metadata = {
            "id": "549cbqdwfu5",
            "name": "ReferentialSelect",
            "exist": True,
            "visible": True,
            "level_id": "702c24c6-55b6-46aa-8c88-e7a3c4afff36",
            "component": "ReferentialSelect",
            "parent_id": "bht7hxtbmb",
            "referential_id": "fa289e2a-07dd-4faf-a75e-1f1669e3d9fd",
        }
        self.multiple_fields = [
            {
                "id": "2qwhbanmbsc",
                "hint": "The general type of a resource.",
                "name": "resourceTypeGeneral",
                "exist": True,
                "label": "resourceTypeGeneral",
                "options": [
                    {
                        "id": "zl8ouisyt1q",
                        "label": "Audiovisual",
                        "owner": "2qwhbanmbsc",
                        "value": "zl8ouisyt1q",
                    },
                    {
                        "id": "6n2eecqvnsd",
                        "label": "Collection",
                        "owner": "2qwhbanmbsc",
                        "value": "6n2eecqvnsd",
                    },
                    {
                        "id": "uj938jmk2rl",
                        "label": "DataPaper",
                        "owner": "2qwhbanmbsc",
                        "value": "uj938jmk2rl",
                    },
                    {
                        "id": "eais7vwztb9",
                        "label": "Dataset",
                        "owner": "2qwhbanmbsc",
                        "value": "eais7vwztb9",
                    },
                    {
                        "id": "q2duprhhsob",
                        "label": "Event",
                        "owner": "2qwhbanmbsc",
                        "value": "q2duprhhsob",
                    },
                    {
                        "id": "qrg4713tgvr",
                        "label": "Image",
                        "owner": "2qwhbanmbsc",
                        "value": "qrg4713tgvr",
                    },
                    {
                        "id": "a2pero9dchm",
                        "label": "InteractiveResource",
                        "owner": "2qwhbanmbsc",
                        "value": "a2pero9dchm",
                    },
                    {
                        "id": "2596bsai2tu",
                        "label": "Model",
                        "owner": "2qwhbanmbsc",
                        "value": "2596bsai2tu",
                    },
                    {
                        "id": "u0cz9bvinwl",
                        "label": "PhysicalObject",
                        "owner": "2qwhbanmbsc",
                        "value": "u0cz9bvinwl",
                    },
                    {
                        "id": "rkmvtr3wosq",
                        "label": "Service",
                        "owner": "2qwhbanmbsc",
                        "value": "rkmvtr3wosq",
                    },
                    {
                        "id": "4wi5c7baejl",
                        "label": "Software",
                        "owner": "2qwhbanmbsc",
                        "value": "4wi5c7baejl",
                    },
                    {
                        "id": "tsheh0mmsc",
                        "label": "Sound",
                        "owner": "2qwhbanmbsc",
                        "value": "tsheh0mmsc",
                    },
                    {
                        "id": "e6esmnvs0rj",
                        "label": "Text",
                        "owner": "2qwhbanmbsc",
                        "value": "e6esmnvs0rj",
                    },
                    {
                        "id": "lqg41cmnwf",
                        "label": "Workflow",
                        "owner": "2qwhbanmbsc",
                        "value": "lqg41cmnwf",
                    },
                    {
                        "id": "gtz6mcevk2",
                        "label": "Other",
                        "owner": "2qwhbanmbsc",
                        "value": "gtz6mcevk2",
                    },
                ],
                "visible": True,
                "multiple": True,
                "parentId": "g8mrg14u4ji",
                "required": True,
                "component": "UsfSelect",
                "parent_id": "g8mrg14u4ji",
                "protected": True,
            }
        ]
        self.csv_row = {
            "6qgq1netsxh": "Creator",
            "5vxode5mjsh": "Item's title",
            "unym14mu68": "Publisher",
            "m1w8jje0wsh": "2021",
            "yxuullxsicj": "video",
            "2qwhbanmbsc": "Audiovisual",
        }

    def test_invalid_option_for_string(self):
        md = Metadata([], template_fields=self.datacite.fields)
        self.csv_row["2qwhbanmbsc"] = "Invalid"
        with self.assertRaises(CSVMetadataError) as ctx:
            md.from_csv_row(self.csv_row, 3)
        self.assertEqual(ctx.exception.line, 3)
        self.assertEqual(
            ctx.exception.error,
            "Value <Invalid> not found in option list for <resourceTypeGeneral>",
        )

    def test_valid_option_for_multiple_option(self):
        self.multiple_fields[0]["multiple"] = True
        self.multiple_fields[0]["repeatable"] = False
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Other|Audiovisual"
        result = md.from_csv_row(self.csv_row, 2)
        self.assertListEqual(result[0]["value"], ["gtz6mcevk2", "zl8ouisyt1q"])

    def test_is_option_is_set_if_option_field(self):
        self.multiple_fields[0]["multiple"] = False
        self.multiple_fields[0]["repeatable"] = False
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Audiovisual"
        result = md.from_csv_row(self.csv_row, 1)
        self.assertTrue(result[0]["is_option"])

    def test_invalid_option_for_multiple_option(self):
        self.multiple_fields[0]["multiple"] = True
        self.multiple_fields[0]["repeatable"] = False
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Otter|Audiovisual"
        with self.assertRaises(CSVMetadataError) as ctx:
            md.from_csv_row(self.csv_row, 4)
        self.assertEqual(ctx.exception.line, 4)
        self.assertEqual(ctx.exception.error, "Value <Otter> not found in option list for <resourceTypeGeneral>")

    def test_valid_option_for_repeatable_option(self):
        self.multiple_fields[0]["multiple"] = False
        self.multiple_fields[0]["repeatable"] = True
        self.multiple_fields[0]["max_repeat"] = 8
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Other;Audiovisual"
        result = md.from_csv_row(self.csv_row, 1)
        self.assertListEqual(result[0]["value"], ["gtz6mcevk2", "zl8ouisyt1q"])

    def test_invalid_option_for_repeatable_field(self):
        self.multiple_fields[0]["multiple"] = False
        self.multiple_fields[0]["repeatable"] = True
        self.multiple_fields[0]["max_repeat"] = 8
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Otter;Audiovisual"
        with self.assertRaises(CSVMetadataError) as ctx:
            md.from_csv_row(self.csv_row, 8)
        self.assertEqual(ctx.exception.line, 8)
        self.assertEqual(ctx.exception.error, "Value <Otter> not found in option list for <resourceTypeGeneral>")

    def test_valid_option_for_multiple_repeatable_option(self):
        self.multiple_fields[0]["multiple"] = True
        self.multiple_fields[0]["repeatable"] = True
        self.multiple_fields[0]["max_repeat"] = 8
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Other|Image;Audiovisual"
        result = md.from_csv_row(self.csv_row, 1)
        self.assertListEqual(result[0]["value"][0], ["gtz6mcevk2"])
        self.assertListEqual(result[0]["value"][1], ["qrg4713tgvr", "zl8ouisyt1q"])

    def test_invalid_option_for_multiple_repeatable_option(self):
        self.multiple_fields[0]["multiple"] = True
        self.multiple_fields[0]["repeatable"] = True
        self.multiple_fields[0]["max_repeat"] = 8
        md = Metadata([], template_fields=self.multiple_fields)
        self.csv_row["2qwhbanmbsc"] = "Otter|Image;Audiovisual"
        with self.assertRaises(CSVMetadataError) as ctx:
            md.from_csv_row(self.csv_row, 1)
        self.assertEqual(ctx.exception.line, 1)
        self.assertEqual(ctx.exception.error, "Value <Otter> not found in option list for <resourceTypeGeneral>")

    def test_metadata_has_length_value(self):
        m = Metadata([], template_fields=self.datacite.fields)
        self.assertEqual(6, len(m))
        self.assertEqual(6, len(m.template_fields))

    def test_metadata_entry_is_an_option(self):
        m = Metadata([], template_fields=self.datacite.fields)
        resource_type_general = m.get_entry_by_name("resourceTypeGeneral")
        self.assertTrue(resource_type_general.is_option)
        resource_type = m.get_entry_by_name("resourceType")
        self.assertFalse(resource_type.is_option)

    def test_metadata_entry_is_a_referential(self):
        template_fields = self.datacite.fields
        template_fields.append(self.referential_metadata)
        m = Metadata([], template_fields=template_fields)
        entry = m.get_entry_by_name("ReferentialSelect")
        self.assertTrue(entry.is_referential_value)

        creator = m.get_entry_by_name("creator")
        self.assertFalse(creator.is_referential_value)

    def test_referential_id_of_template_is_returned_if_not_set_in_metadata(self):
        md, entry = self._get_referential_entry()
        self.assertEqual(entry.referential_id, "fa289e2a-07dd-4faf-a75e-1f1669e3d9fd")
        self.assertEqual(entry.level_id, "702c24c6-55b6-46aa-8c88-e7a3c4afff36")

    def _get_referential_entry(self) -> MetadataEntry:
        template_fields = [
            {
                "id": "y7bm6598g8p",
                "name": "Maison 2",
                "exist": True,
                "fields": [
                    {
                        "id": "d8e60vk3f19",
                        "name": "Pièce",
                        "exist": True,
                        "visible": True,
                        "level_id": "702c24c6-55b6-46aa-8c88-e7a3c4afff36",
                        "component": "ReferentialSelect",
                        "parent_id": "y7bm6598g8p",
                        "referential_id": "fa289e2a-07dd-4faf-a75e-1f1669e3d9fd",
                    },
                    {
                        "id": "yvqlemx6w",
                        "name": "Meuble",
                        "exist": True,
                        "visible": True,
                        "level_id": "c16c3cb2-9364-434e-8b4d-6aff94382f7e",
                        "component": "ReferentialSelect",
                        "parent_id": "y7bm6598g8p",
                        "referential_id": "fa289e2a-07dd-4faf-a75e-1f1669e3d9fd",
                    },
                ],
                "visible": True,
                "component": "ReferentialComponent",
                "parent_id": "root",
                "referential_id": "fa289e2a-07dd-4faf-a75e-1f1669e3d9fd",
            }
        ]
        md = Metadata([], template_fields=template_fields)
        entry = md.get_entry_by_id("d8e60vk3f19")
        return md, entry

    def test_lookup_option_id_for_referential_value(self):
        md, entry = self._get_referential_entry()
        with patch(
            "pount.apps.api.libs.referential.get_referential_node_id_for_node_value"
        ) as get_referential_node_id_for_node_value:
            get_referential_node_id_for_node_value.return_value = "ac6200f4-d381-4b4b-8add-b7b03c73413c"
            self.assertEqual(
                "ac6200f4-d381-4b4b-8add-b7b03c73413c", md.lookup_referential_node_id_from_value(entry, "node value")
            )

    def test_lookup_option_id_for_option_value(self):
        md = Metadata([], template_fields=self.datacite.fields)
        entry = md.get_entry_by_id("2qwhbanmbsc")
        self.assertEqual("gtz6mcevk2", md.lookup_option_id_from_value(entry, "Other"))

    def test_lookup_option_id_for_list_of_values(self):
        md = Metadata([], template_fields=self.datacite.fields)
        entry = md.get_entry_by_id("2qwhbanmbsc")
        self.assertListEqual(
            ["gtz6mcevk2", "zl8ouisyt1q"], md.lookup_option_id_from_value(entry, ["Other", "Audiovisual"])
        )

    def test_lookup_option_id_for_nested_list_of_values(self):
        md = Metadata([], template_fields=self.datacite.fields)
        entry = md.get_entry_by_id("2qwhbanmbsc")
        self.assertListEqual(
            [
                "gtz6mcevk2",
                [
                    "zl8ouisyt1q",
                    "qrg4713tgvr",
                ],
            ],
            md.lookup_option_id_from_value(entry, ["Other", ["Audiovisual", "Image"]]),
        )


class TemplateEntryTest(TestCase):
    entry = {
        "id": "2qwhbanmbsc",
        "hint": "The general type of a resource.",
        "name": "resourceTypeGeneral",
        "exist": True,
        "label": "resourceTypeGeneral",
        "options": [
            {"id": "zl8ouisyt1q", "label": "Audiovisual", "owner": "2qwhbanmbsc", "value": "zl8ouisyt1q"},
            {"id": "6n2eecqvnsd", "label": "Collection", "owner": "2qwhbanmbsc", "value": "6n2eecqvnsd"},
        ],
        "visible": True,
        "multiple": False,
        "required": True,
        "component": "UsfSelect",
        "parent_id": "g8mrg14u4ji",
        "protected": True,
        "default_option": None,
    }

    def test_get_option_by_id(self):
        entry = TemplateEntry(self.entry)
        self.assertEqual("zl8ouisyt1q", entry.get_option_id("Audiovisual"))

    def test_get_option_id_returns_empty_string_if_not_found(self):
        entry = TemplateEntry(self.entry)
        self.assertEqual("", entry.get_option_id("Label does not exist"))
