import json
from pathlib import Path
from unittest.mock import patch

from pount.apps.api.libs.metadata.processors import ReferentialValuesProcessor, SelectValuesProcessor
from pount.apps.api.models import Template
from pount.apps.api.models.init import initialize_templates
from pount.tests import TestCase


def get_fixture_content(filename):
    filename = Path(__file__).parent.parent.parent.resolve() / "fixtures" / filename
    with open(filename) as f:
        return f.read()


class ReferentialValuesProcessorTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_templates()
        cls.metadata = json.loads(get_fixture_content("referential_metadata.json"))
        cls.template_fields = json.loads(get_fixture_content("template_fields.json"))

    def test_label_is_added_by_processor(self):
        processor = ReferentialValuesProcessor()
        with patch("pount.apps.api.libs.metadata.processors.get_node_translations") as get_node_translations:
            get_node_translations.return_value = {"fr": "Libellé français", "en": "English label"}
            metadata = processor.process(self.metadata, self.template_fields)

        referential_metadata_level1 = next(item for item in metadata if item["id"] == "aqe9k7ro2u")
        referential_metadata_level2 = next(item for item in metadata if item["id"] == "hpn3pe7sj9t")
        referential_metadata_level3 = next(item for item in metadata if item["id"] == "e87zh8vfr4r")

        self.assertDictEqual(
            referential_metadata_level1["label"],
            {referential_metadata_level1["value"][0]: {"fr": "Libellé français", "en": "English label"}},
        )
        self.assertDictEqual(
            referential_metadata_level2["label"],
            {referential_metadata_level2["value"][0]: {"fr": "Libellé français", "en": "English label"}},
        )
        self.assertDictEqual(
            referential_metadata_level3["label"],
            {referential_metadata_level3["value"][0]: {"fr": "Libellé français", "en": "English label"}},
        )

    def test_label_for_other_field_is_not_changed(self):
        processor = ReferentialValuesProcessor()
        with patch("pount.apps.api.libs.metadata.processors.get_node_translations") as get_node_translations:
            get_node_translations.return_value = {"fr": "Libellé français", "en": "English label"}
            metadata = processor.process(self.metadata, self.template_fields)

        text_field = next(item for item in metadata if item["id"] == "5vxode5mjsh")
        with self.assertRaises(KeyError):
            text_field["label"]


class SelectValuesProcessorTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_templates()
        cls.datacite: Template = Template.objects.get(name="datacite")

    def test_available_languages_in_template_messages(self):
        processor = SelectValuesProcessor(self.datacite.messages)
        self.assertListEqual(processor.available_languages, ["en", "fr"])

    def test_translate_option_returns_default_if_does_not_exist(self):
        processor = SelectValuesProcessor(self.datacite.messages)
        self.assertEqual(
            processor.translate_option("idontexist", "fr", "Default label"),
            "Default label",
        )

    def test_label_is_added_for_usf_select_field(self):
        metadata = [
            {"id": "yxuullxsicj", "name": "resourceType", "value": "Book"},
            {
                "id": "2qwhbanmbsc",
                "name": "resourceTypeGeneral",
                "value": ["e6esmnvs0rj", "uj938jmk2rl"],
                "is_option": True,
                "is_multiple": True,
            },
        ]
        processor = SelectValuesProcessor(self.datacite.messages)
        processed_metadata = processor.process(metadata=metadata, template_fields=self.datacite.fields)
        processed_entry = next(md for md in processed_metadata if md["id"] == "2qwhbanmbsc")
        self.assertDictEqual(
            processed_entry["label"],
            {
                "e6esmnvs0rj": {"en": "Text", "fr": "Texte"},
                "uj938jmk2rl": {"en": "DataPaper", "fr": "DataPaper"},
            },
        )
