from unittest.mock import patch
from uuid import UUID

from pount.apps.api.libs.referential import get_default_language, get_referential_node_id_for_node_value
from pount.apps.api.models import Referential
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.models.referential import DEFAULT_LANGUAGE
from pount.apps.api.tests.factories.referential import PublicReferentialFactory
from pount.exceptions import MetadataOptionValueNotFound
from pount.tests import TestCase


class DefaultLanguageTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()

    def test_return_default_language_defined_in_instance(self):
        referential = PublicReferentialFactory(
            languages={"default": "zz", "available": ["fr", "zz"], "translations": {}}
        )
        default_language = get_default_language.__wrapped__(referential)
        self.assertEqual(
            default_language,
            "zz",
        )

    def test_return_global_default_language_if_not_set_in_instance(self):
        referential = PublicReferentialFactory(languages={})
        default_language = get_default_language.__wrapped__(referential)
        self.assertEqual(
            default_language,
            DEFAULT_LANGUAGE,
        )


class ReferentialNodeLookupTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()

    def test_lookup_node_by_id(self):
        referential: Referential = PublicReferentialFactory()
        with patch("pount.apps.api.libs.referential.mongo_db") as mongo_db:
            mongo_db.return_value.referential_nodes.find_one.return_value = {
                "id": UUID("030de3d9-5404-409a-82f1-0310f3a76b03"),
                "name": "node name",
            }
            node_id = get_referential_node_id_for_node_value(
                referential.id, "b50a12ea-847e-442a-8bed-6b966ee58787", "node name"
            )
        self.assertEqual(
            node_id,
            UUID("030de3d9-5404-409a-82f1-0310f3a76b03"),
        )
        mongo_db.return_value.referential_nodes.find_one.assert_called_once_with(
            {
                "referential_id": referential.id,
                "level_id": UUID("b50a12ea-847e-442a-8bed-6b966ee58787"),
                "name": "node name",
            },
        )

    def test_if_lookup_does_not_find_node_it_raises_error(self):
        referential: Referential = PublicReferentialFactory()
        with patch("pount.apps.api.libs.referential.mongo_db") as mongo_db:
            mongo_db.return_value.referential_nodes.find_one.return_value = None
            with self.assertRaises(MetadataOptionValueNotFound):
                get_referential_node_id_for_node_value(
                    referential.id, "b50a12ea-847e-442a-8bed-6b966ee58787", "node name"
                )
        mongo_db.return_value.referential_nodes.find_one.assert_called_once_with(
            {
                "referential_id": referential.id,
                "level_id": UUID("b50a12ea-847e-442a-8bed-6b966ee58787"),
                "name": "node name",
            },
        )
