from unittest.mock import patch

from pount.apps.api.libs.referential import share_referential
from pount.apps.api.models import Referential
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.tests.factories.project import ProjectFactory, ProjectForManagerFactory
from pount.apps.api.tests.factories.referential import BaseReferentialFactory, PublicReferentialFactory
from pount.apps.api.tests.factories.user import UserFactory
from pount.tests import TestCase


class ShareReferentialTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.manager = UserFactory()

    def test_base_referential_s_scope_can_not_be_changed(self):
        referential = BaseReferentialFactory()
        with self.assertRaises(ValueError) as ctx:
            share_referential(referential, new_scope=Referential.Scope.SCOPE_PUBLIC)
        self.assertEqual(
            "Can not change a base referential's scope",
            str(ctx.exception),
        )

    def test_public_referential_used_in_other_project_can_not_be_made_public(self):
        project = ProjectForManagerFactory(managers=[self.manager])
        referential = PublicReferentialFactory(creator=self.manager, project=project)
        other_project = ProjectFactory()
        other_project.external_referentials.add(referential)
        with self.assertRaises(ValueError) as ctx:
            share_referential(referential, new_scope=Referential.Scope.SCOPE_PROJECT)
        self.assertEqual(
            "Referential is used in other project",
            str(ctx.exception),
        )

    def test_change_referential_s_scope(self):
        project = ProjectForManagerFactory(managers=[self.manager])
        referential = PublicReferentialFactory(creator=self.manager, project=project)
        with patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save"):
            with patch(
                "pount.apps.api.models.referential.handle_build_node_lists_on_referential_save"
            ) as build_node_lists_task:
                share_referential(referential, new_scope=Referential.Scope.SCOPE_PROJECT)
        referential.refresh_from_db(fields=["scope"])
        build_node_lists_task.delay.assert_called_with(referential.id)
        self.assertEqual(
            referential.scope,
            Referential.Scope.SCOPE_PROJECT,
        )
