from unittest.mock import patch

from pount.apps.api.libs.set import (
    DATACITE_SEARCH_CONFIG,
    add_or_replace_datacite_fields,
    get_mapped_field_names,
    order_fields_in_template_order,
)
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.tests.factories.set import SetFactory
from pount.tests import TestCase


class ListConfiguredFieldsTest(TestCase):
    def setUp(self) -> None:
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()
        initialize_roles()
        initialize_templates()

    def tearDown(self) -> None:
        self.patcher1.stop()

    def test_get_list_of_configured_metadata_fields(self):
        set_model = SetFactory(
            search_config=[
                {
                    "name": "first_configured_field",
                    "label": {"en": "Label"},
                    "config": {
                        "mapping_type": "strings",
                        "use_in_search": True,
                        "use_in_aggregation": False,
                    },
                },
                {
                    "name": "second_configured_field",
                    "label": {"en": "Label"},
                    "config": {
                        "mapping_type": "strings",
                        "use_in_search": True,
                        "use_in_aggregation": False,
                    },
                },
                {
                    "name": "third_configured_field",
                    "label": {"en": "Label"},
                    "config": {
                        "mapping_type": "bools",
                        "use_in_search": False,
                        "use_in_aggregation": False,
                    },
                },
            ]
        )
        self.assertListEqual(
            ["first_configured_field", "second_configured_field", "third_configured_field"],
            get_mapped_field_names(set_model.search_config),
        )

    def test_fields_without_name_are_not_returned(self):
        search_config = [
            {
                "name": "field_with_name",
                "config": {
                    "mapping_type": "strings",
                },
            },
            {
                "config": {
                    "mapping_type": "strings",
                }
            },
        ]
        self.assertListEqual(["field_with_name"], get_mapped_field_names(search_config))


class TestAddOrReplaceDataciteFieldsTest(TestCase):
    def test_replace_datacite_fields(self):
        set_model = {
            "search_config": [
                {"name": "creator", "config": {"key1": "value1"}},
                {"name": "title", "config": {"key2": "value2"}},
                {"name": "publisher", "config": {"key3": "value3"}},
                {"name": "publicationYear", "config": {"key4": "value4"}},
                {"name": "resourceType", "config": {"key5": "value5"}},
                {"name": "resourceTypeGeneral", "config": {"key6": "value6"}},
            ]
        }

        set_model["search_config"] = add_or_replace_datacite_fields(set_model["search_config"])

        expected_search_config = [
            {"name": "creator", "config": DATACITE_SEARCH_CONFIG["creator"]},
            {"name": "title", "config": DATACITE_SEARCH_CONFIG["title"]},
            {"name": "publisher", "config": DATACITE_SEARCH_CONFIG["publisher"]},
            {"name": "publicationYear", "config": DATACITE_SEARCH_CONFIG["publicationYear"]},
            {"name": "resourceType", "config": DATACITE_SEARCH_CONFIG["resourceType"]},
            {"name": "resourceTypeGeneral", "config": DATACITE_SEARCH_CONFIG["resourceTypeGeneral"]},
        ]

        self.assertEqual(set_model["search_config"], expected_search_config)

    def test_add_datacite_fields(self):
        set_model = {
            "search_config": [
                {"name": "creator", "config": {"key1": "value1"}},
                {"name": "title", "config": {"key2": "value2"}},
                {"name": "publicationYear", "config": {"key4": "value4"}},
                {"name": "resourceType", "config": {"key5": "value5"}},
            ]
        }

        set_model["search_config"] = add_or_replace_datacite_fields(set_model["search_config"])

        expected_search_config = [
            {"name": "creator", "config": DATACITE_SEARCH_CONFIG["creator"]},
            {"name": "title", "config": DATACITE_SEARCH_CONFIG["title"]},
            {
                "name": "publisher",
                "config": DATACITE_SEARCH_CONFIG["publisher"],
                "label": {"en": "publisher", "fr": "publisher"},
            },
            {"name": "publicationYear", "config": DATACITE_SEARCH_CONFIG["publicationYear"]},
            {"name": "resourceType", "config": DATACITE_SEARCH_CONFIG["resourceType"]},
            {
                "name": "resourceTypeGeneral",
                "config": DATACITE_SEARCH_CONFIG["resourceTypeGeneral"],
                "label": {"en": "resourceTypeGeneral", "fr": "resourceTypeGeneral"},
            },
        ]

        self.assertListEqual(
            sorted(set_model["search_config"], key=lambda field: field["name"]),
            sorted(expected_search_config, key=lambda field: field["name"]),
        )


class TestOrderFieldsInTemplateOrderTest(TestCase):
    class TemplateFields:
        fieldnames = [
            "field1",
            "field2",
            "field3",
        ]

    def test_order_fields(self):
        search_config = [
            {"name": "field2", "config": {"key2": "value2"}},
            {"name": "field1", "config": {"key1": "value1"}},
            {"name": "field3", "config": {"key3": "value3"}},
        ]

        template_fields = self.TemplateFields()

        ordered_search_config = order_fields_in_template_order(search_config, template_fields)

        expected_search_config = [
            {"name": "field1", "config": {"key1": "value1"}},
            {"name": "field2", "config": {"key2": "value2"}},
            {"name": "field3", "config": {"key3": "value3"}},
        ]

        self.assertEqual(ordered_search_config, expected_search_config)
