from unittest.mock import call, patch

from django.conf import settings

from pount.apps.api.libs.cas import get_affiliation_info, update_or_create_user
from pount.apps.api.libs.structure import Structure
from pount.apps.api.models import User
from pount.tests import TestCase


class UpdateUserFromCasTest(TestCase):

    def test_get_affiliation_info(self):
        attributes = {
            "first_name": ["First name"],
            "last_name": ["Last name"],
            "email": ["email@unistra.fr"],
            "primary_affiliation": ["UMR123"],
            "affiliations": ["UMR123", "CODE"],
        }

        with patch("pount.apps.api.libs.cas.structure_reference") as structure_reference:
            structure_reference.get.side_effect = [
                Structure(
                    id=123,
                    nature="UR",
                    nature_type="UMR",
                    label="UMR123",
                    acronym="ACRO",
                    long_name="Long name",
                    middle_name="Middle name",
                ),
                Structure(
                    id=456,
                    nature="DIRE",
                    nature_type="DIRE",
                    label="CODE",
                    acronym="CODE",
                    long_name="Long name",
                    middle_name="Middle name",
                ),
            ]
            get_affiliation_info(attributes)
            self.assertEqual(structure_reference.get.call_count, 2)
            self.assertEqual(structure_reference.get.mock_calls, [call("UMR123"), call("CODE")])

    def test_if_code_is_not_found_error_is_logged(self):
        attributes = {
            "first_name": ["First name"],
            "last_name": ["Last name"],
            "email": ["email@unistra.fr"],
            "primary_affiliation": ["UMR123"],
            "affiliations": ["XXX"],
        }
        with patch("pount.apps.api.libs.cas.structure_reference") as structure_reference:
            structure_reference.get.side_effect = ValueError("Structure not found")
            with patch("pount.apps.api.libs.cas.logger") as logger:
                get_affiliation_info(attributes)
                logger.exception.assert_called_with("Structure code XXX not found")

    def test_create_user_from_attributes(self):
        attrs = {
            "username": "username",
            "first_name": ["first_name"],
            "last_name": ["last_name"],
            "email": ["email"],
            "affiliations": ["udsTopUnitCode"],
            "primary_affiliation": ["udsTopUnitCode"],
        }
        with patch("pount.apps.api.libs.cas.structure_reference") as structure_reference:
            structure_reference.get.return_value = Structure(
                id=123,
                nature="UR",
                nature_type="UMR",
                label="UMR123",
                acronym="ACRO",
                long_name="Long name",
            )
            update_or_create_user(attrs)
            username = settings.CAS_USERNAME_FORMAT("username")
            user = User.objects.get(username=username)
            self.assertDictEqual(
                user.info,
                {
                    "institution": "Université de Strasbourg",
                    "affiliations": [],
                    "research_labs": [
                        {
                            "id": 123,
                            "label": "UMR123",
                            "nature": "UR",
                            "acronym": "ACRO",
                            "category": "",
                            "long_name": "Long name",
                            "short_name": "",
                            "middle_name": "",
                            "nature_type": "UMR",
                        }
                    ],
                },
            )

    def test_update_user_from_attributes(self):
        user = User.objects.create(
            username=settings.CAS_USERNAME_FORMAT("username"),
            settings={"refresh_info": True},
        )
        attrs = {
            "username": "username",
            "first_name": ["first_name"],
            "last_name": ["last_name"],
            "email": ["email"],
            "affiliations": ["udsTopUnitCode"],
            "primary_affiliation": ["udsTopUnitCode"],
        }
        with patch("pount.apps.api.libs.cas.structure_reference") as structure_reference:
            structure_reference.get.return_value = Structure(
                id=123,
                nature="UR",
                nature_type="UMR",
                label="UMR123",
                acronym="ACRO",
                long_name="Long name",
            )
            update_or_create_user(attrs)
        user.refresh_from_db()
        self.assertDictEqual(
            user.info,
            {
                "institution": "Université de Strasbourg",
                "bio": "",
                "affiliations": [],
                "other_affiliations": [],
                "research_labs": [
                    {
                        "id": 123,
                        "label": "UMR123",
                        "nature": "UR",
                        "acronym": "ACRO",
                        "category": "",
                        "long_name": "Long name",
                        "short_name": "",
                        "middle_name": "",
                        "nature_type": "UMR",
                    }
                ],
            },
        )
