import datetime
from datetime import timedelta
from io import StringIO
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.test import TestCase
from django.utils import timezone

from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.models.item import Item, MediaFile
from pount.apps.api.models.login import Ticket
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.models.user import AuthenticationProvider, Invitation
from pount.apps.api.tests import util

User = get_user_model()


class CleanExpiredTicketsTest(TestCase):
    def setUp(self) -> None:
        self.user = util.create_user("user", is_creator=False)
        self.expired_ticket = self._create_ticket(timezone.now() - timedelta(seconds=120))
        self.valid_ticket = self._create_ticket()

    def _create_ticket(self, expiration: datetime = None):
        return Ticket.objects.create(
            user=self.user,
            provider=AuthenticationProvider.CAS.value,
            expires_at=expiration,
        )

    def test_expired_tickets_are_deleted(self):
        call_command("clean_expired_tickets")
        with self.assertRaises(Ticket.DoesNotExist):
            self.expired_ticket.refresh_from_db()

    def test_valid_tickets_are_not_deleted(self):
        call_command("clean_expired_tickets")
        self.valid_ticket.refresh_from_db()
        self.assertFalse(self.valid_ticket.is_expired)


class CleanExpiredInvitationsTest(TestCase):
    def setUp(self) -> None:
        initialize_roles()
        self.project = Project.objects.create(
            name="Project name",
        )
        self.expired_invitation: Invitation = Invitation.objects.create(
            token="token", expires_at=timezone.now() - timedelta(days=1)
        )
        self.valid_invitation: Invitation = Invitation.objects.create(
            token="token",
        )

    def test_expired_invitations_are_deleted(self):
        call_command("clean_expired_invitations")
        with self.assertRaises(Invitation.DoesNotExist):
            self.expired_invitation.refresh_from_db()

    def test_non_expired_invitations_are_not_deleted(self):
        call_command("clean_expired_invitations")
        self.valid_invitation.refresh_from_db()
        self.assertFalse(self.valid_invitation.is_expired)


class CleanInactiveUsersTest(TestCase):
    def setUp(self) -> None:
        self.active_internal_user = User.objects.create_user(
            "active_internal_user",
            last_login=timezone.now() - timedelta(days=1),
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.absent_internal_user = User.objects.create_user(
            "absent_internal_user",
            is_active=True,
            last_login=timezone.now() - timedelta(days=380),
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.inactive_internal_user = User.objects.create_user(
            "inactive_internal_user",
            is_active=False,
            last_login=timezone.now() - timedelta(days=380),
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.active_external_user = User.objects.create_user(
            "active_external",
            is_active=True,
            last_login=timezone.now() - timedelta(days=1),
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )
        self.inactive_external_user = User.objects.create_user(
            "inactive_external",
            is_active=False,
            last_login=timezone.now() - timedelta(days=1),
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )
        self.absent_external_user = User.objects.create_user(
            "absent_active_external",
            is_active=True,
            last_login=timezone.now() - timedelta(days=380),
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )
        self.absent_inactive_external_user = User.objects.create_user(
            "absent_inactive_external",
            is_active=False,
            last_login=timezone.now() - timedelta(days=380),
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )

    def test_inactive_internal_user_is_not_deleted(self):
        with patch("pount.apps.api.management.commands.clean_inactive_users.logger") as logger:
            call_command("clean_inactive_users")
            logger.info.assert_called_with(
                "External inactive user <LOCAL/absent_inactive_external> deleted"  # noqa: E501
            )
        with self.assertRaises(User.DoesNotExist):
            self.absent_inactive_external_user.refresh_from_db()
        self.assertEqual(User.objects.count(), 6)


class CleanDeletedObjectsTest(TestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock):
        initialize_roles()
        initialize_templates()
        user = User.objects.create_user("username")
        project = Project.objects.create(
            name="Project name",
            description="",
        )
        project.add_member(user)
        project.add_manager(user)
        self.deleted_set = Set.objects.create(
            title="Deleted set",
            description="",
            creator=user,
            template=Template.objects.first(),
            project=project,
        )
        self.deleted_item = Item.objects.create(
            title="Deleted item",
            description="",
            creator=user,
            set=self.deleted_set,
            metadata={},
        )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_recently_soft_deleted_set_is_not_deleted(self, update_mock):
        self.deleted_set.delete()
        call_command("clean_expired_objects", stdout=StringIO())
        self.deleted_set.refresh_from_db()

    def test_soft_deleted_set_older_than_x_days_is_deleted(self):
        self.deleted_set.is_deleted = True
        self.deleted_set.deleted_at = timezone.now() - timedelta(days=90)
        self.deleted_set.save()
        call_command("clean_expired_objects", stdout=StringIO())
        with self.assertRaises(Set.DoesNotExist):
            self.deleted_set.refresh_from_db()

    def test_recently_soft_deleted_item_is_not_deleted(self):
        self.deleted_item.delete()
        call_command("clean_expired_objects", stdout=StringIO())
        self.deleted_item.refresh_from_db()

    def test_soft_deleted_item_older_than_x_days_is_deleted(self):
        self.deleted_item.is_deleted = True
        self.deleted_item.deleted_at = timezone.now() - timedelta(days=90)
        self.deleted_item.save()
        call_command("clean_expired_objects", stdout=StringIO())
        with self.assertRaises(Item.DoesNotExist):
            self.deleted_item.refresh_from_db()

    def test_associated_mediafiles_are_deleted_on_item_delete(self):
        self.deleted_item.is_deleted = True
        self.deleted_item.deleted_at = timezone.now() - timedelta(days=90)
        self.deleted_item.save()
        mediafile = MediaFile.objects.create(
            item=self.deleted_item,
            filename="/path/to/file",
        )
        with patch("pount.apps.api.models.item.MediaFile.delete_file") as delete_mock:
            delete_mock.return_value = None
            call_command("clean_expired_objects", stdout=StringIO())
            with self.assertRaises(MediaFile.DoesNotExist):
                mediafile.refresh_from_db()
            self.assertEqual(delete_mock.call_count, 1)

    def test_associated_template_is_deleted_on_set_delete(self):
        template = self.deleted_set.template
        self.deleted_set.is_deleted = True
        self.deleted_set.deleted_at = timezone.now() - timedelta(days=90)
        self.deleted_set.save(update_fields=["is_deleted", "deleted_at"])
        call_command("clean_expired_objects", stdout=StringIO())
        with self.assertRaises(Template.DoesNotExist):
            template.refresh_from_db()
