from unittest.mock import patch

from django.db import IntegrityError
from django.test import TestCase

from pount.apps.api.models import Referential
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.tests.factories.referential import PublicReferentialFactory
from pount.apps.api.tests.factories.user import UserFactory


class ReferentialTest(TestCase):
    def setUp(self) -> None:
        # Mock the build node list task
        patcher1 = patch("pount.apps.api.models.referential.handle_build_node_lists_on_referential_save")
        self.mock_build_node_lists = patcher1.start()
        self.addCleanup(patcher1.stop)

        # Mock the Item update task
        patcher2 = patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save")
        self.mock_update_items = patcher2.start()
        self.addCleanup(patcher2.stop)

        # Signal patchers
        self.redis_clean_patcher = patch("pount.apps.api.signals.clear_cache_for_referential")
        self.mongo_clean_patcher = patch("pount.apps.api.signals.delete_referential_in_mongodb")

    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.creator = UserFactory()

    def test_can_not_create_a_referential_with_same_name(self):
        PublicReferentialFactory(
            name="Referential Name",
            version="20221101120000",
            creator=self.creator,
        )

        with self.assertRaises(IntegrityError):
            PublicReferentialFactory(
                name="Referential Name",
                version="20221101120000",
                creator=self.creator,
            )

    def test_can_create_another_version_of_referential(self):
        root_referential = PublicReferentialFactory(
            name="Referential Name",
            version="20221101120000",
            creator=self.creator,
        )

        new_version = Referential(
            name=root_referential.name,
            root_version=root_referential,
            version="20221102150000",
            scope=Referential.Scope.SCOPE_PROJECT,
            definition={},
            languages={"default": "", "available": [], "translations": {}},
            lists=[],
            creator=self.creator,
        )
        self.assertEqual(
            new_version.root_version.id,
            root_referential.id,
        )

    def test_can_not_create_two_referentials_with_same_name_and_version(self):
        root_referential = PublicReferentialFactory(
            name="Same referential name",
            version="202201",
            creator=self.creator,
        )

        with self.assertRaises(IntegrityError):
            PublicReferentialFactory(
                name="Same referential name",
                version="202201",
                root_version=root_referential,
                creator=self.creator,
            )

    def test_display_name_is_a_concatenation_of_version_and_name(self):
        referential = PublicReferentialFactory()
        self.assertEqual(
            referential.display_name,
            referential.name + "." + referential.version,
        )

    def test_build_node_lists_is_called_on_save(self):
        """By default, node lists are build and stored in lists field"""
        referential = PublicReferentialFactory()
        referential.save()
        self.mock_build_node_lists.delay.assert_called_with(referential.id)

    def test_save_without_build_node_lists(self):
        referential = PublicReferentialFactory()
        referential.save(build_node_lists=False)
        self.mock_build_node_lists.delay.assert_not_called()

    def test_metadata_of_items_using_referential_in_templates_is_processed_on_save(self):
        referential: Referential = PublicReferentialFactory()
        referential.save()
        self.mock_update_items.delay.assert_called_with(referential.id)

    def test_mongodb_is_cleaned_on_referential_save(self):
        """When a referential is deleted, nodes and levels in MongoDB need to be deleted too"""
        referential = PublicReferentialFactory()
        self.redis_clean_patcher.start()
        with patch("pount.apps.api.signals.delete_referential_in_mongodb") as delete_mongodb_data:
            referential.delete()
            delete_mongodb_data.assert_called_with(referential=referential)
        self.redis_clean_patcher.stop()

    def test_redis_cache_is_cleaned_on_referential_save(self):
        """When a referential is deleted, data that is cached needs to be deleted too"""
        referential = PublicReferentialFactory()

        self.mongo_clean_patcher.start()
        with patch("pount.apps.api.signals.clear_cache_for_referential") as clear_cache_data:
            referential.delete()
            clear_cache_data.assert_called_with(referential=referential)
        self.mongo_clean_patcher.stop()
