from pount.apps.api.models import User
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.tests.factories.project import ProjectFactory
from pount.apps.api.tests.factories.user import UserFactory
from pount.tests import TestCase


class UserProjectMemberTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.project = ProjectFactory()
        cls.member = UserFactory()

    def test_user_is_not_a_member_of_any_project(self):
        self.assertFalse(self.member.is_project_member())

    def test_user_is_a_member_of_some_project(self):
        self.project.add_member(self.member)
        self.assertTrue(self.member.is_project_member())

    def test_user_is_a_member_of_another_project(self):
        another_project = ProjectFactory()
        another_project.add_member(self.member)

        self.assertTrue(self.member.is_project_member())
        self.assertFalse(self.member.is_project_member(project=self.project))
        self.assertFalse(self.member.is_project_member(project=self.project.id))
        self.assertFalse(self.member.is_project_member(project=str(self.project.id)))


class UserNameHasBeenChangedTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.member: User = UserFactory(first_name="John", last_name="Doe")

    def test_user_first_name_has_been_changed(self):
        self.assertFalse(self.member.name_has_changed)
        self.member.first_name = "Jane"
        self.assertTrue(self.member.name_has_changed)

    def test_user_last_name_has_been_changed(self):
        self.member.last_name = "Meyer"
        self.assertTrue(self.member.name_has_changed)

    def test_user_email_has_been_changed(self):
        self.member.email = "user@pount.org"
        self.assertFalse(self.member.name_has_changed)

    def test_user_username_has_been_changed(self):
        self.member.username = "jdoe"
        self.assertFalse(self.member.name_has_changed)

    def test_after_save_name_has_changed_is_false(self):
        self.member.first_name = "Jane"
        self.member.save()
        self.assertFalse(self.member.name_has_changed)


class AdminListTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.admin1 = UserFactory(is_superuser=True)
        cls.admin2 = UserFactory(is_superuser=True)
        cls.disabled_admin = UserFactory(is_superuser=True, is_active=False)

    def test_list_of_admins_contains_all_admins(self):
        self.assertEqual(User.objects.admins().count(), 2)
        self.assertListEqual([self.admin1, self.admin2], list(User.objects.admins()))

    def test_list_of_admins_does_not_contain_disabled_admins(self):
        self.assertNotIn(self.disabled_admin, User.objects.admins())
