from django.db import IntegrityError
from django.test import TestCase

from pount.apps.api.models.group import Group
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.models.project import Project


class GroupTest(TestCase):
    def setUp(self) -> None:
        initialize_roles()
        self.project = Project.objects.create(
            name="Project name",
        )

    def test_group_name_must_be_unique_for_project(self):
        Group.objects.create(project=self.project, name="Group name", settings={})
        with self.assertRaises(IntegrityError):
            # We cannot create another group with the same name for the project
            Group.objects.create(project=self.project, name="Group name", settings={})

    def test_group_must_be_related_to_project(self):
        with self.assertRaises(IntegrityError):
            Group.objects.create(
                name="Group name",
                settings={},
            )

    def test_default_group_can_not_be_deleted(self):
        for name in [Group.MANAGERS, Group.MEMBERS]:
            group = self.project.groups.get(name=name)
            with self.assertRaises(ValueError) as ctx:
                group.delete()
            self.assertIn("Default group cannot be deleted", str(ctx.exception))

    def test_default_group_is_deleted_when_project_is_deleted(self):
        managers = self.project.groups.get(name=Group.MANAGERS)
        members = self.project.groups.get(name=Group.MEMBERS)
        self.project.delete()
        with self.assertRaises(Group.DoesNotExist) as ctx:
            managers.refresh_from_db()
        self.assertIn("Group matching query does not exist", str(ctx.exception))
        with self.assertRaises(Group.DoesNotExist) as ctx:
            members.refresh_from_db()
        self.assertIn("Group matching query does not exist", str(ctx.exception))
