from datetime import datetime
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.test import TestCase

from pount.apps.api.libs.file import item_mediafile_fullname
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.models.item import Item, MediaFile
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.storage_backends import ItemMediaFileStorage

User = get_user_model()


class SetUpMixin:
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        self.creator = User.objects.create_user("creator")
        initialize_templates()
        initialize_roles()
        self.metadata = [
            {
                "id": "6qgq1netsxh",
                "name": "creator",
                "value": ["Creator 1", "Creator 2", "Creator 3"],
                "multiple": False,
                "is_option": False,
                "repeatable": True,
            },
            {
                "id": "5vxode5mjsh",
                "name": "title",
                "value": "Title of the item",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
            {
                "id": "unym14mu68",
                "name": "publisher",
                "value": "Université de Strasbourg",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
            {
                "id": "m1w8jje0wsh",
                "name": "publicationYear",
                "value": "2021",
                "multiple": False,
                "is_option": False,
                "repeatable": False,
            },
        ]
        project = Project.objects.create(
            name="Project name",
            description="Project description",
        )
        self.set = Set.objects.create(
            title="Set title",
            creator=self.creator,
            template=Template.objects.first(),
            project=project,
        )
        self.item = Item.objects.create(
            title="Item title",
            description="Item description",
            metadata=[],
            creator=self.creator,
            set=self.set,
        )
        self.file = MediaFile.objects.create(filename="pount.jpg", item=self.item, info={})


class ItemTest(SetUpMixin, TestCase):
    def test_string_representation_of_item_is_its_title(self):
        self.assertEqual(str(self.item), self.item.title)

    def test_item_is_editable_by_creator(self):
        self.assertTrue(self.item.is_user_editor(self.creator))

    def test_item_is_editable_by_project_manager(self):
        manager = User.objects.create_user("manager")
        self.set.project.add_member(manager)
        self.set.project.add_manager(manager)
        self.assertTrue(self.item.is_user_editor(manager))


class ItemUrlTest(SetUpMixin, TestCase):
    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.url")
    def test_public_url_property_uses_public_expiration_delay(self, url_mock):
        url = f"http://media.pount.org/{self.file.filename}"
        url_mock.return_value = url
        returned_url = self.file.public_url
        url_mock.assert_called_with(
            item_mediafile_fullname(self.file.filename, self.item, ignore_location=True),
            expire=ItemMediaFileStorage.PUBLIC_EXPIRATION_DELAY,
        )
        self.assertEqual(url, returned_url)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.url")
    def test_public_url_is_stored_in_mediafile_object(self, url_mock):
        url_mock.return_value = f"http://media.pount.org/{self.file.filename}"
        url = self.file.public_url
        self.assertEqual(self.file.info["public_url"]["url"], url)

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.url")
    def test_public_url_stored_in_mediafile_object_is_returned_if_present_and_not_expired(  # noqa: E501
        self, url_mock
    ):
        url_mock.return_value = f"http://media.pount.org/{self.file.filename}"
        self.file.info["public_url"] = {
            "url": "http://stored_url/",
            "expires": datetime.now().timestamp() + 3600,
        }
        self.file.save()
        url = self.file.public_url
        self.assertEqual(self.file.info["public_url"]["url"], url)
        self.assertEqual(url, "http://stored_url/")
        url_mock.assert_not_called()

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.url")
    def test_public_url_stored_in_mediafile_object_is_not_returned_if_expired(self, url_mock):
        url_mock.return_value = f"http://media.pount.org/{self.file.filename}"
        self.file.info["public_url"] = {
            "url": "http://stored_url/",
            "expires": datetime.now().timestamp() - 3600,
        }
        self.file.save()
        url = self.file.public_url
        self.assertEqual(self.file.info["public_url"]["url"], url)
        self.assertEqual(url, "http://media.pount.org/pount.jpg")
        url_mock.assert_called_with(
            item_mediafile_fullname(self.file.filename, self.item, ignore_location=True),
            expire=ItemMediaFileStorage.PUBLIC_EXPIRATION_DELAY,
        )

    @patch("pount.apps.api.storage_backends.ItemMediaFileStorage.url")
    def test_url_property_uses_default_expiration_delay(self, url_mock):
        url = f"http://media.pount.org/{self.file.filename}"
        url_mock.return_value = url
        returned_url = self.file.url
        url_mock.assert_called_with(
            item_mediafile_fullname(self.file.filename, self.item, ignore_location=True),
            expire=None,
        )
        self.assertEqual(url, returned_url)


class SoftDeleteItemTest(SetUpMixin, TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.items = []
        for i in range(5):
            self.items.append(
                Item.objects.create(
                    title=f"Item {i}",
                    description="",
                    metadata={},
                    set=self.set,
                    creator=self.creator,
                )
            )

    def test_delete_queryset_of_items_soft_deletes_them(self):
        self.assertEqual(Item.objects.all().count(), 6)
        self.assertEqual(Item.deleted_objects.all().count(), 0)
        Item.objects.all().delete()
        self.assertEqual(Item.objects.all().count(), 0)
        self.assertEqual(Item.deleted_objects.all().count(), 6)


class CitationTest(SetUpMixin, TestCase):
    @patch("pount.apps.api.models.item.make_item_citation")
    def test_citation_is_computed_on_save(self, make_citation_mock):
        make_citation_mock.return_value = "Text of citation"
        self.item.metadata = self.metadata
        self.item.save()
        self.assertEqual(self.item.citation, "Text of citation")
        make_citation_mock.assert_called_once()

    @patch("pount.apps.api.models.item.make_item_citation")
    def test_citation_is_stored_in_settings(self, make_citation_mock):
        make_citation_mock.return_value = "Text of citation"
        self.assertEqual(self.item.settings["citation"], "")
        self.item.save()
        self.assertEqual(self.item.settings["citation"], "Text of citation")

    @patch("pount.apps.api.models.item.make_item_citation")
    def test_citation_is_computed_on_access_if_settings_is_none(self, make_citation_mock):
        # if item.settings is None, on access, citation is computed and saved
        make_citation_mock.return_value = "Text of citation"
        self.item.settings = None
        self.assertEqual(self.item.citation, "Text of citation")
        self.assertEqual(make_citation_mock.call_count, 1)
        self.assertEqual(self.item.settings, {"citation": "Text of citation"})

    @patch("pount.apps.api.models.item.make_item_citation")
    def test_citation_is_computed_on_access_if_not_saved_in_settings(self, make_citation_mock):
        # item.settings is not None but does not contain citation
        make_citation_mock.return_value = "Text of citation"
        self.item.settings = {"other_setting": "Other value"}
        self.assertEqual(self.item.citation, "Text of citation")
        self.assertEqual(make_citation_mock.call_count, 1)
        self.assertDictEqual(
            self.item.settings,
            {"other_setting": "Other value", "citation": "Text of citation"},
        )
