import datetime as dt
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone

from pount.apps.api.models import Project, Set, Template
from pount.apps.api.models.init import initialize_roles, initialize_templates

User = get_user_model()


class ThumbnailUrlTest(TestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        initialize_roles()
        initialize_templates()
        creator = User.objects.create_user("creator")
        project = Project.objects.create(
            name="Project name",
            description="",
        )
        project.add_member(creator)
        project.add_manager(creator)
        self.set = Set.objects.create(
            title="Set title",
            description="",
            project=project,
            template=Template.objects.get(name="datacite"),
            creator=creator,
        )

    def test_size_must_be_a_correct_value(self):
        with self.assertRaises(ValueError) as ctx:
            self.set.thumbnail_url("x")
        self.assertEqual(str(ctx.exception), "Not a valid size")

    def test_return_url_if_present_and_not_expired(self):
        self.set.settings = {
            "thumbnails": {
                "s": {
                    "url": "http://s3.pount.org/thumbnail_url.png",
                    "expires": (timezone.now() + dt.timedelta(days=3)).timestamp(),
                    "filename": "thumbnail_url.png",
                }
            }
        }
        self.assertEqual(
            self.set.thumbnail_url("s"),
            "http://s3.pount.org/thumbnail_url.png",
        )

    def test_url_is_generated_and_stored_if_not_present(self):
        self.set.settings = {"thumbnails": {"s": {"filename": "/path/to/filename.png"}}}
        with patch("pount.apps.api.models.mixins.ThumbnailFileStorage.url") as mock_url:
            mock_url.return_value = "http://s3.pount.org/presigned_url"
            url = self.set.thumbnail_url("s")
            self.assertEqual(url, "http://s3.pount.org/presigned_url")
            self.assertEqual(
                self.set.settings["thumbnails"]["s"]["url"],
                "http://s3.pount.org/presigned_url",
            )
