from django.contrib.auth import get_user_model
from django.test import TestCase

from pount.apps.api.models import AuthenticationProvider

User = get_user_model()


class UserTest(TestCase):
    def test_string_representation_uses_name_if_defined_or_username(
        self,
    ):
        user_without_name = User.objects.create_user(
            "username",
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )
        user_with_name = User.objects.create_user("username2", first_name="First name", last_name="Last name")
        self.assertEqual(str(user_without_name), "username")
        self.assertEqual(str(user_with_name), "First name Last name")

    def test_is_external_property(self):
        external_user = User.objects.create_user(
            "username",
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )
        self.assertTrue(external_user.is_external)
        internal_user = User.objects.create_user(
            "username",
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.assertFalse(internal_user.is_external)


class CreateUserTest(TestCase):
    def test_create_user_requires_a_non_empty_username(self):
        with self.assertRaises(ValueError):
            User.objects.create_user("")

    def test_create_user_prefixes_username_with_authentication_provider(self):
        user = User.objects.create_user(
            "username",
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.assertEqual(user.username, f"{AuthenticationProvider.CAS.value}/username")

    def test_create_user_uses_cas_as_default_authentication_provider(self):
        user = User.objects.create_user("username")
        self.assertEqual(user.username, f"{AuthenticationProvider.LOCAL.value}/username")


class CreateSuperUserTest(TestCase):
    def test_create_superuser_has_is_staff_true(self):
        user = User.objects.create_superuser("username")
        self.assertTrue(user.is_staff)

    def test_create_superuser_has_is_admin_true(self):
        user = User.objects.create_superuser("username")
        self.assertTrue(user.is_superuser)

    def test_create_superuser_uses_cas_as_default_authentication_provider(
        self,
    ):
        user = User.objects.create_superuser("username")
        self.assertEqual(user.username, f"{AuthenticationProvider.CAS.value}/username")

    def test_create_superuser_requires_a_non_empty_username(self):
        with self.assertRaises(ValueError):
            User.objects.create_superuser("")

    def test_create_superuser_requires_staff_status(self):
        with self.assertRaises(ValueError):
            User.objects.create_superuser("username", is_staff=False)

    def test_create_superuser_requires_superuser_status(self):
        with self.assertRaises(ValueError):
            User.objects.create_superuser("username", is_superuser=False)


class UsernameTest(TestCase):
    def setUp(self) -> None:
        self.authentication_provider = AuthenticationProvider.CAS
        self.user = User.objects.create_user(
            "username",
            authentication_provider=self.authentication_provider.value,
        )

    def test_get_authentication_provider_on_user(self):
        self.assertEqual(
            self.user.authentication_provider,
            self.authentication_provider.value,
        )

    def test_authentication_provider_returns_none_when_username_is_none(self):
        self.user.username = None
        self.assertIsNone(self.user.authentication_provider)

    def test_get_base_username_on_user(self):
        self.assertEqual(self.user.base_username, "username")

    def test_get_base_username_returns_none_when_username_is_none(self):
        self.user.username = None
        self.assertIsNone(self.user.base_username)
