from pathlib import Path
from unittest.mock import patch

from rest_framework import status
from rest_framework.exceptions import ErrorDetail
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models.user import AuthenticationProvider, User
from pount.apps.api.views.mixins import ThumbnailMixin


class ProfileTest(APITestCase):
    def setUp(self) -> None:
        self.internal_user: User = User.objects.create_user(
            "username",
            email="internal-user@pount.org",
            authentication_provider=AuthenticationProvider.CAS.value,
        )
        self.external_user: User = User.objects.create_user(
            username="external-user@pount.org",
            email="external-user@pount.org",
            authentication_provider=AuthenticationProvider.LOCAL.value,
        )

    def test_set_first_name_and_last_name(self):
        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse("user-profile"),
            {
                "first_name": "First name",
                "last_name": "Last name",
            },
            format="json",
        )
        self.internal_user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.internal_user.first_name, "First name")
        self.assertEqual(self.internal_user.last_name, "Last name")

    def test_email_is_read_only_on_profile_view(self):
        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse("user-profile"),
            {
                "first_name": "New first name",
                "email": "new-email@pount.org",
            },
            format="json",
        )
        self.internal_user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.internal_user.first_name, "New first name")
        # email is read-only, it's value remains unchanged
        self.assertEqual(self.internal_user.email, "internal-user@pount.org")

    def test_username_is_readonly_on_profile_view(self):
        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse("user-profile"),
            {
                "first_name": "New first name",
                "username": "new-username",
            },
            format="json",
        )
        self.internal_user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(self.internal_user.first_name, "New first name")
        # username is read-only, it's value remains unchanged
        self.assertEqual(self.internal_user.username, "CAS/username")

    def test_user_profile_keys(self):
        self.client.force_login(self.internal_user)
        response = self.client.get(reverse("user-profile"), format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            sorted(list(response.data.keys())),
            [
                "authentication_provider",
                "base_username",
                "email",
                "first_name",
                "id",
                "info",
                "is_admin",
                "is_contributor",
                "is_external",
                "is_project_creator",
                "last_name",
                "name",
                "thumbnails",
                "username",
            ],
        )

    def test_change_email_for_internal_user(self):
        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse("user-profile-change-email"),
            {
                "email": "new-email@pount.org",
            },
            format="json",
        )
        self.internal_user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["email"], self.internal_user.email)
        self.assertEqual(self.internal_user.username, "CAS/username")

    def test_change_email_for_external_user_updates_username(self):
        self.client.force_login(self.external_user)
        response = self.client.patch(
            reverse("user-profile-change-email"),
            {
                "email": "new-email@pount.org",
            },
            format="json",
        )
        self.external_user.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["email"], "new-email@pount.org")
        self.assertEqual(self.external_user.username, "LOCAL/new-email@pount.org")
        self.assertListEqual(
            sorted(list(response.data.keys())),
            sorted(["id", "username", "email", "base_username"]),
        )

    def test_email_is_validated(self):
        self.client.force_login(self.external_user)
        response = self.client.patch(
            reverse("user-profile-change-email"),
            {
                "email": "invalid email",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIsInstance(response.data["email"][0], ErrorDetail)

    def test_email_address_is_unique(self):
        self.client.force_login(self.external_user)
        response = self.client.patch(
            reverse("user-profile-change-email"),
            {
                "email": self.internal_user.email,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data["email"][0].code, "unique")

    def test_settings_is_read_only(self):
        self.client.force_login(self.external_user)
        response = self.client.patch(
            reverse("user-profile"),
            {
                "last_name": "New last name",
                "settings": {
                    "project_creator": True,
                },
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.external_user.refresh_from_db()
        self.assertEqual("New last name", self.external_user.last_name)
        self.assertEqual({}, self.external_user.settings)

    @patch("pount.apps.api.models.mixins.ThumbnailUrlMixin.thumbnail_url")
    @patch("pount.apps.api.storage_backends.ThumbnailFileStorage.save")
    def test_upload_profile_picture(self, storage_mock, url_mock):
        storage_mock.return_value = "thumbnail_filename.png"
        url_mock.return_value = "http://s3.pount.org/generated_thumbnail_filename.png"
        self.client.force_login(self.internal_user)
        file = open(
            Path(__file__).resolve(strict=True).parent.parent / "fixtures/thumbnail.png",
            "rb",
        )
        response = self.client.post(
            reverse("user-profile-thumbnail"),
            {"file": file},
        )
        file.close()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(storage_mock.call_count, len(ThumbnailMixin.SIZES))
        self.assertEqual(
            response.data["thumbnails"]["s"],
            "http://s3.pount.org/generated_thumbnail_filename.png",
        )

    @patch("pount.apps.api.storage_backends.ThumbnailFileStorage.bucket")
    def test_delete_profile_picture(self, bucket_mock):
        bucket_mock.Object.return_value.delete.return_value = None
        self.client.force_login(self.internal_user)
        self.internal_user.settings = {"thumbnails": {"s": {"filename": "filename.png"}}}
        self.internal_user.save()
        response = self.client.delete(reverse("user-profile-thumbnail"))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(bucket_mock.Object.return_value.delete.call_count, 1)
        self.internal_user.refresh_from_db()
        self.assertEqual(self.internal_user.settings, {})
