import uuid
from unittest.mock import patch

from django.contrib.auth import get_user_model
from rest_framework.reverse import reverse

from pount.apps.api.models import Project, Referential, Template
from pount.apps.api.models.init import initialize_roles
from pount.apps.api.tests.factories.project import ProjectFactory, ProjectForManagerFactory, ProjectForMemberFactory
from pount.apps.api.tests.factories.referential import (
    BaseReferentialFactory,
    ProjectReferentialFactory,
    PublicReferentialFactory,
    get_definition_fixture,
)
from pount.apps.api.tests.factories.template import TemplateFactory
from pount.apps.api.tests.factories.user import SuperUserFactory, UserFactory
from pount.exceptions import LevelNotFoundError
from pount.tests import APITestCase

User = get_user_model()


class ListNestedReferentialTest(APITestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.member = UserFactory()
        cls.project = ProjectForMemberFactory(members=cls.member)

    def test_anonymous_access_is_not_authorized(self):
        response = self.get(reverse("nested_referential-list", kwargs={"project_pk": self.project.id}))
        self.response_unauthorized(response)

    def test_member_can_list_project_referentials(self):
        for i in range(3):
            ProjectReferentialFactory(creator=self.member, project=self.project)
        response = self._get_list_response()
        self.response_ok(response)
        self.assertEqual(
            response.data["count"],
            3,
        )

    def _get_list_response(self):
        response = self.get(
            reverse("nested_referential-list", kwargs={"project_pk": self.project.id}),
            user=self.member,
        )
        return response

    def test_list_returns_global_referentials_used_in_project(self):
        global_ref = BaseReferentialFactory()
        project_ref = self._get_project_referential()
        self.project.external_referentials.add(global_ref)
        response = self._get_list_response()
        self.response_ok(response)
        self.assertEqual(response.data["count"], 2)
        self.assertListEqual(
            sorted([_["id"] for _ in response.data["results"]]),
            sorted([str(global_ref.id), str(project_ref.id)]),
        )

    def _get_project_referential(self):
        project_ref = ProjectReferentialFactory(creator=self.member, project=self.project)
        return project_ref

    def test_list_returns_public_referential_used_in_project(self):
        public_ref = PublicReferentialFactory()
        project_ref = self._get_project_referential()
        self.project.external_referentials.add(public_ref)
        response = self._get_list_response()
        self.response_ok(response)
        self.assertEqual(response.data["count"], 2)
        self.assertListEqual(
            sorted([_["id"] for _ in response.data["results"]]),
            sorted([str(public_ref.id), str(project_ref.id)]),
        )

    def test_return_404_if_project_pk_is_not_a_uuid(self):
        response = self.get(
            reverse("nested_referential-list", kwargs={"project_pk": "undefined"}),
            user=self.member,
        )
        self.response_not_found(response)

    def test_project_pk_is_not_a_valid_project_id(self):
        response = self.get(
            reverse("nested_referential-list", kwargs={"project_pk": uuid.uuid4()}),
            user=self.member,
        )
        self.response_not_found(response)


class ListReferentialTest(APITestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.member = UserFactory()
        cls.project = ProjectForMemberFactory(members=cls.member)

    def test_anonymous_access_is_not_authorized(self):
        response = self.get(reverse("referential-list"))
        self.response_unauthorized(response)

    def test_user_can_list_public_referentials(self):
        PublicReferentialFactory()
        PublicReferentialFactory()
        response = self.get(reverse("referential-list"), user=self.member)
        self.response_ok(response)
        self.assertEqual(response.data["count"], 2)

    def test_user_can_list_base_referentials(self):
        BaseReferentialFactory()
        response = self.get(reverse("referential-list"), user=self.member)
        self.response_ok(response)
        self.assertEqual(response.data["count"], 1)


class CreateReferentialTest(APITestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.manager = UserFactory()
        cls.project = ProjectForManagerFactory(managers=[cls.manager])
        cls.definition = get_definition_fixture()

    def test_anonymous_access_is_not_authorized(self):
        response = self.post(reverse("referential-list"), user=None)
        self.response_unauthorized(response)

    def test_member_is_not_authorized_to_create_a_referential(self):
        member = UserFactory()
        self.project.add_member(member)
        response = self.post(reverse("referential-list"), user=member)
        self.response_forbidden(response)

    def test_manager_can_create_a_referential(self):
        data = {
            "name": "A referential name",
            "project_id": self.project.id,
            "scope": Referential.Scope.SCOPE_PROJECT,
            "definition": self.definition,
            "languages": {"available": [], "default": "fr", "translations": {}},
        }
        with patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save"):
            with patch("pount.apps.api.models.referential.handle_build_node_lists_on_referential_save"):
                response = self.post(
                    reverse("referential-list"),
                    data=data,
                    user=self.manager,
                    format="json",
                )
        self.response_created(response)

    def test_private_referential_must_be_linked_to_a_project(self):
        response = self.post(
            reverse("referential-list"),
            user=self.manager,
            data={
                "name": "A referential name",
                "scope": Referential.Scope.SCOPE_PROJECT,
                "definition": self.definition,
                "languages": {"available": [], "default": "fr", "translations": {}},
            },
            format="json",
        )
        self.response_bad_request(response)
        self.assertIn(
            "project_id is required",
            str(response.data["project_id"][0]),
        )

    def test_public_referential_must_be_linked_to_a_project(self):
        response = self.post(
            reverse("referential-list"),
            user=self.manager,
            data={
                "name": "A referential name",
                "scope": Referential.Scope.SCOPE_PUBLIC,
                "definition": self.definition,
                "languages": {"available": [], "default": "fr", "translations": {}},
            },
            format="json",
        )
        self.response_bad_request(response)
        self.assertIn(
            "project_id is required",
            str(response.data["project_id"][0]),
        )

    def test_manager_can_not_create_a_referential_with_global_scope(self):
        response = self.post(
            reverse("referential-list"),
            user=self.manager,
            data={
                "name": "A referential name",
                "project_id": self.project.id,
                "scope": Referential.Scope.SCOPE_BASE,
                "definition": {},
                "languages": {"available": [], "default": "fr", "translations": {}},
            },
            format="json",
        )
        self.response_forbidden(response)

    def test_superadmin_can_create_a_referential_with_global_scope(self):
        superadmin = UserFactory(is_superuser=True)

        with patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save"):
            with patch("pount.apps.api.models.referential.handle_build_node_lists_on_referential_save"):
                response = self.post(
                    reverse("referential-list"),
                    user=superadmin,
                    data={
                        "name": "A referential name",
                        "project_id": None,
                        "scope": Referential.Scope.SCOPE_BASE,
                        "definition": self.definition,
                        "languages": {"available": [], "default": "fr", "translations": {}},
                    },
                    format="json",
                )
        self.response_created(response)

    def test_base_referential_can_not_have_a_project(self):
        admin = UserFactory(is_superuser=True)
        response = self.post(
            reverse("referential-list"),
            user=admin,
            data={
                "name": "A base referential",
                "project_id": self.project.id,
                "scope": Referential.Scope.SCOPE_BASE,
                "definition": self.definition,
                "languages": {"available": [], "default": "fr", "translations": {}},
            },
            format="json",
        )
        self.response_bad_request(response)
        self.assertIn(
            "project_id is not allowed",
            str(response.data["project_id"][0]),
        )

    def test_referential_name_must_be_unique(self):
        referential = ProjectReferentialFactory(creator=self.manager, project=self.project)
        response = self.post(
            reverse("referential-list"),
            user=self.manager,
            data={
                "name": referential.name,
                "project_id": self.project.id,
                "scope": Referential.Scope.SCOPE_PROJECT,
                "definition": self.definition,
                "languages": {"available": [], "default": "fr", "translations": {}},
            },
            format="json",
        )
        self.response_bad_request(response)
        self.assertIn(
            "Referential with this Name already exists",
            str(response.data["name"][0]),
        )


class UpdateReferentialTest(APITestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.member = UserFactory()
        cls.project = ProjectForMemberFactory(members=cls.member)

    def test_anonymous_access_if_not_authorized(self):
        referential = ProjectReferentialFactory(creator=self.member, project=self.project)
        response = self._get_response(referential, user=None)
        self.response_unauthorized(response)

    def _get_response(self, referential: Referential, user: User = None, data: dict = None):
        data = data or {}
        with patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save"):
            with patch("pount.apps.api.models.referential.handle_build_node_lists_on_referential_save"):
                response = self.patch(
                    reverse("referential-detail", kwargs={"pk": referential.id}),
                    data=data,
                    user=user,
                    format="json",
                )
        return response

    def test_member_is_not_allowed_to_update_referential(self):
        referential = ProjectReferentialFactory(project=self.project)
        response = self._get_response(referential, user=self.member)
        self.response_forbidden(response)

    def test_manager_can_update_private_referential(self):
        self.project.add_manager(self.member)
        referential = ProjectReferentialFactory(project=self.project)
        data = {
            "name": "Changed name",
        }
        response = self._get_response(referential, user=self.member, data=data)
        referential.refresh_from_db(fields=["name"])
        self.response_ok(response)
        self.assertEqual(
            referential.name,
            "Changed name",
        )

    def test_manager_can_not_update_global_referential(self):
        self.project.add_manager(self.member)
        referential = BaseReferentialFactory()
        data = {
            "name": "Changed name",
        }
        response = self._get_response(referential, user=self.member, data=data)
        self.response_forbidden(response)

    def test_superadmin_can_update_a_base_referential(self):
        admin = UserFactory(is_superuser=True)
        referential = BaseReferentialFactory()
        data = {
            "name": "Changed name",
        }
        response = self._get_response(referential, user=admin, data=data)
        self.response_ok(response)
        referential.refresh_from_db(fields=["name"])
        self.assertEqual(
            referential.name,
            "Changed name",
        )

    def test_definition_json_is_validated(self):
        invalid_definition = {"A": "B"}
        self.project.add_manager(self.member)
        referential = ProjectReferentialFactory(project=self.project)
        data = {"definition": invalid_definition}
        response = self._get_response(referential, user=self.member, data=data)
        self.response_bad_request(response)
        self.assertIn(
            "JSON Schema validation error",
            str(response.data["definition"]),
        )


class DeleteReferentialTest(APITestCase):
    def setUp(self) -> None:
        # Mock the MongoDB
        patcher1 = patch("pount.apps.api.signals.delete_referential_in_mongodb")
        self.mock_delete_referential_data_in_mongo = patcher1.start()
        self.addCleanup(patcher1.stop)

        patcher2 = patch("pount.apps.api.signals.clear_cache_for_referential")
        self.mock_clear_cache = patcher2.start()
        self.addCleanup(patcher2.stop)

    @classmethod
    def setUpTestData(cls):
        initialize_roles()

    def test_anonymous_access_is_not_authorized(self):
        referential = PublicReferentialFactory()
        response = self._get_response(referential)
        self.response_unauthorized(response)

    def _get_response(self, referential: Referential, user: User = None):
        response = self.delete(
            reverse("referential-detail", kwargs={"pk": referential.id}),
            user=user,
            format="json",
        )
        return response

    def test_project_member_access_is_forbidden(self):
        member = UserFactory()
        project = ProjectForMemberFactory(members=member)
        referential = PublicReferentialFactory(project=project)
        response = self._get_response(referential, user=member)
        self.response_forbidden(response)

    def test_base_referential_can_be_deleted_by_superadmin(self):
        admin = SuperUserFactory()
        referential = BaseReferentialFactory()
        response = self._get_response(referential, user=admin)
        self.response_no_content(response)
        with self.assertRaises(Referential.DoesNotExist):
            referential.refresh_from_db()

    def test_base_referential_can_not_be_deleted_by_manager(self):
        manager = UserFactory()
        ProjectForManagerFactory(managers=[manager])
        referential = BaseReferentialFactory()
        response = self._get_response(referential, manager)
        self.response_forbidden(response)

    def test_manager_can_not_delete_referential_from_another_project(self):
        manager = UserFactory()
        ProjectForManagerFactory(managers=[manager])
        other_project = ProjectFactory()
        referential = ProjectReferentialFactory(project=other_project)
        response = self._get_response(referential, manager)
        self.response_forbidden(response)

    def test_manager_can_delete_project_referential(self):
        manager = UserFactory()
        project = ProjectForManagerFactory(managers=[manager])
        referential = ProjectReferentialFactory(project=project)
        response = self._get_response(referential, manager)
        self.response_no_content(response)
        with self.assertRaises(Referential.DoesNotExist):
            referential.refresh_from_db()

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_manager_can_not_delete_referential_if_used_in_a_template(self, update_mock):
        manager = UserFactory()
        project = ProjectForManagerFactory(managers=[manager])
        referential = ProjectReferentialFactory(project=project)
        template: Template = TemplateFactory(project=project)
        template.used_referentials.add(referential)
        response = self._get_response(referential, manager)
        self.response_forbidden(response)

    def test_manager_can_not_delete_referential_if_used_in_another_project(self):
        manager = UserFactory()
        project = ProjectForManagerFactory(managers=[manager])
        referential = ProjectReferentialFactory(project=project, scope=Referential.Scope.SCOPE_PUBLIC)
        other_project = ProjectFactory()
        other_project.external_referentials.add(referential)
        response = self._get_response(referential, manager)
        self.response_forbidden(response)


class UseReferentialInProjectTest(APITestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.project: Project = ProjectFactory()
        cls.public_referential: Referential = PublicReferentialFactory()

    def test_anonymous_access_is_not_authorized(self):
        response = self.patch(self._get_url(), user=None)
        self.response_unauthorized(response)

    def test_member_is_not_allowed_to_use_referential_in_project(self):
        member = UserFactory()
        self.project.add_member(member)
        response = self.patch(
            self._get_url(),
            user=member,
            format="json",
        )
        self.response_forbidden(response)

    def _get_url(self, referential: Referential = None) -> str:
        referential = referential or self.public_referential
        url = reverse(
            "nested_referential-use-referential-in-project",
            kwargs={"project_pk": self.project.id, "pk": referential.pk},
        )
        return url

    def test_manager_is_allowed_to_use_referential_in_project(self):
        manager = self._get_manager()
        response = self.patch(
            self._get_url(),
            user=manager,
            format="json",
        )
        self.response_ok(response)
        self.assertIn(
            self.public_referential,
            self.project.external_referentials.all(),
        )

    def _get_manager(self) -> User:
        manager = UserFactory()
        self.project.add_member(manager)
        self.project.add_manager(manager)
        return manager

    def test_manager_can_not_use_private_referential_in_project(self):
        manager = self._get_manager()
        private_referential = ProjectReferentialFactory()
        response = self.patch(
            self._get_url(referential=private_referential),
            user=manager,
            format="json",
        )
        self.response_bad_request(response)
        self.assertIn(
            "Referential is not public",
            str(response.data[0]),
        )

    def test_anonymous_user_is_not_authorized_to_remove(self):
        response = self._get_delete_response(user=None)
        self.response_unauthorized(response)

    def _get_delete_response(self, user: User):
        response = self.delete(
            reverse(
                "nested_referential-unuse-referential-in-project",
                kwargs={"project_pk": self.project.id, "pk": self.public_referential.id},
            ),
            user=user,
        )
        return response

    def test_member_can_not_remove_external_referential_from_project(self):
        member = UserFactory()
        self.project.add_member(member)
        self.project.external_referentials.add(self.public_referential)
        response = self._get_delete_response(user=member)
        self.response_forbidden(response)

    def test_manager_can_remove_external_referential_from_project(self):
        manager = self._get_manager()
        self.project.external_referentials.add(self.public_referential)
        response = self._get_delete_response(user=manager)
        self.response_no_content(response)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_manager_can_not_remove_external_referential_if_used_in_any_template(self, update_mock):
        template: Template = TemplateFactory()
        manager = self._get_manager()
        self.project.templates.add(template)
        self.project.external_referentials.add(self.public_referential)
        template.used_referentials.add(self.public_referential)
        response = self._get_delete_response(user=manager)
        self.response_forbidden(response)
        self.assertIn(
            "Referential is used in one or more templates",
            str(response.data),
        )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_manager_can_remove_external_referential_if_used_in_template_of_another_project(self, update_mock):
        template: Template = TemplateFactory()
        manager = self._get_manager()
        other_project = ProjectForManagerFactory(managers=[manager])
        other_project.templates.add(template)
        other_project.external_referentials.add(self.public_referential)
        template.used_referentials.add(self.public_referential)
        self.project.external_referentials.add(self.public_referential)
        response = self._get_delete_response(user=manager)
        self.response_no_content(response)

    def test_use_project_referential_as_external_does_nothing(self):
        manager = UserFactory()
        project = ProjectForManagerFactory(managers=[manager])
        referential = PublicReferentialFactory(creator=manager, project=project)
        response = self.patch(
            reverse(
                "nested_referential-use-referential-in-project",
                kwargs={"project_pk": project.id, "pk": referential.pk},
            ),
            user=manager,
            format="json",
        )
        self.response_ok(response)
        self.assertNotIn(
            referential,
            self.project.external_referentials.all(),
        )


class ChangeReferentialScopeTest(APITestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.manager = UserFactory()
        cls.project = ProjectForManagerFactory(managers=[cls.manager])
        cls.referential = ProjectReferentialFactory(creator=cls.manager, project=cls.project)

    def test_make_referential_public(self):
        with patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save"):
            with patch("pount.apps.api.models.referential.handle_build_node_lists_on_referential_save"):
                response = self.patch(
                    reverse("referential-share", kwargs={"pk": self.referential.id}),
                    data={"scope": Referential.Scope.SCOPE_PUBLIC},
                    user=self.manager,
                    format="json",
                )
        self.referential.refresh_from_db(fields=["scope"])
        self.response_ok(response)
        self.assertEqual(
            self.referential.scope,
            Referential.Scope.SCOPE_PUBLIC,
        )

    def test_manager_can_not_make_referential_global(self):
        response = self.patch(
            reverse("referential-share", kwargs={"pk": self.referential.id}),
            data={"scope": Referential.Scope.SCOPE_BASE},
            user=self.manager,
            format="json",
        )
        self.response_forbidden(response)

    def test_admin_can_make_referential_global(self):
        admin = UserFactory(is_superuser=True)
        with patch("pount.apps.api.models.referential.handle_update_item_metadata_on_referential_save"):
            with patch("pount.apps.api.models.referential.handle_build_node_lists_on_referential_save"):
                response = self.patch(
                    reverse("referential-share", kwargs={"pk": self.referential.id}),
                    data={"scope": Referential.Scope.SCOPE_BASE},
                    user=admin,
                    format="json",
                )
        self.response_ok(response)

    def test_member_can_not_change_referential_scope(self):
        member = UserFactory()
        self.project.add_member(member)
        response = self.patch(
            reverse("referential-share", kwargs={"pk": self.referential.id}),
            data={"scope": Referential.Scope.SCOPE_BASE},
            user=member,
            format="json",
        )
        self.response_forbidden(response)

    def test_global_referential_can_not_be_made_public_or_private(self):
        admin = UserFactory(is_superuser=True)
        referential = BaseReferentialFactory()
        response = self.patch(
            reverse("referential-share", kwargs={"pk": referential.id}),
            data={"scope": Referential.Scope.SCOPE_PUBLIC},
            user=admin,
            format="json",
        )
        self.response_bad_request(response)

    def test_if_referential_is_used_in_another_project_it_cannot_be_made_private(self):
        manager = UserFactory()
        project: Project = ProjectForManagerFactory(managers=[manager])
        referential: Referential = PublicReferentialFactory(creator=manager, project=project)
        other_project: Project = ProjectFactory()
        other_project.external_referentials.add(referential)
        response = self.patch(
            reverse("referential-share", kwargs={"pk": referential.id}),
            data={"scope": Referential.Scope.SCOPE_PROJECT},
            user=manager,
            format="json",
        )
        self.response_bad_request(response)
        self.assertIn(
            "Referential is used in other project",
            response.data["scope"],
        )


class ListOfLanguagesTest(APITestCase):
    def test_return_list_of_languages_for_a_referential(self):
        referential = BaseReferentialFactory(languages={"default": "fr", "available": ["fr", "cs"]})
        response = self.get(reverse("referential-languages", kwargs={"pk": referential.id}), format="json", user=None)
        self.response_ok(response)
        self.assertEqual(
            [
                {"code": "fr", "name": "Français"},
                {"code": "cs", "name": "Čeština"},
            ],
            response.data,
        )

    def test_anonymous_access_to_a_list_of_languages_of_a_private_referential_is_forbidden(self):
        initialize_roles()
        referential = ProjectReferentialFactory()
        response = self.get(reverse("referential-languages", kwargs={"pk": referential.id}), format="json", user=None)
        self.response_forbidden(response)


class NodeForLevelTest(APITestCase):
    def test_return_404_if_level_does_not_exist(self):
        referential = BaseReferentialFactory()
        level_id = "b73bdfe3-abcd-4162-b91b-24adcd1cfb11"
        ancestor_id = "e53dec8a-7a27-48c5-aa4b-57cb591f21a6"
        with patch("pount.apps.api.views.referential.get_node_list") as node_list:
            node_list.side_effect = LevelNotFoundError(f"Level {level_id} not found")
            response = self.get(
                f'{reverse("referential-nodes", kwargs={"pk": referential.pk})}?level_id={level_id}=fr&ancestor={ancestor_id}',  # noqa: E501
                format="json",
                user=None,
            )
        self.response_not_found(response)
        node_list.assert_called_once()
        self.assertIn(f"Level {level_id} not found", str(response.data["detail"]))


class UseReferentialInTemplateTest(APITestCase):
    def test_anonymous_access_is_not_allowed(self):
        referential = BaseReferentialFactory()
        response = self.get(reverse("referential-templates", kwargs={"pk": referential.pk}), format="json", user=None)
        self.response_unauthorized(response)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_manager_can_list_templates_in_which_the_referential_is_used(self, update_mock):
        manager = UserFactory()
        referential: Referential = BaseReferentialFactory()
        template: Template = TemplateFactory()
        template.used_referentials.set([referential])
        response = self.get(
            reverse("referential-templates", kwargs={"pk": referential.pk}),
            user=manager,
            format="json",
        )
        self.response_ok(response)
        self.assertListEqual(
            [_["id"] for _ in response.data],
            [str(template.id)],
        )
