from unittest.mock import patch

from rest_framework.reverse import reverse

from pount.apps.api.libs.set import update_set_search_config_on_template_save
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.tests.factories.set import SetFactory
from pount.tests import APITestCase


class UpdateSearchConfigTest(APITestCase):
    def setUp(self) -> None:
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()
        self.patcher2 = patch("pount.apps.api.views.set.cache")
        self.cache_mock = self.patcher2.start()
        initialize_roles()
        initialize_templates()
        self.set = SetFactory()
        self.set.project.add_member(self.set.creator)
        self.set.project.add_manager(self.set.creator)
        update_set_search_config_on_template_save(self.set.template)

    def tearDown(self) -> None:
        self.patcher1.stop()
        self.patcher2.stop()

    def test_anonymous_access_is_not_authorized(self):
        response = self.put(
            reverse("set-search-config", kwargs={"pk": self.set.id}),
            format="json",
            user=None,
        )
        self.response_unauthorized(response)

    def test_invalid_search_config_raises_validation_error(self):
        response = self.put(
            reverse("set-search-config", kwargs={"pk": self.set.id}),
            data=[
                {
                    "name": "creator",
                    "label": {"fr": "Créateur", "en": "Creator"},
                    "config": {
                        "mapping_type": "invalid_type",
                    },
                }
            ],
            format="json",
            user=self.set.creator,
        )
        self.response_bad_request(response)
        self.assertIn("JSON Schema validation error", str(response.data))

    def test_configured_fields_must_match_template_fields(self):
        response = self.put(
            reverse("set-search-config", kwargs={"pk": self.set.id}),
            data=[
                {
                    "name": "not_in_template",
                    "label": {"fr": "français", "en": "english"},
                    "config": {
                        "mapping_type": None,
                        "use_in_search": False,
                        "use_in_aggregation": False,
                    },
                }
            ],
            format="json",
            user=self.set.creator,
        )
        self.response_bad_request(response)
        self.assertIn("search_config and template don't have the same set of fields", str(response.data))

    def test_if_template_field_has_no_name_search_config_is_set_without_that_field(self):
        # If a template field has no name it will be excluded when the view checks if all fields are present
        # when submitting seach_config
        self.set.template.fields.append(
            {
                "id": "abcdef123",
                "type": "text",
                "label": "Label",
                "required": False,
                "visible": False,
                "component": "UsfText",
            }
        )
        self.set.template.save()
        data = self.set.search_config
        response = self.put(
            reverse("set-search-config", kwargs={"pk": self.set.id}),
            data=data,
            format="json",
            user=self.set.creator,
        )
        self.response_ok(response)

    def test_update_search_config(self):
        data = self.set.search_config
        data[0]["config"]["mapping_type"] = "strings_fr"
        data[0]["config"]["use_in_search"] = True
        response = self.put(
            reverse("set-search-config", kwargs={"pk": self.set.id}),
            data=data,
            format="json",
            user=self.set.creator,
        )
        self.response_ok(response)
        self.set.refresh_from_db()
        self.assertEqual(self.set.search_config[0]["config"]["mapping_type"], "strings_fr")
        self.assertEqual(self.set.search_config[0]["config"]["use_in_search"], True)

    def test_cache_is_cleaned_on_update_search_config(self):
        self.set.search_config[0]["config"]["mapping_type"] = "strings_fr"
        self.set.search_config[0]["config"]["use_in_search"] = True
        response = self.put(
            reverse("set-search-config", kwargs={"pk": self.set.id}),
            data=self.set.search_config,
            format="json",
            user=self.set.creator,
        )
        self.response_ok(response)
        self.assertEqual(self.cache_mock.delete_pattern.call_count, 3)
        self.cache_mock.delete_pattern.assert_any_call(f"aggregations_for_set_{self.set.id}_*")
        self.cache_mock.delete_pattern.assert_any_call("aggregations_default_*")
        self.cache_mock.delete_pattern.assert_any_call(f"search_fields_for_set_{self.set.id}")
