from __future__ import annotations

import json
from pathlib import PosixPath
from typing import TypeVar

from django.conf import settings
from django.contrib.auth.password_validation import get_default_password_validators
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from jsonschema.exceptions import ValidationError as SchemaValidationError
from jsonschema.validators import validate
from rest_framework import serializers

T = TypeVar("T")


class PasswordValidator:
    """
    Apply Django Password validators configured in settings
    AUTH_PASSWORD_VALIDATORS
    """

    def __init__(self):
        self.validators = get_default_password_validators()

    def __call__(self, value, *args):
        for validator in self.validators:
            try:
                validator.validate(value, *args)
            except ValidationError as error:
                # convert Django ValidationError to DRF ValidationError
                raise serializers.ValidationError(error.message)


class JSONSchemaValidator:
    schema: PosixPath

    def __init__(self, schema: str | PosixPath):
        if not isinstance(schema, PosixPath):
            schema = settings.JSON_SCHEMA_DIR / schema
        if not schema.is_file():
            raise ValueError(_("Schema is not a file or is not readable"))
        self.schema = schema

    def __call__(self, value: T) -> T:
        with open(self.schema) as schema:
            try:
                validate(value, schema=json.load(schema))
            except SchemaValidationError as err:
                raise serializers.ValidationError("JSON Schema validation error\n" + str(err))
        return value
