from django.contrib.auth import get_user_model
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, inline_serializer
from rest_framework import mixins, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rules.contrib.views import objectgetter, permission_required

from pount.apps.api import rules
from pount.apps.api.libs.group import add_set_to_group, remove_user_from_group
from pount.apps.api.libs.notifications import notify_group, notify_group_contribution_to_set, notify_group_membership
from pount.apps.api.models.group import Group
from pount.apps.api.models.permission import Permission, Role
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.permissions import GroupPermissions
from pount.apps.api.serializers.group import CreateGroupSerializer, GroupSerializer
from pount.apps.api.views.util import dataobjectgetter

User = get_user_model()


class GroupViewSet(
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    viewsets.GenericViewSet,
):
    queryset = Group.objects.all()
    serializer_class = GroupSerializer
    permission_classes = [GroupPermissions]

    def update(self, request, *args, **kwargs):
        response = super().update(request, *args, **kwargs)
        notify_group(request.user, self.get_object(), "group_update")
        return response

    def destroy(self, request, *args, **kwargs):
        group = self.get_object()
        response = super().destroy(request, *args, **kwargs)
        notify_group(request.user, group, "group_delete")
        return response

    @extend_schema(
        request=CreateGroupSerializer(),
        responses={
            status.HTTP_201_CREATED: CreateGroupSerializer,
        },
    )
    @method_decorator(
        permission_required(
            rules.PROJECT_EDIT, fn=dataobjectgetter(Project, attr_name="project_id"), raise_exception=True
        )
    )
    @action(methods=["POST"], detail=False)
    def create_for_project(self, request):
        project = get_object_or_404(Project, pk=request.data.get("project_id"))
        serializer = CreateGroupSerializer(
            data=request.data,
            project=project,
            context=self.get_serializer_context(),
        )
        if serializer.is_valid(raise_exception=True):
            instance: Group = serializer.save()
            notify_group(request.user, instance, "group_create")
            for user in instance.users.distinct():
                notify_group_membership(request.user, user, instance, "group_user_add")
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @extend_schema(
        request=inline_serializer(
            "GroupAddUserRequestSerializer",
            fields={
                "user_id": serializers.UUIDField(),
            },
        ),
        responses={
            status.HTTP_200_OK: GroupSerializer,
        },
    )
    @method_decorator(permission_required(rules.GROUP_EDIT, fn=objectgetter(Group), raise_exception=True))
    @action(methods=["POST"], detail=True)
    def add_user(self, request, pk):
        user = get_object_or_404(User, pk=request.data.get("user_id"))
        group = get_object_or_404(Group, pk=pk)
        if not user.is_active:
            raise ValidationError(_("User is not active."))
        if not user.has_perm(rules.PROJECT_VIEW, group.project):
            raise ValidationError(_("User is not a project member."))
        group.users.add(user)
        notify_group_membership(request.user, user, group, "group_user_add")
        serializer = self.get_serializer(instance=group)
        return Response(serializer.data)

    @extend_schema(
        request=inline_serializer(
            "GroupRemoveUserRequestSerializer",
            fields={
                "user_id": serializers.UUIDField(),
            },
        ),
        responses={
            status.HTTP_200_OK: inline_serializer(
                "GroupRemoveUserResponseSerializer",
                fields={
                    "email": serializers.CharField(help_text="Whether an email has been sent to the user"),
                    "group": GroupSerializer(),
                },
            ),
        },
    )
    @method_decorator(permission_required(rules.GROUP_EDIT, fn=objectgetter(Group), raise_exception=True))
    @action(methods=["DELETE"], detail=True)
    def remove_user(self, request, pk):
        """
        Remove user from a group.

        If the user is removed from the Members group, he is removed from the whole project.
        If the user is external and isn't member of any other group then he is disabled
        """
        # @todo don't pass user_id in request.data
        user = get_object_or_404(User, pk=request.data.get("user_id"))
        group: Group = get_object_or_404(Group, pk=pk)

        if group.name == Group.MANAGERS and len(group.users.exclude(id=user.id)) == 0:
            raise ValidationError(_("Group must have at least one manager."))

        email_result = remove_user_from_group(group, user, removed_by=request.user)

        if user.is_external and user.project_groups.count() == 0:
            # if user is external and isn't member of any other group
            # then disable him
            user.is_active = False
            user.save()

        serializer = self.get_serializer(instance=group)
        return Response({"email": email_result, "group": serializer.data})

    @extend_schema(
        request=inline_serializer(
            "GroupAddSetRequestSerializer",
            fields={
                "set_id": serializers.UUIDField(),
            },
        ),
        responses={status.HTTP_204_NO_CONTENT: None},
    )
    @method_decorator(permission_required(rules.GROUP_EDIT, fn=objectgetter(Group), raise_exception=True))
    @action(methods=["POST"], detail=True)
    def add_set(self, request, pk):
        group: Group = get_object_or_404(Group, pk=pk)
        set: Set = get_object_or_404(Set, project=group.project, id=request.data.get("set_id"))
        add_set_to_group(group, set)
        notify_group_contribution_to_set(
            actor=request.user,
            group=group,
            set=set,
            verb="group_contribute_to_set",
        )
        return Response(status=status.HTTP_204_NO_CONTENT)

    @extend_schema(
        request=inline_serializer(
            "GroupRemoveSetRequestSerializer",
            fields={
                "set_id": serializers.UUIDField(),
            },
        ),
        responses={
            status.HTTP_204_NO_CONTENT: None,
        },
    )
    @method_decorator(
        permission_required(
            rules.GROUP_EDIT,
            fn=objectgetter(Group),
            raise_exception=True,
        )
    )
    @action(methods=["DELETE"], detail=True)
    def remove_set(self, request, pk):
        # @todo don't pass set_id in request.data
        group: Group = get_object_or_404(Group, pk=pk)
        set: Set = get_object_or_404(Set, project=group.project, id=request.data.get("set_id"))
        Permission.objects.filter(
            group=group,
            role=Role.objects.get_contributor(),
            set__id=set.id,
        ).delete()
        notify_group_contribution_to_set(
            actor=request.user,
            group=group,
            set=set,
            verb="group_uncontribute_to_set",
        )
        return Response(status=status.HTTP_204_NO_CONTENT)
