import logging
import os.path
import zipfile

from django.contrib.auth import get_user_model
from django.db import IntegrityError, transaction
from django.http import FileResponse, HttpResponse
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view, inline_serializer
from rest_framework import exceptions, filters, mixins, serializers, status, viewsets
from rest_framework.decorators import action, api_view
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rules.contrib.views import objectgetter, permission_required

from pount.apps.api import rules
from pount.apps.api.libs.export import add_content_disposition_header_for_attachment, datacite_xml
from pount.apps.api.libs.file import item_mediafile_fullname, upload_rti_zipfile
from pount.apps.api.libs.item import import_items_from_csv
from pount.apps.api.models import Item, MediaFile, Set
from pount.apps.api.pagination import StandardPagination
from pount.apps.api.permissions import ItemPermission
from pount.apps.api.serializers.item import (
    ItemListSerializer,
    ItemSerializer,
    MediaFileSerializer,
    UpdateMediaFileSerializer,
)
from pount.apps.api.storage_backends import ItemMediaFileStorage
from pount.apps.api.views.mixins import ThumbnailMixin
from pount.exceptions import CSVImportError, FileNameExistsError

User = get_user_model()
logger = logging.getLogger(__file__)


class ItemViewSet(ThumbnailMixin, viewsets.ModelViewSet):
    queryset = Item.objects.all()
    permission_classes = [ItemPermission]

    serializer_class = ItemSerializer
    pagination_class = StandardPagination

    filter_backends = [DjangoFilterBackend, filters.OrderingFilter]
    filterset_fields = ["set_id"]
    ordering_fields = ["created_at", "updated_at", "title", "is_public"]

    def get_queryset(self):
        queryset = super().get_queryset()
        # There is no need to filter. If the user has access to the set, they can access the items
        # queryset = queryset.is_public_or_user_is_project_member(self.request.user)
        if self.action == "retrieve":
            # we need the parent objects to determine if it can be shared
            queryset = queryset.select_related("set", "set__project")
        return queryset

    def check_permissions(self, request):
        if not bool(request.user and request.user.is_authenticated):
            raise exceptions.AuthenticationFailed()
        if request.method == "POST" and self.action != "thumbnail":
            # for object creation, permission is checked on set
            try:
                set: Set = Set.objects.get(pk=request.data.get("set_id", None))
            except Set.DoesNotExist:
                raise exceptions.PermissionDenied()
            if not request.user.has_perm(rules.ITEM_CREATE, set):
                raise exceptions.PermissionDenied()
        return super().check_permissions(request)

    def get_serializer_class(self):
        if self.action in ["list", "import_csv"]:
            return ItemListSerializer
        return super().get_serializer_class()

    @extend_schema(
        responses={
            (status.HTTP_200_OK, "text/xml"): OpenApiTypes.BINARY,
        }
    )
    @action(methods=["GET"], detail=True, url_path="export/datacite")
    def export_datacite(self, request, pk):
        item = self.get_object()
        xml = datacite_xml(item)
        response = HttpResponse(
            xml.encode("utf-8"),
            status=status.HTTP_200_OK,
            content_type="application/xml",
        )
        response = add_content_disposition_header_for_attachment(f"{item.id}-datacite.xml", response)
        return response

    @extend_schema(
        request=inline_serializer(
            "ImportItemsFromCSVFileSerializer",
            fields={
                "set_id": serializers.UUIDField(),
                "csv": serializers.FileField(),
            },
        ),
        responses={
            status.HTTP_201_CREATED: inline_serializer(
                "ImportItemsFromCSVFileResponseSerializer",
                fields={
                    "created_items": serializers.IntegerField(),
                },
            )
        },
    )
    @action(methods=["POST"], detail=False)
    def import_csv(self, request):
        set = get_object_or_404(Set, pk=request.data.get("set_id", ""))
        csv_file = request.FILES.get("csv", None)
        if not csv_file:
            return Response(
                {"error": _("Missing file")},
                status=status.HTTP_400_BAD_REQUEST,
            )
        try:
            with transaction.atomic():
                # If one record fails, we reject the whole export
                created_items = import_items_from_csv(csv_file, set, request.user)
        except CSVImportError as e:
            logger.exception(_("Error processing CSV import"))
            return Response(
                {"error": e.error or _("Error processing CSV import")},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response({"created_items": created_items}, status=status.HTTP_201_CREATED)

    @extend_schema(
        request=inline_serializer(
            "DeleteItemRequestSerializer",
            fields={
                "item_ids": serializers.ListField(child=serializers.UUIDField()),
            },
        ),
    )
    def destroy(self, request, *args, **kwargs):
        """Delete one item (detail = True) or bulk delete items (detail = False)"""
        # @todo don't pass item_ids in request.data
        if self.detail:
            # Delete one single Item
            return super().destroy(request, *args, **kwargs)
        else:
            # Bulk delete objects
            items = Item.objects.is_user_creator_or_project_manager(request.user).filter(
                pk__in=request.data.get("item_ids")
            )
            items.delete()
            return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(
    # tags=["api", "Item", "File"],
    request=inline_serializer(
        "PostUrlRequest",
        fields={
            "filename": serializers.CharField(),
            "content_type": serializers.CharField(),
        },
    ),
    responses={
        status.HTTP_200_OK: OpenApiTypes.STR,
    },
)
@api_view(["POST"])
@permission_required(
    rules.ITEM_EDIT,
    fn=objectgetter(Item),
    raise_exception=True,
)
def post_url(request, pk):
    """
    Build the URL and form fields that can be used for a presigned S3 post.
    After uploading the file to S3, a MediaFile object must be created to associate the file to the Item.
    If a file with the same name already exists for the item, it must be deleted first.
    """
    item = get_object_or_404(Item, pk=pk)
    filename = request.data.get("filename", None)
    content_type = request.data.get("content_type", None)
    if filename is None or content_type is None:
        return Response(
            "filename or content_type are missing",
            status=status.HTTP_400_BAD_REQUEST,
        )
    try:
        url = ItemMediaFileStorage().post_url(item_mediafile_fullname(filename, item), content_type)
    except FileNameExistsError:
        raise exceptions.ValidationError({"filename": _("A file with this name already exists for this item")})
    return Response(data=url, status=status.HTTP_200_OK)


@extend_schema_view(
    retrieve=extend_schema(
        parameters=[OpenApiParameter("items_pk", OpenApiTypes.UUID, OpenApiParameter.PATH)],
    ),
    create=extend_schema(
        parameters=[OpenApiParameter("items_pk", OpenApiTypes.UUID, OpenApiParameter.PATH)],
        description=_("Associate an uploaded file to the item"),
    ),
    update=extend_schema(
        parameters=[OpenApiParameter("items_pk", OpenApiTypes.UUID, OpenApiParameter.PATH)],
        request=UpdateMediaFileSerializer,
        responses={
            status.HTTP_200_OK: UpdateMediaFileSerializer,
        },
    ),
    partial_update=extend_schema(
        parameters=[OpenApiParameter("items_pk", OpenApiTypes.UUID, OpenApiParameter.PATH)],
        request=UpdateMediaFileSerializer,
        responses={
            status.HTTP_200_OK: UpdateMediaFileSerializer,
        },
    ),
)
class MediaFileViewSet(
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.CreateModelMixin,
    mixins.DestroyModelMixin,
    viewsets.GenericViewSet,
):
    queryset = MediaFile.objects.all()
    serializer_class = MediaFileSerializer

    def check_permissions(self, request):
        # Permission must be checked on the parent Item object
        item = get_object_or_404(Item, pk=self.kwargs.get("items_pk"))
        if not all(
            [
                request.user.is_authenticated,
                request.user.has_perm(rules.ITEM_EDIT, item),
            ]
        ):
            self.permission_denied(request)

    def get_serializer_class(self):
        if self.action in ["update", "partial_update"]:
            return UpdateMediaFileSerializer
        return super().get_serializer_class()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["item_id"] = self.kwargs["items_pk"]
        return context

    def create(self, request, *args, **kwargs):
        try:
            response = super().create(request, *args, **kwargs)
        except IntegrityError:
            raise exceptions.ValidationError({"filename": _("A file with this name already exists for this item")})
        return response

    @extend_schema(
        parameters=[
            OpenApiParameter("items_pk", OpenApiTypes.UUID, OpenApiParameter.PATH),
        ]
    )
    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        if instance.info.get("is_rti", True):
            # For RTI files, all related files must be deleted too
            files_to_delete = MediaFile.objects.filter(item=instance.item, info__is_rti=True, is_hidden=True)
            for _file in files_to_delete:
                _file.delete()
        return super().destroy(request, *args, **kwargs)

    @extend_schema(
        parameters=[
            OpenApiParameter("items_pk", OpenApiTypes.UUID, OpenApiParameter.PATH),
        ]
    )
    @action(methods=["POST"], detail=False)
    def upload_rti(self, request, items_pk):
        file_obj = request.FILES.get("file", None)
        component = request.POST.get("component", None)
        if not file_obj:
            return Response(_("Missing file"), status=status.HTTP_400_BAD_REQUEST)

        try:
            zfile = zipfile.ZipFile(file_obj)
        except zipfile.BadZipfile:
            return Response(_("File is not a zip file"), status=status.HTTP_400_BAD_REQUEST)

        if "info.json" not in zfile.namelist():
            return Response(
                _("Missing info.json"),
                status=status.HTTP_400_BAD_REQUEST,
            )

        item = get_object_or_404(Item, id=items_pk)
        mediafile = upload_rti_zipfile(zfile, item, component)
        serializer = MediaFileSerializer(instance=mediafile, context={"request": request})

        return Response(serializer.data, status=status.HTTP_201_CREATED)


@extend_schema(
    responses={
        status.HTTP_200_OK: OpenApiTypes.BINARY,
    }
)
@api_view(["GET"])
# TODO: Check permission
# @permission_required(
#     rules.ITEM_EDIT,
#     fn=objectgetter(Item),
#     raise_exception=True,
# )
def get_rti_file(request, pk, filename):
    media_storage = ItemMediaFileStorage()
    key = os.path.join(f"{pk}/rti/", filename)
    file = media_storage.open(key)
    return FileResponse(file)
