import smtplib
import uuid
from datetime import timedelta

from django.contrib.auth import get_user_model, logout
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import update_last_login
from django.core import signing
from django.http import Http404, HttpResponseBadRequest, HttpResponseRedirect
from django.views.decorators.csrf import csrf_exempt
from drf_spectacular.utils import extend_schema, inline_serializer
from rest_framework import serializers, status
from rest_framework.decorators import api_view
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from pount.apps.api.libs import login
from pount.apps.api.models import AuthenticationProvider, Ticket
from pount.apps.api.serializers.user import ResetPasswordSerializer

User = get_user_model()


@login_required(login_url="/cas/login/")
def login_cas(request):
    """
    Authenticate the user with CAS authentication server, create an
    authentication ticket and redirect to client with ticket id
    """
    redirect_url = request.GET.get("redirect", None)
    if redirect_url is None:
        return HttpResponseBadRequest()
    ticket = login.create_ticket(request.user, provider=AuthenticationProvider.CAS.value)
    redirect_url = login.add_ticket_to_url(redirect_url, ticket.id)
    return HttpResponseRedirect(redirect_url)


def user_logout(request):
    redirect_url = request.GET.get("redirect", None)
    if redirect_url is None:
        return HttpResponseBadRequest()
    logout(request)
    return HttpResponseRedirect(redirect_url)


@extend_schema(
    request=inline_serializer("JWTTokenSerializer", fields={"ticket": serializers.CharField()}),
    responses={
        status.HTTP_200_OK: inline_serializer(
            "TokenPairSerializer", fields={"access": serializers.CharField(), "refresh": serializers.CharField()}
        ),
    },
)
@api_view(["POST"])
@csrf_exempt
def jwt_token(request):
    """
    Get a JWT Authentication token in exchange for an authentication ticket
    """
    ticket_id = request.data.get("ticket", None)
    if ticket_id is None:
        return Response(
            {"error": "missing_ticket"},
            status=status.HTTP_400_BAD_REQUEST,
        )
    try:
        ticket = Ticket.objects.get(pk=ticket_id)
        if ticket.is_expired:
            response = Response(
                {"error": "expired"},
                status=status.HTTP_401_UNAUTHORIZED,
            )
        else:
            tokens = login.make_jwt_token_pair(ticket.user, request)
            update_last_login(None, ticket.user)
            response = Response(tokens)
        ticket.delete()
        return response
    except Ticket.DoesNotExist:
        return Response(
            {"error": "does_not_exist"},
            status=status.HTTP_404_NOT_FOUND,
        )
    except ValidationError:
        return Response(
            {"error": "invalid_user"},
            status=status.HTTP_401_UNAUTHORIZED,
        )


@extend_schema(
    request=inline_serializer(
        "PasswordResetInitializeRequest",
        fields={
            "email": serializers.EmailField(),
            "url": serializers.CharField(),
        },
    ),
    responses={
        status.HTTP_204_NO_CONTENT: None,
    },
)
@api_view(["POST"])
def password_reset_initialize(request):
    """
    Send email with a link to reset the password
    """
    email = request.data.get("email")
    try:
        user = User.objects.get_by_natural_key(email, AuthenticationProvider.LOCAL.value)
    except User.DoesNotExist:
        raise Http404
    reset_id = uuid.uuid4()
    token = signing.dumps({"email": email}, salt="password-reset")
    user.settings["password_reset"] = {
        "token": token,
        "reset_id": str(reset_id),
    }
    try:
        login.send_reset_password_email(user, request)
        user.save()
        return Response(status=status.HTTP_204_NO_CONTENT)
    except ValueError:
        return Response(
            {"error": "missing_url"},
            status=status.HTTP_400_BAD_REQUEST,
        )
    except smtplib.SMTPException:
        return Response(
            {"error": "send_email_failed"},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )


@extend_schema(
    request=inline_serializer(
        "PasswordResetSigninRequest",
        fields={
            "token": serializers.CharField(),
            "rid": serializers.UUIDField(),
            "new_password": serializers.CharField(),
            "new_password2": serializers.CharField(),
        },
    ),
    responses={
        status.HTTP_200_OK: inline_serializer(
            "PasswordResetTokenPairSerializer",
            fields={"access": serializers.CharField(), "refresh": serializers.CharField()},
        ),
    },
)
@api_view(["POST"])
def password_reset_signin(request):
    try:
        raw_token = request.data.get("token")
        reset_id = request.data.get("rid")
        expiration_delay = timedelta(hours=24)  # token is valid for 24 hours
        decoded_token = signing.loads(raw_token, salt="password-reset", max_age=expiration_delay)
        user = User.objects.get(
            email=decoded_token["email"],
            settings__password_reset__token=raw_token,
            settings__password_reset__reset_id=reset_id,
        )
    except (
        KeyError,
        TypeError,
        User.DoesNotExist,
        signing.BadSignature,
    ):
        return Response({"error": "invalid_token"}, status=status.HTTP_403_FORBIDDEN)
    except signing.SignatureExpired:
        return Response(
            {"error": "expired"},
            status=status.HTTP_403_FORBIDDEN,
        )
    del user.settings["password_reset"]
    serializer = ResetPasswordSerializer(
        instance=user,
        data={
            "new_password": request.data.get("new_password"),
            "new_password_2": request.data.get("new_password2"),
        },
    )
    if serializer.is_valid(raise_exception=True):
        user = serializer.save()
    tokens = login.make_jwt_token_pair(user, request)
    update_last_login(None, user)
    response = Response(tokens)
    return response
