from django.db import models
from PIL import Image
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.exceptions import ParseError
from rest_framework.parsers import MultiPartParser
from rest_framework.response import Response

from pount.apps.api.libs.file import make_thumbnail, resize_min_height, upload_thumbnail
from pount.apps.api.storage_backends import ThumbnailFileStorage


class UnsupportedFileFormat(ParseError):
    pass


class ThumbnailMixin:
    SIZES = [
        ("s", (48, 48), make_thumbnail),
        ("m", (64, 64), make_thumbnail),
        ("l", (128, 128), make_thumbnail),
        ("embed", (None, 500), resize_min_height),
    ]

    @action(
        methods=["POST", "DELETE"],
        detail=True,
        parser_classes=[MultiPartParser],
    )
    def thumbnail(self, request, pk):
        instance = self.get_object()

        if request.method == "DELETE":
            return self.perform_delete(instance)

        if request.method == "POST":
            try:
                return self.perform_post(request, instance)
            except UnsupportedFileFormat:
                return Response(
                    {"error": "bad_format"},
                    status=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
                )
            except ParseError:
                return Response(status=status.HTTP_400_BAD_REQUEST)

    def perform_post(self, request, instance):
        model_name = instance._meta.model_name
        if "file" not in request.data:
            raise ParseError("Empty content")
        uploaded_file = request.data["file"]

        try:
            image: Image = Image.open(uploaded_file)
        except Exception as e:  # noqa: F841
            raise UnsupportedFileFormat("Unsupported image type")

        # width, height = image.size
        # if width != height:
        #     raise ParseError("Image width and height are not the same")

        storage = ThumbnailFileStorage()

        thumbnails = {}

        for size_code, size_dimensions, resize_function in self.SIZES:
            thumbnail = resize_function(image, size_dimensions)
            if thumbnail:
                filename = upload_thumbnail(
                    thumbnail,
                    f"{model_name}/{str(instance.id)}-{size_code}.png",
                    storage,
                )
                thumbnails[size_code] = {"filename": filename}
            else:
                thumbnails[size_code] = None

        if hasattr(instance, "settings"):
            if instance.settings is None:
                instance.settings = {}
            instance.settings["thumbnails"] = thumbnails
        else:
            if instance.info is None:
                instance.info = {}
            instance.info["thumbnails"] = thumbnails
        instance.save()

        return Response(
            {
                "thumbnails": {
                    size_code: instance.thumbnail_url(size_code)
                    for size_code, dimensions, resize_function in self.SIZES
                }
            }
        )

    def perform_delete(self, instance: models.Model):
        if hasattr(instance, "settings"):
            field_name = "settings"
            settings = instance.settings
        else:
            field_name = "info"
            settings = instance.info

        if settings is None or settings.get("thumbnails", None) is None:
            return Response(
                {"error": "No thumbnails for object"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        storage = ThumbnailFileStorage()
        for size in settings.get("thumbnails", {}).keys():
            storage.delete(settings["thumbnails"][size]["filename"])
        del settings["thumbnails"]

        setattr(instance, field_name, settings)
        instance.save()

        return Response(status=status.HTTP_204_NO_CONTENT)
