from django.contrib.auth import get_user_model
from django.db.models import QuerySet
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as df_filters
from drf_spectacular.utils import extend_schema, inline_serializer
from rest_framework import filters, serializers, status
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404
from rest_framework.parsers import MultiPartParser
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet
from rules.contrib.views import objectgetter, permission_required

from pount.apps.api import rules
from pount.apps.api.libs.email import send_email
from pount.apps.api.models.project import Project, ProjectQuerySet
from pount.apps.api.models.template import Template
from pount.apps.api.pagination import StandardPagination
from pount.apps.api.permissions import ProjectPermission
from pount.apps.api.serializers.project import ProjectGroupsSerializer, ProjectListSerializer, ProjectSerializer
from pount.apps.api.serializers.template import CompactTemplateSerializer, TemplateSerializer
from pount.apps.api.views.mixins import ThumbnailMixin

User = get_user_model()


class ImageUploadParser(MultiPartParser):
    media_type = "image/*"


class ProjectFilter(df_filters.FilterSet):
    is_manager = df_filters.BooleanFilter(field_name="is_manager")

    class Meta:
        model = Project
        fields = ["is_manager"]


class ProjectViewSet(ThumbnailMixin, ModelViewSet):
    queryset = Project.objects.all()
    serializer_class = ProjectSerializer
    permission_classes = [ProjectPermission]
    pagination_class = StandardPagination

    filter_backends = [df_filters.DjangoFilterBackend, filters.OrderingFilter]
    filterset_class = ProjectFilter
    ordering = ["-is_manager", "name"]
    ordering_fields = ["is_manager", "name", "updated_at", "is_public"]

    def get_queryset(self):
        queryset = super().get_queryset()
        queryset = queryset.annotate_with_is_manager(self.request.user)
        queryset = queryset.annotate_with_set_count()
        return queryset.for_member(self.request.user)

    def get_serializer_class(self):
        if self.action == "list":
            return ProjectListSerializer
        return super().get_serializer_class()

    @extend_schema(
        responses={
            status.HTTP_200_OK: ProjectGroupsSerializer,
        }
    )
    @action(methods=["GET"], detail=True)
    def groups(self, request, pk):
        project = get_object_or_404(Project, pk=pk)
        serializer = ProjectGroupsSerializer(instance=project)
        return Response(data=serializer.data)

    @extend_schema(
        responses={
            status.HTTP_200_OK: TemplateSerializer(many=True),
        },
        filters=False,
    )
    @method_decorator(
        permission_required(
            rules.PROJECT_LIST_TEMPLATES,
            fn=objectgetter(Project),
            raise_exception=True,
        )
    )
    @action(methods=["GET"], detail=True)
    def templates(self, request, pk):
        project = get_object_or_404(Project, pk=pk)
        serializer = TemplateSerializer(instance=project.templates.locals(), many=True)
        return Response(data=serializer.data)

    @extend_schema(
        filters=False,
        responses={
            status.HTTP_200_OK: CompactTemplateSerializer(many=True),
        },
    )
    @method_decorator(
        permission_required(
            rules.PROJECT_LIST_TEMPLATES,
            fn=objectgetter(Project),
            raise_exception=True,
        )
    )
    @action(methods=["GET"], detail=True)
    def list_templates(self, request, pk):
        project = get_object_or_404(Project, pk=pk)
        templates = (project.templates.locals() | Template.objects.globals()).order_by("name")
        serializer = CompactTemplateSerializer(instance=templates, many=True)
        return Response(data=serializer.data)

    @extend_schema(
        request=inline_serializer(
            "ProjectRequestDeletionSerializer",
            fields={
                "project_ids": serializers.ListField(child=serializers.UUIDField()),
            },
        ),
    )
    @action(methods=["DELETE"], detail=False)
    def request_deletion(self, request) -> Response:
        """
        Projects can not be deleted by a manager. A request is mode to an admin
        """
        # From given ids, select projects the current user is a manager of
        project_ids: list[str] = request.data.get("project_ids", [])
        projects = self._get_manager_projects(request.user, project_ids)

        if self._has_non_empty_projects(projects):
            # At least one project is not empty, the operation is cancelled
            return Response(
                {"error": "project_not_empty"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        admins = self._get_admin_emails()
        self._make_project_private(projects)
        self._send_deletion_request_emails(request.user, projects, admins, request)

        return Response(status=status.HTTP_204_NO_CONTENT)

    def _get_manager_projects(self, user: User, project_ids: list[str]) -> ProjectQuerySet:
        return self.queryset.for_manager(user).filter(id__in=project_ids)

    @staticmethod
    def _has_non_empty_projects(projects: ProjectQuerySet) -> bool:
        return projects.annotate_with_set_count().filter(num_sets__gt=0).exists()

    @staticmethod
    def _get_admin_emails() -> list[str]:
        return list(User.objects.admins().values_list("email", flat=True))

    @staticmethod
    def _make_project_private(projects: QuerySet) -> None:
        projects.update(is_public=False)

    @staticmethod
    def _send_deletion_request_emails(manager: User, projects: QuerySet, admins: list[str], request: Request) -> None:
        for project in projects:
            send_email(
                subject=_("Request to delete a project"),
                template="email/project/request_deletion",
                context={
                    "manager": manager,
                    "project": project,
                    "admin_url": request.build_absolute_uri(
                        reverse("admin:api_project_change", kwargs={"object_id": project.id})
                    ),
                },
                recipients=admins,
                sender=manager.email,
            )
