from django.core.exceptions import ValidationError
from django.http import Http404
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view, inline_serializer
from rest_framework import exceptions, mixins, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound
from rest_framework.filters import OrderingFilter, SearchFilter
from rest_framework.generics import get_object_or_404
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rules.contrib.rest_framework import AutoPermissionViewSetMixin

from pount.apps.api.libs.referential import get_languages_list, get_level_list, get_node_list, share_referential
from pount.apps.api.models import Project, Referential
from pount.apps.api.models.referential import DEFAULT_LANGUAGE
from pount.apps.api.pagination import StandardPagination
from pount.apps.api.rules import MANAGE_REFERENTIAL_IN_PROJECT
from pount.apps.api.serializers.referential import (
    ReferentialDetailSerializer,
    ReferentialEditSerializer,
    ReferentialLanguagesSerializer,
    ReferentialLevelsSerializer,
    ReferentialListSerializer,
    ReferentialNodesSerializer,
    TemplateListSerializer,
)
from pount.exceptions import LevelNotFoundError


@extend_schema_view(
    retrieve=extend_schema(responses={status.HTTP_200_OK: ReferentialDetailSerializer}),
    list=extend_schema(responses={status.HTTP_200_OK: ReferentialListSerializer}),
    create=extend_schema(
        request=ReferentialEditSerializer,
        responses={status.HTTP_201_CREATED: ReferentialEditSerializer},
    ),
    update=extend_schema(
        request=ReferentialEditSerializer,
        responses={status.HTTP_200_OK: ReferentialDetailSerializer},
    ),
)
class ReferentialViewSet(AutoPermissionViewSetMixin, viewsets.ModelViewSet):
    queryset = Referential.objects.current_version()
    serializer_class = ReferentialListSerializer
    pagination_class = StandardPagination

    permission_classes = [IsAuthenticated]
    permission_type_map = {
        **AutoPermissionViewSetMixin.permission_type_map,
        "share": "change",
        "nodes": "view_nodes",
        "levels": "view_nodes",
        "languages": "view_nodes",
        "templates": "view",
    }

    filter_backends = [DjangoFilterBackend, OrderingFilter, SearchFilter]
    ordering_fields = ["name", "scope", "updated_at"]
    filterset_fields = ["scope", "project_id"]
    search_fields = ["name", "description"]

    def get_serializer_class(self):
        if self.action in ["retrieve"]:
            return ReferentialDetailSerializer
        if self.action in ["create", "update", "partial_update", "share"]:
            return ReferentialEditSerializer
        return super().get_serializer_class()

    def check_permissions(self, request):
        super().check_permissions(request)
        if (
            self.action in ["create", "update", "partial_update", "destroy"]
            and request.data.get("scope") == Referential.Scope.SCOPE_BASE
            and not request.user.is_superuser
        ):
            self.permission_denied(request)

        if self.action == "share":
            if request.data.get("scope") == Referential.Scope.SCOPE_BASE and not request.user.is_superuser:
                self.permission_denied(request)
            else:
                return

        if project_id := request.data.get("project_id", None):
            project = Project.objects.get(pk=project_id)
            if not request.user.has_perm(MANAGE_REFERENTIAL_IN_PROJECT, project):
                self.permission_denied(request)

    def get_queryset(self):
        qs = super().get_queryset()
        if self.action == "list":
            qs = qs.public()
        return qs

    def destroy(self, request, *args, **kwargs):
        referential: Referential = self.get_object()
        if referential.scope == Referential.Scope.SCOPE_PUBLIC:
            if Project.objects.filter(external_referentials=referential).exclude(id=referential.project.id).count():
                # Referential is used in another project
                self.permission_denied(request, _("Referential is used in one or more projects"))
        if referential.templates.count():
            # Referential is used in at least one template
            self.permission_denied(request, _("Referential is used in one or more templates"))
        return super().destroy(request, *args, **kwargs)

    @extend_schema(
        parameters=[
            OpenApiParameter(
                "level_id",
                OpenApiTypes.UUID,
                OpenApiParameter.QUERY,
                required=True,
                description=_("UUID string identifying the referential's Level"),
            ),
            OpenApiParameter(
                "lang",
                OpenApiTypes.STR,
                OpenApiParameter.QUERY,
                required=False,
                description=_("String identifying the language (2 letter ISO code)"),
                default=DEFAULT_LANGUAGE,
            ),
            OpenApiParameter(
                "parent",
                OpenApiTypes.UUID,
                OpenApiParameter.QUERY,
                required=False,
                description=_("UUID string identifying the parent node"),
            ),
            OpenApiParameter(
                "ancestor",
                OpenApiTypes.UUID,
                OpenApiParameter.QUERY,
                required=False,
                description=_("UUID string identifying an ancestor node"),
            ),
        ],
        responses={
            status.HTTP_200_OK: ReferentialNodesSerializer(many=True),
        },
    )
    @action(
        methods=["GET"],
        detail=True,
        permission_classes=[AllowAny],
        pagination_class=None,
        filter_backends=[],
    )
    def nodes(self, request: Request, pk):
        """
        Get nodes from a referential's level
        """
        referential = self.get_object()
        params = {
            "level_id": request.query_params.get("level_id", None),
            "language": request.query_params.get("lang", DEFAULT_LANGUAGE),
            "parent": request.query_params.get("parent", None),
            "ancestor": request.query_params.get("ancestor", None),
        }
        try:
            nodes = get_node_list(referential=referential, params=params)
        except LevelNotFoundError as err:
            raise NotFound(str(err))
        serializer = ReferentialNodesSerializer(instance=nodes, many=True)
        return Response(serializer.data)

    @extend_schema(
        parameters=[
            OpenApiParameter(
                "lang",
                OpenApiTypes.STR,
                OpenApiParameter.QUERY,
                required=False,
                description=_("String identifying the language (2 letter ISO code)"),
                default=DEFAULT_LANGUAGE,
            ),
        ],
        responses={
            status.HTTP_200_OK: ReferentialLevelsSerializer(many=True),
        },
    )
    @action(
        methods=["GET"],
        detail=True,
        permission_classes=[AllowAny],
        filter_backends=[],
        pagination_class=None,
    )
    def levels(self, request, pk):
        referential = self.get_object()
        params = {
            "language": request.query_params.get("lang", DEFAULT_LANGUAGE),
        }
        levels = get_level_list(referential=referential, params=params)
        serializer = ReferentialLevelsSerializer(instance=levels, many=True)
        return Response(serializer.data)

    @action(
        methods=["GET"],
        detail=True,
        permission_classes=[AllowAny],
        filter_backends=[],
        pagination_class=None,
    )
    def languages(self, request, pk):
        """List of available languages"""
        referential = self.get_object()
        languages = get_languages_list(referential=referential)
        serializer = ReferentialLanguagesSerializer(instance=languages, many=True)
        return Response(serializer.data)

    @extend_schema(
        request=inline_serializer(
            "ReferentialShareSerializer",
            fields={"scope": serializers.ModelField(model_field=Referential()._meta.get_field("scope"))},
        ),
        responses={status.HTTP_200_OK: ReferentialEditSerializer},
    )
    @action(methods=["PATCH"], detail=True)
    def share(self, request, pk):
        """Change a referential's scope"""
        referential = self.get_object()
        scope = request.data.get("scope")
        try:
            referential = share_referential(referential, scope)
        except ValueError as e:
            raise exceptions.ValidationError({"scope": str(e)})
        serializer = self.get_serializer(instance=referential)
        return Response(serializer.data)

    @extend_schema(
        responses={status.HTTP_200_OK: TemplateListSerializer(many=True)},
    )
    @action(methods=["GET"], detail=True, pagination_class=None, filter_backends=[])
    def templates(self, request, pk):
        """List of all templates using a referential"""
        referential: Referential = self.get_object()
        templates = referential.templates.all()
        serializer = TemplateListSerializer(instance=templates, many=True)
        return Response(serializer.data)


class NestedReferentialViewSet(
    AutoPermissionViewSetMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """
    List referentials defined or used in the project
    """

    queryset = Referential.objects.current_version()
    serializer_class = ReferentialListSerializer
    pagination_class = StandardPagination
    permission_classes = [IsAuthenticated]
    permission_type_map = {
        **AutoPermissionViewSetMixin.permission_type_map,
        "use_referential_in_project": "use",
        "unuse_referential_in_project": "use",
    }

    filter_backends = [DjangoFilterBackend, OrderingFilter, SearchFilter]
    ordering_fields = ["name", "scope", "updated_at"]
    filterset_fields = ["scope", "project_id"]
    search_fields = ["name", "description"]

    def get_queryset(self):
        try:
            project = Project.objects.get(pk=self.kwargs.get("project_pk"))
        except (Project.DoesNotExist, ValidationError):
            raise Http404

        if self.action == "use_referential_in_project":
            return self.queryset

        # We select project referentials and external referentials
        qs = self.queryset.filter(project_id=project.id) | self.queryset.filter(projects=project.id)
        return qs.distinct()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["project_id"] = self.kwargs["project_pk"]
        return context

    def check_permissions(self, request):
        super().check_permissions(request)
        if self.action in ["use_referential_in_project", "unuse_referential_in_project"]:
            project: Project = get_object_or_404(Project.objects.all(), pk=self.kwargs["project_pk"])
            if not request.user.has_perm(MANAGE_REFERENTIAL_IN_PROJECT, project):
                self.permission_denied(request)

    @extend_schema(request=None, responses={status.HTTP_200_OK: ReferentialDetailSerializer})
    @action(methods=["PATCH"], detail=True, url_path="bind")
    def use_referential_in_project(self, request, pk, project_pk):
        """Use an external (public) or a base referential in the current project"""
        project: Project = get_object_or_404(Project.objects.all(), pk=project_pk)
        referential: Referential = self.get_object()
        try:
            project.add_external_referential(referential)
        except ValueError:
            raise exceptions.ValidationError(_("Referential is not public"))
        serializer = ReferentialDetailSerializer(instance=referential)
        return Response(serializer.data)

    @extend_schema(request=None, responses={status.HTTP_204_NO_CONTENT: None})
    @action(methods=["DELETE"], detail=True, url_path="unbind")
    def unuse_referential_in_project(self, request, pk, project_pk):
        """Stop using an external or base referential in the current project (must not be used in a template)"""
        project: Project = get_object_or_404(Project.objects.all(), pk=project_pk)
        referential: Referential = self.get_object()
        if referential.templates.filter(project=project).count():
            self.permission_denied(request, _("Referential is used in one or more templates"))
        project.external_referentials.remove(referential)
        return Response(status=status.HTTP_204_NO_CONTENT)
