from django.core.exceptions import ValidationError
from django.http import HttpResponse
from django.utils.translation import get_language_from_request
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, inline_serializer
from rest_framework import serializers, status
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from pount.apps.api.libs.export import add_content_disposition_header_for_attachment
from pount.apps.api.libs.search_results_export import (
    export_search_results,
    get_item_data,
    get_project_data,
    get_set_data,
)
from pount.apps.api.serializers.search import SearchResultSerializer
from pount.apps.search.query import Query, QueryError


@extend_schema(
    request=inline_serializer(
        "SearchExportRequest",
        fields={
            "ids": serializers.CharField(help_text=_("Comma separated list of UUIDs")),
            "type": serializers.CharField(help_text=_("Type of object (item, set or project)")),
        },
    ),
    responses={
        (status.HTTP_200_OK, "text/csv"): OpenApiTypes.BINARY,
    },
)
@api_view(http_method_names=["POST"])
@permission_classes([IsAuthenticated])
@csrf_exempt
def search_export(request):
    try:
        ids = request.data.get("ids", None).split(",")
    except AttributeError:
        return HttpResponse(_("IDs not given"), status=status.HTTP_400_BAD_REQUEST)

    model = request.data.get("type", None)
    if model not in ["item", "set", "project"]:
        return HttpResponse(_("Not a valid type"), status=status.HTTP_400_BAD_REQUEST)

    try:
        if model == "item":
            fieldnames, results = get_item_data(ids)
        elif model == "set":
            fieldnames, results = get_set_data(ids)
        elif model == "project":
            fieldnames, results = get_project_data(ids)
    except ValidationError:
        # IDS are not valid UUIDs
        return HttpResponse(_("Not valid UUIDs"), status=status.HTTP_400_BAD_REQUEST)

    response = HttpResponse(content_type="text/csv")
    response = add_content_disposition_header_for_attachment("search-results.csv", response)
    response = export_search_results(response, fieldnames, results)

    return response


@extend_schema(
    responses={
        status.HTTP_200_OK: SearchResultSerializer(),
    },
    parameters=[
        OpenApiParameter(
            name="q",
            description=_("The search query"),
            required=True,
            type=OpenApiTypes.STR,
        ),
        OpenApiParameter(
            name="project_ids",
            type=OpenApiTypes.UUID,
            location=OpenApiParameter.QUERY,
            description=_("Restrict results to comma separated list of project UUIDs"),
            required=False,
            style="simple",
            many=True,
        ),
        OpenApiParameter(
            name="set_ids",
            type=OpenApiTypes.UUID,
            location=OpenApiParameter.QUERY,
            description=_("Restrict results to comma separated list of set UUIDs"),
            required=False,
            style="simple",
            many=True,
        ),
        OpenApiParameter(
            name="page",
            type=OpenApiTypes.INT,
            location=OpenApiParameter.QUERY,
            description=_("Page number"),
            required=False,
            style="simple",
            default=1,
        ),
        OpenApiParameter(
            name="page_size",
            type=OpenApiTypes.INT,
            location=OpenApiParameter.QUERY,
            description=_("Number of results per page"),
            required=False,
            style="simple",
            default=10,
        ),
    ],
)
@api_view(http_method_names=["GET"])
@permission_classes([IsAuthenticated])
def global_search(request):
    params = request.query_params.dict()
    query = params.pop("q", "")
    language: str = get_language_from_request(request)
    try:
        search_query = Query(language=language, user=request.user, **params)
        results = search_query.execute(query=query)
    except QueryError as e:
        return Response(str(e), status=status.HTTP_400_BAD_REQUEST)

    serializer = SearchResultSerializer(
        instance=results,
        context={
            "query": query,
            "aggregation_fields": search_query.aggregation_fields,
        },
    )

    return Response(serializer.data, status=status.HTTP_200_OK)
