import logging

from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import exceptions, filters, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from pount.apps.api import rules
from pount.apps.api.libs.metadata import TemplateFields
from pount.apps.api.libs.set import add_or_replace_datacite_fields, order_fields_in_template_order
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.pagination import StandardPagination
from pount.apps.api.permissions import SetPermission
from pount.apps.api.serializers.set import SearchConfigSerializer, SetListSerializer, SetSerializer
from pount.apps.api.validators import JSONSchemaValidator
from pount.apps.api.views.mixins import ThumbnailMixin

User = get_user_model()
logger = logging.getLogger(__name__)


class SetViewSet(ThumbnailMixin, viewsets.ModelViewSet):
    queryset = Set.objects.all()
    serializer_class = SetSerializer
    pagination_class = StandardPagination
    permission_classes = [SetPermission]

    filter_backends = [DjangoFilterBackend, filters.OrderingFilter]
    filterset_fields = ["project_id"]
    ordering_fields = ["created_at", "updated_at", "title", "is_public"]
    ordering = ["title"]

    def get_queryset(self, pks: list = None):
        queryset = super().get_queryset()
        queryset = queryset.is_public_or_user_is_project_member(self.request.user)
        queryset = queryset.annotate_with_item_count()
        return queryset

    def check_permissions(self, request):
        if not bool(request.user and request.user.is_authenticated):
            raise exceptions.AuthenticationFailed()
        if request.method == "POST" and self.action == "import_csv":
            return request.user.has_perm(rules.SET_EDIT, self.get_object())
        if request.method == "POST" and self.action != "thumbnail":
            # For object creation, permission is checked on project
            try:
                project: Project = Project.objects.get(pk=request.data.get("project_id", None))
            except Project.DoesNotExist:
                raise exceptions.PermissionDenied()
            if not request.user.has_perm(rules.SET_CREATE, project):
                raise exceptions.PermissionDenied()
        return super().check_permissions(request)

    def get_serializer_class(self):
        if self.action == "list":
            return SetListSerializer
        return super().get_serializer_class()

    @extend_schema(
        parameters=[
            OpenApiParameter(
                "set_ids",
                OpenApiTypes.UUID,
                OpenApiParameter.QUERY,
                many=True,
                description="Set ids when bulk deleting (detail = False)",
            )
        ]
    )
    def destroy(self, request, *args, **kwargs):
        """Delete one set (detail = True) or bulk delete items (detail = False)"""
        # @todo dont pass set_ids in request.data
        if self.detail:
            instance = self.get_object()
            if instance.items.count():
                return Response(
                    {"error": "set_not_empty"},
                    status=status.HTTP_400_BAD_REQUEST,
                )
            self.perform_destroy(instance)
            return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            # Bulk delete objects
            sets = self.get_queryset().annotate_with_item_count().filter(pk__in=request.data.get("set_ids"))
            if sets.filter(num_items__gt=0).count():
                # One or more sets still have items, cancel the delete request
                return Response(
                    {"error": "set_not_empty"},
                    status=status.HTTP_400_BAD_REQUEST,
                )
            sets.delete()
            return Response(status=status.HTTP_204_NO_CONTENT)

    @extend_schema(
        description="Get and set search configuration for the set with the given id",
        request=SearchConfigSerializer(many=True),
        responses={
            status.HTTP_200_OK: SearchConfigSerializer,
        },
    )
    @action(methods=["GET", "PUT"], detail=True, url_path="search-config")
    def search_config(self, request, pk):
        set_model: Set = self.get_object()
        if request.method == "GET":
            serializer = SearchConfigSerializer(instance=set_model.search_config, many=True)
        else:
            JSONSchemaValidator("set-search-config.schema.json")(request.data)
            context = {"request": request, "instance": set_model}
            serializer = SearchConfigSerializer(data=request.data, many=True, context=context)
            if serializer.is_valid(raise_exception=True):
                fields_in_template = {
                    str(field)
                    for field in TemplateFields.from_template(set_model.template)
                    if str(field) != "Unknown name"  # When a template has a field with no name set, ignore it
                }
                fields_submitted = {field.get("name") for field in request.data}
                if not fields_in_template == fields_submitted:
                    raise serializers.ValidationError(
                        _("search_config and template don't have the same set of fields")
                    )
                search_config = serializer.validated_data
                search_config = add_or_replace_datacite_fields(search_config)
                search_config = order_fields_in_template_order(
                    search_config, TemplateFields(set_model.template.fields)
                )
                set_model.search_config = search_config
                set_model.save()
                # Invalidate caches
                for _key in [
                    f"aggregations_for_set_{set_model.pk}_*",
                    "aggregations_default_*",
                    f"search_fields_for_set_{set_model.pk}",
                ]:
                    cache.delete_pattern(_key)
        return Response(data=serializer.data)
