from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, status, viewsets
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404
from rest_framework.permissions import IsAuthenticated

from pount.apps.api import rules
from pount.apps.api.libs.export import add_content_disposition_header_for_attachment
from pount.apps.api.libs.template import export_template_as_csv
from pount.apps.api.models import Project, Template
from pount.apps.api.permissions import TemplatePermission
from pount.apps.api.serializers.template import TemplateSerializer


class TemplateViewSet(viewsets.ModelViewSet):
    queryset = Template.objects.all()
    serializer_class = TemplateSerializer
    permission_classes = [TemplatePermission]
    pagination_class = None

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.action == "list":
            queryset = queryset.globals()
        return queryset

    def check_permissions(self, request):
        if request.method in ["PATCH", "DELETE"] and self.get_object().name == "datacite":
            raise exceptions.PermissionDenied("Datacite template can not be edited or deleted")
        if request.method == "POST":
            # For template creation, permission is checked on project
            scope = request.data.get("scope", Template.SCOPE_PROJECT)
            if scope in [Template.SCOPE_BASE, Template.SCOPE_GLOBAL] and not request.user.is_superuser:
                raise exceptions.PermissionDenied()
            else:
                project: Project = get_object_or_404(Project, pk=request.data.get("project_id"))
                if not request.user.has_perm(rules.PROJECT_EDIT, project):
                    raise exceptions.PermissionDenied()
        else:
            if self.action == "export_as_csv":
                if not request.user.is_authenticated or not request.user.has_perm(
                    rules.TEMPLATE_EDIT, self.get_object()
                ):
                    self.permission_denied(request, _("Not a project manager"))
            super().check_permissions(request)

    @extend_schema(
        responses={
            (status.HTTP_200_OK, "text/csv"): OpenApiTypes.BINARY,
        }
    )
    @action(methods=["GET"], detail=True, permission_classes=[IsAuthenticated])
    def export_as_csv(self, request, pk):
        template = self.get_object()
        response = HttpResponse(content_type="text/csv")
        response = add_content_disposition_header_for_attachment(f"{str(template.name)}.csv", response)
        response = export_template_as_csv(response, template)

        return response
