import logging
from smtplib import SMTPException

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.http import Http404, HttpResponseBadRequest
from django.utils.encoding import force_bytes, force_str
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema, inline_serializer
from rest_framework import generics, mixins, serializers, status, viewsets
from rest_framework.decorators import api_view, permission_classes
from rest_framework.exceptions import ParseError
from rest_framework.exceptions import ValidationError as DRFValidationError
from rest_framework.parsers import MultiPartParser
from rest_framework.permissions import IsAdminUser, IsAuthenticated
from rest_framework.response import Response
from rules.contrib.views import permission_required

from pount.apps.api import rules
from pount.apps.api.libs.invite import (
    make_invitation_token,
    send_internal_invitation_email,
    send_invitation_email,
    validate_invitation,
)
from pount.apps.api.libs.notifications import notify_join_project
from pount.apps.api.models.project import Project
from pount.apps.api.models.user import AuthenticationProvider, Invitation
from pount.apps.api.pagination import StandardPagination
from pount.apps.api.serializers.group import GroupMemberSerializer
from pount.apps.api.serializers.user import (
    ChangeEmailSerializer,
    ChangePasswordSerializer,
    ProfileSerializer,
    UserCreationSerializer,
)
from pount.apps.api.views.mixins import ThumbnailMixin, UnsupportedFileFormat
from pount.apps.api.views.util import dataobjectgetter
from pount.exceptions import PountError

User = get_user_model()
logger = logging.getLogger(__file__)


class UserViewSet(mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet):
    """
    API endpoint that allows users to be viewed, but not edited.
    """

    queryset = User.objects.all().order_by("username")
    serializer_class = GroupMemberSerializer
    pagination_class = StandardPagination
    permission_classes = [IsAdminUser]


class ProfileView(generics.RetrieveUpdateAPIView):
    """
    API to view and update user's profile
    """

    serializer_class = ProfileSerializer
    permission_classes = [IsAuthenticated]
    http_method_names = ["get", "patch", "put"]

    def get_object(self):
        return self.request.user


# class ProfileActions(GenericViewSet):
#
#     serializer_class = ProfileSerializer
#     permission_classes = [IsAuthenticated]
#     queryset = User.objects.none()
#
#     @extend_schema(
#         request=ChangeEmailSerializer,
#         responses={
#             status.HTTP_200_OK: ChangeEmailSerializer,
#         }
#     )
#     @action(methods=["PATCH"], detail=False, url_path="cemail")
#     def change_mail(self, request):
#         user = self.request.user
#         serializer = ChangeEmailSerializer(instance=user, data=request.data, context={"request": request})
#         serializer.is_valid(raise_exception=True)
#         serializer.save()
#         return Response(data=serializer.data)


class ThumbnailView(ThumbnailMixin, generics.GenericAPIView):
    permission_classes = [IsAuthenticated]
    http_method_names = ["post", "delete"]
    parser_classes = [MultiPartParser]

    @extend_schema(
        request=inline_serializer(
            "UploadProfileThumbnailSerializer",
            fields={
                "file": serializers.FileField(),
            },
        ),
        responses={
            status.HTTP_201_CREATED: inline_serializer(
                "ProfileThumbnailsSerializer",
                fields={"thumbnails": serializers.DictField(child=serializers.CharField())},
            ),
        },
    )
    def post(self, request, *args, **kwargs):
        """Upload a picture to be used as the user's profile picture"""
        instance = request.user
        try:
            return self.perform_post(request, instance)
        except UnsupportedFileFormat:
            return Response(
                {"error": "bad_format"},
                status=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            )
        except ParseError:
            return Response(status=status.HTTP_400_BAD_REQUEST)

    @extend_schema(
        request=None,
        responses={
            status.HTTP_204_NO_CONTENT: None,
        },
    )
    def delete(self, request, *args, **kwargs):
        """Delete a user's profile thumbnails"""
        instance = request.user
        return self.perform_delete(instance)


@extend_schema(
    request=ChangeEmailSerializer,
    responses={
        status.HTTP_200_OK: ChangeEmailSerializer,
    },
)
@api_view(["PATCH"])
@permission_classes([IsAuthenticated])
def change_email(request):
    user = request.user
    serializer = ChangeEmailSerializer(instance=user, data=request.data, context={"request": request})
    try:
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(data=serializer.data)
    except DRFValidationError:
        return Response(
            serializer.errors,
            status=status.HTTP_400_BAD_REQUEST,
        )


@extend_schema(
    request=None,
    responses={
        status.HTTP_200_OK: OpenApiTypes.NONE,
        status.HTTP_400_BAD_REQUEST: OpenApiTypes.NONE,
    },
)
@api_view(["PATCH"])
@permission_classes([IsAuthenticated])
def ask_for_ldap_info_update(request):
    try:
        request.user.update_on_next_connection()
    except PountError:
        return HttpResponseBadRequest()
    return Response(status=status.HTTP_200_OK)


@extend_schema(
    request=ChangePasswordSerializer,
    responses={
        status.HTTP_200_OK: ChangePasswordSerializer,
    },
)
@api_view(["PATCH"])
@permission_classes([IsAuthenticated])
def change_password(request):
    user = request.user
    serializer = ChangePasswordSerializer(
        instance=user,
        data=request.data,
        context={"request": request},
    )
    try:
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(data=serializer.data)
    except DRFValidationError:
        return Response(
            serializer.errors,
            status=status.HTTP_400_BAD_REQUEST,
        )


@extend_schema(
    request=inline_serializer(
        "SignupFromInvitationRequest",
        fields={
            "invitation_id": serializers.UUIDField(),
            "first_name": serializers.CharField(),
            "last_name": serializers.CharField(),
            "email": serializers.EmailField(),
            "new_password": serializers.CharField(min_length=8),
            "new_password2": serializers.CharField(min_length=8),
        },
    ),
    responses={
        status.HTTP_201_CREATED: UserCreationSerializer,
    },
)
@api_view(["POST"])
def signup_from_invitation(request):
    invitation: Invitation = generics.get_object_or_404(
        Invitation.objects.all(), pk=request.data.get("invitation_id", "")
    )
    invitation_data, error = validate_invitation(invitation)
    if error:
        return Response(
            {"error": invitation_data["error_text"]},
            status=status.HTTP_400_BAD_REQUEST,
        )

    project = Project.objects.get(pk=invitation_data["project_id"])
    serializer = UserCreationSerializer(
        data=request.data,
        invitation=invitation_data,
        context={"request": request, "project": project},
    )
    if serializer.is_valid():
        user = serializer.save()
        notify_join_project(user, project)
        invitation.delete()
        return Response(serializer.data, status=status.HTTP_201_CREATED)
    return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


@extend_schema(
    request=inline_serializer(
        "InviteInternalUserRequest",
        fields={
            "username": serializers.CharField(),
            "project_id": serializers.UUIDField(),
        },
    ),
    responses={
        status.HTTP_200_OK: inline_serializer(
            "InviteInternalUserResponse",
            fields={
                "email": serializers.CharField(help_text="Result of send email to user (no, sent or error)"),
                "user": GroupMemberSerializer(),
            },
        )
    },
)
@api_view(["POST"])
@permission_required(
    rules.PROJECT_EDIT,
    fn=dataobjectgetter(Project, attr_name="project_id"),
    raise_exception=True,
)
def invite_internal_user(request):
    try:
        user = User.objects.get_by_natural_key(
            request.data.get("username", None),
            AuthenticationProvider.CAS.value,
        )
    except User.DoesNotExist:
        raise Http404
    project = generics.get_object_or_404(Project.objects.all(), pk=request.data.get("project_id", None))
    project.add_member(user)
    user_serializer = GroupMemberSerializer(instance=user)
    email_result = "no"
    if user.email:
        try:
            send_internal_invitation_email(user.email, project)
            email_result = "sent"
        except (SMTPException, ConnectionRefusedError):
            email_result = "error"
    notify_join_project(user, project)
    return Response(
        data={
            "email": email_result,
            "user": user_serializer.data,
        }
    )


@extend_schema(
    request=inline_serializer(
        "InviteUserRequest",
        fields={
            "project_id": serializers.UUIDField(),
            "email": serializers.EmailField(),
            "url": serializers.URLField(),
        },
    ),
    responses={
        status.HTTP_201_CREATED: None,
        status.HTTP_400_BAD_REQUEST: OpenApiTypes.OBJECT,
        status.HTTP_500_INTERNAL_SERVER_ERROR: OpenApiTypes.OBJECT,
    },
)
@api_view(["POST"])
@permission_required(
    rules.PROJECT_EDIT,
    fn=dataobjectgetter(Project, attr_name="project_id"),
    raise_exception=True,
)
def invite_user(request):
    email = request.data.get("email", None)
    try:
        validate_email(email)
    except ValidationError:
        return Response({"error": "invalid_email"}, status=status.HTTP_400_BAD_REQUEST)
    try:
        user = User.objects.get_by_natural_key(email, AuthenticationProvider.LOCAL.value)
    except User.DoesNotExist:
        user = None
    project = Project.objects.get(pk=request.data.get("project_id", None))
    token = make_invitation_token(email, project.id, user)
    invitation = Invitation.objects.create(token=token)
    oid = urlsafe_base64_encode(force_bytes(invitation.id))
    invitation_url = request.data.get("url")
    invitation_url = f"{invitation_url}?oid={oid}&token={token}"
    try:
        send_invitation_email(email, project, invitation_url)
    except SMTPException as error:
        logger.exception(error)
        return Response(
            {"error": "send_email_failed"},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )
    return Response(status=status.HTTP_201_CREATED)


@extend_schema(
    request=inline_serializer(
        "ValidateInvitation",
        fields={
            "oid": serializers.CharField(help_text="Encoded invitation id"),
            "token": serializers.CharField(help_text="Signed invitation token"),
        },
    ),
    responses=inline_serializer(
        "ValidateInvitationResult",
        fields={
            "result": serializers.CharField(),
            "data": inline_serializer(
                "ValidateInvitationResultData",
                fields={
                    "email": serializers.EmailField(),
                    "invitation_id": serializers.UUIDField(),
                    "project_id": serializers.UUIDField(),
                },
            ),
        },
    ),
)
@api_view(["POST"])
def invite_validate(request):
    """
    Validate an open Invitation for and external User
    requires a token and oid
    returns the email, invitation_id and project_id
    """
    oid = request.data.get("oid", "")
    invitation_id = force_str(urlsafe_base64_decode(oid))
    token = request.data.get("token", "")
    try:
        invitation = Invitation.objects.get(
            id=invitation_id,
            token=token,
        )
    except (Invitation.DoesNotExist, ValidationError):
        return Response(
            {"error": "invitation_not_found"},
            status=status.HTTP_404_NOT_FOUND,
        )

    data, error = validate_invitation(invitation)

    if error:
        return Response({"error": data["error_text"]}, status=status.HTTP_400_BAD_REQUEST)

    if data["user_id"] is not None:
        # The user already exists, add it to the project's default group
        project = Project.objects.get(pk=data["project_id"])
        user = User.objects.get(pk=data["user_id"])
        user.is_active = True
        user.save()
        project.add_member(user)
        notify_join_project(user, project)
        invitation.delete()
        return Response(data={"result": "added_to_project", "data": data})
    else:
        return Response(data={"result": "valid", "data": data})
