from django.core.exceptions import FieldError, ImproperlyConfigured
from django.shortcuts import get_object_or_404


def dataobjectgetter(model, attr_name="pk", field_name="pk"):
    """
    Helper that returns a function suitable for use as the ``fn`` argument
    to the ``rules.contrib.views.permission_required`` decorator. Internally
    uses ``get_object_or_404``, so keep in mind that this may raise
    ``Http404``.

    Adapted from rules.contrib.views.objectgetter to use request.data rather
    than request function argument.

    ``model`` can be a model class, manager or queryset.

    ``attr_name`` is the name of the ``request.data`` attribute.

    ``field_name`` is the model's field name by which the lookup is made, eg.
    "id", "slug", etc.
    """

    def _getter(request, *view_args, **view_kwargs):
        if attr_name not in request.data.keys():
            raise ImproperlyConfigured(
                f"Key {attr_name} is not available in [{', '.join(request.data.keys())}]"  # noqa:E501
            )
        try:
            return get_object_or_404(model, **{field_name: request.data[attr_name]})
        except FieldError:
            raise ImproperlyConfigured(f"Model {model} has no field named {field_name}")

    return _getter
