from botocore.exceptions import ClientError
from django.contrib.auth import get_user_model
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.reverse import reverse

from pount.apps.api.libs.file import item_mediafile_filename, sizeof_format
from pount.apps.api.models.item import Item, MediaFile
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.serializers.mixins import ThumbnailUrlMixin


class PublicNestedSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = Set
        fields = ["id", "title", "description", "template_id"]


class PublicUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = [
            "id",
            "name",
            "email",
            "info",
        ]


class NestedUserSerializer(PublicUserSerializer):
    class Meta(PublicUserSerializer.Meta):
        fields = ["id", "name"]


class PublicProjectSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    sets = serializers.SerializerMethodField()
    managers = NestedUserSerializer(many=True)
    thumbnails = serializers.SerializerMethodField()

    class Meta:
        model = Project
        fields = [
            "id",
            "name",
            "description",
            "created_at",
            "updated_at",
            "sets",
            "managers",
            "thumbnails",
        ]

    @extend_schema_field(PublicNestedSetSerializer)
    def get_sets(self, instance: Project):
        sets = instance.public_sets()
        serializer = PublicNestedSetSerializer(sets, many=True)
        return serializer.data


class PublicProjectListSerializer(PublicProjectSerializer):
    class Meta(PublicProjectSerializer.Meta):
        fields = [
            "id",
            "name",
            "description",
            "created_at",
            "updated_at",
            "managers",
            "thumbnails",
        ]


class NestedProjectSerializer(PublicProjectSerializer):
    class Meta(PublicProjectSerializer.Meta):
        fields = ["id", "name"]


class PublicNestedMediaFileSerializer(serializers.ModelSerializer):
    name = serializers.SerializerMethodField()
    size = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = MediaFile
        fields = [
            "id",
            "name",
            "url",
            "size",
            "content_type",
            "is_downloadable",
            "is_visualized",
            "component",
        ]

    def get_name(self, instance: MediaFile) -> str:
        return item_mediafile_filename(instance.filename)

    def get_size(self, instance: MediaFile) -> str:
        return sizeof_format(instance.size)

    def get_url(self, instance: MediaFile) -> str:
        if instance.info.get("is_rti"):
            url = reverse(
                "public:item-get-rti",
                kwargs={
                    "pk": str(instance.item.pk),
                    "filename": item_mediafile_filename(instance.filename),
                },
            )
            return self.context["request"].build_absolute_uri(url)
        try:
            return instance.public_url
        except ClientError:
            return None


class PublicItemSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    set = PublicNestedSetSerializer()
    project = serializers.SerializerMethodField()
    creator = NestedUserSerializer()
    media_files = serializers.SerializerMethodField()
    metadata = serializers.SerializerMethodField()
    thumbnails = serializers.SerializerMethodField()

    class Meta:
        model = Item
        fields = [
            "id",
            "title",
            "description",
            "metadata",
            "created_at",
            "updated_at",
            "set",
            "project",
            "media_files",
            "creator",
            "thumbnails",
            "citation",
        ]

    @extend_schema_field(NestedProjectSerializer)
    def get_project(self, instance: Item):
        project = instance.set.project
        return NestedProjectSerializer(project).data

    @extend_schema_field(OpenApiTypes.OBJECT)
    def get_metadata(self, instance: Item):
        template: Template = instance.set.template
        if isinstance(template.private_fields, list) and isinstance(instance.metadata, list):
            # Filter out private fields
            return [field for field in instance.metadata if field["id"] not in template.private_fields]
        return instance.metadata

    @extend_schema_field(PublicNestedMediaFileSerializer(many=True))
    def get_media_files(self, instance: Item):
        mediafiles = instance.media_files.filter(is_hidden=False)
        serializer = PublicNestedMediaFileSerializer(instance=mediafiles, many=True, context=self.context)
        return serializer.data


class PublicItemListSerializer(PublicItemSerializer):
    class Meta(PublicItemSerializer.Meta):
        fields = [
            "id",
            "title",
            "description",
            "thumbnails",
            "citation",
        ]


class PublicSetSerializer(ThumbnailUrlMixin, serializers.ModelSerializer):
    project = NestedProjectSerializer()
    creator = NestedUserSerializer()
    thumbnails = serializers.SerializerMethodField()

    class Meta:
        model = Set
        fields = [
            "id",
            "title",
            "description",
            "publisher",
            "project",
            "creator",
            "created_at",
            "updated_at",
            "thumbnails",
        ]


class PublicTemplateSerializer(serializers.ModelSerializer):
    fields = serializers.SerializerMethodField(method_name="get_metadata_fields")

    class Meta:
        model = Template
        fields = [
            "id",
            "name",
            "fields",
            "messages",
        ]

    def get_metadata_fields(self, instance: Template) -> list[dict]:
        # we need to recursively filter out fields whose "id" is in the
        # instance's private_fields array
        def filter_fields(fields: list, private_ids: list):
            filtered = []
            for f in fields:
                if f["id"] not in private_ids:
                    if "fields" in f.keys():
                        # we need to check recursively
                        f["fields"] = filter_fields(f["fields"], private_ids)
                    filtered.append(f)
            return filtered

        if isinstance(instance.private_fields, list) and instance.private_fields:
            return filter_fields(instance.fields, instance.private_fields)
        return instance.fields
