from unittest.mock import patch

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models import Item
from pount.apps.api.models.init import initialize_roles
from pount.apps.public.tests import factories
from pount.apps.public.tests.setup import SetUpMixin


class ItemTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()
        self.item = Item.objects.public().first()

    def test_list_of_items_is_accessible(self):
        response = self.client.get(reverse("public:item-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_of_items_keys(self):
        response = self.client.get(reverse("public:item-list"))
        self.assertListEqual(
            list(response.data["results"][0].keys()),
            ["id", "title", "description", "thumbnails", "citation"],
        )

    def _get_detail_response(self, item):
        response = self.client.get(reverse("public:item-detail", kwargs={"pk": item.id}))
        return response

    def test_detail_of_item_is_accessible(self):
        response = self._get_detail_response(self.item)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["title"], self.item.title)

    def test_detail_of_item_keys(self):
        response = self._get_detail_response(self.item)
        self.assertListEqual(
            list(response.data.keys()),
            [
                "id",
                "title",
                "description",
                "metadata",
                "created_at",
                "updated_at",
                "set",
                "project",
                "media_files",
                "creator",
                "thumbnails",
                "citation",
            ],
        )

    def test_set_data_includes_template_id(self):
        response = self._get_detail_response(self.item)
        self.assertListEqual(
            list(response.data["set"].keys()),
            ["id", "title", "description", "template_id"],
        )

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_private_fields_are_filtered_out(self, update_mock):
        template = self.item.set.template
        template.private_fields = ["abc", "def"]
        template.save()
        self.item.metadata = [
            {
                "id": "abc",
                "name": "Private field 1",
                "value": "abc",
                "multiple": False,
                "repeatable": False,
            },
            {
                "id": "def",
                "name": "Private field 2",
                "value": "def",
                "multiple": False,
                "repeatable": False,
            },
            {
                "id": "ghi",
                "name": "Public field 1",
                "value": "Public field 1 value",
                "multiple": False,
                "repeatable": False,
            },
            {
                "id": "jkl",
                "name": "Public field 2",
                "value": "Public field 2 value",
                "multiple": False,
                "repeatable": False,
            },
        ]
        self.item.save()
        response = self.client.get(reverse("public:item-detail", kwargs={"pk": self.item.id}))
        self.assertListEqual([_["id"] for _ in response.data["metadata"]], ["ghi", "jkl"])

    def test_detail_of_item_contains_mediafiles(self):
        response = self._get_detail_response(self.items[0])
        self.assertIsInstance(response.data["media_files"], list)
        self.assertDictEqual(
            dict(response.data["media_files"][0]),
            {
                "component": "Image",
                "id": str(self.mediafile.id),
                "is_downloadable": self.mediafile.is_downloadable,
                "is_visualized": self.mediafile.is_visualized,
                "name": self.mediafile.filename,
                "size": "1.0 KiB",
                "url": "http://media.pount.org/pount.png",
                "content_type": self.mediafile.info["content_type"],
            },
        )

    def test_list_of_items_can_be_filtered_by_item_ids(self):
        # items can be filtered by passing a comma separated list of
        # items ids as param
        items = Item.objects.public().all()[:3]
        ids = [str(_.id) for _ in items]
        response = self.client.get(reverse("public:item-list") + "?item_ids=" + ",".join(ids))
        self.assertEqual(response.data["count"], 3)
        self.assertListEqual(sorted(_["id"] for _ in response.data["results"]), sorted(ids))

    def test_detail_of_private_item_is_not_found(self):
        item = Item.objects.filter(is_public=False).first()
        response = self._get_detail_response(item)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_detail_of_public_item_of_private_set_is_not_found(self):
        response = self._get_detail_response(self.public_item_of_private_set)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_detail_of_public_item_of_public_set_of_private_project_is_not_found(  # noqa: E501
        self,
    ):
        response = self._get_detail_response(self.public_item_of_public_set_of_private_project)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_post_method_not_allowed(self):
        response = self.client.post(reverse("public:item-list", {}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete_method_not_allowed(self):
        response = self.client.delete(reverse("public:item-detail", kwargs={"pk": self.item.id}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_put_patch_method_not_allowed(self):
        response = self.client.put(reverse("public:item-detail", kwargs={"pk": self.item.id}), {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        response = self.client.patch(reverse("public:item-detail", kwargs={"pk": self.item.id}), {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_deleted_items_are_not_listed(self):
        response = self.client.get(reverse("public:item-list"))
        self.assertEqual(len(response.data["results"]), 5)
        self.assertNotIn(self.deleted_item.id, [_["id"] for _ in response.data["results"]])

    def test_deleted_item_is_not_found(self):
        response = self.client.get(reverse("public:item-detail", kwargs={"pk": self.deleted_item.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class RtiMediaFileTest(APITestCase):
    def setUp(self) -> None:
        initialize_roles()

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_rti_file_view_requires_public_item(self, update_mock):
        item = factories.ItemFactory(is_public=False)
        response = self.client.get(
            reverse(
                "public:item-get-rti",
                kwargs={"pk": str(item.id), "filename": "file.ext"},
            )
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_rti_file_view_for_public_item(self, update_mock):
        item = factories.ItemFactory()
        with patch("pount.apps.public.views.ItemMediaFileStorage") as storage:
            storage.return_value.open.return_value = "https://pount.unistra.fr/rti-file.ext"
            response = self.client.get(
                reverse(
                    "public:item-get-rti",
                    kwargs={"pk": str(item.id), "filename": "file.ext"},
                )
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(storage.return_value.open.call_count, 1)
