from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.public.tests.setup import SetUpMixin

User = get_user_model()


class ProjectTest(SetUpMixin, APITestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_project_list_is_accessible(self):
        response = self.client.get(reverse("public:project-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)

    def test_public_project_detail_is_accessible(self):
        response = self.client.get(reverse("public:project-detail", kwargs={"pk": self.public_project.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_private_project_detail_is_not_found(self):
        response = self.client.get(reverse("public:project-detail", kwargs={"pk": self.private_project.pk}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_children_sets_are_only_public_sets(self):
        response = self.client.get(reverse("public:project-detail", kwargs={"pk": self.public_project.pk}))
        self.assertEqual(len(response.data["sets"]), 2)
        self.assertListEqual(
            [_["id"] for _ in response.data["sets"]],
            [str(self.public_set1.pk), str(self.public_set2.pk)],
        )

    def test_project_data_includes_managers(self):
        response = self.client.get(reverse("public:project-detail", kwargs={"pk": self.public_project.pk}))
        self.assertEqual(len(response.data["managers"]), 1)
        self.assertEqual(response.data["managers"][0]["id"], str(self.manager.pk))

    def test_post_method_not_allowed(self):
        response = self.client.post(reverse("public:project-list", {}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_put_patch_method_not_allowed(self):
        response = self.client.put(
            reverse("public:project-detail", kwargs={"pk": self.public_project.pk}),
            {},
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        response = self.client.patch(
            reverse("public:project-detail", kwargs={"pk": self.public_project.pk}),
            {},
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete_method_not_allowed(self):
        response = self.client.delete(reverse("public:project-detail", kwargs={"pk": self.public_project.pk}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_deleted_set_is_not_included_in_project_sets(self):
        response = self.client.get(reverse("public:project-detail", kwargs={"pk": self.public_project.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["sets"]), 2)
        self.assertNotIn(self.deleted_set.id, [_["id"] for _ in response.data["sets"]])
