from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.public.tests.setup import SetUpMixin


class SetTest(SetUpMixin, APITestCase):
    def test_list_of_sets_is_accessible(self):
        response = self.client.get(reverse("public:set-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_detail_of_set_is_accessible(self):
        response = self.client.get(reverse("public:set-detail", kwargs={"pk": self.public_set1.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_public_set_of_private_project_is_not_visible(self):
        response = self.client.get(
            reverse(
                "public:set-detail",
                kwargs={"pk": self.public_set_of_private_project.id},
            )
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_public_set_of_public_project_is_visible(self):
        response = self.client.get(reverse("public:set-detail", kwargs={"pk": self.public_set1.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["project"]["id"], str(self.public_project.id))

    def test_list_contains_only_public_sets_of_public_templates(self):
        response = self.client.get(
            reverse("public:set-list"),
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            sorted([_["id"] for _ in response.data["results"]]),
            sorted([str(self.public_set1.id), str(self.public_set2.id)]),
        )

    def test_list_is_paginated(self):
        response = self.client.get(reverse("public:set-list"))
        self.assertListEqual(
            list(response.data.keys()),
            ["count", "next_page", "previous_page", "results"],
        )

    def test_post_method_not_allowed(self):
        response = self.client.post(reverse("public:set-list"), {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_put_patch_methods_not_allowed(self):
        response = self.client.put(
            reverse("public:set-detail", kwargs={"pk": self.public_set1.id}),
            {},
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        response = self.client.patch(
            reverse("public:set-detail", kwargs={"pk": self.public_set1.id}),
            {},
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete_method_not_allowed(self):
        response = self.client.delete(reverse("public:set-detail", kwargs={"pk": self.public_set1.id}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_deleted_set_is_not_found(self):
        response = self.client.get(reverse("public:set-detail", kwargs={"pk": self.deleted_set.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_deleted_item_is_not_included_in_list(self):
        response = self.client.get(reverse("public:set-list"))
        self.assertEqual(len(response.data["results"]), 2)
        self.assertNotIn(self.deleted_set.id, [_["id"] for _ in response.data["results"]])
