from unittest.mock import patch

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from pount.apps.api.models import Template
from pount.apps.public.tests.factories import SetFactory, TemplateFactory
from pount.apps.public.tests.setup import SetUpMixin


class TemplatesTest(SetUpMixin, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock) -> None:
        super().setUp()
        datacite = Template.objects.get(name="datacite")
        self.template = TemplateFactory(fields=datacite.fields, scope=Template.SCOPE_SET)
        self.set = SetFactory(template=self.template)

    def _get_response(self):
        response = self.client.get(reverse("public:template-detail", kwargs={"pk": self.template.id}))
        return response

    def test_template_view(self):
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["name"], self.template.name)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_template_serializer_removes_private_fields(self, update_mock):
        self.template.private_fields = ["6qgq1netsxh", "m1w8jje0wsh"]
        self.template.save(update_fields=["private_fields"])
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        private_fields = [_f for _f in response.data["fields"] if _f["id"] in self.template.private_fields]
        self.assertEqual(len(private_fields), 0)

    def test_set_must_be_public(self):
        self.set.is_public = False
        self.set.save()
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_project_must_be_public(self):
        self.set.project.is_public = False
        self.set.project.save()
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_template_must_be_of_scope_set(self, update_mock):
        self.template.scope = Template.SCOPE_PROJECT
        self.template.save()
        response = self._get_response()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class TemplatesPrivateFieldsFilterTest(SetUpMixin, APITestCase):
    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def setUp(self, update_mock):
        super().setUp()
        self.fields = [
            {"id": "one", "name": "One"},
            {
                "id": "two",
                "name": "Two",
                "fields": [
                    {"id": "sub1", "name": "Sub field 1"},
                    {
                        "id": "sub2",
                        "name": "Sub field 2",
                        "fields": [
                            {
                                "id": "sub2sub1",
                                "name": "Sub field of sub field 2",
                            }
                        ],
                    },
                ],
            },
        ]
        self.template = TemplateFactory(fields=self.fields, scope=Template.SCOPE_SET)
        self.set = SetFactory(template=self.template)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_empty_private_field_list_removes_nothing(self, update_mock):
        self.template.private_fields = []
        self.template.save()
        response = self.client.get(reverse("public:template-detail", kwargs={"pk": self.template.id}))
        self.assertListEqual(response.data["fields"], self.fields)

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_private_field_is_removed(self, update_mock):
        self.template.private_fields = ["one"]
        self.template.save()
        response = self.client.get(reverse("public:template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(len(response.data["fields"]), 1)
        self.assertListEqual(response.data["fields"], self.fields[1:])

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_private_field_is_removed_with_children(self, update_mock):
        self.template.private_fields = ["two"]
        self.template.save()
        response = self.client.get(reverse("public:template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(len(response.data["fields"]), 1)
        self.assertListEqual(response.data["fields"], self.fields[:1])

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    def test_nested_private_field_is_removed(self, update_mock):
        self.template.private_fields = ["sub2sub1"]
        self.template.save()
        response = self.client.get(reverse("public:template-detail", kwargs={"pk": self.template.id}))
        self.assertEqual(len(response.data["fields"]), 2)
        self.assertListEqual(response.data["fields"][1]["fields"][1]["fields"], [])
