import os

from django.contrib.auth import get_user_model
from django.http import FileResponse
from django.utils.translation import get_language_from_request
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import filters, mixins, status
from rest_framework.decorators import api_view
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from pount.apps.api.models.item import Item
from pount.apps.api.models.project import Project
from pount.apps.api.models.set import Set
from pount.apps.api.models.template import Template
from pount.apps.api.serializers.search import SearchResultSerializer
from pount.apps.api.storage_backends import ItemMediaFileStorage
from pount.apps.public import serializers
from pount.apps.search.query import Query, QueryError

User = get_user_model()


class ProjectViewSet(mixins.ListModelMixin, mixins.RetrieveModelMixin, GenericViewSet):
    model = Project
    serializer_class = serializers.PublicProjectSerializer
    queryset = Project.objects.public().order_by("-updated_at")
    filter_backends = [filters.OrderingFilter]
    ordering_fields = ["created_at", "updated_at", "name"]

    def get_serializer_class(self):
        if self.action in ["list"]:
            return serializers.PublicProjectListSerializer
        return super().get_serializer_class()


class SetViewSet(mixins.ListModelMixin, mixins.RetrieveModelMixin, GenericViewSet):
    model = Set
    serializer_class = serializers.PublicSetSerializer
    queryset = Set.objects.public().order_by("-updated_at")
    filter_backends = [DjangoFilterBackend, filters.OrderingFilter]
    filterset_fields = ["project_id"]
    ordering_fields = ["created_at", "updated_at", "title"]


class ItemViewSet(mixins.ListModelMixin, mixins.RetrieveModelMixin, GenericViewSet):
    model = Item
    serializer_class = serializers.PublicItemSerializer
    queryset = Item.objects.public().order_by("-updated_at")
    filter_backends = [DjangoFilterBackend, filters.OrderingFilter]
    filterset_fields = ["set_id"]
    ordering_fields = ["created_at", "updated_at", "title"]

    def get_serializer_class(self):
        if self.action == "list":
            return serializers.PublicItemListSerializer
        return super().get_serializer_class()

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.action == "list" and self.request.query_params.get("item_ids"):
            # items can be filtered by passing a comma separated list of item
            # ids as param
            ids = self.request.query_params.get("item_ids", []).split(",")
            queryset = queryset.filter(id__in=ids)
        return queryset


@extend_schema(
    responses={
        status.HTTP_200_OK: OpenApiTypes.BINARY,
    }
)
@api_view(["GET"])
def get_rti_file(request, pk, filename):
    item = get_object_or_404(Item.objects.public(), pk=pk)
    media_storage = ItemMediaFileStorage()
    key = os.path.join(f"{str(item.id)}/rti/", filename)
    file = media_storage.open(key)
    return FileResponse(file)


class TemplateViewSet(mixins.RetrieveModelMixin, GenericViewSet):
    model = Template
    serializer_class = serializers.PublicTemplateSerializer
    queryset = Template.objects.filter(
        set__is_public=True,
        set__project__is_public=True,
        scope=Template.SCOPE_SET,
    )


class UserViewSet(mixins.RetrieveModelMixin, GenericViewSet):
    model = User
    serializer_class = serializers.PublicUserSerializer
    queryset = User.objects.filter(is_active=True)


@extend_schema(
    responses={
        status.HTTP_200_OK: SearchResultSerializer(),
    },
    parameters=[
        OpenApiParameter(
            name="q",
            description=_("The search query"),
            required=True,
            type=OpenApiTypes.STR,
        ),
        OpenApiParameter(
            name="project_ids",
            type=OpenApiTypes.UUID,
            location=OpenApiParameter.QUERY,
            description=_("Restrict results to comma separated list of project UUIDs"),
            required=False,
            style="simple",
            many=True,
        ),
        OpenApiParameter(
            name="set_ids",
            type=OpenApiTypes.UUID,
            location=OpenApiParameter.QUERY,
            description=_("Restrict results to comma separated list of set UUIDs"),
            required=False,
            style="simple",
            many=True,
        ),
        OpenApiParameter(
            name="page",
            type=OpenApiTypes.INT,
            location=OpenApiParameter.QUERY,
            description=_("Page number"),
            required=False,
            style="simple",
            default=1,
        ),
        OpenApiParameter(
            name="page_size",
            type=OpenApiTypes.INT,
            location=OpenApiParameter.QUERY,
            description=_("Number of results per page"),
            required=False,
            style="simple",
            default=10,
        ),
    ],
)
@api_view(http_method_names=["GET"])
def global_search(request):
    params = request.query_params.dict()
    query = params.pop("q", "")
    language: str = get_language_from_request(request)
    try:
        search_query = Query(language=language, user=None, **params)
        search_query.exclude(
            [
                "is_public",
                "sets_count",
                "items_count",
            ]
        )
        results = search_query.execute(query=query, exclude_private_fields=True)
    except QueryError as e:
        return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

    serializer = SearchResultSerializer(
        instance=results,
        context={
            "query": query,
            "aggregation_fields": search_query.aggregation_fields,
        },
    )

    return Response(serializer.data, status=status.HTTP_200_OK)
