from typing import Any

from elasticsearch_dsl import Boolean, Date, Document, Keyword, MetaField, Object, Text

from pount.apps.api.libs.metadata import Metadata
from pount.apps.api.models import Group, Item
from pount.apps.search.elasticsearch.documents import THUMBNAIL_EXPIRATION, IcuCollationKeyword, pount_standard

# Chat du 19 sept.
# Index would be created separately, e.g.
# item_index = Index("items-{id}")
# item_index.settings(
#     analysis={
#         "analyzer": {"pount_french_analyzer": "...", "etc": "..."}
#     }
# )
# item_index.create()

# Then we could save items like this:
# _s = ItemDoc.from_model(i)
# _s.save(index=item_index)


class ItemDoc(Document):
    id = Keyword()
    title = Text(
        analyzer=pount_standard,
        fields={
            "sort": IcuCollationKeyword(),
        },
    )
    description = Text(
        analyzer=pount_standard,
    )
    creator = Object(
        properties={
            "name": Text(
                analyzer=pount_standard,
                fields={
                    "keyword": Keyword(ignore_above=256),
                },
            ),
            "id": Keyword(),
        }
    )
    project = Object(
        properties={
            "name": Text(
                analyzer=pount_standard,
            ),
            "id": Keyword(),
        },
    )
    set = Object(
        properties={
            "title": Text(
                analyzer=pount_standard,
            ),
            "id": Keyword(),
        },
    )
    managers = Text(
        analyzer=pount_standard,
        fields={
            "keyword": Keyword(
                ignore_above=256,
            )
        },
        multi=True,
    )
    groups = Keyword(multi=True)
    publication_date = Date(format="date_optional_time")
    is_public = Boolean()
    thumbnail = Keyword()
    metadata = Object(dynamic=True)

    class Meta:
        dynamic = MetaField("strict")

    class Index:
        name = "items"
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 0,
        }

    @classmethod
    def from_model(cls, item: Item) -> "ItemDoc":
        _doc = prepare_item(item)
        return cls(**_doc)


def prepare_metadata(item: Item) -> dict[str, Any]:
    md = Metadata(item.metadata, item.set.template.fields)

    # We construct a list of tuples with the field name and the mapping type if the field has a mapping type
    # and is used in the search or the aggregation
    indexed_fields: list[tuple[str, str]] = [
        (_field.get("name"), _field.get("config", {}).get("mapping_type"))
        for _field in item.set.search_config
        if _field.get("config", {}).get("mapping_type")
        and (_field.get("config", {}).get("use_in_search") or _field.get("config", {}).get("use_in_aggregation"))
    ]

    # If the entry is a referential value or an option, use the field name as is,
    # otherwise suffix it with the mapping type
    metadata: dict[str, Any] = {}
    for _field_name, _mapping_type in indexed_fields:
        _field_entry = md.get_entry_by_name(_field_name)
        if _field_entry.is_referential_value or _field_entry.is_option:
            _keyname = _field_name
        else:
            _keyname = f"{_field_name}_{_mapping_type}"
        metadata[_keyname] = _field_entry.value_to_index

    return metadata


def prepare_item(item: Item) -> dict[str, Any]:
    doc = {
        "_id": str(item.id),
        "doc_type": "item",
        "id": str(item.id),
        "title": item.title,
        "description": item.description,
        "creator": {
            "id": str(item.creator.id),
            "name": item.creator.name,
        },
        "project": {
            "id": str(item.set.project.id),
            "name": item.set.project.name,
        },
        "set": {
            "id": str(item.set.id),
            "title": item.set.title,
        },
        "managers": [manager.name for manager in item.set.project.managers.all()],
        "groups": list(item.set.project.groups.filter(name=Group.MEMBERS).values_list("id", flat=True)),
        "thumbnail": item.thumbnail_url("s", THUMBNAIL_EXPIRATION),
        "is_public": item.is_public and item.set.is_public and item.set.project.is_public,
        "publication_date": item.publication_date,
        "metadata": prepare_metadata(item),
    }
    return doc
