from typing import Any

from elasticsearch_dsl import Boolean, Document, Integer, Keyword, Text

from pount.apps.api.models import Group, Item, Project
from pount.apps.search.elasticsearch.documents import THUMBNAIL_EXPIRATION, IcuCollationKeyword, pount_standard


class ProjectDoc(Document):
    id = Keyword()
    name = Text(
        analyzer=pount_standard,
        fields={"sort": IcuCollationKeyword()},
    )
    description = Text(
        analyzer=pount_standard,
    )
    managers = Text(
        analyzer=pount_standard,
        fields={
            "keyword": Keyword(ignore_above=256),
        },
    )
    thumbnail = Keyword()
    groups = Keyword(multi=True)
    is_public = Boolean()
    sets_count = Integer()
    public_sets_count = Integer()
    items_count = Integer()
    public_items_count = Integer()

    class Index:
        name = "projects"
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 0,
        }

    @classmethod
    def from_model(cls, project: Project) -> "ProjectDoc":
        _doc = prepare_project(project)
        return cls(**_doc)


def prepare_project(project: Project) -> dict[str, Any]:
    def prepare_public_sets_count() -> int:
        try:
            return project.public_sets().count()
        except AttributeError:
            # public_sets() returns None
            return 0

    def prepare_items_count() -> int:
        return Item.objects.filter(set__project=project).distinct().count()

    def prepare_public_items_count() -> int:
        try:
            return Item.objects.filter(set__in=project.public_sets(), is_public=True).distinct().count()
        except TypeError:
            # project.public_sets() returns None
            return 0

    doc = {
        "_id": project.id,
        "doc_type": "project",
        "id": project.id,
        "name": project.name,
        "description": project.description,
        "managers": [manager.name for manager in project.managers.all()],
        "thumbnail": project.thumbnail_url("s", expiration=THUMBNAIL_EXPIRATION),
        "groups": list(project.groups.filter(name=Group.MEMBERS).values_list("id", flat=True)),
        "is_public": project.is_public,
        "sets_count": project.sets.count(),
        "public_sets_count": prepare_public_sets_count(),
        "items_count": prepare_items_count(),
        "public_items_count": prepare_public_items_count(),
    }

    return doc
