from typing import Any

from elasticsearch_dsl import Boolean, Date, Document, Integer, Keyword, Object, Text

from pount.apps.api.models import Group, Set
from pount.apps.search.elasticsearch.documents import THUMBNAIL_EXPIRATION, IcuCollationKeyword, pount_standard


class SetDoc(Document):
    id = Keyword()
    title = Text(
        analyzer=pount_standard,
        fields={
            "sort": IcuCollationKeyword(),
        },
    )
    description = Text(
        analyzer=pount_standard,
    )
    creator = Object(
        properties={
            "name": Text(
                analyzer=pount_standard,
                fields={
                    "keyword": Keyword(ignore_above=256),
                },
            ),
            "id": Keyword(),
        }
    )
    project = Object(
        properties={
            "name": Text(
                analyzer=pount_standard,
            ),
            "id": Keyword(),
        },
    )
    managers = Text(
        analyzer=pount_standard,
        fields={
            "keyword": Keyword(
                ignore_above=256,
            )
        },
        multi=True,
    )
    groups = Keyword(multi=True)
    publication_date = Date(format="date_optional_time")
    publisher = Text(
        analyzer=pount_standard,
        fields={
            "keyword": Keyword(
                ignore_above=256,
            )
        },
    )
    is_public = Boolean()
    items_count = Integer()
    public_items_count = Integer()
    thumbnail = Keyword()

    class Index:
        name = "sets"
        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 0,
        }

    @classmethod
    def from_model(cls, set_model: Set) -> "SetDoc":
        _doc = prepare_set(set_model)
        return cls(**_doc)


def prepare_set(set_model: Set) -> dict[str, Any]:
    def prepare_public_items_count() -> int:
        try:
            return set_model.public_items().count()
        except AttributeError:
            # public_items returns None
            return 0

    doc = {
        "_id": str(set_model.id),
        "doc_type": "set",
        "id": str(set_model.id),
        "title": set_model.title,
        "description": set_model.description,
        "creator": {
            "id": set_model.creator.id,
            "name": set_model.creator.name,
        },
        "project": {
            "id": str(set_model.project.id),
            "name": set_model.project.name,
        },
        "managers": [manager.name for manager in set_model.project.managers.all()],
        "groups": list(set_model.project.groups.filter(name=Group.MEMBERS).values_list("id", flat=True)),
        "publication_date": set_model.publication_date,
        "publisher": set_model.publisher,
        "is_public": set_model.is_public and set_model.project.is_public,
        "items_count": set_model.items.count(),
        "public_items_count": prepare_public_items_count(),
        "thumbnail": set_model.thumbnail_url("s", expiration=THUMBNAIL_EXPIRATION),
    }
    return doc
