import djclick as click
from elasticsearch import RequestError

from pount.apps.api.models import Project, Set
from pount.apps.search.elasticsearch.documents.item import ItemDoc
from pount.apps.search.elasticsearch.documents.project import ProjectDoc
from pount.apps.search.elasticsearch.documents.set import SetDoc
from pount.apps.search.elasticsearch.index import (
    create_index,
    create_projects_index,
    create_sets_index,
    get_settings_and_mappings,
)


@click.group(invoke_without_command=True)
def search_index():
    click.echo("To manage search index, use one of the sub commands")


@search_index.command(help="Create the indexes for the configured sets")
@click.option(
    "--recreate",
    is_flag=True,
    default=False,
    help="Delete the indexes if they already exists and recreate them",
)
def create_set_indices(recreate):
    click.secho("Creating the set indices", fg="green")
    _do_create_set_indices(recreate)


def _do_create_set_indices(recreate: bool = False):
    sets = Set.objects.filter(search_config__isnull=False)
    for set_model in sets:
        create_index(
            set_model.items_index_name,
            get_settings_and_mappings("items-index-settings.json"),
            delete_if_exists=recreate,
        )


@search_index.command(help="Create the indexes for the configured projects")
def create_indices():
    click.secho("Create elasticsearch projects index", fg="green")
    create_projects_index()
    click.secho("Create elasticsearch sets index", fg="green")
    create_sets_index()


@search_index.command(help="Delete the indices and then recreate and populate them")
def rebuild_indices():
    click.secho("Rebuild elasticsearch indices", fg="green")

    # Delete and recreate the indices
    click.secho("Delete and recreate the indices", fg="green")
    create_projects_index(delete_if_exists=True)
    create_sets_index(delete_if_exists=True)
    _do_create_set_indices(recreate=True)

    # Populate the indices
    click.secho("Repopulate the indices", fg="magenta")
    for project in Project.objects.all():
        ProjectDoc.from_model(project).save()

        for set_model in project.sets.all():
            SetDoc.from_model(set_model).save()

            _index_set_items(set_model)

    click.secho("Indices recreated and populated", bg="light_green", fg="black")


def _index_set_items(set_model: Set):
    for item in set_model.items.all():
        try:
            ItemDoc.from_model(item).save(index=set_model.items_index_name)
        except RequestError as e:
            click.secho(
                f"Could not save item {item.id} in index {set_model.items_index_name} ({str(e.error)})",
                fg="red",
                err=True,
            )
            continue


@search_index.command(help="Rebuild a specific set index")
@click.argument("set_id")
def rebuild_set_index(set_id):
    try:
        set_model = Set.objects.get(pk=set_id, search_config__isnull=False)
    except Set.DoesNotExist:
        click.secho("Set does not exist", fg="red")
        return

    click.secho(f"Delete and recreate the index for set {set_model.id} {set_model.title}", fg="green")
    create_index(
        set_model.items_index_name,
        get_settings_and_mappings("items-index-settings.json"),
        delete_if_exists=True,
    )
    _index_set_items(set_model)

    click.secho("Index recreated and populated", fg="black", bg="bright_green")
