from __future__ import annotations

import logging

from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from elasticsearch_dsl import A, Q, Search
from elasticsearch_dsl.response import Response

from pount.apps.api.libs.metadata import TemplateFields
from pount.apps.api.libs.set import DATACITE_SEARCH_CONFIG, add_or_replace_datacite_fields
from pount.apps.api.models import Group, Set, Template, User
from pount.exceptions import PountError
from pount.settings.common import DEFAULT_LANGUAGE

logger = logging.getLogger(__name__)


MAX_PAGES = 100  # We won't allow to fetch more than 100 pages
CACHE_TTL = 60 * 60 * 24  # 24 hours
BASE_FIELDS = ["title^5", "name^5", "description"]


class QueryError(PountError):
    pass


class Query:
    def __init__(self, **kwargs):
        """
        Initializes a new instance of the class.

        Args:
            **kwargs: Keyword arguments that can be used to initialize the instance.
                - language (str): The language of the instance. Defaults to settings.DEFAULT_LANGUAGE
                - user (User): If the user is provided, the query will return only items that can be accessed by the
                    user, otherwise it returns only public items.
                - project_ids (str): Limit the search to the given project IDs.
                - set_ids (str): Limit the search to the given set IDs.
                - page (int): The page number.
                - page_size (int): The page size. Defaults to 10.

        Raises:
            QueryError: If the page number is greater than MAX_PAGES.

        Returns:
            None
        """
        self.language = kwargs.pop("language", DEFAULT_LANGUAGE)
        self.aggregation_fields = {}
        self.excluded_fields = []
        self.exclude_private_fields = False

        self.user: User = kwargs.pop("user", None)

        project_ids = kwargs.pop("project_ids", None)
        self.project_ids: list[str] = project_ids.split(",") if project_ids else []

        set_ids = kwargs.pop("set_ids", None)
        self.set_ids: list[str] = set_ids.split(",") if set_ids else []

        self.page: int | None = kwargs.pop("page", None)
        self.page_size = int(kwargs.pop("page_size", 10))

        if self.page and int(self.page) > MAX_PAGES:
            raise QueryError(_("Page number must be lower than 101"))

        self.filters = {filter_name: filter_value.split(",") for filter_name, filter_value in kwargs.items()}

    def __repr__(self):
        kwargs = {
            "user": self.user,
            "project_ids": self.project_ids,
            "set_ids": self.set_ids,
            "page": self.page,
            "page_size": self.page_size,
            "language": self.language,
            "filters": self.filters,
        }
        return f"{self.__class__.__name__}(**{kwargs})"

    def execute(self, query: str, exclude_private_fields: bool = False) -> Response:
        """
        Executes a search query using Elasticsearch.

        Args:
            query (str): The search query.
            exclude_private_fields (bool, optional): Whether to exclude private fields from the search results.
              Defaults to False.

        Returns:
            Response: The search response from Elasticsearch.
        """
        self.exclude_private_fields = exclude_private_fields

        s = Search()
        s = s.index("projects", "sets", "items-*")

        filters = self.add_filters()
        if len(filters):
            s = s.filter("bool", must=filters)

        match_query = Q(
            "multi_match",
            query=query,
            fields=self.get_search_fields(),
            type="most_fields",
            fuzziness="auto",
            minimum_should_match="75%",
        )
        s = s.query(match_query)

        for bucket_name, agg in self.get_aggregation_fields():
            self.aggregation_fields[bucket_name] = agg
            s.aggs.bucket(bucket_name, agg)

        if self.page:
            start, end = self._get_pagination()
            s = s[start:end]

        if self.excluded_fields and isinstance(self.excluded_fields, list):
            s = s.source(excludes=self.excluded_fields)

        logging.info(s.query.to_dict())

        response = s.execute()

        return response

    def exclude(self, fields: list[str]) -> None:
        """Excludes the given fields from the search results."""
        self.excluded_fields = fields

    def add_filters(self) -> list[Q]:
        filters = []

        if self.project_ids:
            filters.append(Q("terms", project__id=self.project_ids) | Q("terms", id=self.project_ids))

        if self.set_ids:
            filters.append(Q("terms", set__id=self.set_ids))

        filters = self.add_user_filter(filters)

        if self.filters:
            for filter_name, filter_value in self.filters.items():
                if len(filter_value) == 1:
                    filters.append(Q("term", **{filter_name: filter_value[0]}))
                else:
                    filters.append(Q("terms", **{filter_name: filter_value}))

        return filters

    def add_user_filter(self, filters: list[Q]) -> list[Q]:
        if self.user:
            user_project_groups = self.get_groups_for_user()
            filters.append(Q("terms", groups=list(user_project_groups)))
        else:
            # Make sure that if we have no user, we only search through public objects
            filters.append(Q("term", is_public=True))
        return filters

    def get_groups_for_user(self):
        cache_key = f"project_groups_for_user_{self.user.pk}"
        if user_project_groups := cache.get(cache_key):
            return user_project_groups

        user_project_groups = list(self.user.project_groups.filter(name=Group.MEMBERS).values_list("id", flat=True))

        cache.set(cache_key, user_project_groups, 60 * 15)
        return user_project_groups

    def get_search_fields(self) -> list[str]:
        base_fields = BASE_FIELDS

        if self.set_ids and len(self.set_ids) == 1:
            for field in self.get_search_fields_for_set(self.set_ids[0]):
                base_fields.append(field)
        else:
            base_fields += self.get_default_search_fields()
        return base_fields

    def get_aggregation_fields(self) -> list[tuple[str, A]]:
        aggregations = [
            ("doc_type", A("terms", field="doc_type")),
            ("is_public", A("terms", field="is_public")),
        ]
        if len(self.set_ids) == 1:
            aggregations += self.get_aggregations_from_set(self.set_ids[0])
        else:
            aggregations += self.get_default_aggregations()

        return aggregations

    def get_aggregations_from_set(self, set_id: str) -> list[tuple[str, A]]:
        cache_key = f"aggregations_for_set_{set_id}_{self.language}"
        if aggregations := cache.get(cache_key):
            return aggregations

        set_model = Set.objects.select_related("template").get(pk=set_id)
        search_config = set_model.search_config
        template_fields = TemplateFields(set_model.template.fields)
        private_field_names = self._get_private_field_names(set_model, template_fields)

        aggregations = []

        self._build_aggregations(aggregations, search_config, template_fields, private_field_names)
        cache.set(cache_key, aggregations, CACHE_TTL)
        return aggregations

    def _build_aggregations(self, aggregations, search_config, template_fields, private_field_names) -> None:
        for config in search_config:
            if config["config"]["use_in_aggregation"] and config["name"] not in private_field_names:
                field = template_fields.get_field(config["name"])
                if field.component == "UsfSelect":
                    agg_name = config["name"]
                    field_name = f"metadata.{field.name}.opt_{self.language}"
                elif field.component == "ReferentialSelect":
                    agg_name = config["name"]
                    field_name = f"metadata.{field.name}.opt_{self.language}"
                else:
                    agg_name = config["name"]
                    field_name = f"metadata.{field.name}_{config['config']['mapping_type']}"
                if config["config"]["mapping_type"] in ["strings", "strings_fr"]:
                    field_name += ".keyword"
                aggregations.append((agg_name, A("terms", field=field_name)))

    def get_default_aggregations(self) -> list[tuple[str, A]]:
        cache_key = f"aggregations_default_{self.language}"
        if aggregations := cache.get(cache_key):
            return aggregations

        search_config = add_or_replace_datacite_fields(search_config=[])
        template_fields = TemplateFields(Template.objects.get(name="datacite").fields)
        aggregations = []
        private_field_names = []

        self._build_aggregations(aggregations, search_config, template_fields, private_field_names)

        cache.set(cache_key, aggregations, CACHE_TTL)
        return aggregations

    def get_default_search_fields(self) -> list[str]:
        return [
            f"metadata.{key}_{field['mapping_type']}"
            for key, field in DATACITE_SEARCH_CONFIG.items()
            if field.get("use_in_search")
        ]

    def get_search_fields_for_set(self, set_id: str) -> list[str]:
        cache_key = f"search_fields_for_set_{set_id}"
        if fields := cache.get(cache_key):
            return fields

        set_model = Set.objects.select_related("template").get(pk=set_id)
        template_fields = TemplateFields(set_model.template.fields)
        private_fields_names = self._get_private_field_names(set_model, template_fields)

        field_names = []

        for config in set_model.search_config:
            if not config["config"]["use_in_search"] or config["name"] in private_fields_names:
                continue

            field = template_fields.get_field(config["name"])

            field_name = self._get_field_name(config, field)
            field_names.append(field_name)

        cache.set(cache_key, field_names, CACHE_TTL)
        return field_names

    def _get_field_name(self, config, field) -> str:
        if field.component == "UsfSelect":
            field_name = f"metadata.{field.name}.opt_*"
        elif field.component == "ReferentialSelect":
            field_name = f"metadata.{field.name}.opt_*"
        else:
            field_name = f"metadata.{field.name}_{config['config']['mapping_type']}"
        return field_name

    def _get_private_field_names(self, set_model, template_fields) -> list[str]:
        private_fields_names = []
        if self.exclude_private_fields:
            private_fields = set_model.template.private_fields
            private_fields_names = [template_fields.get_field_by_id(field_id).name for field_id in private_fields]
        return private_fields_names

    def _get_pagination(self) -> tuple[int, int]:
        start: int = (int(self.page) - 1) * self.page_size
        end: int = start + self.page_size
        return start, end
