from django.db.models import signals

from pount.apps.api import models
from pount.apps.search import tasks as search_tasks
from pount.apps.search.tasks import update_objects_on_manager_list_change

WATCHED_MODELS = (
    models.Project,
    models.Set,
    models.Item,
    models.Group,
    models.User,
)


def update_project_in_index_when_model_is_saved(sender, instance: models.project.Project, **kwargs):
    search_tasks.update_project_in_index.delay(instance.id)


def delete_project_from_index_when_model_is_deleted(sender, instance: models.project.Project, **kwargs):
    search_tasks.delete_project_in_index.delay(instance.id)


def manage_search_index_when_a_set_is_created_or_updated(sender, instance: models.set.Set, created, **kwargs):
    if created:
        # Create an Index for the items of the set
        search_tasks.create_items_index.delay(instance.id)
        return
    if getattr(instance, "is_deleted", False):
        # Delete the Index for the items of the set
        search_tasks.delete_items_index.delay(instance.items_index_name)
        return
    search_tasks.update_set_in_index.delay(instance.id)


def update_item_in_index_when_model_is_saved(sender, instance: models.item.Item, **kwargs):
    if instance.is_deleted:
        # Items are soft deleted
        search_tasks.delete_item_from_index.delay(instance.id, instance.set.items_index_name)
        return
    search_tasks.update_item_in_index.delay(instance.id)


def delete_item_from_index_when_model_is_deleted(sender, instance: models.item.Item, **kwargs):
    index_name = instance.set.items_index_name
    search_tasks.delete_item_from_index.delay(instance.id, index_name)


def delete_search_index_when_a_set_is_deleted(sender, instance: models.set.Set, **kwargs):
    # Delete the items index when a Set is deleted
    search_tasks.delete_items_index.delay(instance.items_index_name)
    # Delete the set document in the Sets index
    search_tasks.delete_set_from_index(instance.id)


def update_objects_on_user_change(sender, instance: models.user.User, **kwargs):
    if instance.name_has_changed:
        # Update the user's name in all projects, sets and items
        search_tasks.update_objects_on_user_change.delay(instance.id)


def update_objects_on_group_change(sender, instance: models.group.Group, **kwargs):
    # https://docs.djangoproject.com/en/3.2/ref/signals/#m2m-changed
    user_ids = kwargs.pop("pk_set")
    action = kwargs.pop("action")
    model = kwargs.pop("model")

    # Users added or removed from Managers:
    if action in ["post_add", "pre_remove", "pre_clear"] and model == models.user.User:
        project_ids = (
            models.project.Project.objects.filter(
                groups__users__in=user_ids,
                groups__name=models.group.Group.MANAGERS,
            )
            .distinct()
            .values_list("id", flat=True)
        )
        update_objects_on_manager_list_change.delay(list(project_ids))


class BaseSignalProcessor:
    def __init__(self):
        self.setup()

    def setup(self):
        pass

    def teardown(self):
        pass


class ElasticSearchSignalProcessor(BaseSignalProcessor):
    def setup(self):
        # Set
        signals.post_save.connect(manage_search_index_when_a_set_is_created_or_updated, sender=models.set.Set)
        signals.post_delete.connect(delete_search_index_when_a_set_is_deleted, sender=models.set.Set)

        # Project
        signals.post_save.connect(update_project_in_index_when_model_is_saved, sender=models.project.Project)
        signals.post_delete.connect(delete_project_from_index_when_model_is_deleted, sender=models.project.Project)

        # Item
        signals.post_save.connect(update_item_in_index_when_model_is_saved, sender=models.item.Item)
        signals.post_delete.connect(delete_item_from_index_when_model_is_deleted, sender=models.item.Item)

        signals.post_save.connect(update_objects_on_user_change, sender=models.user.User)

        signals.m2m_changed.connect(update_objects_on_group_change, sender=models.group.Group.users.through)

    def teardown(self):
        # Set
        signals.post_save.disconnect(manage_search_index_when_a_set_is_created_or_updated)
        signals.post_delete.disconnect(delete_search_index_when_a_set_is_deleted)

        # Project
        signals.post_save.disconnect(update_project_in_index_when_model_is_saved)
        signals.post_delete.disconnect(delete_search_index_when_a_set_is_deleted)

        # Item
        signals.post_save.disconnect(update_item_in_index_when_model_is_saved)
        signals.post_delete.disconnect(delete_item_from_index_when_model_is_deleted)

        # User
        signals.post_save.disconnect(update_objects_on_user_change)

        signals.m2m_changed.disconnect(update_objects_on_group_change)


class DummySignalProcessor(BaseSignalProcessor):
    """
    Used for unittests: we don't want model to ES sync signals to be processed
    """

    pass
