from __future__ import annotations

from typing import Any
from uuid import UUID

from celery import shared_task
from django.db.models import Q
from elasticsearch import NotFoundError

from pount.apps.api.models import Group, Item, Project, Set, User
from pount.apps.search.elasticsearch.documents.item import ItemDoc
from pount.apps.search.elasticsearch.documents.project import ProjectDoc
from pount.apps.search.elasticsearch.documents.set import SetDoc
from pount.apps.search.elasticsearch.index import create_index, delete_index, get_settings_and_mappings


@shared_task
def create_items_index(set_id: UUID) -> None:
    try:
        set_model = Set.objects.get(pk=set_id)
    except Set.DoesNotExist:
        return
    settings_and_mappings = get_settings_and_mappings("items-index-settings.json")
    index_name = set_model.items_index_name
    create_index(index_name, settings_and_mappings)


@shared_task
def delete_items_index(index_name: str) -> dict[str, Any]:
    return delete_index(index_name)


@shared_task
def update_project_in_index(project_id: UUID) -> None:
    try:
        project = Project.objects.get(pk=project_id)
    except Project.DoesNotExist:
        return
    doc = ProjectDoc.from_model(project)
    doc.save()

    for set_model in project.sets.all():
        update_set_in_index(set_model.id)


@shared_task
def delete_project_in_index(project_id: UUID) -> None:
    try:
        doc = ProjectDoc.get(project_id)
        doc.delete()
    except NotFoundError:
        pass


@shared_task
def update_set_in_index(set_id: UUID) -> None:
    try:
        set_model = Set.objects.get(pk=set_id)
    except Set.DoesNotExist:
        return
    doc = SetDoc.from_model(set_model)
    doc.save()

    project_doc = ProjectDoc.from_model(set_model.project)
    project_doc.save()

    # Update related items
    item_index_name = set_model.items_index_name
    for item in set_model.items.all():
        item_doc = ItemDoc.from_model(item)
        item_doc.save(index=item_index_name)


@shared_task
def delete_set_from_index(set_id: UUID) -> None:
    try:
        doc = SetDoc.get(set_id)
        doc.delete()
    except NotFoundError:
        pass


@shared_task
def update_item_in_index(item_id: UUID) -> None:
    try:
        item = Item.objects.get(pk=item_id)
    except Item.DoesNotExist:
        return
    doc = ItemDoc.from_model(item)
    doc.save(index=item.set.items_index_name)


@shared_task
def delete_item_from_index(item_id: UUID, index_name: str) -> None:
    try:
        doc = ItemDoc.get(item_id, index=index_name)
        doc.delete(index=index_name)
    except NotFoundError:
        pass


@shared_task
def update_objects_on_user_change(user_id: UUID) -> None:
    try:
        user = User.objects.get(pk=user_id)
    except User.DoesNotExist:
        return

    # Update projects where user is a manager
    projects = Project.objects.filter(groups__in=Group.objects.filter(name=Group.MANAGERS, users=user))
    for project in projects:
        doc = ProjectDoc.from_model(project)
        doc.save()

    # Update Sets where user is a creator or a project manager
    sets = Set.objects.filter(Q(creator=user) | Q(project__in=projects))
    for set_model in sets:
        doc = SetDoc.from_model(set_model)
        doc.save()

    # Update Items where user is a creator or a project manager
    items = Item.objects.filter(Q(creator=user) | Q(set__project__in=projects))
    for item in items:
        doc = ItemDoc.from_model(item)
        doc.save(index=item.set.items_index_name)


@shared_task
def update_objects_on_manager_list_change(project_ids: list[UUID]) -> None:
    # User's have been added or removed from the list of managers
    for project in Project.objects.filter(id__in=project_ids):
        project_doc = ProjectDoc.from_model(project)
        project_doc.save()

        for set_model in project.sets.all():
            set_doc = SetDoc.from_model(set_model)
            set_doc.save()

            for item in set_model.items.all():
                item_doc = ItemDoc.from_model(item)
                item_doc.save(index=item.set.items_index_name)
