import uuid
from unittest.mock import patch

from elasticsearch_dsl import A, Q

from pount.apps.api.libs.metadata import TemplateEntry, TemplateFields
from pount.apps.api.libs.set import DATACITE_SEARCH_CONFIG
from pount.apps.api.models import Group, Project, User
from pount.apps.api.models.init import initialize_roles, initialize_templates
from pount.apps.api.tests.factories.project import ProjectFactory
from pount.apps.api.tests.factories.set import SetFactory
from pount.apps.api.tests.factories.user import UserFactory
from pount.apps.search.query import BASE_FIELDS, MAX_PAGES, Query, QueryError
from pount.tests import TestCase


class ExcludeFieldsTest(TestCase):
    def test_exclude_fields(self):
        user = UserFactory()
        query = Query(
            user=user,
        )
        query.exclude(["name"])
        self.assertEqual(query.excluded_fields, ["name"])


class UserFilterTest(TestCase):
    def test_if_user_is_set_apply_user_s_groups_filter(self):
        initialize_roles()
        user: User = UserFactory()
        project: Project = ProjectFactory()
        project.add_member(user)
        project.add_manager(user)
        query = Query(
            user=user,
        )
        self.assertListEqual(
            [
                Q(
                    "terms",
                    groups=list(project.groups.filter(name=Group.MEMBERS, users=user).values_list("id", flat=True)),
                )
            ],
            query.add_user_filter([]),
        )

    def test_if_use_is_not_set_apply_filter_for_public_objects_only(self):
        query = Query(user=None)
        self.assertListEqual(
            [Q("term", is_public=True)],
            query.add_user_filter([]),
        )


class QueryFilterTest(TestCase):
    def setUp(self):
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()

    def tearDown(self):
        self.patcher1.stop()

    def test_project_id_filter(self):
        initialize_roles()
        project = ProjectFactory()
        query = Query(
            project_ids=str(project.id),
        )
        self.assertListEqual(
            [
                (Q("terms", project__id=[str(project.id)]) | Q("terms", id=[str(project.id)])),
                Q("term", is_public=True),
            ],
            query.add_filters(),
        )

    def test_multiple_project_id_filter(self):
        initialize_roles()
        project1 = ProjectFactory()
        project2 = ProjectFactory()
        query = Query(project_ids=",".join([str(project1.id), str(project2.id)]))
        self.assertListEqual(
            [
                (
                    Q("terms", project__id=[str(project1.id), str(project2.id)])
                    | Q("terms", id=[str(project1.id), str(project2.id)])
                ),
                Q("term", is_public=True),
            ],
            query.add_filters(),
        )

    def test_set_id_filter(self):
        initialize_roles()
        set_model = SetFactory()
        query = Query(
            set_ids=str(set_model.id),
        )
        self.assertListEqual(
            [
                Q("terms", set__id=[str(set_model.id)]),
                Q("term", is_public=True),
            ],
            query.add_filters(),
        )

    def test_multiple_set_id_filter(self):
        initialize_roles()
        set_model1 = SetFactory()
        set_model2 = SetFactory()
        query = Query(set_ids=",".join([str(set_model1.id), str(set_model2.id)]))
        self.assertListEqual(
            [
                Q("terms", set__id=[str(set_model1.id), str(set_model2.id)]),
                Q("term", is_public=True),
            ],
            query.add_filters(),
        )

    def test_remaining_filters_are_applied(self):
        query = Query(
            some_filter_name="some_value",
        )

        self.assertListEqual(
            [
                Q("term", is_public=True),
                Q("term", some_filter_name="some_value"),
            ],
            query.add_filters(),
        )

    def test_mutliple_remaining_filters_are_applied(self):
        query = Query(
            some_filter_name="first_value,second_value,third_value",
            some_other_filter_name="some_other_value",
        )
        self.assertListEqual(
            [
                Q("term", is_public=True),
                Q("terms", some_filter_name=["first_value", "second_value", "third_value"]),
                Q("term", some_other_filter_name="some_other_value"),
            ],
            query.add_filters(),
        )


class PaginationTest(TestCase):
    def test_max_number_of_pages(self):
        with self.assertRaises(QueryError) as e:
            Query(page=MAX_PAGES + 1)
        self.assertIn(f"Page number must be lower than {MAX_PAGES + 1}", str(e.exception))

    def test_get_start_and_end(self):
        query = Query(page=2)
        self.assertTupleEqual(
            query._get_pagination(),
            (10, 20),
        )

        query = Query(page=1)
        self.assertTupleEqual(
            query._get_pagination(),
            (0, 10),
        )


class TestSearchFields(TestCase):
    def setUp(self):
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()

    def tearDown(self):
        self.patcher1.stop()

    def _get_default_fields(self):
        return [
            f"metadata.creator_{DATACITE_SEARCH_CONFIG['creator']['mapping_type']}",
            f"metadata.title_{DATACITE_SEARCH_CONFIG['title']['mapping_type']}",
            f"metadata.publisher_{DATACITE_SEARCH_CONFIG['publisher']['mapping_type']}",
            f"metadata.resourceType_{DATACITE_SEARCH_CONFIG['resourceType']['mapping_type']}",
        ]

    def test_get_default_search_fields(self):
        default_search_fields = Query(language="fr").get_default_search_fields()
        self.assertListEqual(
            self._get_default_fields(),
            default_search_fields,
        )

    def test_get_search_fields(self):
        self.assertListEqual(
            BASE_FIELDS + self._get_default_fields(),
            Query(language="fr").get_search_fields(),
        )

    def test_search_fields_for_set(self):
        set_id: str = str(uuid.uuid4())
        with patch("pount.apps.search.query.Query.get_search_fields_for_set") as get_search_fields_for_set:
            get_search_fields_for_set.return_value = ["first_field", "second_field"]
            self.assertListEqual(
                BASE_FIELDS + ["first_field", "second_field"],
                Query(set_ids=set_id).get_search_fields(),
            )
        get_search_fields_for_set.assert_called_once_with(set_id)

    def test_get_search_fields_for_set(self):
        initialize_roles()
        set_model = SetFactory()
        query = Query(language="fr")
        fieldnames = query.get_search_fields_for_set(str(set_model.id))
        self.assertListEqual(
            self._get_default_fields(),
            fieldnames,
        )

    def test_get_field_name_for_select(self):
        config = {}
        field = TemplateEntry(
            entry={
                "id": "2qwhbanmbsc",
                "name": "resourceTypeGeneral",
                "options": [],
                "multiple": False,
                "visible": True,
                "required": True,
                "component": "UsfSelect",
                "protected": True,
            }
        )
        self.assertEqual(
            "metadata.resourceTypeGeneral.opt_*",
            Query()._get_field_name(config, field),
        )

    def test_get_field_name_for_referential_select(self):
        config = {}
        field = TemplateEntry(
            entry={
                "id": "2qwhbanmbsc",
                "name": "fieldname",
                "component": "ReferentialSelect",
                "options": [],
            }
        )
        self.assertEqual(
            "metadata.fieldname.opt_*",
            Query()._get_field_name(config, field),
        )

    def test_get_field_name_for_any_field(self):
        config = {"config": {"mapping_type": "bools"}}
        field = TemplateEntry(
            entry={
                "id": "2qwhbanmbsc",
                "name": "fieldname",
                "component": "text",
            }
        )
        self.assertEqual(
            "metadata.fieldname_bools",
            Query()._get_field_name(config, field),
        )

    def test_get_privage_field_names(self):
        initialize_roles()
        set_model = SetFactory()
        set_model.template.private_fields = ["2qwhbanmbsc"]
        set_model.template.save()
        template_fields = TemplateFields(set_model.template.fields)
        query = Query()
        query.exclude_private_fields = True
        self.assertListEqual(
            ["resourceTypeGeneral"],
            query._get_private_field_names(set_model, template_fields),
        )


class AggregationFieldTest(TestCase):
    def setUp(self):
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()

    def tearDown(self):
        self.patcher1.stop()

    def test_get_default_aggregation_field_names(self):
        initialize_roles()
        initialize_templates()
        aggregations = Query().get_default_aggregations()
        self.assertListEqual(
            [
                ("publicationYear", A("terms", field="metadata.publicationYear_longs")),
                ("resourceTypeGeneral", A("terms", field="metadata.resourceTypeGeneral.opt_fr.keyword")),
            ],
            aggregations,
        )

    def test_get_aggregations_from_set(self):
        initialize_roles()
        set_model = SetFactory()
        aggregations = Query().get_aggregations_from_set(str(set_model.id))
        self.assertListEqual(
            [
                ("publicationYear", A("terms", field="metadata.publicationYear_longs")),
                ("resourceTypeGeneral", A("terms", field="metadata.resourceTypeGeneral.opt_fr.keyword")),
            ],
            aggregations,
        )

    def test_get_aggregation_fields_without_set_id(self):
        with patch("pount.apps.search.query.Query.get_default_aggregations") as get_default_aggregations:
            get_default_aggregations.return_value = [("agg_name", A("terms", field="agg_field"))]
            aggregations = Query().get_aggregation_fields()
        self.assertListEqual(
            [
                ("doc_type", A("terms", field="doc_type")),
                ("is_public", A("terms", field="is_public")),
                ("agg_name", A("terms", field="agg_field")),
            ],
            aggregations,
        )

    def test_get_aggregation_fields_with_set_id(self):
        initialize_roles()
        initialize_templates()
        with patch("pount.apps.search.query.Query.get_aggregations_from_set") as get_aggregations_from_set:
            get_aggregations_from_set.return_value = [("set_agg_name", A("terms", field="set_agg_field"))]
            aggregations = Query(set_ids=str(uuid.uuid4())).get_aggregation_fields()
        self.assertListEqual(
            [
                ("doc_type", A("terms", field="doc_type")),
                ("is_public", A("terms", field="is_public")),
                ("set_agg_name", A("terms", field="set_agg_field")),
            ],
            aggregations,
        )
