from unittest.mock import call, patch
from uuid import uuid4

from elasticsearch import NotFoundError

from pount.apps.api.models.init import initialize_roles
from pount.apps.api.tests.factories.item import ItemFactory
from pount.apps.api.tests.factories.project import ProjectFactory
from pount.apps.api.tests.factories.set import SetFactory
from pount.apps.api.tests.factories.user import UserFactory
from pount.apps.search.tasks import (
    delete_item_from_index,
    delete_project_in_index,
    delete_set_from_index,
    update_item_in_index,
    update_objects_on_manager_list_change,
    update_objects_on_user_change,
    update_project_in_index,
    update_set_in_index,
)
from pount.tests import TestCase


class UpdateProjectInIndexTest(TestCase):
    def test_project_doc_is_created_from_model_id(self):
        initialize_roles()
        project = ProjectFactory()
        with patch("pount.apps.search.tasks.ProjectDoc") as ProjectDoc:
            update_project_in_index(project.id)
        ProjectDoc.from_model.assert_called_once_with(project)
        ProjectDoc.from_model.return_value.save.assert_called_once_with()

    def test_project_doc_is_not_created_if_project_does_not_exist(self):
        with patch("pount.apps.search.tasks.ProjectDoc") as ProjectDoc:
            update_project_in_index(uuid4())
        ProjectDoc.from_model.assert_not_called()

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    @patch("pount.apps.search.tasks.ProjectDoc")
    def test_project_update_also_updates_related_sets(self, ProjectDoc, handle_update_mock):
        initialize_roles()
        project = ProjectFactory()
        set_model1 = SetFactory(project=project)
        set_model2 = SetFactory(project=project)
        with patch("pount.apps.search.tasks.update_set_in_index") as update_set_in_index:
            update_project_in_index(project.id)
        self.assertEqual(update_set_in_index.call_count, 2)
        update_set_in_index.assert_has_calls(
            [
                call(set_model1.id),
                call(set_model2.id),
            ]
        )


class DeleteProjectInIndexTest(TestCase):
    def test_project_doc_is_deleted_from_model_id(self):
        initialize_roles()
        project = ProjectFactory()
        with patch("pount.apps.search.tasks.ProjectDoc") as ProjectDoc:
            delete_project_in_index(project.id)
        ProjectDoc.get.assert_called_once_with(project.id)
        ProjectDoc.get.return_value.delete.assert_called_once_with()

    def test_project_doc_is_not_deleted_if_project_does_not_exist(self):
        with patch("pount.apps.search.tasks.ProjectDoc") as ProjectDoc:
            ProjectDoc.get.return_value.delete.side_effect = NotFoundError
            delete_project_in_index(uuid4())
        ProjectDoc.from_model.assert_not_called()


class UpdateSetInIndexTest(TestCase):
    def setUp(self):
        self.patcher1 = patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
        self.patcher1.start()

    def tearDown(self):
        self.patcher1.stop()

    @patch("pount.apps.search.tasks.ProjectDoc")
    @patch("pount.apps.search.tasks.SetDoc")
    @patch("pount.apps.search.tasks.ItemDoc")
    def test_set_doc_is_updated_from_model(self, ItemDoc, SetDoc, ProjectDoc):
        initialize_roles()
        set_model = SetFactory()
        update_set_in_index(set_model.id)
        SetDoc.from_model.assert_called_once_with(set_model)
        SetDoc.from_model.return_value.save.assert_called_once_with()

    @patch("pount.apps.search.tasks.Set.objects")
    def test_set_doc_is_not_updated_if_set_does_not_exist(self, objects):
        with patch("pount.apps.search.tasks.SetDoc") as SetDoc:
            from pount.apps.api.models import Set

            objects.get.side_effect = Set.DoesNotExist
            update_set_in_index(uuid4())
        SetDoc.from_model.assert_not_called()

    @patch("pount.apps.search.tasks.ProjectDoc")
    @patch("pount.apps.search.tasks.SetDoc")
    def test_update_set_updates_associated_project(self, SetDoc, ProjectDoc):
        initialize_roles()
        set_model = SetFactory()
        update_set_in_index(set_model.id)
        ProjectDoc.from_model.assert_called_once_with(set_model.project)
        ProjectDoc.from_model.return_value.save.assert_called_once_with()

    @patch("pount.apps.search.tasks.ProjectDoc")
    @patch("pount.apps.search.tasks.SetDoc")
    def test_update_set_updates_associated_items(self, SetDoc, ProjectDoc):
        initialize_roles()
        set_model = SetFactory()
        item_model = ItemFactory(set=set_model)
        with patch("pount.apps.search.tasks.ItemDoc") as ItemDoc:
            update_set_in_index(set_model.id)
        ItemDoc.from_model.assert_has_calls(
            [
                call(item_model),
            ]
        )

        ItemDoc.from_model.return_value.save.assert_has_calls(
            [
                call(index=item_model.set.items_index_name),
            ]
        )


class DeleteSetFromIndexTest(TestCase):
    def test_set_doc_is_deleted_from_model_id(self):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_roles()
            set_model = SetFactory()
            with patch("pount.apps.search.tasks.SetDoc") as SetDoc:
                delete_set_from_index(set_model.id)
            SetDoc.get.assert_called_once_with(set_model.id)
            SetDoc.get.return_value.delete.assert_called_once_with()

    def test_set_doc_is_not_deleted_if_set_does_not_exist(self):
        with patch("pount.apps.search.tasks.SetDoc") as SetDoc:
            SetDoc.get.return_value.delete.side_effect = NotFoundError
            delete_set_from_index(uuid4())
        SetDoc.from_model.assert_not_called()


class UpdateItemInIndexTest(TestCase):
    def test_item_doc_is_updated_from_model_id(self):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_roles()
            item = ItemFactory()
            with patch("pount.apps.search.tasks.ItemDoc") as ItemDoc:
                update_item_in_index(item.id)
            ItemDoc.from_model.assert_called_once_with(item)
            ItemDoc.from_model.return_value.save.assert_called_once_with(index=item.set.items_index_name)

    @patch("pount.apps.search.tasks.Item.objects")
    def test_item_doc_is_not_updated_if_item_does_not_exist(self, objects):
        with patch("pount.apps.search.tasks.ItemDoc") as ItemDoc:
            from pount.apps.api.models import Item

            objects.get.side_effect = Item.DoesNotExist
            update_item_in_index(uuid4())
        ItemDoc.from_model.assert_not_called()


class DeleteItemFromIndexTest(TestCase):
    def test_item_doc_is_deleted_from_model_id(self):
        with patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save"):
            initialize_roles()
            item = ItemFactory()
            with patch("pount.apps.search.tasks.ItemDoc") as ItemDoc:
                delete_item_from_index(item.id, item.set.items_index_name)
            ItemDoc.get.assert_called_once_with(item.id)
            ItemDoc.get.return_value.delete.assert_called_once_with(index=item.set.items_index_name)

    def test_item_doc_is_not_deleted_if_item_does_not_exist(self):
        with patch("pount.apps.search.tasks.ItemDoc") as ItemDoc:
            ItemDoc.get.return_value.delete.side_effect = NotFoundError
            delete_item_from_index(uuid4(), "index-name")
        ItemDoc.from_model.assert_not_called()


class UpdateObjectsOnUserChangeTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.user = UserFactory()

    @patch("pount.apps.search.tasks.ItemDoc")
    @patch("pount.apps.search.tasks.SetDoc")
    @patch("pount.apps.search.tasks.ProjectDoc")
    def test_objects_are_not_updated_if_user_does_not_exist(self, ProjectDoc, SetDoc, ItemDoc):
        update_objects_on_user_change(uuid4())
        ProjectDoc.assert_not_called()

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    @patch("pount.apps.search.tasks.ItemDoc")
    @patch("pount.apps.search.tasks.SetDoc")
    @patch("pount.apps.search.tasks.ProjectDoc")
    def test_projects_where_user_is_a_manager_are_updated(self, ProjectDoc, SetDoc, ItemDoc, handle_update_mock):
        project = ProjectFactory()
        project.add_member(self.user)
        project.add_manager(self.user)
        set_model = SetFactory(project=project)
        item = ItemFactory(set=set_model)
        update_objects_on_user_change(self.user.id)
        ProjectDoc.from_model.assert_called_once_with(project)
        ProjectDoc.from_model.return_value.save.assert_called_once_with()
        SetDoc.from_model.assert_called_once_with(set_model)
        SetDoc.from_model.return_value.save.assert_called_once_with()
        ItemDoc.from_model.assert_called_once_with(item)
        ItemDoc.from_model.return_value.save.assert_called_once_with(index=item.set.items_index_name)


class UpdateObjectsOnManagerListChangeTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        initialize_roles()
        cls.user = UserFactory()

    @patch("pount.apps.api.models.template.handle_update_item_metadata_on_template_save")
    @patch("pount.apps.search.tasks.ItemDoc")
    @patch("pount.apps.search.tasks.SetDoc")
    @patch("pount.apps.search.tasks.ProjectDoc")
    def test_objects_depending_on_project_are_updated(self, ProjectDoc, SetDoc, ItemDoc, handle_update_mock):
        project = ProjectFactory()
        set_model = SetFactory(project=project)
        ItemFactory(set=set_model)

        update_objects_on_manager_list_change([project.id])
        ProjectDoc.from_model.assert_called_once_with(project)
        ProjectDoc.from_model.return_value.save.assert_called_once_with()

        set_model = project.sets.first()
        SetDoc.from_model.assert_called_once_with(set_model)
        SetDoc.from_model.return_value.save.assert_called_once_with()

        item = set_model.items.first()
        ItemDoc.from_model.assert_called_once_with(item)
        ItemDoc.from_model.return_value.save.assert_called_once_with(index=set_model.items_index_name)
