from __future__ import annotations

import hashlib
import json
from functools import wraps
from json import JSONEncoder as _JSONEncoder
from typing import Any, Callable

from django.core.cache import cache
from django.db.models import Model

DEFAULT_TIMEOUT = 60 * 5


class JSONEncoder(_JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Model):
            return f"{o._meta.label}.{o.pk}"
        return super().default(o)


def cache_get_key(key_prefix: str | Callable, **kwargs) -> str:
    if callable(key_prefix):
        prefix_object = kwargs.pop("prefix_object")
        prefix = key_prefix(kwargs.pop(prefix_object))
    else:
        prefix = str(key_prefix)
    _kwargs = json.dumps(kwargs, sort_keys=True, cls=JSONEncoder)
    param_hash: str = hashlib.md5(_kwargs.encode()).hexdigest()
    return f"{prefix}:{param_hash}"


def cached(
    timeout: int | None = DEFAULT_TIMEOUT,
    key_prefix: str | Callable = None,
    key_object: str | None = None,
    key_name: str | None = None,
):
    """Decorator to cache results of a function"""

    def cached_decorator(func):
        @wraps(func)
        def _wrapped_function(*args, **kwargs):
            if callable(key_prefix) and key_object is None:
                raise ValueError("key_object must be set when using a callable as key_prefix")
            _key_name = key_name or cache_get_key(
                key_prefix=key_prefix,
                prefix_object=key_object,
                **kwargs,
            )
            _result = cache.get(_key_name)

            if not _result:
                _result = func(*args, **kwargs)
                cache.set(_key_name, _result, timeout)

            return _result

        return _wrapped_function

    return cached_decorator
