# flake8: noqa:E501
from os import environ
from pathlib import Path

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from django.core.exceptions import ImproperlyConfigured

######################
# Path configuration #
######################

# See: https://adamj.eu/tech/2020/03/16/use-pathlib-in-your-django-project/
BASE_DIR = Path(__file__).resolve(strict=True).parent.parent
SITE_ROOT = str(BASE_DIR)
SITE_NAME = BASE_DIR.name

JSON_SCHEMA_DIR = BASE_DIR / "schemas"

#########################
# General configuration #
#########################
SITE_ID = 1

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "this_is_my_key_there_are_many_like_it_but_this_one_is_mine"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

# Hosts/domain names that are valid for this site.
# If DEBUG = True and ALLOWED_HOSTS is left empty here, Django will behave
# as if ALLOWED_HOSTS = ['localhost', '127.0.0.1', '[::1]'].
# See: https://docs.djangoproject.com/fr/3.0/ref/settings/#allowed-hosts
ALLOWED_HOSTS = []

ROOT_URLCONF = f"{SITE_NAME}.urls"

WSGI_APPLICATION = f"{SITE_NAME}.wsgi.application"


########################
# Internationalization #
########################
# See: https://docs.djangoproject.com/en/3.0/topics/i18n/
# Language for this installation.
DEFAULT_LANGUAGE = "fr"
LANGUAGES = [("fr","French"), ("en", "English")]
LANGUAGE_CODE = DEFAULT_LANGUAGE

# Local timezone for this installation.
TIME_ZONE = "Europe/Paris"
# If USE_I18N = True, Django will do some optimization regarding i18n loading.
USE_I18N = True
# If USE_L10N = True, Django will format dates, numbers and calendars
# according to the current locale.
# If USE_TZ = True, Django will use timezone-aware datetimes.
USE_TZ = True

LOCALE_PATHS = [BASE_DIR / "locale"]



##############################
# Static files configuration #
##############################

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = BASE_DIR.parent / "assets"

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/site_media/'

# Additional locations of static files
STATICFILES_DIRS = [BASE_DIR / "static"]

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
]

#################
# Authorization #
#################

AUTH_USER_MODEL = "api.User"

AUTHENTICATION_BACKENDS = (
    "rules.permissions.ObjectPermissionBackend",
    "django.contrib.auth.backends.ModelBackend",
    "django_cas.backends.CASBackend",
)

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
]


def username_format(username):
    from pount.apps.api.models.user import AuthenticationProvider

    return f"{AuthenticationProvider.CAS.value}/{username.strip().lower()}"


LOGIN_URL = "/cas/login"
CAS_SERVER_URL = "https://cas.unistra.fr/cas/"
CAS_LOGOUT_REQUEST_ALLOWED = ("cas1.di.unistra.fr", "cas2.di.unistra.fr")
CAS_LOGOUT_COMPLETELY = True
CAS_REDIRECT_URL = "/"
CAS_RETRY_LOGIN = True
CAS_USERNAME_FORMAT = username_format
# On CAS login, we call this callback that retrieves and updates user info
CAS_RESPONSE_CALLBACKS = ('pount.apps.api.libs.cas.update_or_create_user',)

SIMPLE_JWT = {
    "ALGORITHM": "RS256",
    "UPDATE_LAST_LOGIN": True,
}


def load_key(filename, key_type, **kwargs):
    if key_type not in ("SIGNING_KEY", "VERIFYING_KEY"):
        raise ImproperlyConfigured(
            "key_type must be SIGNING_KEY or VERIFYING_KEY"
        )
    path = BASE_DIR.parent / "keys" / filename
    if not path.exists():
        raise ImproperlyConfigured(f'The JWT key "{path}" does not exist.')
    if key_type == "SIGNING_KEY":
        password = kwargs.get("password", None)
        if password is not None:
            password = password.encode()
        file = path.open("rb")
        private_key = serialization.load_pem_private_key(
            file.read(),
            password=password,
            backend=default_backend(),
        )
        file.close()
        SIMPLE_JWT[key_type] = private_key
    else:
        file = path.open("rb")
        SIMPLE_JWT[key_type] = file.read()
        file.close()


load_key(
    "jwt.key",
    "SIGNING_KEY",
    password=environ.get("SIGNING_KEY_PASSWORD", None),
)
load_key("jwt.key.pub", "VERIFYING_KEY")


# ADMINS will be notified of 500 errors by email.
# MANAGERS will be notified of 404 errors (use IGNORABLE_404_URLS to filter).
# See: https://docs.djangoproject.com/en/3.0/howto/error-reporting/
ADMINS = [
    # ('Your Name', 'your_email@example.com'),
]
MANAGERS = ADMINS

POUNT_ADMIN_EMAIL = "pount@unistra.fr"


#########################
# Session configuration #
#########################

SESSION_SERIALIZER = "django.contrib.sessions.serializers.JSONSerializer"


###########
# Logging #
###########

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(levelname)s %(asctime)s %(name)s:%(lineno)s %(message)s"
        },
        "django.server": {
            "()": "django.utils.log.ServerFormatter",
            "format": "[%(server_time)s] %(message)s",
        },
    },
    "filters": {
        "require_debug_false": {
            "()": "django.utils.log.RequireDebugFalse",
        },
        "require_debug_true": {
            "()": "django.utils.log.RequireDebugTrue",
        },
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "filters": ["require_debug_true"],
            "class": "logging.StreamHandler",
        },
        "django.server": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "django.server",
        },
        "mail_admins": {
            "level": "ERROR",
            "filters": ["require_debug_false"],
            "class": "django.utils.log.AdminEmailHandler",
        },
        "file": {
            "level": "INFO",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": "",
            "maxBytes": 209715200,
            "backupCount": 3,
            "formatter": "default",
        },
        "cron": {
            "level": "INFO",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": "",
            "maxBytes": 209715200,
            "backupCount": 3,
            "formatter": "default",
        }
    },
    "loggers": {
        "django": {
            "handlers": ["console", "mail_admins"],
            "level": "INFO",
        },
        "django.server": {
            "handlers": ["django.server"],
            "level": "INFO",
            "propagate": False,
        },
        "pount": {
            "handlers": ["mail_admins", "file"],
            "level": "ERROR",
            "propagate": True,
        },
        "cron": {
            "handlers": ["cron"],
            "level": "INFO",
            "propagate": True,
        }
    },
}

SENTRY_DSN = "https://ee5a1720770a43e5be41a4f934a98c33@sentry.app.unistra.fr/7"


#############
# Databases #
#############

# In your virtualenv, edit the activation file and set
# the environment variables defined in this file
# ex: export DEFAULT_DB_USER='postgresql'
# See: https://docs.djangoproject.com/en/3.0/ref/settings/#databases
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": environ.get("DEFAULT_DB_NAME", "pount"),
        "USER": environ.get("DEFAULT_DB_USER", "pount"),
        "PASSWORD": environ.get("DEFAULT_DB_PASSWORD", "pount"),
        "HOST": environ.get("DEFAULT_DB_HOST", "localhost"),
        "PORT": environ.get("DEFAULT_DB_PORT"),
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"


###############
# Middlewares #
###############

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django_cas.middleware.CASMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]


################
# Applications #
################

DJANGO_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

THIRD_PARTY_APPS = [
    "rest_framework",  # Django REST api
    "rest_framework.authtoken",
    "rules.apps.AutodiscoverRulesConfig",  # for per-object authorization
    "corsheaders",  # CORS
    "django_cas",
    "django_filters",
    "drf_spectacular",
    "health_check",
    "health_check.db",
    "health_check.contrib.rabbitmq",
    "health_check.contrib.s3boto3_storage",
]

LOCAL_APPS = [
    "pount.apps.api",
    "pount.apps.public",
    "pount.apps.search",
]

INSTALLED_APPS = DJANGO_APPS + THIRD_PARTY_APPS + LOCAL_APPS


#############
# Templates #
#############
# Templates are not used by this project, but Django wants its config. :|
# APP_DIRS is True because rest_framework needs its page.
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "DIRS": [BASE_DIR / "templates"],
        "OPTIONS": {
            "debug": DEBUG,
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ],
        },
    }
]


#########
# CORS  #
#########
CORS_ORIGIN_ALLOW_ALL = True
CORS_ALLOW_HEADERS = (
    "x-requested-with",
    "content-type",
    "accept",
    "origin",
    "authorization",
    "x-csrftoken",
    "range",
)


###########################
# Pount-specific settings #
###########################
ITEM_IMPORTERS = {
    "zenodo": "pount.apps.api.importers.zenodo",
    "dataverse": None,  # TODO
}

#########################
# django-rest-framework #
#########################

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework_simplejwt.authentication.JWTAuthentication",
        "rest_framework.authentication.TokenAuthentication",
    ],
    "DEFAULT_RENDERER_CLASSES": (
        "djangorestframework_camel_case.render.CamelCaseJSONRenderer",
        "djangorestframework_camel_case.render.CamelCaseBrowsableAPIRenderer",
    ),
    "DEFAULT_PARSER_CLASSES": (
        "djangorestframework_camel_case.parser.CamelCaseFormParser",
        "djangorestframework_camel_case.parser.CamelCaseMultiPartParser",
        "djangorestframework_camel_case.parser.CamelCaseJSONParser",
    ),
    "JSON_UNDERSCOREIZE": {
        "no_underscore_before_number": True,  # Leave UUIDs in JSONFields alone
    },
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
    "DEFAULT_FILTER_BACKENDS": ["django_filters.rest_framework.DjangoFilterBackend"],
    "DEFAULT_PAGINATION_CLASS": "pount.apps.api.pagination.StandardPagination",
    "PAGE_SIZE": 10,
}

SPECTACULAR_SETTINGS = {
    "TITLE": "Pount API",
    "DESCRIPTION": "API for Pount Project",
    "VERSION": "1.0.0",
    "POSTPROCESSING_HOOKS": [
        "drf_spectacular.hooks.postprocess_schema_enums",
        "drf_spectacular.contrib.djangorestframework_camel_case.camelize_serializer_fields",
    ],
    "COMPONENT_SPLIT_REQUEST": True,
    "ENUM_NAME_OVERRIDES": {
        "ReferentialScopeEnum": "pount.apps.api.models.referential.Referential.Scope",
        "TemplateScopeEnum": "pount.apps.api.models.template.Template.SCOPES",
    }
}

NOTIFICATIONS_NOTIFICATION_MODEL = "api.Notification"


##############
# S3 Storage #
##############
AWS_ACCESS_KEY_ID = environ.get("AWS_ACCESS_KEY_ID")
AWS_SECRET_ACCESS_KEY = environ.get("AWS_SECRET_ACCESS_KEY")
AWS_STORAGE_BUCKET_NAME = "pount"
AWS_S3_ENDPOINT_URL = "https://s3.unistra.fr"
AWS_S3_FILE_OVERWRITE = False
STORAGES = {
    "default": {
        "BACKEND": "pount.apps.api.storage_backends.ItemMediaFileStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
}


HARD_DELETE_AFTER_DAYS = 30


##########
# CACHES #
##########
CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": "redis://127.0.0.1:6379/7",
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        },
        "KEY_PREFIX": "pount",
    },
}

###########
# CAMELOT #
###########
CAMELOT = {
    "TOKEN": environ.get("CAMELOT_TOKEN"),
    "URL": environ.get("CAMELOT_URL", "")
}

#################
# ELASTICSEARCH #
#################
ELASTICSEARCH_DSL = {
    "default": {
        "hosts": environ.get('ELASTICSEARCH_HOST', 'localhost'),
    },
}


##########
# CELERY #
##########
RABBITMQ_USER = environ.get("RABBITMQ_USER", "")
RABBITMQ_PASSWORD = environ.get("RABBITMQ_PASSWORD", "")
RABBITMQ_SERVER = environ.get("RABBITMQ_SERVER", "localhost")
RABBITMQ_VHOST = environ.get("RABBITMQ_VHOST", "")
CELERY_NAME = "pount"
CELERY_BROKER_URL = f"amqp://{RABBITMQ_USER}:{RABBITMQ_PASSWORD}@{RABBITMQ_SERVER}/{RABBITMQ_VHOST}"
CELERY_RESULT_BACKEND = ""


#########
# MONGO #
#########
MONGODB_USER = environ.get("MONGODB_USER", "pount_dev")
MONGODB_PASSWORD = environ.get("MONGODB_PASSWORD", "pount_dev")
MONGODB_HOST = environ.get("MONGODB_HOST", "localhost")
MONGODB_DATABASE = environ.get("MONGODB_DATABASE", "pount_dev")
MONGODB_URI = f"mongodb://{MONGODB_USER}:{MONGODB_PASSWORD}@{MONGODB_HOST}/?retryWrites=true&w=majority&authSource={MONGODB_DATABASE}"
