import typing

from django.http import HttpResponse
from django.test import TestCase as DjangoTestCase
from rest_framework import status
from rest_framework.response import Response
from rest_framework.test import APIClient

ResponseType = typing.Union[HttpResponse, Response]


class TestCase(DjangoTestCase):
    def _perform_request(self, method: str, path, *args, **kwargs) -> ResponseType:
        if user := kwargs.pop("user", None):
            self.client.force_login(user)
        return getattr(self.client, method)(path, *args, **kwargs)

    def get(self, path: str, *args, **kwargs) -> ResponseType:
        return self._perform_request("get", path, *args, **kwargs)

    def post(self, path: str, *args, **kwargs) -> ResponseType:
        return self._perform_request("post", path, *args, **kwargs)

    def patch(self, path: str, *args, **kwargs) -> ResponseType:
        return self._perform_request("patch", path, *args, **kwargs)

    def put(self, path: str, *args, **kwargs) -> ResponseType:
        return self._perform_request("put", path, *args, **kwargs)

    def delete(self, path: str, *args, **kwargs) -> ResponseType:
        return self._perform_request("delete", path, *args, **kwargs)

    def response_ok(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def response_created(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def response_no_content(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def response_bad_request(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def response_unauthorized(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def response_forbidden(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def response_not_found(self, response: ResponseType) -> None:
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class APITestCase(TestCase):
    client_class = APIClient
